IMPLEMENTATION MODULE B2WBoot;  (* Bootstrap for winchester *)

(* This is a very special module which must be linked as the runtime base.
   The initialisation code is never executed, instead the routine 'Boot' is
   so positioned in the code as to be started by the Stride ROM.  Do not
   alter any code in this module, or its linking strategy, without thoroughly
   understanding how the ROM works, and what LUMOS expects once the image is
   loaded. *)

(* we assume an 171 block FAT. The ROM loads the first 64 blocks. *)
(* ultimately, this should be fully general *)

FROM SYSTEM IMPORT ADDRESS, SETREG, CODE, ADR;

VAR     Directory [800H] : ARRAY [0..15] OF 
                              RECORD Name : ARRAY [0..19] OF CHAR;
                                     First : CARDINAL;
                                     Junk : ARRAY [0..9] OF CHAR;
                              END;
        LCB [800H] : RECORD  Flag, Size, Version : CARDINAL;
                             EntryPoint : PROC;
                             LoadKOld, LoadKNew : CARDINAL;
                             FFOld, FirstFree : ADDRESS;
                             Layer : CARDINAL;
                             FirstCode : ADDRESS;
                     END;
        BootDev [0] : ARRAY [0..3] OF CHAR;
        MemTop [0FF8300H] : ADDRESS;  (* where the ROM leaves the size *)

(*$T-*)
(*$S-*)
(*$P-*)

PROCEDURE  Key;

BEGIN   CODE(424FH,4F54H);  (* "BOOT" *)
END Key;
 
PROCEDURE  Boot( unit : CARDINAL);  (* this is the entry point *)

(* this routine is executing on the CMOS stack ! *)

VAR     i, j : CARDINAL;
        a, s : ADDRESS;
        f : BOOLEAN;
        Debug : PROC;

BEGIN   BootDev[0] := "#";
        BootDev[1] := "W";
        BootDev[2] := "0";
        BootDev[3] := 0C;
        Debug := PROC(0FE0010H);  (* re-enter the ROM *)
        WRead(64,400H+64*200H,110*200H);
        i := 174;  (* First block of DIRECTORY.SYS *)
        f := FALSE;
        LOOP WRead(i,800H,200H); (* Read dir block *)
             j := 0;
             LOOP WITH Directory[j] DO 
                    f := (Name[0] = "L") AND
                         (Name[1] = "U") AND
                         (Name[2] = "M") AND
                         (Name[3] = "O") AND
                         (Name[4] = "S") AND
                         (Name[5] = ".") AND
                         (Name[6] = "S") AND
                         (Name[7] = "Y") AND
                         (Name[8] = "S") AND
                         (Name[9] = 0C);
                    IF f THEN i := First END;
                    END;
                  IF f OR (j = 15) THEN EXIT ELSE INC(j) END;
                  END;
             IF f THEN EXIT END;
             i := FAT(i);
             IF i = 65535 THEN Debug END;
             END;
        WRead(i,800H,200H); (* Read the LCB *)
        i := FAT(i);
        a := LCB.FirstCode;
        s := 200H;
        j := i;
        LOOP IF FAT(i) = i+1
               THEN s := s + 200H;
                    i := FAT(i);
               ELSE WRead(j,a,s);
                    a := a + s;
                    i := FAT(i);
                    IF i = 65535 THEN EXIT END;
                    s := 200H;
                    j := i;
               END;
             END;
        SETREG(13,LCB.FirstFree);
        SETREG(15,MemTop);
        LCB.EntryPoint;
END Boot;

PROCEDURE  FAT( n : CARDINAL) : CARDINAL;

(* The FAT is an array of CARDINALS beginning at #0A00.  This function
   finds FAT[n] by computing a pointer. *)

TYPE    cp = POINTER TO CARDINAL;

VAR     a : ADDRESS;
        p : cp;

BEGIN   a := 0A00H + 2*ADDRESS(n);
        p := a;
        RETURN p^;
END FAT;

PROCEDURE  WRead( block : CARDINAL; mem, size : ADDRESS);

BEGIN   SETREG(0,ADDRESS(block));
        SETREG(1,mem);
        SETREG(2,size);
        CODE(027400B);               (* MOVE.L D0,-(A7) *)
        CODE(027401B);               (* MOVE.L D1,-(A7) *)
        CODE(027402B);               (* MOVE.L D2,-(A7) *)
        CODE(047271B,00FEH,003CH);   (* JSR    WDREAD   *)
END WRead;

BEGIN  (* never executed *)
END B2WBoot.
