IMPLEMENTATION MODULE DialProto;

(* This is the top layer - data transparency - of the MMDF dial protocol
   as implemented for LMDF.  Since we are only implementing an end node
   and plan only to talk with PMDF, many assumptions are made to simplify
   the coding.
*)     

FROM PacketDriver IMPORT PutPacket, GetPacket;
FROM LogFile IMPORT LogString, LogLn, Terminate;
FROM R2Strings IMPORT Compare;
FROM R2Format IMPORT FormatCard, ScanCard;

CONST   DATA  = 0;
        XPATH = 2;
        RPATH = 4;
        ESCAP = 6;
        omax = 112;
        imax = 120;
        path = "78FFFFFFFF000000000000000000000008";

PROCEDURE  InitProto;   (* REALLY Kludgy *)

VAR     rep : ARRAY [0..39] OF CHAR;
        eor : BOOLEAN;  

   PROCEDURE  CheckRep( VAR t : ARRAY OF CHAR);

   BEGIN   IF Compare(rep,t) <> 0 THEN
              LogString("Protocols don't agree");
              LogLn; Terminate;
              END;
   END CheckRep;

BEGIN   PutPacket(path,XPATH,FALSE);
        PutPacket(path,RPATH,FALSE);
        GetPacket(rep,XPATH,eor);  
        CheckRep(path);
        GetPacket(rep,RPATH,eor);
        CheckRep(path);
        PutPacket("7E",ESCAP,FALSE);
        GetPacket(rep,ESCAP,eor);
        CheckRep("7E");
END InitProto;

PROCEDURE  SendChar( c : CHAR);

(* Encode c and put it in the current packet.  Send off the packet if full *)

VAR     ns : ARRAY [0..1] OF CHAR;

BEGIN   IF omax-ocnt < 3 THEN Flush(FALSE) END;
        CASE c OF
            11C       : StuffChar("~"); StuffChar("I");
          | 12C       : StuffChar("~"); StuffChar("J");  (* shouldn't see *)
          | 14C       : StuffChar("~"); StuffChar("L");
          | 15C       : StuffChar("~"); StuffChar("J");  (* convert here *)
          | 176C      : StuffChar("~"); StuffChar("~");
          | 40C..175C : StuffChar(c);
          ELSE FormatCard(ORD(c),16,2,ns);
               StuffChar("~"); StuffChar(ns[0]); StuffChar(ns[1]);
          END;
END SendChar;

PROCEDURE  SendRec( VAR r : ARRAY OF CHAR);

(* Send the chars in r and terminate the record *)

VAR     i : CARDINAL;

BEGIN   i := 0;
        WHILE (i <= HIGH(r)) AND (r[i] <> 0C) DO
          SendChar(r[i]);
          INC(i);
          END;
        Flush(TRUE);
END SendRec;

PROCEDURE  RecvChar( VAR c : CHAR) : BOOLEAN;                   

(* Receive one char, return TRUE unless this is the last char in a record *)

VAR     ns : ARRAY [0..1] OF CHAR;
        e : BOOLEAN;
        n : CARDINAL;

BEGIN   e := Fetch(c);
        IF c <> "~" THEN RETURN e END;
        e := Fetch(ns[0]);
        CASE ns[0] OF
            'I','i' : c := 11C;
          | 'J','j' : c := 15C;  (* convert here *)
          | 'L','l' : c := 14C;
          | 'M','m' : c := 15C;
          | '~'     : c := '~';
          ELSE e := Fetch(ns[1]);
               IF NOT ScanCard(ns,16,n) THEN
                 LogString("Invalid escape sequence: ~");
                 LogString(ns); LogLn; Terminate;
                 END;
               c := CHR(n);
          END;
        RETURN e;
END RecvChar;

PROCEDURE  RecvRec( VAR r : ARRAY OF CHAR);

(* Receive until end of record, fill r and discard any surplus *)
                                                                
VAR     i : CARDINAL;
        c : CHAR;

BEGIN   i := 0;
        WHILE RecvChar(c) DO
          IF i <= HIGH(r) THEN
            r[i] := c;
            INC(i);
            END;
          END;
        IF i <= HIGH(r) THEN r[i] := 0C END;
END RecvRec;
             
VAR     obuf : ARRAY [0..omax] OF CHAR;
        ibuf : ARRAY [0..imax] OF CHAR;
        ocnt, icnt : CARDINAL;
        eor : BOOLEAN;

PROCEDURE  StuffChar( c : CHAR);

BEGIN   IF ocnt = omax THEN Flush(FALSE) END;
        obuf[ocnt] := c;
        INC(ocnt);
END StuffChar;

PROCEDURE  Flush( endrec : BOOLEAN);

BEGIN   obuf[ocnt] := 0C;
        PutPacket(obuf,DATA,endrec);
        ocnt := 0;
END Flush;

PROCEDURE  Fetch( VAR c : CHAR ) : BOOLEAN;

BEGIN   WHILE ibuf[icnt] = 0C DO
          IF eor 
            THEN icnt := 0;
                 ibuf[0] := 0C;
                 c := 0C;
                 eor := FALSE;
                 RETURN FALSE;
            ELSE GetPacket(ibuf,DATA,eor);
                 icnt := 0;
            END;
          END;
        c := ibuf[icnt];
        INC(icnt);
        RETURN TRUE;
END Fetch;
                      
BEGIN   ocnt := 0;
        icnt := 0;
        ibuf[0] := 0C;
        eor := FALSE;
END DialProto.
