IMPLEMENTATION MODULE MailInterface;  (* for LMDF *)

FROM R2Streams IMPORT Stream, OpenStream, ReadLine, ResetStream, EndStream,
                      WriteLine, CreateStream, Done, SysErr, CloseStream;
FROM R2Strings IMPORT Copy, Remove, Find, Insert, Compress, Concatenate;
FROM OptionFile IMPORT  SetOptionFile, GetStrOption;
FROM R2Directories IMPORT NextWild, SetWild, FullName;
FROM MailDates IMPORT MakeNewStamp, MakeDate;
FROM R2SysCalls IMPORT Status;
FROM LockSync IMPORT LockWait, LockClear;

VAR     Mailer, Remote, Local : ARRAY [0..79] OF CHAR;

PROCEDURE  NextMessage( VAR msg : Stream;
                        VAR to, from : ARRAY OF CHAR) : BOOLEAN;

(* Each time this is called, the next file addressed to the Mailer
   is opened on msg and parsed for from and to.  Returns false if no such
   file.  msg is opened for writing so that ReleaseStream can be used to
   discard the file when finished with it.
*)        

VAR     ln : ARRAY [0..133] OF CHAR;
        j, k : CARDINAL;
        rs : Status;

BEGIN   rs := NextWild();
        IF rs <> Ok THEN RETURN FALSE END;
        OpenStream(msg,FullName,TRUE);
        to[0] := 0C;
        from[0] := 0C;
        LOOP IF EndStream(msg) THEN EXIT END;
             ReadLine(msg,ln);
             IF ln[0] = 0C THEN EXIT END;
             IF Find("From:",ln) = 0 THEN
               Remove(ln,0,5);
               Compress(ln);
               Copy(ln,from);
               END;
             IF Find("To:",ln) = 0 THEN
               Remove(ln,0,3);
               k := FindCh("(",ln);
               j := FindCh(")",ln);
               IF k < j THEN Remove(ln,k,j-k+1); END;
               k := FindCh("<",ln);
               j := FindCh(">",ln);
               IF k < j THEN
                 Remove(ln,j,133);
                 Remove(ln,0,k+1);
                 END;
               Compress(ln);
               Copy(ln,to);
               END;
             END;
        ResetStream(msg);
        RETURN TRUE;
END NextMessage;                            

PROCEDURE  Deliver( VAR msg : Stream; VAR to : ARRAY OF CHAR) : Status;

(* Create the file to hold a message for 'to' and stamp it with a 
   Received: line.  The caller writes the text onto msg, and then
   closes it.
*)

VAR     s, t : ARRAY [0..133] OF CHAR;
        r : Status;

BEGIN   r := MakeMsg(msg,to);
        IF r <> Ok THEN RETURN r END;
        s := "Received: from ";
        Concatenate(s,Remote,s);
        Concatenate(s,"; ",s);
        MakeDate(t);
        Concatenate(s,t,s);
        WriteLine(msg,s);
        RETURN Ok;
END Deliver;

PROCEDURE  Return( VAR msg : Stream; VAR to, why : ARRAY OF CHAR) : Status;

(* Create a Return-to-sender envelope for 'to' - as above *)

VAR     s, t : ARRAY [0..133] OF CHAR;
        r : Status;

BEGIN   r := MakeMsg(msg,to);
        IF r <> Ok THEN RETURN r END;
        Concatenate("To:      ",to,s);   WriteLine(msg,s);
        Concatenate("From:    ",Mailer,s);   WriteLine(msg,s);
        MakeDate(t);
        Concatenate("Date:    ",t,s);   WriteLine(msg,s);
        WriteLine(msg,"Subject: Undeliverable mail");
        WriteLine(msg,"");
        s := "The following message was not accepted by the mail system at ";
        Concatenate(s,Remote,t);
        WriteLine(msg,t);
        s := "for the following reason: ";
        Concatenate(s,why,t);
        WriteLine(msg,t);
        WriteLine(msg,"");
        RETURN Ok;
END Return;

PROCEDURE  MakeMsg( VAR msg : Stream; VAR to : ARRAY OF CHAR) : Status;

(* Create a message file for 'to' after checking that it is a valid user *)

CONST   pwlck = "passwords.lock";

VAR     ps : Stream;
        s, t : ARRAY [0..79] OF CHAR;
        k : CARDINAL;

BEGIN   Copy(to,s);
        k := FindCh("@",s);                 (* elim @host  *)
        Remove(s,k,79);        
        IF NOT LockWait(pwlck,30) THEN RETURN TimedOut END;
        OpenStream(ps,"$SysDisk-PASSWORDS.SYS",FALSE);
        IF NOT Done THEN RETURN SysErr END;
        LOOP IF EndStream(ps) THEN t[0] := 0C; EXIT END;
             ReadLine(ps,t);
             k := FindCh(" ",t);
             Remove(t,k,79);
             k := 0;
             LOOP IF t[k] = 0C THEN EXIT END;
                  IF CAP(t[k]) <> CAP(s[k]) THEN EXIT END;
                  INC(k);
                  END;
             IF t[k] = 0C THEN EXIT END;
             END;
        CloseStream(ps);
        LockClear(pwlck);
        IF t[0] = 0C THEN RETURN NoSuch END;
        MakeNewStamp(t,s);
        Concatenate("$MAIL-",s,t);
        CreateStream(msg,t);
        IF NOT Done THEN RETURN SysErr END;
        RETURN Ok;
END MakeMsg;

PROCEDURE  FindCh( c : CHAR; VAR s : ARRAY OF CHAR) : CARDINAL;

VAR     k : CARDINAL;

BEGIN   k := 0;
        WHILE (k <= HIGH(s)) AND (s[k] <> 0C) AND (s[k] <> c) DO INC(k) END;
        RETURN k;
END FindCh;                          

VAR     b : BOOLEAN;
        tmp : ARRAY [0..79] OF CHAR;
        rs : Status;

BEGIN   b := SetOptionFile("$MAIL-Config");
        b := GetStrOption("Mailer",Mailer);
        b := GetStrOption("Remote",Remote);
        b := GetStrOption("Local",Local);
        tmp := "$MAIL-N.*";
        Insert(Mailer,tmp,7);
        rs := SetWild(tmp);
END MailInterface.
