IMPLEMENTATION MODULE MailParse;  (* KGRW 1987 *)

FROM SYSTEM IMPORT ADDRESS;
FROM R2SysCalls IMPORT Status;
FROM R2Storage IMPORT ALLOCATE, DEALLOCATE;
FROM R2Strings IMPORT Length;
CONST
   MATCHALL = '*';
   MATCHONE = '?';
   MATCHOPT = '~';
TYPE
   CMDList = POINTER TO CMD;
   CMD = RECORD
            Pat : ARRAY [0..9] OF CHAR;
            Proc : PROC;
            Nxt : CMDList;
            END;

PROCEDURE Exec( VAR cmd : ARRAY OF CHAR; list : CMDList ) : BOOLEAN;
   (* Returns FALSE if the command was not found *)
VAR
   p : CMDList;
BEGIN
   IF list=NIL THEN RETURN FALSE END; (* no commands to execute *)
   IF (ADDRESS(list^.Proc)<>NIL) AND (cmd[0]=0C) THEN
      list^.Proc;
      RETURN TRUE;
      END;
   p := list^.Nxt;
   WHILE p<>NIL DO
      WITH p^ DO
         IF Match( cmd, Pat ) THEN Proc; RETURN TRUE; END;
         p := Nxt;
         END;
      END;
   RETURN FALSE;
END Exec;

PROCEDURE UnSetDefault( list : CMDList );
BEGIN
   IF list<>NIL THEN list^.Proc := PROC(NIL) END;
END UnSetDefault;

PROCEDURE SetDefault( p : PROC; VAR list : CMDList ) : Status;
BEGIN
   IF list=NIL THEN
      NEW(list); IF list=NIL THEN RETURN NoHeap END;
      WITH list^ DO
         Nxt := NIL;
         Pat[0] := 0C;
         END;
      END;
      list^.Proc := p;
   RETURN Ok;
END SetDefault;

PROCEDURE AddCMD( VAR pat : ARRAY OF CHAR; p : PROC; VAR list : CMDList
                 ) : Status;
VAR
   n : CMDList;
BEGIN
   IF list=NIL THEN
      NEW(list); IF list=NIL THEN RETURN NoHeap END;
      WITH list^ DO
         Pat[0] := 0C;
         Proc := PROC(NIL);
         Nxt := NIL;
         END;
      END;
   NEW(n); IF n=NIL THEN RETURN NoHeap END;
   WITH n^ DO
      UCopy( pat, Pat );
      Proc := p;
      Nxt := list^.Nxt;
      list^.Nxt := n;
      END;
   RETURN Ok;
END AddCMD;

PROCEDURE Match( VAR cmd, pat : ARRAY OF CHAR ) : BOOLEAN;
VAR
   ic, ip, lc, lp, nc, np : CARDINAL;
   cc, cp : CHAR;
   st, op : BOOLEAN;
BEGIN
   lc := Length(cmd);
   lp := Length(pat);
   st := FALSE; op := FALSE;
   ip := 0; ic := 0;
   LOOP
      IF ip >= lp THEN cp := 0C ELSE cp := CAP(pat[ip]) END;
      IF cp = MATCHOPT THEN
         op := TRUE; st := FALSE; INC(ip);
         IF ip >= lp THEN cp := 0C ELSE cp := CAP(pat[ip]) END;
         END;
      IF ic >= lc THEN cc := 0C ELSE cc := CAP(cmd[ic]) END;
      IF (cc=0C) AND (op OR (cp=0C)) THEN RETURN TRUE END;
      IF cc=cp THEN
         INC(ic); INC(ip);
      ELSIF (cp=MATCHONE) AND (cc<>0C) THEN
         INC(ip); INC(ic);
      ELSIF cp=MATCHALL THEN
         st := TRUE; INC(ip);
         IF (ip<lp) AND (pat[ip]=MATCHOPT) THEN lp := ip END;
         np := ip; nc := ic;
      ELSIF st THEN
         INC(nc); IF nc>lc THEN RETURN FALSE END;
         ic := nc; ip := np;
      ELSE RETURN FALSE END;
      END;
END Match;

PROCEDURE Component( VAR s, c : ARRAY OF CHAR );
VAR
   i,j : CARDINAL;
BEGIN
   i := 0; j := 0;
   WHILE (i<HIGH(s)) AND (s[i]=' ') DO INC(i) END;
   WHILE (i<HIGH(s)) AND (j<HIGH(c)) AND (s[i]>' ') DO
      c[j] := s[i];
      INC(i); INC(j);
      END;
   WHILE (i<HIGH(s)) AND (s[i]>' ') DO INC(i) END;
   WHILE (i<HIGH(s)) AND (s[i]=' ') DO INC(i) END;
   IF (j<HIGH(c)) THEN c[j] := 0C END;
   j := 0;
   WHILE (i<HIGH(s)) AND (s[i]>=' ') DO
      s[j] := s[i];
      INC(i); INC(j);
      END;
   s[j] := 0C;
END Component;

PROCEDURE UCopy( VAR s, t : ARRAY OF CHAR );
VAR
   i : CARDINAL;
BEGIN
   i := 0;
   WHILE (i<=HIGH(s)) AND (i<=HIGH(t)) AND (s[i]<>0C) DO
      t[i] := s[i];
      INC(i);
      END;
   IF (i<=HIGH(t)) THEN t[i] := 0C END;
END UCopy;

PROCEDURE UCompare( VAR s, t : ARRAY OF CHAR ) : INTEGER;
VAR
   i : CARDINAL;
BEGIN
   i := 0;
   WHILE (i<HIGH(s)) AND (i<HIGH(t)) AND (CAP(s[i])=CAP(t[i])) AND
         (s[i]<>0C) DO INC(i) END;
   IF ((i=HIGH(s)) OR (i=HIGH(t))) AND (CAP(s[i])=CAP(t[i])) THEN
      RETURN Length(s)-Length(t) END;
   RETURN ORD(CAP(s[i]))-ORD(CAP(t[i]));
END UCompare;

BEGIN
   MatchAll[0] := MATCHALL; MatchAll[1] := 0C;
   MatchOne[0] := MATCHONE; MatchOne[1] := 0C;
   MatchOpt[0] := MATCHOPT; MatchOpt[1] := 0C;
END MailParse.
