IMPLEMENTATION MODULE PacketDriver;  (* LMDF packet routines *)

FROM R2SysCalls IMPORT WriteLn, WriteString, Status;
FROM LogFile IMPORT LogLn, LogString, LogCard, Terminate;
FROM TimedRead IMPORT ReadPacket;
FROM R2Strings IMPORT Substring;

CONST   tmo = 1000;  (* 10 seconds *)
        maxtries = 10;

VAR     exp, seq : CARDINAL;
        hexdigs : ARRAY [0..15] OF CHAR;
        
PROCEDURE  PutPacket( VAR p : ARRAY OF CHAR; type : CARDINAL;
                      eof : BOOLEAN);

VAR     header : ARRAY [0..5] OF CHAR;
        reply : ARRAY [0..133] OF CHAR;
        tries, c : CARDINAL;
        resend : BOOLEAN;
        
BEGIN   c := Sum(p,0);
        MakeHeader(eof,seq,type,c,header);
        tries := 0; resend := TRUE;
        LOOP IF tries = maxtries THEN
               LogString("Too many retries in PutPacket"); 
               LogLn; Terminate; 
               END;
             IF resend THEN
               WriteString(header); WriteString(p); 
               WriteLn; INC(tries);
               END;
             IF Expect(reply,type+1,seq,resend) THEN EXIT END;
             END;
        seq := (seq + 1) MOD 4;
END PutPacket;
  
PROCEDURE  GetPacket( VAR p : ARRAY OF CHAR; type : CARDINAL; 
                      VAR eof : BOOLEAN);

VAR     msg : ARRAY [0..133] OF CHAR;
        hdr : ARRAY [0..5] OF CHAR;
        dummy : BOOLEAN;
        tries : CARDINAL;

BEGIN   tries := 0;
        LOOP IF tries = maxtries THEN 
               LogString("Too many retries in GetPacket"); 
               LogLn; Terminate; 
               END;
             INC(tries);
             IF Expect(msg,type,exp,dummy) THEN EXIT END;
             END;
        MakeHeader(FALSE,exp,type+1,0,hdr);
        WriteString(hdr); WriteLn;
        Substring(msg,p,6,133);
        eof := ODD(DeHex(msg[5]) DIV 4);
        exp := (exp + 1) MOD 4;
END GetPacket;

PROCEDURE  Expect( VAR p : ARRAY OF CHAR; tp, sq : CARDINAL;
                   VAR rsn : BOOLEAN ) : BOOLEAN;

VAR     xhdr : ARRAY [0..5] OF CHAR;
        dr4, dr5 : CARDINAL;
        r : Status;

BEGIN   r := ReadPacket(p,tmo);
        rsn := TRUE;
        dr4 := DeHex(p[4]);
        dr5 := DeHex(p[5]);
        IF r <> Ok
          THEN LogString("Error reading packet"); LogLn;
        ELSIF p[0] = 0C
          THEN LogString("Empty packet received"); LogLn; rsn := FALSE;
        ELSIF NOT Check(p) 
          THEN LogString("Checksum error in: "); LogString(p); LogLn;
        ELSIF (dr5 DIV 8) <> 1
          THEN LogString("Origin bit error in: "); LogString(p); 
               LogLn; rsn := FALSE;
        ELSIF (dr5 MOD 4) <> sq
          THEN IF ODD(dr4) THEN
                 LogString("Out of sequence acknowledgement ignored: ");
                 LogString(p); LogLn; rsn := FALSE;
               ELSIF (dr5 MOD 4) <> exp THEN
                 MakeHeader(FALSE,dr5 MOD 4,dr4+1,0,xhdr);
                 WriteString(xhdr); WriteLn;
                 LogString("Out of sequence packet acknowledged: ");
                 LogString(p); LogLn;
               ELSE 
                 LogString("Premature data packet ignored: ");
                 LogString(p); LogLn;
                 END;
        ELSIF dr4 <> tp
          THEN MakeHeader(FALSE,dr5 MOD 4,dr4+1,0,xhdr);
               WriteString(xhdr); WriteLn;
               LogString("Invalid type packet acknowledged: ");
               LogString(p); LogLn;
          ELSE RETURN TRUE
          END;
        RETURN FALSE;
END Expect;

PROCEDURE  MakeHeader( ef : BOOLEAN; sq, tp, cs : CARDINAL; 
                          VAR h : ARRAY OF CHAR );

VAR     i, k : CARDINAL;

BEGIN   IF ef THEN k := 4 + sq ELSE k := sq END;
        h[5] := hexdigs[k];
        h[4] := hexdigs[tp];
        k := Sum(h,4) + cs;
        FOR i := 0 TO 3 DO
          h[3-i] := hexdigs[k MOD 16];
          k := k DIV 16;
          END;
END MakeHeader;

PROCEDURE  Check( VAR p : ARRAY OF CHAR) : BOOLEAN;

VAR     i, x, c : CARDINAL;

BEGIN   c := Sum(p,4);
        x := 0;
        FOR i := 0 TO 3 DO x := x*16 + DeHex(p[i]) END;
        RETURN x=c;
END Check;

PROCEDURE  Sum( VAR a : ARRAY OF CHAR; start : CARDINAL) : CARDINAL;

VAR     i, k : CARDINAL;

BEGIN   k := 0;
        i := start;
        WHILE (i <= HIGH(a)) AND (a[i] <> 0C) DO
          INC(k,ORD(a[i]));
          INC(i);
          END;
        RETURN k;
END Sum;

PROCEDURE  DeHex( c : CHAR) : CARDINAL;

BEGIN   IF c >  'F' THEN RETURN 0 END;
        IF c >= 'A' THEN RETURN ORD(c) - 55 END;
        IF c >  '9' THEN RETURN 0 END;
        IF c >= '0' THEN RETURN ORD(c) - 48 END;
        RETURN 0;
END DeHex;

BEGIN   hexdigs := "0123456789ABCDEF";
        seq := 0;
        exp := 0;
END PacketDriver.
