IMPLEMENTATION MODULE DNS;
FROM R2Strings    IMPORT Length,Copy,Insert,Remove;
FROM R2NewerCalls IMPORT SetUDPPort,Status,Send,Port,IPAddr,Receive,
                         DNSInfo;
FROM R2NetSubs    IMPORT ScanIPAddr, FormatIPAddr, WriteIPAddr;
FROM R2SysCalls   IMPORT WriteString,WriteLn,Write;
FROM R2NewCalls   IMPORT UpTime;
FROM R2NumbIO     IMPORT WriteCard;
FROM SYSTEM       IMPORT ADDRESS;

(*Note: If the caller already has a port open this call will close it and
        open a new one. Only one port per process is allowed*)

CONST
  Idoffset=0;                    (*Offsets in a D. Server Format*)
  Qnumoffset=4;
  ParmHoffset=2;
  ParmLoffset=3;
  Qsectoff=12;
  NumAoff=6;
  NumAuth=8;
  NumAdd=10;
  forport=53;                    (*Dns requests are received on port 53*)
  tmo=400;                       (*waits 4 seconds for a reply*)

VAR
  Server1IP,Server2IP,netmask:IPAddr;
  ps:RESULT;
  ds:BOOLEAN;
  ident:ADDRESS;
  value,tident:CARDINAL;
  Resp:ARRAY[0..511] OF CHAR;  (*maximum size for a UDP message*)
  OurDname,resolver1,resolver2:ARRAY[0..50] OF CHAR;

(*This procedure returns a message in the string given to it according to
  the result encountered*)

PROCEDURE GetMessage(VAR str:ARRAY OF CHAR;rs:RESULT);
BEGIN
  CASE rs OF
   BadDomain: Copy('Bad domain',str); |
   BadQuery: Copy('Bad query sent',str); |
   BadIdent: Copy('Bad identification',str); |
   NotAResponse: Copy('Packet received was not a response',str); |
   ServerFailure: Copy('Server failure',str); |
   NoError: Copy('No Error',str);
 ELSE
   Copy('Failed',str);
 END(*CASE*);
END GetMessage;
                         

(*This procedure makes sure the domain name given to resolve is a character 
  string and not an IP address*)

PROCEDURE IsAddr(dname:ARRAY OF CHAR;len:CARDINAL):BOOLEAN;
VAR
  i:CARDINAL;
BEGIN
  FOR i:=0 TO len-1 DO
    IF ((ORD(dname[i])>=60) AND (ORD(dname[i])<=71)) THEN
       RETURN TRUE;               (*A digit is present*)
    END(*IF*);
  END(*FOR*);
  RETURN FALSE;
END IsAddr;

(*This procedure fills in the constant parts of a query message so that the
  information only has to be filled in once*)

PROCEDURE MakeCQuery(VAR Request:ARRAY OF CHAR);
BEGIN                 
  Request[ParmLoffset]:=0C;        (*sets Response bits*)
  Request[ParmHoffset]:=1C;        (*sets recursion desired bit*)
  Request[Qnumoffset]:=0C;
  Request[Qnumoffset+1]:=1C;       (*sets # of questions field to 1*)
  Request[NumAoff]:=0C;            (*sets # of answers to 0*)
  Request[NumAoff+1]:=0C;  
  Request[NumAuth]:=0C;            (*sets # of authoritive to 0*)
  Request[NumAuth+1]:=0C;
  Request[NumAdd]:=0C;             (*sets # of additional to 0*)
  Request[NumAdd+1]:=0C;
END MakeCQuery;                            

(*This procedure fills in the Query domain name into the packet*)

PROCEDURE InsDname(temp:ARRAY OF CHAR;VAR Request:ARRAY OF CHAR):CARDINAL;
VAR
  i,count,numloc,QToffset,high:CARDINAL;
BEGIN
  i:=0;
  count:=0;                  
  high:=HIGH(temp);
  numloc:=0;                              
  UpTime(ident);                              (*gets clock time in ticks*)
  tident:=CARDINAL(ident MOD 10000H);                 
  Request[Idoffset]:=CHR(tident DIV 100H);    (*sets ident. field*)
  Request[Idoffset+1]:=CHR(tident MOD 100H);
  WHILE (i<=high) AND (temp[i]<>0C) DO
    IF temp[i]='.' THEN
      Request[Qsectoff+numloc]:=CHR(count MOD 100H); (*fits length in 1 byte*)
      numloc:=numloc+count+1;
      count:=0;
    ELSE
      INC(count);
      Request[Qsectoff+numloc+count]:=temp[i];
    END(*IF*);          
    INC(i);
  END(*WHILE*);
  Request[Qsectoff+numloc]:=CHR(count MOD 100H); (*places last length in*)
  Request[Qsectoff+numloc+count+1]:=0C;          (*ends DName with 0C*)
  QToffset:=Qsectoff+numloc+count+2;
  Request[QToffset]:=0C;                (*Inserts query type ='A'*)
  Request[QToffset+1]:=1C;
  Request[QToffset+2]:=0C;              (*Inserts query class ='IN'*)
  Request[QToffset+3]:=1C;
  RETURN QToffset+4;                    (*returns length of Request*)
END InsDname;
                                                                         

(*This makes the domain name less specific when it is not found locally*)

PROCEDURE Parse(VAR temp:ARRAY OF CHAR);                              
VAR
  len,i:CARDINAL;
BEGIN
  len:=Length(temp); 
  i:=1;
  WHILE (temp[i]<>'.') AND (len>i) DO
    INC(i);
  END(*WHILE*);
  Remove(temp,0,i);
END Parse;
               
(*This procedure checks if the response message returned is okay to
  process or has some error*)

PROCEDURE CHECK(Resp:ARRAY OF CHAR):RESULT;                             
VAR
  RespValue:CARDINAL;
BEGIN
  value:=ORD(Resp[0])*100H+ORD(Resp[1]);
  IF value<>tident THEN                      (*checks if ident is zero*)
     RETURN BadIdent;
  ELSIF ORD(Resp[2]) DIV 80H=0 THEN           (*checks if it is a response*)
     RETURN NotAResponse;
  ELSE
     RespValue:=ORD(Resp[3]) MOD 10H;         (*checks for other errors*)
     CASE RespValue OF
       1: RETURN BadQuery; |
       2: RETURN ServerFailure; |
       3: RETURN BadDomain; 
     ELSE
       RETURN NoError;
     END(*CASE*);
  END(*IF*);
END CHECK;

(*This procedure extracts the relevant information from the response
  packet to our DNS query*)

PROCEDURE ExtractInfo(VAR Ipnum:IPAddr;VAR dname:ARRAY OF CHAR;
                                   Resp:ARRAY OF CHAR):RESULT;
VAR
  Ansoffset,RdLoffset,Dlength,Rdoffset,Qtype,RDLength:CARDINAL;
  pos,i,loc,j,Dnpos,count,RespHigh,DHigh,numAns,pointer:CARDINAL;
  ptrflag:BOOLEAN;
BEGIN           
  RespHigh:=HIGH(Resp); 
  ptrflag:=FALSE;
  pos:=Qsectoff;    
  WHILE (pos<=RespHigh) AND (Resp[pos]<>0C) DO     (*skips domain name*)
    INC(pos);
  END(*WHILE*);
  Ansoffset:=pos+5;                                          
   
(*Look for an answer with an 'A' type Query int the answer section.
                           ie with an IP address*)

  FOR i:=1 TO numAns DO    
    j:=Ansoffset;
    pointer:=CARDINAL(Resp[Ansoffset]) DIV 40H;
    IF pointer=3 THEN 
      j:=j+2;                    (*skipping the pointer field for DName*)
    ELSE
      WHILE Resp[j]<>0C DO       (*skips actual Domain name*)
        INC(j);
      END(*WHILE*); 
      INC(j);
    END(*IF*);
    Qtype:=CARDINAL(Resp[j])*100H+CARDINAL(Resp[j+1]);
    IF Qtype<>1 THEN                 (*Not a Query type A Answer*)
       RDLength:=CARDINAL(Resp[j+8])*100H+
                                   CARDINAL(Resp[j+9]);
       Ansoffset:=j+10+RDLength;      (*skips to next answer*)
       IF i=numAns THEN
          RETURN NoAnswer;
       END(*IF*);
    ELSE
       i:=numAns+1;                   (*finishes the loop*)
    END(*IF*);
  END(*FOR*);

  i:=0;loc:=0;     
  Dnpos:=0;                           

(*checks for a pointer. ie first 2 bits are ones*)

  IF (RespHigh>Ansoffset+1) AND ((ORD(Resp[Ansoffset]) DIV 40H) = 3) THEN
    Dnpos:= (ORD(Resp[Ansoffset]) MOD 40H)*100H + ORD (Resp[Ansoffset+1]);
    count:=ORD(Resp[Dnpos]);          (*next 14 bits is an integer ptr*)
    ptrflag:=TRUE;
  ELSE                    
    Dnpos:=Ansoffset;                          (*No pointer in record field*)
    count:= ORD(Resp[Ansoffset]) MOD 40H;
  END(*IF*);    
  DHigh:=HIGH(dname);
  WHILE (RespHigh>Dnpos+i) AND (Resp[Dnpos+i]<>0C) DO
    INC(i);
    FOR j:=0 TO count-1 DO            (*copies each part of DName into string*)
      IF DHigh>loc THEN
        dname[loc]:=Resp[Dnpos+i];
        INC(loc);
      END(*IF*);
      INC(i);
    END(*FOR*);            
    count:=ORD(Resp[Dnpos+i]);
    dname[loc]:='.';
    INC(loc);
  END(*WHILE*);     
  dname[loc-1]:=0C;                               
  IF ptrflag THEN                     (*obtains location of res. data length*)
    RdLoffset:=Ansoffset+10;                       
  ELSE 
    RdLoffset:=Ansoffset+i+9;
  END(*IF*);
  Dlength:=ORD(Resp[RdLoffset])*100H + ORD(Resp[RdLoffset+1]); 
  Rdoffset:=RdLoffset+2;
  IF RdLoffset+3<=RespHigh THEN
    Ipnum:=ADDRESS(Resp[Rdoffset])*1000000H+ADDRESS(Resp[Rdoffset+1])*10000H+
        ADDRESS(Resp[Rdoffset+2])*100H+ADDRESS(Resp[Rdoffset+3]);
  END(*IF*);
  RETURN NoError;
END ExtractInfo;


(*This procedure tries to obtain the IP number for a domain name by
  generating a query packet and sending it onto the network*)

PROCEDURE Resolve(VAR dname:ARRAY OF CHAR;VAR IPnum:IPAddr;
                  VAR Resolver:ARRAY OF CHAR;VAR ResolverIP:IPAddr):RESULT;
VAR
  rs:Status;
  ls:RESULT;
  SenderPort:Port;
  SenderIP:IPAddr;
  dlen,len,i,j:CARDINAL;
  valid:RESULT;
  temp,dstr:ARRAY[0..254] OF CHAR;      (*maximum domain name size*)
  Request:ARRAY[0..511] OF CHAR;        (*maximum UDP message size*)
  QFlag,FinFlag:BOOLEAN;
BEGIN       
   QFlag:=FALSE;                        (*Will indicate if fully qualified*)
   dlen:=Length(dname);
   Copy(resolver1,Resolver);
   ResolverIP:=Server1IP;
   IF IsAddr(dname,dlen) THEN
     RETURN BadDomain;
   END(*IF*);
   IF dname[dlen-1]='.' THEN            (*fully qualified name*)
      QFlag:=TRUE;
      dname[dlen-1]:=0C;
   END(*IF*);
   rs:=SetUDPPort(1005);           (*If not user 0, port # is ignored*)
   MakeCQuery(Request);            (*makes constant part of query*)
   FOR i:=0 TO 1 DO                (*Tries two DNS servers*)
     FinFlag:=FALSE;
     Copy(OurDname,temp);
     WHILE NOT FinFlag DO          (*loops most specific DN->least specific*)
       Copy(dname,dstr);
       IF temp[0]=0C THEN
          FinFlag:=TRUE;
       END(*IF*);
       IF QFlag THEN       (*It is fully qualified*)
         len:=InsDname(dstr,Request);
         FinFlag:=TRUE;
       ELSE                   
         Insert(temp,dstr,dlen);        (*concatenates DN onto string*)
         len:=InsDname(dstr,Request);   (*inserts domain name in Query*)
       END(*IF*);
       IF i=0 THEN
         rs:=Send(Request,len,forport,Server1IP);
       ELSE
         rs:=Send(Request,len,forport,Server2IP);
       END(*IF*);
       IF rs=Ok THEN
         rs:=Receive(Resp,len,tmo,SenderPort,SenderIP);
         IF rs=Ok THEN
           ls:=CHECK(Resp);       (*checks ident field,response field & type*)
           IF ls=NoError THEN
             ps:=ExtractInfo(IPnum,dname,Resp);     (*Gets the IP number*)
             RETURN ps;
           END(*IF*);
         END(*IF*);
       END(*IF*);
       Parse(temp);                     (*makes D. Name less specifec*)
     END(*WHILE*);        
     IF i=2 THEN
       Copy(resolver2,Resolver);
       ResolverIP:=Server2IP;
     END(*IF*);
   END(*FOR*);
   IF rs<>Ok THEN
     RETURN Failed;
   ELSE
     RETURN ls;
   END(*IF*);
END Resolve;

VAR
  hostIP,gateIP:ARRAY[0..3] OF CHAR;

BEGIN
(*This procedure call obtains all the required information for DNS from
  the operating system*)

  DNSInfo(hostIP,gateIP,OurDname,resolver1,resolver2,Server1IP,Server2IP,
             netmask);
END DNS.
