MODULE FServer;

  (* Finger - Server Implementation *)
     
                  IMPORT R2NewerCalls, R2SysCalls;
FROM R2SysCalls   IMPORT GetUser, SetCCMode, Sleep;
FROM R2Strings    IMPORT Length;
FROM R2Format     IMPORT FormatCard;
FROM R2NewCalls   IMPORT GetName;
FROM R2NewerCalls IMPORT TCPListen, TCPClose, TCPRead, TCPWrite, SocHandle,
                         Status;

CONST
  CR         = 15C;
  LF         = 12C;
  FingerPort = 79;
  FirstProc  = 7;

TYPE               (* all information is written to buffer and then we  *)
  Buff  = RECORD   (* just send the buffer .                            *)
            data : ARRAY [0..2047] OF CHAR;
            pos  : CARDINAL;
          END;
VAR          
  buff  : Buff;
  soc   : SocHandle;
  l2res : R2SysCalls.Status;
  n2res : R2NewerCalls.Status;
    
PROCEDURE PutChar( c : CHAR );                                           
  (* place single character into buffer *)
  BEGIN   
    WITH buff DO
      data[pos] := c;
      INC(pos);   
    END; (* WITH *)
  END PutChar;       
               
PROCEDURE PutCRLF;                       
  (* place CR, LF into buffer *)
  BEGIN
    PutChar(CR);
    PutChar(LF);
  END PutCRLF;

PROCEDURE PutStr( VAR s : ARRAY OF CHAR );    
  (* place ascii string into buffer *)
  VAR
    i, len : CARDINAL;
  BEGIN   
    len := Length(s);
    IF len > 0 THEN
      FOR i := 0 TO (len - 1) DO
        PutChar( s[i] );
      END; (* FOR *)
    END; (* IF *)
  END PutStr;

PROCEDURE PutCard( n : CARDINAL );   
  (* convert CARDINAL n into string and place into buffer *)
  VAR
    s : ARRAY [0..2] OF CHAR;
  BEGIN      
    FormatCard( n, 10, 2, s );
    PutChar( s[0] );
    PutChar( s[1] );
  END PutCard;
                 
PROCEDURE PutName( c : CARDINAL );                         
  (* place user name of process 'c' into buffer *)
  VAR
    l2res : R2SysCalls.Status;
    s     : ARRAY [0..19] OF CHAR;
  BEGIN
    PutStr('   ');
    l2res := GetName( c, s );
    IF l2res = R2SysCalls.Ok THEN PutStr( s ); END; 
  END PutName;

PROCEDURE PutUser( c : CARDINAL );               
  (* place user 'c' into buffer *)
  BEGIN                          
    PutStr(' user ');
    PutCard( c );
  END PutUser;

PROCEDURE PutProcess( c : CARDINAL );
  (* place process 'c' into buffer *)
  BEGIN   
    PutStr('Process ');
    PutCard(c);
  END  PutProcess;               

PROCEDURE BuffJobs;                 
  (* buffer information for all active jobs *)
  VAR
    p, u : CARDINAL;
  BEGIN                                      
    buff.pos := 0;
    FOR p := FirstProc TO 40 DO 
      IF GetUser( p, u ) = R2SysCalls.Ok THEN
        PutProcess( p );
        PutUser( u );
        PutName( p );
        PutCRLF;
      END; (* IF *)
    END; (* FOR *)
  END BuffJobs;              

VAR
  proc : CARDINAL;
BEGIN
  SetCCMode( FALSE );
  LOOP                  (* serve forever ! *)
    LOOP
      n2res := TCPListen( FingerPort, soc );
      IF n2res = Ok THEN EXIT; END;              
      l2res := Sleep( 1500 );
    END; (* LOOP *)
    n2res := TCPRead( soc, buff.data, buff.pos, FALSE );
    BuffJobs;
    n2res := TCPWrite( soc, buff.data, buff.pos );
    n2res := TCPClose( soc );
  END; (* LOOP *)
END FServer.
