MODULE PING;
        
FROM CommComp     IMPORT Component;
FROM DNS          IMPORT RESULT, Resolve;
FROM R2Format     IMPORT ScanCard;
FROM R2NetSubs    IMPORT ReadIPAddr, WriteIPAddr, ScanIPAddr;
FROM R2NewerCalls IMPORT IPAddr, Ticks, Status, Ping;
FROM R2NumbIO     IMPORT WriteCard, WriteAddr;            
FROM R2SysCalls   IMPORT GetCommand, Sleep, WriteString, WriteLn;
FROM SYSTEM       IMPORT ADDRESS;
                  IMPORT R2SysCalls;
                                   
CONST
  DefTmo      = 500;   (* Default time out *)
  DefDataSize = 54;    (* Default data size *)
  DefPkts     = 1;     (* Default # of pings *)

PROCEDURE Error( s : ARRAY OF CHAR );
  BEGIN
    WriteLn;
    WriteString(s); 
    WriteLn;
  END Error;
                     
PROCEDURE DoPing( daddr : IPAddr; numdata : CARDINAL );
 (* Perform a singular ping of host specified by daddr *)
  VAR
    nres : Status;
    rtt  : Ticks;
  BEGIN
    WriteLn;
    WriteString(' PING : [');
    WriteIPAddr( daddr, 1 );
    WriteString('] '); 
    WriteCard(numdata,1);    
    WriteString(' data bytes. ');
    nres := Ping( daddr, DefTmo, numdata, rtt );
    WriteLn; 
    WriteString(' '); 
    WriteIPAddr( daddr, 1 );
    IF nres = Ok THEN    
      WriteString(' is alive!');
    ELSIF nres <> TimedOut THEN
      WriteString(' ...system error');
    ELSE
      WriteString(' did not respond (Timed Out).');
    END; (* IF *)   
    WriteLn;
  END DoPing;

PROCEDURE DoPings( daddr : IPAddr; numdata, numpkts : CARDINAL );
 (* Does multiple pings and prints out detailed results such as success 
   ratio and avg. rtt. *)
  VAR
    nres       : Status;
    lres       : R2SysCalls.Status;
    total, rtt : Ticks;
    i, good    : CARDINAL;
  BEGIN               
    total := 0;
    good := 0;
    WriteLn;
    WriteString('  PING : [');
    WriteIPAddr( daddr, 1 );
    WriteString('] '); 
    WriteCard(numdata,1);    
    WriteString(' data bytes, ');
    WriteCard( numpkts, 1 );
    WriteString(' times, once per second.');
    WriteLn; WriteLn; 
    FOR i := 1 TO numpkts DO
      IF numpkts > 1 THEN
        WriteString('   sent PING # '); 
        WriteCard( i, 1 );
      END; (* IF *)
      nres := Ping( daddr, DefTmo, numdata, rtt );

    (* report results of individual pings *)

      IF nres = Ok THEN                          
        WriteString(',  PING receipt # '); 
        WriteCard( i, 2 );
        WriteString(',  rtt : ');
        WriteAddr(rtt,10,3);
        WriteString('/100 seconds'); 
        INC(good);
        total := total + rtt;
        lres := Sleep(100);
      ELSIF nres = TimedOut THEN
        WriteString(' ...timed out');
      ELSE
        WriteString(' ...system error');
      END; (* IF *)
      WriteLn;
    END; (* FOR *) 

   (* report cumalitive results and averages *)
  
    WriteString('  PING Statistics'); WriteLn;
    IF good <> 0 THEN
      WriteString('   Sent       : '); WriteCard( i, 1 ); WriteLn;
      WriteString('   Received   : '); WriteCard( good, 1 ); WriteLn;
      WriteString('   Success    : '); 
      WriteCard(CARDINAL(ADDRESS(((1000 * good DIV i)+5)DIV 10)), 1 );
      WriteString(' %'); WriteLn;
      WriteString('   Avg. rtt   : '); 
      WriteAddr( ( total + Ticks( good DIV 2 )) DIV Ticks( good ), 10, 1 ); 
      WriteString('/100 seconds'); WriteLn;
    ELSE
      WriteString('  No response from ');
      WriteIPAddr( ip, 1 ); WriteLn;
    END; (* IF *)
  END DoPings;

PROCEDURE GetParams( VAR ip : IPAddr; 
                     VAR numdata, numpkts : CARDINAL ) : BOOLEAN;
  VAR
    str1, str2, str3 : ARRAY [0..79] OF CHAR;
    ip3 : IPAddr;
  BEGIN        
    GetCommand( str1 );
    Component( str1, str2 );                  
    IF str2[0] = 0C THEN
      Error('Usage: PING host [data size [# packets]]');
      RETURN FALSE;
    END; (* IF *) 
    IF ( NOT ScanIPAddr( str2, ip )) AND
       ( Resolve( str2, ip, str3, ip3 ) <> NoError ) THEN
      Error('INVALID HOST ADDRESS/NAME');
      RETURN FALSE;
    END; (* IF *)
    Component( str1, str2 );
    IF ( str2[0] = 0C )  OR ( NOT ScanCard( str2, 10, numdata ) ) THEN
      numdata := DefDataSize;
      numpkts := DefPkts;
      RETURN TRUE;      
    END; (* IF *)
    Component( str1, str2 );
    IF ( str2[0] = 0C ) OR ( NOT ScanCard( str2, 10, numpkts ) ) THEN
      numpkts := DefPkts;
    END; (* IF *)
    RETURN TRUE; 
  END GetParams;

VAR
  ip               : IPAddr;
  numdata, numpkts : CARDINAL;
  res              : Status;              
  rtt              : Ticks;
BEGIN
  IF GetParams( ip, numdata, numpkts ) THEN
    IF numpkts > 1 THEN
      DoPings( ip, numdata, numpkts );
    ELSE                             
      DoPing( ip, numdata );
    END; (* IF *)
  END; (* IF *)
END PING.
