MODULE TFTP;
                     
                  IMPORT R2SysCalls, DNS;
FROM R2NewerCalls IMPORT SetUDPPort, Send, Receive, Port, IPAddr,
                         Status;   
FROM R2SysCalls   IMPORT Exit, WriteString, WriteLn, Result, Read, Write,
                         ReadString;
FROM R2Streams    IMPORT OpenStream, CreateStream, CloseStream, Done,
                         EndStream, ReadChar, WriteChar, Stream;
FROM R2Strings    IMPORT Length, Copy;
FROM R2NetSubs    IMPORT WriteIPAddr, ScanIPAddr;
FROM DNS          IMPORT Resolve, RESULT;
FROM R2NumbIO     IMPORT WriteCard;

CONST
  RReq = 1C;
  WReq = 2C;
  Data = 3C;
  Ack  = 4C;
  Err  = 5C;
  TMO  = 1000;
  MaxRetries = 4;
  DPort = 999;

VAR
  mode  : ARRAY [0..29] OF CHAR;
  fd    : Stream;
  fHost : RECORD
            port : Port;
            addr : IPAddr;
          END;
  eof   : BOOLEAN;

PROCEDURE Insert( VAR s,t : ARRAY OF CHAR; pos : CARDINAL );
  VAR 
    i,j : CARDINAL;
  BEGIN
    i := pos;
    j := 0;
    LOOP
      IF (i > HIGH(t)) OR (j > HIGH(s)) THEN RETURN; END;
      t[i] := s[j];
      IF t[i] = 0C THEN RETURN; END;
      INC( i );
      INC( j );
    END; 
  END Insert; 

PROCEDURE Stop( s : ARRAY OF CHAR );
  VAR
    n2res : Status;
  BEGIN
    n2res := SetUDPPort( 0 );
    CloseStream( fd );
    WriteLn;    
    Message('Error : ',s,'');
    Exit( R2SysCalls.Status( -1 ) );
  END Stop;                        

PROCEDURE RcvdError( m : ARRAY OF CHAR );
  VAR
    i : CARDINAL;
  BEGIN        
    WriteString('Received error message [');
    WriteCard(ORD(m[3]),1);
    WriteString('] : ');
    i := 4;
    WHILE (m[i] <> 0C) AND (i<30) DO
     Write(m[i]);
     INC(i);
    END; (* WHILE *)
    Stop('');
  END RcvdError;

PROCEDURE PutError( p : Port; a : IPAddr; c : CARDINAL; m : ARRAY OF CHAR );
  VAR
    n2res : Status;
    s     : ARRAY [0..50] OF CHAR;
    len   : CARDINAL;
  BEGIN
    s[0] := 0C;
    s[1] := Err;
    s[2] := 0C;
    s[3] := CHR( c MOD 256 );
    CASE c OF
      1 : Insert('File not found', s, 4 ); len := 19;      |
      2 : Insert('Access violation', s, 4 );  len := 21;   |
      3 : Insert('Disk full', s, 4 ); len := 14;           |
      4 : Insert('Illegal TFTP op', s, 4 ); len := 20;     |
      5 : Insert('Unknown transfer ID', s, 4 ); len := 24; |
      6 : Insert('File exists', s, 4 ); len := 16;
    ELSE
      s[3] := 0C;
      Insert( m, s, 4 );
      len := 5 + Length( m );
    END;
    n2res := Send( s, len, p, a );
  END PutError;

PROCEDURE GetAck( ackNum : CARDINAL ) : BOOLEAN;
  VAR
    n2res : Status;
    fp    : Port;
    fa    : IPAddr;
    s     : ARRAY [0..30] OF CHAR;
    len   : CARDINAL;
  BEGIN
    LOOP
      n2res := Receive( s, len, TMO, fp, fa );
      IF n2res = Ok THEN
        IF ( fa <> fHost.addr ) THEN
          PutError( fp, fa, 5, '' );
        ELSIF ( s[1] = Err ) THEN    (* bad op *)
          RcvdError( s );
        ELSIF ( s[1] <> Ack ) THEN   (* bad op *)
          PutError( fp, fa, 4, '' );
          Stop('Illegal operation by foreign host');
        ELSIF ( ORD( s[2] ) * 256 + ORD( s[3] )) <> ackNum THEN
          PutError( fp, fa, 4, '' );
          Stop('Received bad acknowledgement');
        ELSE
          RETURN TRUE;
        END; (* IF *)
      ELSE
        RETURN FALSE;
      END; (* IF *)
    END; (* LOOP *)
  END GetAck;

PROCEDURE AckIt( p : Port; a : IPAddr; c : CARDINAL );
  VAR
    s     : ARRAY [0..3] OF CHAR;
    n2res : Status;
  BEGIN    
    s[0] := 0C;
    s[1] := Ack;
    s[2] := CHR( c DIV 256 );
    s[3] := CHR( c MOD 256 );
    n2res := Send( s, 4, p, a );
  END AckIt;

PROCEDURE GetWrPermission( f : ARRAY OF CHAR );
  VAR
    s     : ARRAY [0..130] OF CHAR;
    n2res : Status;
    i     : CARDINAL;
  BEGIN
    s[0] := 0C;
    s[1] := WReq;
    Insert( f, s, 2 );
    Insert( mode, s, Length( f ) + 3 );  
    FOR i := 1 TO MaxRetries DO
      n2res := Send( s, Length( f ) + Length( mode ) + 4, fHost.port, fHost.addr );
      IF GetAck( 0 ) THEN RETURN; END;
    END; (* FOR *)
    Stop('Could not connect');
  END GetWrPermission;
         
PROCEDURE PutRdReq( f : ARRAY OF CHAR );
  VAR
    s     : ARRAY [0..139] OF CHAR;
    n2res : Status;
  BEGIN   
    s[0] := 0C;
    s[1] := RReq;
    Insert( f, s, 2 );
    Insert( mode, s, Length( f ) + 3 );
    n2res := Send( s, Length( f ) + Length( mode ) + 4, 
                                     fHost.port, fHost.addr );
  END PutRdReq;
         
VAR
  cr : BOOLEAN;

PROCEDURE RdBlock( VAR pkt : ARRAY OF CHAR; VAR len : CARDINAL );
  VAR
    i : CARDINAL;
  BEGIN
    i := 4;
    LOOP
      IF EndStream( fd ) AND NOT cr THEN EXIT; END;
      IF i = 516 THEN EXIT; END;
      IF cr THEN
        pkt[i] := CHR(10);
        cr := FALSE;
      ELSE
        ReadChar( fd, pkt[i] );
        cr := pkt[i] = CHR(13);
      END; (* IF *)
      INC( i );
    END; (* LOOP *)
    len := i;
    eof := EndStream( fd ) AND ( len <> 516 );
    IF eof AND ( len = 0 ) THEN
      len := 5;
      pkt[4] := 0C;
    END; (* IF *)
  END RdBlock;

PROCEDURE PutBlock( pkt : ARRAY OF CHAR; len, ack : CARDINAL );
  VAR
   p2pkt : ARRAY [0..3] OF CHAR;
   i, p2len : CARDINAL;
   fp    : Port;
   fa    : IPAddr;
   n2res : Status;
  BEGIN  
    pkt[0] := 0C;
    pkt[1] := Data;
    pkt[2] := CHR( ack DIV 256 );
    pkt[3] := CHR( ack MOD 256 );
    FOR i := 1 TO MaxRetries DO
      n2res := Send( pkt, len, fHost.port, fHost.addr );
      IF GetAck( ack ) THEN RETURN; END;
    END; (* FOR *)
    Stop('Could not get ack for data block');
  END PutBlock;

PROCEDURE PutFile( f1, f2 : ARRAY OF CHAR );
  VAR   
    pkt : ARRAY [0..515] OF CHAR;
    ack, len : CARDINAL;
    n2res : Status;
  BEGIN
    OpenStream( fd, f1, FALSE );
    IF NOT Done THEN Stop('Could not open file'); END;
    Message('File ',f1,' opened for transfer');
    n2res := SetUDPPort( DPort );
    IF n2res <> Ok THEN Stop('Could not open port'); END;
    GetWrPermission( f2 );    
    ack := 1;
    LOOP
      RdBlock( pkt, len );
      PutBlock( pkt, len, ack );
      INC( ack );
      IF eof THEN EXIT; END;
    END; (* LOOP *)
    Message('Transfer of file : ',f1,' ...completed successfully');     
    Message('Successful completion','','');
    CloseStream( fd );
    n2res := SetUDPPort( 0 );
  END PutFile;             

PROCEDURE WriteBlock( pkt : ARRAY OF CHAR; len : CARDINAL );
  VAR
    i : CARDINAL;
  BEGIN 
    FOR i := 4 TO ( len - 1 ) DO
      IF pkt[i] <> CHR(10) THEN
        WriteChar( fd, pkt[i] );
      END; (* IF *)
    END; (* FOR *)
  END WriteBlock;

PROCEDURE GetBlock( VAR pkt : ARRAY OF CHAR; VAR len : CARDINAL; 
                                                 num : CARDINAL  );
  VAR
    p2pkt : ARRAY [0..3] OF CHAR;
    p2len, num2, i : CARDINAL;
    fp    : Port;
    fa    : IPAddr;
    n2res : Status;
  BEGIN
    FOR i := 1 TO MaxRetries DO
      n2res := Receive( pkt, len, TMO, fp, fa );
      IF n2res = Ok THEN                       
WriteString('len : '); WriteCard(len,2); WriteLn;
        IF ( fa <> fHost.addr ) THEN
          PutError( fp, fa, 5, '' );
        ELSIF ( pkt[1] = Err ) THEN
          RcvdError( pkt );
        ELSIF ( pkt[1] <> Data ) THEN
          PutError( fp, fa, 4, '' );
          Stop('Received bad packet');
        ELSE 
          num2 := ORD( pkt[2] ) * 256 + ORD( pkt[3] );               
          AckIt( fp, fa, num2 );
          IF num2 = num THEN
            eof := len < 516;
            RETURN;
          END; (* IF *)
        END; (* IF *)
      END; (* IF *)
    END; (* FOR *)
    Stop('Connection was lost (timed out)');
  END GetBlock;

PROCEDURE GetFile( f1, f2 : ARRAY OF CHAR );
  VAR
    n2res    : Status;
    len, num : CARDINAL;
    pkt      : ARRAY [0..520] OF CHAR;
  BEGIN
    CreateStream( fd, f1 );
    IF NOT Done THEN Stop('Could not create file'); END;
    Message('Local file : ',f1,', created');
    n2res := SetUDPPort( DPort );
    IF n2res <> Ok THEN Stop('Could not open port'); END;
    PutRdReq( f2 );           
    num := 1;
    LOOP
      GetBlock( pkt, len, num );
      WriteBlock( pkt, len );
      IF eof THEN EXIT; END;
      INC( num );
    END; (* LOOP *)                 
    Message(f2,' ...received as : ',f1);
    Message('Job completed','','');
    CloseStream( fd );
    n2res := SetUDPPort( 0 );
  END GetFile;
           
PROCEDURE Message( s1,s2,s3 : ARRAY OF CHAR );
  BEGIN
    WriteString( s1 ); WriteString(s2); WriteString(s3);
    WriteLn;
  END Message;

PROCEDURE Prompt( s : ARRAY OF CHAR );
  BEGIN  
    WriteString(s); WriteString(' : ');
  END Prompt;
            
PROCEDURE GetAns( VAR b : BOOLEAN );
  VAR 
    ch : CHAR;
  BEGIN              
    LOOP
      Read( ch, FALSE );
      IF Result() = R2SysCalls.Ok THEN
        IF ( ch = 'p' ) OR ( ch = 'P') THEN 
          Write( ch ); WriteLn;
          b := TRUE; 
          RETURN; 
        END;
        IF ( ch = 'g' ) OR ( ch = 'G') THEN 
          Write( ch ); WriteLn;
          b := FALSE;
          RETURN; 
        END;
      END;
    END;    
  END GetAns;

PROCEDURE DoJob;
  VAR
    put : BOOLEAN;              
    s, s2, f1, f2 : ARRAY [0..99] OF CHAR;
    ip2 : IPAddr;
    r : RESULT;
  BEGIN
    Prompt('Remote Host');
    ReadString( s );
    r := Resolve( s, fHost.addr, s2, ip2 );
    IF r <> DNS.NoError THEN
      IF NOT ScanIPAddr( s, fHost.addr ) THEN
        Stop('Name error');
      END;
    END;                     
    WriteString('Connected to : ');
    WriteIPAddr( fHost.addr, 2 );   WriteLn; 
    Prompt('Get or Put');
    GetAns( put );
    Prompt('Remote file name');
    ReadString(f2);
    Prompt('Local file name (CR if same)');
    ReadString(f1);
    IF f1[0] = 0C THEN Copy( f2, f1 ); END;
    IF put THEN 
      PutFile( f1, f2 )
    ELSE                
      GetFile( f1, f2 )
    END;
  END DoJob;
        
BEGIN           
  cr := FALSE;
  Copy( 'octet', mode );
  fHost.port := 69;
  DoJob;
END TFTP.
