MODULE TFTPS;

  (*  Trivial File Transfer Protocol - Server Implementation *)
                     
                  IMPORT R2SysCalls, R2Streams;
FROM R2NewerCalls IMPORT SetUDPPort, Send, Receive, Port, IPAddr,
                         Status;   
FROM R2SysCalls   IMPORT Result, Read, Write, Handle, Create,
                         GetPID, GetUser, SetUser, Process, FileMode,
                         ReadBlock, WriteBlock, Open, Close;
FROM R2Streams    IMPORT OpenStream, CreateStream, CloseStream, Done,
                         EndStream, ReadChar, WriteChar, Stream;
FROM R2Strings    IMPORT Length, Copy;

CONST

   (* op codes *)

  RReq = 1C;
  WReq = 2C;
  Data = 3C;
  Ack  = 4C;
  Err  = 5C;

   (* other *)

  TMO  = 1000;     (* time to wait for an acknowlegement *)
  MaxRetries = 4;
  DPort = 69;
  UseNum = 0FFFFH; 

TYPE
  Mode = ( NETASCII, OCTET );   (* transfer modes ... mail not supported *)

VAR
  mode  : Mode;
  fd    : Stream;
  h     : Handle;
  fHost : RECORD
            port : Port;
            addr : IPAddr;
          END;
  err,                     (* flag *)
  eof   : BOOLEAN;         (* flag *)


PROCEDURE Insert( VAR s,t : ARRAY OF CHAR; pos : CARDINAL );
 (* not quite like R2Strings !! ie t[0] is always 0C *)
  VAR 
    i,j : CARDINAL;
  BEGIN
    i := pos;
    j := 0;
    LOOP
      IF (i > HIGH(t)) OR (j > HIGH(s)) THEN RETURN; END;
      t[i] := s[j];
      IF t[i] = 0C THEN RETURN; END;
      INC( i );
      INC( j );
    END; (* LOOP *) 
  END Insert; 

PROCEDURE PutError( p : Port; a : IPAddr; c : CARDINAL; m : ARRAY OF CHAR );
 (* send error message *)
  VAR
    n2res : Status;
    s     : ARRAY [0..50] OF CHAR;
    len   : CARDINAL;
  BEGIN
    s[0] := 0C;
    s[1] := Err;
    s[2] := 0C;
    s[3] := CHR( c MOD 256 );
    CASE c OF
      1 : Insert('File not found', s, 4 ); len := 19;      |
      2 : Insert('Access violation', s, 4 );  len := 21;   |
      3 : Insert('Disk full', s, 4 ); len := 14;           |
      4 : Insert('Illegal TFTP op', s, 4 ); len := 20;     |
      5 : Insert('Unknown transfer ID', s, 4 ); len := 24; |
      6 : Insert('File exists', s, 4 ); len := 16;
    ELSE
      s[3] := 0C;
      Insert( m, s, 4 );           (* error message provided by caller *)
      len := 5 + Length( m );
    END;
    n2res := Send( s, len, p, a );
  END PutError;

PROCEDURE GetAck( ackNum : CARDINAL ) : BOOLEAN;                        
 (* get acknowledgement for last block sent *)
  VAR
    n2res : Status;
    fp    : Port;
    fa    : IPAddr;
    s     : ARRAY [0..30] OF CHAR;
    tries,
    len   : CARDINAL;                                                  
   
  BEGIN  
    tries := 0;
    LOOP
      n2res := Receive( s, len, TMO, fp, fa );
      IF n2res = Ok THEN
        IF ( fa <> fHost.addr ) OR ( fp <> fHost.port ) THEN
          PutError( fp, fa, 5, '' );
        ELSIF ( s[1] = Err ) THEN    (* bad op *)
          err := TRUE;
        ELSIF ( s[1] <> Ack ) THEN   (* bad op *)
          PutError( fp, fa, 4, '' );
          err := TRUE;
        ELSIF ( ORD( s[2] ) * 256 + ORD( s[3] )) <> ackNum THEN
          PutError( fp, fa, 4, '' );
          err := TRUE;
        ELSE
          RETURN TRUE;
        END; (* IF *)
      ELSE
        RETURN FALSE;
      END; (* IF *)
      IF err THEN RETURN TRUE; END;    (* need to pass error further *)
      INC( tries );
      IF tries = MaxRetries THEN RETURN FALSE; END; (* give up! *)
    END; (* LOOP *)
  END GetAck;

PROCEDURE AckIt( p : Port; a : IPAddr; c : CARDINAL );                
 (* acknowledge the last block received *)
  VAR
    s     : ARRAY [0..3] OF CHAR;
    n2res : Status;
  BEGIN    
    s[0] := 0C;
    s[1] := Ack;
    s[2] := CHR( c DIV 256 );
    s[3] := CHR( c MOD 256 );
    n2res := Send( s, 4, p, a );
  END AckIt;

PROCEDURE GetReq( VAR put : BOOLEAN; VAR f : ARRAY OF CHAR );
 (* block till we get a request  *)
  VAR
    n2res : Status;
    i, op, len : CARDINAL;
    s : ARRAY [0..129] OF CHAR;
  BEGIN     
    LOOP
      n2res := Receive( s, len, 0, fHost.port, fHost.addr );
      IF ( n2res = Ok ) AND ( len > 4 ) THEN
        op := ORD( s[0] ) * 256  + ORD( s[1] );
        IF ( op = ORD(RReq) ) OR ( op = ORD(WReq) ) THEN
          IF op = ORD(WReq) THEN AckIt( fHost.port, fHost.addr, 0 ); END;
          put := op = ORD(RReq);
          i := 2;
          LOOP                     (* get file name *)       
            f[i-2] := s[i];
            IF ( i > len ) OR ( s[i] = 0C ) THEN INC( i ); EXIT; END;
            INC( i );
          END; (* LOOP *)                           
          IF ( s[i] = 'o') OR ( s[i] = 'O') THEN      (* get transfer mode *)
            mode := OCTET;
          ELSE
            mode := NETASCII;
          END; (* IF *)
          RETURN;
        END; (* IF *)
      END; (* IF *)
    END; (* LOOP *)
  END GetReq;
                  
VAR
  cr : BOOLEAN;

PROCEDURE RdBlock( VAR pkt : ARRAY OF CHAR; VAR len : CARDINAL );
  VAR    
    l2res : R2SysCalls.Status;
    i : CARDINAL;
  BEGIN         
    IF mode = OCTET THEN                            (* binary file *)
      l2res := ReadBlock( h, pkt );
      eof := l2res = R2SysCalls.EndFile;
      IF NOT eof THEN
        FOR i := 4 TO 515 DO pkt[i] := pkt[i-4]; END; 
        len := 516;
      ELSE
        len := 4;
      END; (* IF *)
    ELSE                                             (* text file *)
      i := 4;
      LOOP
        IF EndStream( fd ) AND NOT cr THEN EXIT; END;
        IF i = 516 THEN EXIT; END;
        IF cr THEN
          pkt[i] := CHR(10);
          cr := FALSE;
        ELSE
          ReadChar( fd, pkt[i] );
          cr := pkt[i] = CHR(13);
        END; (* IF *)
        INC( i );
      END; (* LOOP *)
      len := i;
      eof := EndStream( fd ) AND ( len <> 516 );   (* eof is len < 516 ! *)
    END; (* IF *)
  END RdBlock;

PROCEDURE PutBlock( pkt : ARRAY OF CHAR; len, ack : CARDINAL );         
 (* send tftp data packet *)
  VAR
   p2pkt : ARRAY [0..3] OF CHAR;
   i, p2len : CARDINAL;
   fp    : Port;
   fa    : IPAddr;
   n2res : Status;
  BEGIN
    pkt[0] := 0C;
    pkt[1] := Data;
    pkt[2] := CHR( ack DIV 256 );
    pkt[3] := CHR( ack MOD 256 );
    FOR i := 1 TO MaxRetries DO
      n2res := Send( pkt, len, fHost.port, fHost.addr );
      IF GetAck( ack ) THEN RETURN; END;
    END; (* FOR *)
    err := TRUE;
  END PutBlock;

PROCEDURE PutFile( f1 : ARRAY OF CHAR );
  VAR 
    l2res : R2SysCalls.Status;  
    pkt : ARRAY [0..515] OF CHAR;
    ack, len : CARDINAL;
  BEGIN                        
    IF mode = NETASCII THEN
      OpenStream( fd, f1, FALSE );
      IF NOT Done THEN 
        R2SysCalls.GetMessage( R2Streams.SysErr, pkt ); 
        PutError( fHost.port, fHost.addr, 0, pkt );
        RETURN; 
      END; (* IF *)
    ELSE
      l2res := Open( f1, FALSE, h );   
      IF l2res <> R2SysCalls.Ok THEN
        R2SysCalls.GetMessage( R2Streams.SysErr, pkt ); 
        PutError( fHost.port, fHost.addr, 0, pkt );
        RETURN; 
      END; (* IF *)
    END; (* IF *)
    ack := 1;
    LOOP
      RdBlock( pkt, len );
      PutBlock( pkt, len, ack );
      INC( ack );
      IF err OR eof THEN EXIT; END;
    END; (* LOOP *)                
    IF mode = NETASCII THEN
      CloseStream( fd );
    ELSE
      l2res := Close( h, TRUE );
    END; (* IF *)
  END PutFile;             

PROCEDURE WrBlock( pkt : ARRAY OF CHAR; len : CARDINAL );
  VAR
    i : CARDINAL;  
    l2res : R2SysCalls.Status;
  BEGIN  
    IF mode = OCTET THEN                             (* binary file *)
      FOR i := 4 TO len-1 DO pkt[i-4] := pkt[i]; END; 
      IF len <> 4 THEN l2res := WriteBlock( h, pkt, FALSE ); END;
    ELSE
      FOR i := 4 TO ( len - 1) DO          (* text file *)
        IF pkt[i] <> CHR(10) THEN
          WriteChar( fd, pkt[i] );
        END; (* IF *)
      END; (* FOR *)
    END; (* IF *)
  END WrBlock;

PROCEDURE GetBlock( VAR pkt : ARRAY OF CHAR; VAR len : CARDINAL; 
                                                 num : CARDINAL  );
 (* receive next data block *)
  VAR
    p2pkt : ARRAY [0..3] OF CHAR;
    p2len, num2, i : CARDINAL;
    fp    : Port;
    fa    : IPAddr;
    n2res : Status;
  BEGIN
    FOR i := 1 TO MaxRetries DO
      n2res := Receive( pkt, len, TMO, fp, fa );
      IF n2res = Ok THEN
        IF ( fa <> fHost.addr ) OR ( fp <> fHost.port ) THEN
          PutError( fp, fa, 5, '' );
        ELSIF ( pkt[1] = Err ) THEN
          err := TRUE;
        ELSIF ( pkt[1] <> Data ) THEN
          PutError( fp, fa, 4, '' );
          err := TRUE;
        ELSE 
          num2 := ORD( pkt[2] ) * 256 + ORD( pkt[3] );               
          IF num2 < num THEN  (* they did not get our last ack, rxmit *)
            AckIt( fp, fa, num2 );
          ELSIF num2 = num THEN
            eof := len < 516;        (* End of transmission ! *)
            RETURN;
          ELSE     (* must have lost some data blocks !! *)
            PutError( fp, fa, 0, 'out of synch!'); 
            err := TRUE;
          END; (* IF *)
        END; (* IF *)
      END; (* IF *)
      IF err THEN RETURN; END;      (* need to propogate it! *)
    END; (* FOR *)
    err := TRUE;
  END GetBlock;

PROCEDURE GetFile( f1 : ARRAY OF CHAR );
  VAR
    l2res : R2SysCalls.Status;
    len, num : CARDINAL;
    pkt      : ARRAY [0..515] OF CHAR;
  BEGIN         
    IF mode = NETASCII THEN
      CreateStream( fd, f1 );
      IF NOT Done THEN 
        R2SysCalls.GetMessage( R2Streams.SysErr, pkt ); 
        PutError( fHost.port, fHost.addr, 0, pkt );
        RETURN;                    
      END; (* IF *)
    ELSE
      l2res := Create( f1, FileMode(0B0H), h ); 
      IF l2res <> R2SysCalls.Ok THEN
        R2SysCalls.GetMessage( R2Streams.SysErr, pkt ); 
        PutError( fHost.port, fHost.addr, 0, pkt );
        RETURN; 
      END; (* IF *)
    END; (* IF *)
    num := 1;
    LOOP
      GetBlock( pkt, len, num );
      IF err THEN EXIT; END;
      WrBlock( pkt, len );
      IF eof THEN EXIT; END;
      INC( num );
    END; (* LOOP *)                 
    IF mode = NETASCII THEN
      CloseStream( fd );
    ELSE
      l2res := Close( h, TRUE );
    END; (* IF *)
  END GetFile;
           
PROCEDURE DoJob;
  VAR
    put : BOOLEAN;
    f : ARRAY [0..99] OF CHAR;
  BEGIN
    LOOP          
      GetReq( put, f );
      err := FALSE;
      cr := FALSE;
      IF put THEN 
        PutFile( f );
      ELSE                
        GetFile( f );
      END; (* IF *)
    END; (* LOOP *)
  END DoJob;
                              
VAR
  n2res : Status;
  l2res : R2SysCalls.Status;
  proc  : Process;
  user  : CARDINAL;
BEGIN    
  l2res := GetPID( proc );
  l2res := GetUser( proc, user );
  IF user = 0 THEN
    n2res := SetUDPPort( DPort );
    IF n2res = Ok THEN 
      l2res := SetUser( proc, UseNum );
      IF l2res = R2SysCalls.Ok THEN DoJob; END; 
    END; (* IF *)
  END; (* IF *)
END TFTPS.
