MODULE TNetServer;

  (* Telnet - Server Implementation *)

FROM TelOpts      IMPORT IAC, DONT, DOIT, WONT, WILL, RDont, RDo,
                         RWill, RWont, SDo, SDont, SWill, SWont,
                         Echo, SuppressGA, RegisterSoc, TInit;
FROM R2NewCalls   IMPORT AttachTerminal, GetJobInfo;
FROM R2SysCalls   IMPORT Sleep, Result, SetCCMode, WriteString, GetTime, 
                         Ticks, Kill;
FROM R2NewerCalls IMPORT TCPListen, TCPWrite, TCPRead, TCPClose,
                         TCPAttach, TCPDetach, Port, SocHandle,
                         Status, IPAddr, VTIn, VTOut, VTAllocate,
                         VTDeallocate, SetPriority;
FROM R2Strings    IMPORT Copy, Length;
                  IMPORT R2SysCalls;

CONST
  TelnetPort = 23;    (* server port number *)
  tmo        = 100;
  CtrlC      = 3C;
  CR         = 15C;
  LF         = 12C;
  NoOp       = 0C;

VAR
  pos, len, unit : CARDINAL;
  s              : ARRAY[0..2048] OF CHAR;
  CrLf, logout   : ARRAY[0..20] OF CHAR;
  Fiac, Fcmnd, 
  OptMode        :BOOLEAN;
  forHost        : IPAddr;
  port           : Port;
  soc            : SocHandle;
  greeting       : ARRAY [0..50] OF CHAR;
  ch             : CHAR;

PROCEDURE DoOpt();
VAR
  cmnd, opt:CHAR;
BEGIN
  IF NOT Fiac THEN
    INC( pos );
    Fiac := TRUE;
  END; (* IF *)
  IF ( pos < len ) AND NOT Fcmnd THEN
     cmnd := s[pos];
     INC( pos );
     Fcmnd := TRUE;
  END; (* IF *)

  IF ( pos < len ) AND Fcmnd THEN
    IF ( ORD(cmnd) > 250 ) AND ( cmnd <> IAC ) THEN
      opt := s[pos];
      INC( pos );
      CASE cmnd OF
        DONT : RDont( opt );        |
        DOIT : RDo( opt );          |
        WONT : RWont( opt );        |
        WILL : RWill( opt );
      ELSE;
      END; (* CASE *)
    END; (* IF *)
    OptMode := FALSE;
    Fiac := FALSE;
    Fcmnd := FALSE;
  END(*IF*);
END DoOpt;

PROCEDURE RdCh( VAR ch : CHAR ) : BOOLEAN;
  BEGIN
    IF pos < len THEN
      OptMode := ( s[pos] = IAC ) OR OptMode;
      IF OptMode THEN
         DoOpt();
         RETURN RdCh( ch );
      END; (* IF *)
      ch := s[pos];
      INC( pos );
      RETURN TRUE;
    ELSE
      pos := 0;
      RETURN FALSE;
    END; (* IF *)
  END RdCh;

PROCEDURE InitTel() : BOOLEAN;  (* change this, need not be a func *)
  BEGIN
    SetCCMode( FALSE );
    Copy('logout',logout);
    Copy('LUMOS-2 Telnet Server ( Proto - Type VER # 1.1 )',greeting);
    CrLf[0] := CR;
    CrLf[1] := LF;
    RETURN TRUE;
  END InitTel;

PROCEDURE InitConn();
  VAR
    l2res : R2SysCalls.Status;
  BEGIN
    n2res := TCPAttach( soc );    
    TInit;
    RegisterSoc( soc );
    l2res := SetPriority(3);
    Fiac := FALSE;
    Fcmnd := FALSE;
    OptMode := FALSE;
    pos := 0;
    SWill( Echo );
    SDo( SuppressGA );
    SDont( Echo );
    l2res := R2SysCalls.Ok;
    WHILE l2res = R2SysCalls.Ok DO
      l2res := VTOut( unit, ch );
    END; (* WHILE *)
END InitConn;

PROCEDURE CleanUp;
  (* kill off any jobs running on the now closed Virtual Terminal *)
  VAR
    i, pgs, ti, proc : CARDINAL;
    im : ARRAY [0..2] OF CHAR;
  BEGIN
    FOR i := 7 TO 40 DO
      l2res := GetJobInfo( i, pgs, ti, proc, im );
      IF l2res = R2SysCalls.Ok THEN
        IF ti = unit THEN
          l2res :=  Kill( i, R2SysCalls.Status( -1 ));
        END; (* IF *)
      END; (* IF *)
    END; (* FOR *)
  END CleanUp;

PROCEDURE Serve();                   
  (* performance is enhanced by toggling priorities *)
  VAR    
    i : CARDINAL;
  BEGIN                      

    l2res := SetPriority( 4 );
    n2res := TCPWrite( soc, greeting, Length( greeting ) );
    n2res := TCPWrite( soc, CrLf, 2 );
    i := 0; 
    len := 0;     
    l2res := VTIn( unit, CtrlC );

    LOOP
      n2res:=TCPRead( soc, s, len, FALSE );
      i := 0;
      LOOP
        IF i = len THEN EXIT; END;
        IF ( s[i] <> NoOp ) AND ( s[i] <> LF ) THEN
          l2res := VTIn( unit, s[i] );
        END;
        INC(i);
      END;    

      l2res := SetPriority( 3 );
      l2res := Sleep( tmo );
      i := 0;
      l2res := VTOut( unit, ch );
      LOOP
        s[i] := ch;
        IF ( l2res <> R2SysCalls.Ok ) OR ( i = 2048 ) THEN EXIT; END; 
        l2res := VTOut( unit, ch );
        INC(i);
      END;                                   
                           
        (* check if remote user logged out! *)
      IF l2res = R2SysCalls.NotActive THEN RETURN; END;

      l2res := SetPriority( 4 );
      IF i <> 0 THEN n2res:=TCPWrite( soc, s, i ); END;
      IF n2res <> Ok THEN EXIT; END;
    END;

    WHILE l2res = R2SysCalls.Ok DO l2res := VTOut( unit, ch ); END;
  END Serve;

VAR      
  l2res   : R2SysCalls.Status;
  n2res   : Status;
  i       : CARDINAL;

BEGIN (* MAIN *)
  IF InitTel() THEN
    LOOP

      LOOP  (* get a socket *)
        n2res := TCPListen( TelnetPort, soc );
        IF n2res = Ok THEN
          EXIT;
        END; (* IF *)
        l2res := Sleep( tmo );
      END; (* LOOP *)

      InitConn();            (* sends our setup requests *)
      l2res := Sleep( 100 );
      n2res := TCPRead( soc, s, len, FALSE );
      LOOP                       (* deal with their setup requests/replies *)
        IF n2res = Ok THEN 
          LOOP
            IF NOT RdCh(ch) THEN EXIT; END;
          END(*LOOP*);
        END; (* IF *)
        l2res := Sleep( 200 );
        n2res := TCPRead( soc, s, len, FALSE );
        IF len = 0 THEN EXIT; END;
      END; (* LOOP *)          

      (* ready to start up user connection *)

      IF VTAllocate( unit ) = R2SysCalls.Ok THEN
        Serve();
      END; (* IF *)

      VTDeallocate( unit );
      n2res := TCPClose( soc );
      CleanUp;   (* kill of any jobs still running on VT *)
    END; (* LOOP *)
  END; (* IF *)
END TNetServer.
