IMPLEMENTATION MODULE Commands;  (* General command line parser for LUMOS-2 *)

(* G.R. Guenther 1988 05 28  *)
(* Modified:     1989 01 03  LUMOS-2 version *)

(* This module provides a standard way of getting arguments from the
   command line.  It parses the command into fields, identifies value
   assignments and prompts for missing information.
*)

FROM R2SysCalls IMPORT GetCommand, WriteString, ReadString, WriteLn;
FROM R2Strings IMPORT Insert, Substring, Find, Remove, Length;
FROM Errors IMPORT Error;

CONST   Ermsg = "Command line syntax error";

VAR     Com : ARRAY [0..79] OF CHAR;
        first, alwaysprompt : BOOLEAN;

PROCEDURE  GetArg( VAR promptstring, arg, val : ARRAY OF CHAR;
                   required, acceptval : BOOLEAN );

(* If required is TRUE and the command line is currently empty,
   promptstring is printed and a new command line is read in.
   (The command line is initially the tail of the command used to
   invoke the program, if it is available.)  If the command line is
   still empty, GetArg returns without touching the parameters
   arg or val.  If the line is not empty, the first argument is
   removed and assigned to arg.  If acceptval is true and this
   argument has an associated value, it is separated and returned
   in val.

   Arguments are separated by white space and values are quoted
   strings attached to the argument with an equals sign.
   Command line syntax errors are fatal; they cause the program
   to exit (with an appropriate message).
*)     

VAR     i, j : CARDINAL;
        temp : ARRAY [0..79] OF CHAR;
        promptonce : BOOLEAN;
        eq, qu : ARRAY [0..1] OF CHAR;

BEGIN   eq[0] := '='; eq[1] := 0C; qu[0] := '"'; qu[1] := 0C;
        promptonce := alwaysprompt AND (promptstring[0] <> 0C);
        LOOP i := 0;
             WHILE (i <= HIGH(Com)) AND (Com[i] = ' ') DO INC(i) END;
             Remove(Com,0,i);
             IF (Com[0] = 0C) AND (required OR promptonce)
               THEN WriteString(promptstring);
                    WriteString(": ");
                    ReadString(Com);
                    IF first THEN alwaysprompt := TRUE END;
                    promptonce := FALSE;
               ELSE EXIT
               END;
             END;
        first := FALSE;
        IF Com[0] = 0C THEN RETURN END;
        i := 0;
        WHILE (i <= HIGH(Com)) AND (Com[i] > ' ') DO INC(i) END;
        j := Find(eq,Com);
        IF (NOT acceptval) OR (j >= i) THEN
          Substring(Com,arg,0,i);
          Remove(Com,0,i);
          RETURN;
          END;
        Substring(Com,arg,0,j);
        temp[0] := 0C; Insert(Com,temp,0);
        Remove(Com,0,j+1);
        IF (j > 0) AND (Com[0] = '"') THEN
          Remove(Com,0,1);
          i := Find(qu,Com);
          IF i < Length(Com) THEN
            Substring(Com,val,0,i);
            Remove(Com,0,i+1);
            RETURN;
            END;
          END;
        Error(Ermsg,temp);
END GetArg;

BEGIN   GetCommand(Com);
        first := TRUE;
        alwaysprompt := FALSE;
END Commands.
