IMPLEMENTATION MODULE Errors;  (* Error message formatter *)

(* G.R. Guenther 1988 05 28  *)
(* Modified: GRG 1989 01 03  -- LUMOS-2 version *)

(* This module provides a simple formatter for error messages.
   It enforces a uniform style of display.
*)

FROM R2SysCalls IMPORT WriteString, WriteLn, Status, Exit;
FROM R2Strings IMPORT Insert;

VAR     Prog : ARRAY [0..19] OF CHAR;  (* let's be reasonable *)

PROCEDURE  SetName( VAR progname : ARRAY OF CHAR);

(* Assigns the name to be used as the prefix of all error messages. *)

BEGIN   Prog[0] := 0C; Insert(progname,Prog,0);
END SetName;

PROCEDURE  Error( VAR msg, suppl : ARRAY OF CHAR);

(* Displays an error message in standard form with the text from msg.
   If suppl is not blank, it is appended to the message along with a
   separator. The program is terminated after the message is printed,
   and a failure status is passed back to the operating system. 
*)

VAR     m, s : ARRAY [0..131] OF CHAR;  (* avoid short allocation errors *)
                          
BEGIN   m[0] := 0C; Insert(msg,m,0);
        s[0] := 0C; Insert(suppl,s,0);
        WriteLn;
        WriteString(Prog);
        WriteString(" -- ");
        WriteString(m);
        IF s[0] <> 0C THEN
          WriteString(": ");
          WriteString(s);
          END;
        WriteLn;
        Exit(Status(-1));
END Error;

PROCEDURE  Assert( assertion : BOOLEAN; VAR ident : ARRAY OF CHAR);

(* The assertion is normally a boolean expression that is expected
   to be TRUE.  If it isn't, Error("Internal error",ident) is called.
*)

BEGIN   IF assertion THEN RETURN END;
        Error("Internal error",ident);
END Assert;

BEGIN   Prog := "NoName";
END Errors.
