IMPLEMENTATION MODULE OneWay;

(*$T-*)  (* Don't check subscripts, etc. *)

PROCEDURE  Cipher( VAR str : ARRAY OF CHAR);

(* Do a one-way encryption of str onto itself.  All bits in the first
   16 bytes of str are significant.  The output will consist of non-blank
   printable characters only.  This algorithm will work on any machine for
   which the s and b fields of equiv occupy the same memory.  It will
   produce different results on big-endian and little-endian machines.
*)

TYPE    b16 = SET OF [0..15];
        line = ARRAY [0..15] OF CHAR;
        equiv = RECORD CASE BOOLEAN OF
                         TRUE  : s : line |
                         FALSE : b : ARRAY [0..7] OF b16
                         END;
                END;

VAR     text : line;
        old, new : equiv;
        ind : ARRAY [0..15] OF CARDINAL;
        n, i, j, k : CARDINAL;
        c : CHAR;

BEGIN   i := 0;
        WHILE (i <= 15) AND (i <= HIGH(str)) AND (str[i] <> 0C) DO
          text[i] := str[i];
          INC(i);
          END;
        FOR k := i TO 15 DO text[k] := 0C END;
        FOR n := 1 TO 3 DO
          FOR i := 0 TO 15 DO
            text[i] := CHR((ORD(text[i])*37+51*i+29) MOD 256);
            END;
          old.s := text;
          FOR i := 0 TO 15 DO ind[i] := i END;
          FOR i := 0 TO 14 DO
            FOR j := 0 TO 14-i DO
              IF (text[j] < text[j+1]) = ODD(n) THEN
                k := ind[j]; ind[j] := ind[j+1]; ind[j+1] := k;
                c := text[j]; text[j] := text[j+1]; text[j+1] := c;
                END;
              END;
            END;
          FOR i := 0 TO 7 DO
            new.b[i] := b16{};
            FOR j := 0 TO 15 DO
              IF j IN old.b[i] THEN INCL(new.b[i],ind[j]) END;
              END;
            END;
          FOR i := 0 TO 15 DO
            text[ind[i]] := new.s[i];
            END;
          END;
        FOR i := 0 TO 15 DO 
          k := ORD(text[i]);
          IF k <= 32 THEN INC(k,256) END;
          k := (((k - 33) DIV 3) + 33);
          text[i] := CHR(k);
          END;
        i := 0;
        WHILE (i <= 15) AND (i <= HIGH(str)) DO
          str[i] := text[i];
          INC(i);
          END;
        IF i <= HIGH(str) THEN str[i] := 0C END;
END Cipher;

END OneWay.
