IMPLEMENTATION MODULE R2Directories;

FROM R2SysCalls IMPORT Handle, Close, Open, Translate, ReadBlock, Time, Status,
                       ModeBits, FileMode, GetMode, DevAttr, DevAttrSet, 
                       GetDevAttr;
FROM R2Strings IMPORT Length, Copy, Insert, Substring, Remove;

TYPE    DirRec = RECORD Name : ARRAY [0..19] OF CHAR;
                        First : CARDINAL;
                        Date : Time;
                        Owner, Mode : CARDINAL;
                 END;

VAR     db : RECORD 
               CASE BOOLEAN OF
                 TRUE  : a : ARRAY [0..511] OF CHAR  |
                 FALSE : d : ARRAY [0..15] OF DirRec;
                 END;
               END;
        rnum : CARDINAL;
        setup : BOOLEAN;
        dh : Handle;

PROCEDURE  LastHyph( VAR s : ARRAY OF CHAR) : CARDINAL;

VAR     i, h : CARDINAL;

BEGIN   h := Length(s);
        i := 0;
        WHILE (i <= HIGH(s)) AND (s[i] <> 0C) DO
          IF s[i] = '-' THEN h := i END;
          INC(i);
          END;
        RETURN h;
END LastHyph;  

PROCEDURE  TestWild( VAR p : ARRAY OF CHAR) : BOOLEAN;
       
VAR     ws, ph : ARRAY [0..79] OF CHAR;
        rs : Status;
        i, hp : CARDINAL;        
        a : DevAttrSet;

BEGIN   Copy(p,ws);
        rs := Translate(ws,ph);
        IF rs <> Ok THEN RETURN FALSE END;
        hp := LastHyph(ph);
        Substring(ph,ws,0,hp);  (* GetDevAttr doesn't like * & ? *)
        rs := GetDevAttr(ws,i,a);
        IF rs <> Ok THEN RETURN FALSE END;
        IF NOT (BLK IN a) THEN RETURN FALSE END;
        Substring(ph,ws,hp+1,80);
        IF hp >= Length(ph) THEN RETURN TRUE END;
        hp := Length(ws);
        IF hp = 0 THEN RETURN TRUE END;
        FOR i := 0 TO hp-1 DO
          IF (ws[i] = '?') OR (ws[i] = '*')
            THEN RETURN TRUE END;
          END;
        RETURN FALSE;
END TestWild;
        
PROCEDURE  SetWild( VAR p : ARRAY OF CHAR) : Status;

VAR     ws : ARRAY [0..79] OF CHAR;
        hp, ow : CARDINAL;
        rs : Status;
        ti : Time;
        fm : FileMode;

BEGIN   IF setup THEN rs := Close(dh,FALSE) END;
        Copy(p,ws);
        rs := Translate(ws,Path);
        IF rs <> Ok THEN RETURN rs END;
        hp := LastHyph(Path);
        Substring(Path,Pattern,hp+1,80);
        IF hp >= Length(Path) THEN Pattern[0] := 0C END;
        Remove(Path,hp,80);
        IF Length(Pattern) = 0 THEN 
          Pattern[0] := '*';
          Pattern[1] := 0C;
          END;
        Copy(Path,ws);                
        hp := LastHyph(ws);
        IF hp = Length(ws) THEN Insert("-DIRECTORY.SYS",ws,hp) END;
        rs := Open(ws,FALSE,dh);
        IF rs <> Ok THEN RETURN rs END;
        rs := GetMode(dh,ti,fm,ow);
        IF NOT (DIR IN fm) THEN 
          rs := Close(dh,FALSE);
          RETURN BadDirect;
          END;
        setup := TRUE;
        rnum := 16;       (* Force read *)
        RETURN Ok;
END SetWild;
        
PROCEDURE  NextWild() : Status;

VAR     rs, xs : Status;

BEGIN   LOOP IF rnum = 16 THEN
               rs := ReadBlock(dh,db.a);
               IF rs <> Ok THEN 
                 setup := FALSE;
                 xs := Close(dh,FALSE);
                 RETURN rs; 
                 END;
               rnum := 0;
               END;
             WITH db.d[rnum] DO
               IF (Name[0] <> 0C) AND WildMatch(Pattern,Name) THEN
                 Copy(Name,NName);
                 FullName[0] := '-'; FullName[1] := 0C;
                 Insert(Name,FullName,1);
                 Insert(Path,FullName,0);
                 NDate := Date;
                 NFirst := First;
                 NMode := Mode;
                 NOwner := Owner;
                 INC(rnum);
                 RETURN Ok;
                 END;
               END;
             INC(rnum);
             END;
END NextWild;

PROCEDURE  WildMatch( VAR p, s : ARRAY OF CHAR) : BOOLEAN;

VAR     ls, lp, is, ip, ns, np : CARDINAL;
        cs, cp : CHAR;       
        st : BOOLEAN;

BEGIN   ls := 0;
        WHILE (ls <= HIGH(s)) AND (s[ls] <> 0C) DO INC(ls) END;
        lp := 0;
        WHILE (lp <= HIGH(p)) AND (p[lp] <> 0C) DO INC(lp) END;
        st := FALSE;
        ip := 0;
        is := 0;
        LOOP IF ip >= lp THEN cp := 0C ELSE cp := p[ip] END;
             IF is >= ls THEN cs := 0C ELSE cs := s[is] END;
             IF (cs = cp) AND (cs = 0C) THEN RETURN TRUE END;
             IF cs = cp THEN 
               INC(ip); 
               INC(is);
             ELSIF (cp = '?') AND (cs <> 0C) THEN 
               INC(ip); 
               INC(is);
             ELSIF cp = '*' THEN 
               st := TRUE;
               INC(ip);
               np := ip;
               ns := is;
             ELSIF st THEN
               INC(ns);
               IF ns > ls THEN RETURN FALSE END;
               is := ns;
               ip := np;
             ELSE RETURN FALSE END;
             END;
END WildMatch;
             
BEGIN   setup := FALSE;
END R2Directories.
