IMPLEMENTATION MODULE R2Streams;

FROM SYSTEM IMPORT ADDRESS, WORD;
FROM R2SysCalls IMPORT Status, Close, Open, Create, Handle, ReadBlock,
                       WriteBlock, GetDevAttr, DevAttr, DevAttrSet,
                       Seek, FileMode, Size;
FROM R2Storage IMPORT ALLOCATE, DEALLOCATE;

TYPE    Stream = POINTER TO SCB;
        SCB = RECORD H : Handle;
                     B : ARRAY [0..511] OF CHAR;
                     Record, Dirty, Modified, Created, 
                     PutBack, Write : BOOLEAN;
                     BackChar : CHAR;
                     Cbk, Cby, Ebk, Eby : CARDINAL;
              END;
        CtoW = RECORD CASE BOOLEAN OF
                  TRUE  : w : WORD |
                  FALSE : hi, lo : CHAR
                  END;
               END;

PROCEDURE  CreateStream( VAR s : Stream; VAR filename : ARRAY OF CHAR);

VAR     ds : CARDINAL;
        da : DevAttrSet;

BEGIN   Done := TRUE;
        SysErr := GetDevAttr(filename,ds,da);
        IF SysErr <> Ok THEN Done := FALSE; RETURN END;
        NEW(s);
        IF s = NIL THEN Done := FALSE; END;
        WITH s^ DO
          Record := NOT (BLK IN da);
          IF Record 
            THEN SysErr := Open(filename,TRUE,H)
            ELSE SysErr := Create(filename,CreateMode,H)
            END;
          Modified := TRUE;  (* FALSE *)
          Dirty := FALSE;
          Created := TRUE;
          Write := TRUE;
          PutBack := FALSE;
          Eby := 0; Ebk := 0;
          Cby := 0; Cbk := 0;
          END;
        IF SysErr <> Ok THEN Done := FALSE; DISPOSE(s); END;
END CreateStream;

PROCEDURE  OpenStream( VAR s : Stream; VAR filename : ARRAY OF CHAR; 
                       modify : BOOLEAN);

VAR     ds : CARDINAL;
        da : DevAttrSet;
        rs : Status;

BEGIN   Done := TRUE;
        SysErr := GetDevAttr(filename,ds,da);
        IF SysErr <> Ok THEN Done := FALSE; RETURN END;
        NEW(s);
        IF s = NIL THEN Done := FALSE; END;
        WITH s^ DO
          Record := NOT (BLK IN da);
          Write := modify;
          SysErr := Open(filename,modify,H);
          IF SysErr <> Ok THEN Done := FALSE; DISPOSE(s); RETURN END;
          Modified := FALSE;
          Dirty := FALSE;
          Created := FALSE;
          PutBack := FALSE;
          Cbk := 65535; Cby := 0;
          Ebk := 65535; Eby := 0;
          SetStream(s,0,0);
          END;
        IF NOT Done THEN DISPOSE(s); RETURN END;
END OpenStream;

PROCEDURE  CloseStream( VAR s : Stream);

VAR     k : CARDINAL;
        nexterr : Status;

BEGIN   WITH s^ DO
          Done := TRUE;
          IF Created OR Modified THEN
            SetStream(s,Eby,Ebk);
            IF (Eby <> 0) AND Write THEN 
              IF NOT Record THEN PutChar(s,32C); END;
              FOR k := Cby TO 511 DO PutChar(s,0C) END;
              END;
            END;
          nexterr := Close(H,(*NOT Created OR Modified*) TRUE);
          IF SysErr = Ok THEN SysErr := nexterr END;
          END;
        DISPOSE(s);
        IF SysErr <> Ok THEN Done := FALSE END;
END CloseStream;

PROCEDURE  ReleaseStream( VAR s : Stream);

BEGIN   SysErr := Close(s^.H,FALSE);
        DISPOSE(s);
        Done := TRUE;
END ReleaseStream;

PROCEDURE  GetPos( VAR s : Stream; VAR pos : ADDRESS);

BEGIN   WITH s^ DO 
          Done := NOT Record;
          pos := ADDRESS(Cby) + 512*ADDRESS(Cbk); 
          IF PutBack THEN DEC(pos) END;
          END;
END GetPos;

PROCEDURE  SetPos( VAR s : Stream; pos : ADDRESS);

VAR     rs : Status;
        bc, k : CARDINAL;

BEGIN   Done := NOT s^.Record;
        IF Done AND (pos = 0FFFFFFFFH) THEN
          IF s^.Created THEN 
            SetStream(s,s^.Eby,s^.Ebk);
            s^.PutBack := FALSE;
            RETURN;
            END;
          rs := Size(s^.H,bc);
          IF bc = 0 THEN RETURN (* We are already at end *) END;
          SetStream(s,0,bc-1);
          k := 511;
          WHILE (k > 0) AND (s^.B[k] = 0C) DO DEC(k) END;
          IF s^.B[k] <> 32C
            THEN SetStream(s,0,bc)
            ELSE s^.Cby := k;
            END;
          SetEnd(s);
          s^.PutBack := FALSE;
          RETURN;
          END;
        IF Done
          THEN SetStream(s,CARDINAL(pos MOD 512),CARDINAL(pos DIV 512)); 
               s^.PutBack := FALSE;
          END;
END SetPos;

PROCEDURE  ResetStream( VAR s : Stream);

BEGIN   SetPos(s,0);
END ResetStream;

PROCEDURE  EndStream( VAR s : Stream) : BOOLEAN;

VAR     c, e : ADDRESS;

BEGIN   WITH s^ DO 
          c := ADDRESS(Cby)+512*ADDRESS(Cbk);
          e := ADDRESS(Eby)+512*ADDRESS(Ebk);
          IF PutBack THEN DEC(c) END;
          RETURN c >= e;
          END;
END EndStream;

PROCEDURE  ReadChar( VAR s : Stream; VAR ch : CHAR);

VAR     nx : CHAR;

BEGIN   IF s^.Record AND s^.Write THEN Done := FALSE; RETURN END;
        Done := TRUE;
        IF NOT EndStream(s)
          THEN GetChar(s,ch);
               IF (ch = EOL) AND NOT EndStream(s) THEN 
                 GetChar(s,nx);
                 IF nx <> 12C THEN PutBackChar(s,nx) END;
                 END;
               IF NOT EndStream(s) THEN
                 GetChar(s,nx);
                 IF nx = 32C THEN SetEnd(s) ELSE PutBackChar(s,nx) END;
                 END;
          ELSE ch := 0C;
          END;
END ReadChar;

PROCEDURE  ReadWord( VAR s : Stream; VAR w : WORD);

VAR     k : CtoW;

BEGIN   IF s^.Record AND s^.Write THEN Done := FALSE; RETURN END;
        Done := TRUE;
        Even(s);
        GetChar(s,k.hi);
        GetChar(s,k.lo);
        w := k.w;
END ReadWord;

PROCEDURE  ReadLine( VAR s : Stream; VAR l : ARRAY OF CHAR);

VAR     i : CARDINAL;
        ch : CHAR;

BEGIN   IF s^.Record AND s^.Write THEN Done := FALSE; RETURN END;
        i := 0;
        LOOP ReadChar(s,ch);
             IF (ch = EOL) OR (ch = 0C) THEN 
               IF i <= HIGH(l) THEN l[i] := 0C END;
               EXIT;
               END;
             IF i <= HIGH(l) THEN 
               l[i] := ch; 
               INC(i);
               END;
             END;
END ReadLine;

PROCEDURE  WriteChar( VAR s : Stream; ch : CHAR);

BEGIN   IF s^.Record AND NOT s^.Write THEN Done := FALSE; RETURN END;
        Done := TRUE;
        PutChar(s,ch);
        IF ch = EOL THEN PutChar(s,12C) END;
        SetEnd(s);
END WriteChar;

PROCEDURE  WriteWord( VAR s : Stream; w : WORD);

BEGIN   ModifyWord(s,w);
        SetEnd(s);
END WriteWord;

PROCEDURE  ModifyWord( VAR s : Stream; w : WORD);

VAR     k : CtoW;

BEGIN   IF s^.Record AND NOT s^.Write THEN Done := FALSE; RETURN END;
        Done := TRUE;
        Even(s);
        k.w := w;
        PutChar(s,k.hi);
        PutChar(s,k.lo);
END ModifyWord;

PROCEDURE  WriteLine( VAR s : Stream; VAR l : ARRAY OF CHAR);

VAR     i : CARDINAL;

BEGIN   IF s^.Record AND NOT s^.Write THEN Done := FALSE; RETURN END;
        i := 0;
        WHILE (i <= HIGH(l)) AND (l[i] <> 0C) DO
          WriteChar(s,l[i]);
          INC(i);
          END;
        WriteChar(s,EOL);
END WriteLine;

PROCEDURE  Even ( VAR s : Stream);

VAR     ch : CHAR;

BEGIN   WITH s^ DO IF ODD(Cby) THEN GetChar(s,ch) END; END;
END Even;

PROCEDURE  SetEnd( VAR s : Stream);

BEGIN   WITH s^ DO
          Ebk := Cbk;
          Eby := Cby;
          END;
END SetEnd;

PROCEDURE  PutBackChar( VAR s : Stream; ch : CHAR);

BEGIN   WITH s^ DO
          PutBack := TRUE;
          BackChar := ch;
          END;
END PutBackChar;

PROCEDURE  GetChar( VAR s : Stream; VAR ch : CHAR);

BEGIN   WITH s^ DO
          IF PutBack THEN
            ch := BackChar;
            PutBack := FALSE;
            RETURN
            END;
          ch := B[Cby];
          INC(Cby);
          IF Record AND (ch = 0C) THEN
            ch := EOL;
            Cby := 512;
            END;
          IF Cby > 511 THEN SetStream(s,0,Cbk+1) END;
          END;
END GetChar;

PROCEDURE  PutChar( VAR s : Stream; ch : CHAR);

BEGIN   WITH s^ DO
          Dirty := TRUE;
          Modified := TRUE;
          B[Cby] := ch;
          INC(Cby);
          IF Cby > 511 THEN SetStream(s,0,Cbk+1) END;
          END;
END PutChar;

PROCEDURE  SetStream( VAR s : Stream; sby, sbk : CARDINAL);

VAR     k : CARDINAL;
        tr : BOOLEAN;

BEGIN   WITH s^ DO
          SysErr := Ok;
          IF (sbk <> Cbk) AND Dirty THEN
            IF NOT Record THEN SysErr := Seek(H,Cbk); END;
            IF SysErr <> Ok THEN Done := FALSE; RETURN END;
            tr := (Cbk=Ebk) OR ((Cbk+1=Ebk) AND (Eby = 0));
            SysErr := WriteBlock(H,B,tr); (* Trunc at eof *)
            IF SysErr <> Ok THEN Done := FALSE; RETURN END;
            END;
          IF (sbk <> Cbk) AND NOT (Record AND Write) THEN
            IF NOT Record THEN SysErr := Seek(H,sbk); END;
            IF SysErr <> Ok THEN Done := FALSE; RETURN END;
            B[0] := 0C;
            IF NOT (Record AND (sbk >= Ebk))
              THEN SysErr := ReadBlock(H,B) END;
            END;
          IF SysErr = EscKey THEN SysErr := Ok END;
          IF SysErr = EndFile THEN
            Ebk := sbk;
            Eby := 0;
            IF Record AND (B[0] <> 0C) THEN INC(Ebk) END;
            SysErr := Ok;
            END;
          IF SysErr <> Ok THEN Done := FALSE; RETURN END;
          Cbk := sbk;
          Cby := sby;
          Dirty := FALSE;
          END;
END SetStream;

BEGIN   CreateMode := FileMode(0B0H);
END R2Streams.
