IMPLEMENTATION MODULE R2Strings;

(*  Strings are truncated if they do not fit into the output parameter.
    All strings are assumed to be 0 offset.  Positions number from 0 to
    Length(s), 0 being before the first character, Length(s) being after
    the last.  *)

PROCEDURE  Copy ( VAR s, t : ARRAY OF CHAR);

VAR     i, k : CARDINAL;

BEGIN   k := HIGH(s);
        IF k > HIGH(t) THEN k := HIGH(t) END;
        FOR i := 0 TO k DO t[i] := s[i] END;
        FOR i := k+1 TO HIGH(t) DO t[i] := 0C END;
END Copy;

PROCEDURE  Compare( VAR a, b : ARRAY OF CHAR) : INTEGER;

  (*  returns a value < 0 if a < b, = 0 if a = b, > 0 if a > b  *)

VAR     i : CARDINAL;

BEGIN   i := 0;
        WHILE (i < HIGH(a)) AND (i < HIGH(b)) AND (a[i] = b[i])
              AND (a[i] <> 0C) DO INC(i) END;
        IF ((i = HIGH(a)) OR (i = HIGH(b))) AND (a[i] = b[i])
          THEN RETURN Length(a) - Length(b) END;
        RETURN ORD(a[i]) - ORD(b[i]);
END Compare;

PROCEDURE  Find( VAR pat, s : ARRAY OF CHAR) : CARDINAL;

(* Exact match is required *)

VAR     i, j : CARDINAL;

BEGIN   i := 0;
        WHILE (i <= HIGH(s)) AND (s[i] <> 0C) DO
          j := 0;
          WHILE (j <= HIGH(pat)) AND (pat[j] <> 0C) AND (s[i+j] = pat[j])
            DO INC(j) END;
          IF (j > HIGH(pat)) OR (pat[j] = 0C) THEN RETURN i END;
          INC(i);
          END;
        RETURN i;
END Find;

PROCEDURE  Substring( VAR s, t : ARRAY OF CHAR; pos, len : CARDINAL);

VAR     i : CARDINAL;

BEGIN   i := 0;
        WHILE (i <= HIGH(t)) AND (pos <= HIGH(s)) AND (len > 0) DO
          t[i] := s[pos];
          INC(pos);
          DEC(len);
          INC(i);
          END;
        IF i <= HIGH(t) THEN t[i] := 0C END;
END Substring;

PROCEDURE  Insert( VAR s, t : ARRAY OF CHAR; pos : CARDINAL);

VAR     i, j, k, n, m : CARDINAL;

BEGIN   n := Length(t);
        IF pos > n THEN RETURN END;
        k := Length(s);
        m := n + k;
        j := pos + k;
        IF m > HIGH

) THEN m := HIGH(t) END;
        FOR i := m TO j BY -1 DO t[i] := t[i-k] END;
        IF j > HIGH(t)+1 THEN j := HIGH(t)+1 END;
        IF j >= pos+1 THEN                              (* Cardinal junk *)
          FOR i := pos TO j-1 DO t[i] := s[i-pos] END;
          END;
END Insert;

PROCEDURE  Length( VAR s : ARRAY OF CHAR) : CARDINAL;

VAR     i : CARDINAL;

BEGIN   i := 0;
        WHILE (i <= HIGH(s)) AND (s[i] <> 0C) DO INC(i) END;
        RETURN i;
END Length;

PROCEDURE  Remove( VAR s : ARRAY OF CHAR; pos, len : CARDINAL);

VAR     i, j : CARDINAL;

BEGIN   j := Length(s);
        IF (pos >= j) THEN RETURN END;
        IF pos+len > j THEN len := j-pos END;
        j := j-len;
        IF j >= pos+1 THEN
          FOR i := pos TO j-1 DO s[i] := s[i+len] END;
          END;
        IF j <= HIGH(s) THEN s[j] := 0C END;
END Remove;

PROCEDURE  Concatenate( VAR s1, s2, t : ARRAY OF CHAR);

BEGIN   Copy(s1,t);
        Insert(s2,t,Length(s1));
END Concatenate;

PROCEDURE  Compress( VAR s : ARRAY OF CHAR);

(*  Remove all blanks from s *)

VAR     i, j : CARDINAL;

BEGIN   i := 0;
        j := 0;
        WHILE (i <= HIGH(s)) AND (s[i] <> 0C) DO
          IF s[i] > ' ' THEN
            s[j] := s[i];
            INC(j);
            END;
          INC(i);
          END;
        IF j <= HIGH(s) THEN s[j] := 0C END;
END Compress;

END R2Strings.
