IMPLEMENTATION MODULE R2System; 

(*$T-*)
(*$S-*)

(*  R2System.MOD -- Minimal support for M2-68K -- LUMOS runtime version
 
A simplified startup module derived from Smiler-2.  We assume that 
D6 and D7 are not used by the compiler, but instead contain the FirstFree
and StackLimit values respectively.  32 bit signed
arithmetic is not supported, neither are the process management
primitives.  All unsupported features cause an illegal instruction trap.
A7, D6 and D7 are set by the loader.  A display is created on A5 and
the stack is set up so that an Exit(Ok) is done on return from the program
module.  Stack checking compares the SP and D7, if insufficient space
remains the stack is automatically extended by calling ExpandStack and 
correcting D7.  The compiler generated halts (HALTX) cause an Exit with 
an appropriate status. *)

FROM SYSTEM IMPORT CODE, REGISTER, ADDRESS, PROCESS, SETREG;

(*$P-*)

PROCEDURE CASEX;

BEGIN   CODE(020137B);       (*       MOVEA.L (SP)+,A0      *)
        CODE(110220B);       (*       SUB.L   (A0),D0       *)
        CODE(130250B,4);     (*       CMP.L   4(A0),D0      *)
        CODE(061014B);       (*       BHI     ELSE          *)
        CODE(161500B);       (*       ASL.L   #1,D0         *)
        CODE(150250B,8);     (*       ADD.L   8(A0),D0      *)
        CODE(150360B,4000B); (*       ADDA.W  (A0,D0.L),A0  *)
        CODE(047320B);       (*       JMP     (A0)          *)
        CODE(150750B,8);     (* ELSE: ADDA.L  8(A0),A0      *)
        CODE(054610B);       (*       SUBQ.L  #4,A0         *)
        CODE(047320B);       (*       JMP     (A0)          *)
END CASEX;

(*$P-*)
 
PROCEDURE HALTX;             (*       Exit(Halt+D0)         *)

BEGIN   CODE(0D07CH,59);     (*       ADD.W   #59,D0        *)
        CODE(01F00H);        (*       MOVE.B  D0,-(SP)      *)
        CODE(07202H);        (*       MOVEQ.L #2,D1         *)
        CODE(04E4FH);        (*       TRAP    #15           *)
END HALTX;

(*$P-*)

PROCEDURE STACKTEST; (* Expand the stack if necessary and possible *)

BEGIN   CODE(0204FH);        (*       MOVE.L  SP,A0        *)
        CODE(041E8H,-128);   (*       LEA     -128(A0),A0  *)
        CODE(091C0H);        (*       SUB.L   D0,A0        *)
        CODE(0B1C7H);        (*       CMP.L   D7,A0        *)
        CODE(06222H);        (*       BHI     OK           *)
        CODE(03F00H);        (*       MOVE.W  D0,-(SP)     *)
        CODE(04267H);        (*       CLR.W   -(SP)        *)
        CODE(03F3CH,1);      (*       MOVE.W  #1,-(SP)     *)
        CODE(0323CH,41);     (*       MOVE.W  #41,D1       *)
        CODE(04E4FH);        (*       TRAP    #15          *)
        CODE(0321FH);        (*       MOVE.W  (SP)+,D1     *)
        CODE(0321FH);        (*       MOVE.W  (SP)+,D1     *)
        CODE(0301FH);        (*       MOVE.W  (SP)+,D0     *)
        CODE(04A41H);        (*       TST     D1           *)
        CODE(067DCH);        (*       BEQ     STACKTEST    *)
        CODE(01F3CH,62);     (*       MOVE.B  #62,-(SP)    *)
        CODE(0323CH,2);      (*       MOVE.W  #2,D1        *)
        CODE(04E4FH);        (*       TRAP    #15          *)
        CODE(047165B);       (* OK:   RTS                  *)
END STACKTEST;

PROCEDURE TERMINATE;
BEGIN   CODE(04AFCH);
END TERMINATE;

PROCEDURE NEWPROCESS (PROCESSCODE : PROC;
                      FWAWSP : ADDRESS; LENGTHWSP : ADDRESS;
                      VAR PROCESSDESCRIPTOR : PROCESS;
                      INITIALPRIO : ADDRESS);
BEGIN   CODE(04AFCH);
END NEWPROCESS;

PROCEDURE TRANSFER;
BEGIN   CODE(04AFCH);
END TRANSFER;

PROCEDURE IOTRANSFER;
BEGIN   CODE(04AFCH);
END IOTRANSFER;

PROCEDURE MODCALL;
BEGIN   CODE(04AFCH);
END MODCALL;

PROCEDURE SYSCALL; 
BEGIN   CODE(04AFCH);
END SYSCALL;

PROCEDURE MULU32     (MULTIPLICAN, MULTIPLIER : ADDRESS);
BEGIN
    CODE(044347B,07C00H);
    CODE(021056B,8); 
    CODE(022056B,12); 
    CODE(02601H,02801H,02A02H,04844H,04845H);
    CODE(0C2C2H,0C4C4H,0C6C5H,0C8C5H,04841H);
    CODE(0D242H,04285H,0D985H,0D243H,0D985H);
    CODE(04841H,04242H,04243H,04842H,04843H);
    CODE(0D483H,0D484H);
    CODE(026501B,12);
    CODE(026502B,8);
    CODE(046337B,03EH);
END MULU32;

PROCEDURE DIVU32     (DIVIDEND, DIVISOR       : ADDRESS);
BEGIN
    CODE(044347B,0F000H);
    CODE(021056B,12);
    CODE(020056B,8);
    CODE(04283H,082C0H,06800H,00010H,02401H);
    CODE(04241H,04841H,082C0H,03601H,03202H);
    CODE(082C0H,02001H,04841H,03203H,04841H);
    CODE(04240H,04840H);
    CODE(026501B,12);
    CODE(026500B,8);
    CODE(046337B,0FH);
END DIVU32;

PROCEDURE MULS32     (MULTIPLICAN, MULTIPLIER : ADDRESS);
BEGIN   CODE(04AFCH);
END MULS32;

PROCEDURE DIVS32     (DIVIDEND, DIVISOR       : ADDRESS);
BEGIN   CODE(04AFCH);
END DIVS32;

(* Floating point emulator (?) lifted from original SYSTEMX *)

(* Currently, FPE errors are reported as Range Errors  *)

CONST   MOVEQ = 070000B;

  PROCEDURE FADD       (ADDER,       ADDEND     : REAL);

  BEGIN
    CODE ( 48E7H, 9F00H);
    CODE ( 282EH, 000CH, 2C2EH, 0008H,0E39CH, 0E39EH);
    CODE ( 2004H, 9086H, 6402H,0C946H);
    CODE ( 1604H,0C63CH, 0001H, 0800H, 0000H, 6704H, 08C3H, 0001H);
    CODE (0E19CH, 4245H, 1A04H, 56C4H,0E29CH, 4204H,0E19EH, 4247H,
           1E06H, 56C6H,0E29EH, 4206H);
    CODE ( 3005H, 9047H,0B07CH, 0018H, 622EH,0E0AEH);
    CODE ( 0803H, 0001H, 660AH);
    CODE (0D886H, 6404H,0E294H, 5245H, 601CH);
    CODE ( 9886H, 6B18H, 6716H,0B8BCH, 0000H,0FFFFH, 6208H, 9A7CH,
           0010H, 4844H, 6B06H, 5345H,0E38CH, 6AFAH);
    CODE (0D8BCH, 0000H, 0080H, 6404H, 0E294H, 5245H);
    CODE ( 4204H, 4A84H,
           6602H, 4245H, 4A45H, 6E04H, 4284H, 6018H,0BA7CH, 00FFH,
           6D08H);
    CODE (MOVEQ + 4); HALTX;
    CODE (0E38CH, 1805H,0E09CH,0E213H,0E294H);
    CODE ( 2D44H, 000CH, 4CDFH, 00F9H);
  END FADD;

  (*$P- jumps to FADD *)
  PROCEDURE FSUB       (MINUEND,     SUBTRAHEND : REAL);

  BEGIN
    SETREG(8, FADD);
    CODE ( 086FH, 0007H, 0004H);  (* CHANGE SIGN OF SOURCE OPERAND *)
    CODE ( 4ED0H);   (* JUMP (A0) *)
  END FSUB;

  PROCEDURE FMUL       (MULTIPLICAN, MULTIPLIER : REAL);

  BEGIN
    CODE ( 48E7H,0CF00H);
    CODE ( 282EH, 000CH, 2C2EH, 0008H);
    CODE ( 2E04H,0BD87H);
    CODE ( 4844H, 3A04H,0CA7CH, 7F80H, 676CH,0EE4DH,
          0C87CH, 007FH,
           887CH, 0080H, 4846H, 3E06H,0CE7CH, 7F80H, 6758H,0EE4FH,
          0CC7CH, 007FH, 8C7CH, 0080H);
    CODE (0DA47H, 9A7CH, 007EH, 6D46H);
    CODE ( 2004H, 4840H, 4846H,0C0C6H, 4240H, 4840H);
    CODE ( 3204H,0C2C6H,0D081H);
    CODE ( 2204H, 4841H, 4846H,0C2C6H,0D081H);
    CODE (0C8C6H, 4844H,0D880H);
    CODE ( 6B04H,0E38CH, 5345H);
    CODE (0D8BCH, 0000H, 0080H, 6404H,0E294H, 5245H);
    CODE ( 4A45H, 6B0EH,0BA7CH, 00FFH, 6D0CH);
    CODE (MOVEQ + 4); HALTX;
    CODE ( 4284H, 600AH);
    CODE (0E38CH, 1805H,0E09CH,0E397H,0E294H);
    CODE ( 2D44H, 000CH, 4CDFH, 00F3H);
  END FMUL;

  PROCEDURE FDIV       (DIVIDEND,    DIVISOR    : REAL);

  BEGIN
    CODE ( 48E7H, 8F00H);
    CODE ( 282EH, 000CH, 2C2EH, 0008H);
    CODE ( 2E04H,0BD87H);
    CODE (0E38EH,0E19EH, 4247H, 1E06H, 675EH, 1C3CH, 0001H,0E29EH,
          0E38CH,0E19CH, 4245H, 1A04H, 6756H, 183CH, 0001H,0E29CH);
    CODE ( 9A47H,0DA7CH, 007FH);
    CODE ( 4846H,0E28CH, 88C6H, 3004H, 4244H, 88C6H, 4844H, 3800H);
    CODE ( 2006H, 4840H,0C0C4H,0E288H, 80C6H,0C0BCH, 0000H,0FFFFH,
          0E388H, 4844H, 9880H);
    CODE ( 6B04H,0E38CH, 5345H);
    CODE (0D8BCH, 0000H, 0080H, 6404H,0E294H, 5245H);
    CODE ( 4A45H, 6B0EH,0BA7CH, 00FFH, 6D0CH);
    CODE (MOVEQ + 4); HALTX;
    CODE ( 4284H, 600AH);
    CODE (0E39CH, 1805H,0E09CH,0E397H,0E294H);
    CODE ( 2D44H, 000CH, 4CDFH, 00F1H);
  END FDIV;

  PROCEDURE FCMP       (DESTINATION, SOURCE     : REAL);

  (* RESULT IS SET IN CONDITION-CODE-REGISTER N Z V BITS *)

  BEGIN
    CODE ( 48E7H,0E000H);
    CODE ( 202EH, 000CH, 2400H, 6A06H, 0A80H, 7FFFH,0FFFFH,
          0C4BCH, 7F80H, 0000H, 6602H, 4280H);
    CODE ( 222EH, 0008H, 2401H, 6A06H, 0A81H, 7FFFH,0FFFFH,
          0C4BCH, 7F80H, 0000H, 6602H, 4281H);
    CODE (0B081H);
    CODE ( 4CDFH, 0007H);
  END FCMP;

  PROCEDURE FTST       (TOTEST                  : REAL);

  (* RESULT IS SET IN CONDITION-CODE-REGISTER Z N BITS *)

  BEGIN
    CODE ( 48E7H,0A000H);
    CODE ( 202EH, 0008H);
    CODE ( 2400H,0C4BCH, 7F80H, 0000H, 6602H, 4280H);
    CODE ( 4A80H);
    CODE ( 4CDFH, 0005H);
  END FTST;

  PROCEDURE FLOATX     (TOCONVERT               : ADDRESS);

  BEGIN
    CODE ( 48E7H,0C000H);
    CODE ( 202EH, 0008H, 672AH, 323CH, 009EH);
    CODE (0B0BCH, 0000H,0FFFFH, 6206H, 4840H, 927CH, 0010H);
    CODE (0B0BCH, 00FFH,0FFFFH, 6204H,0E188H, 5141H);
    CODE (0E388H, 55C9H,0FFFCH);
    CODE ( 1001H,0E098H,0E288H);
    CODE ( 2D40H, 0008H, 4CDFH, 0003H);
  END FLOATX;

  PROCEDURE TRUNCX     (TOCONVERT               : REAL);

  BEGIN
    CODE ( 48E7H,0C000H);
    CODE ( 202EH, 0008H);
    CODE ( 6A0AH,0C0BCH, 7F80H, 0000H, 6716H, 6020H);
    CODE (0E388H,0E198H, 4241H, 1200H, 103CH, 0001H,0E298H);
    CODE ( 927CH, 007FH, 6C04H, 4280H, 601AH);
    CODE ( 927CH, 001FH, 4441H, 6C08H);
    CODE (MOVEQ + 4); HALTX;
    CODE (0E2A8H);
    CODE (0B0BCH, 0000H,0FFFFH, 62EEH);
    CODE ( 2D40H, 0008H, 4CDFH, 0003H);
  END TRUNCX;

(* end of floating point emulator *)

(*$P-*)

BEGIN   CODE(117374B,36);      (*    SUBA.W   #36,SP  ; Create display *)
        CODE(025117B);         (*    MOVEA.L  SP,A5   ; Save PD        *)
        CODE(044172B,4);       (*    PEA      4(PC)   ; Create return  *)
        CODE(060006B);         (*    BRA      +6      ; Leave SYSTEMX  *)
        CODE(071002B);         (*    MOVEQ    #2,D1   ; Exit           *)
        CODE(041147B);         (*    CLR.W    -(SP)   ; Ok = 0         *)
        CODE(047117B);         (*    TRAP     #15     ; Exit(Ok)       *)
END R2System.
