IMPLEMENTATION MODULE L2Cache;

(* An LRU cache intended (initially) for cacheing directory blocks
   as an enhancement to the file system.
*)                                  

FROM SYSTEM IMPORT ADR, ADDRESS;
FROM L2BlockMove IMPORT FastCopy;
FROM L2Device IMPORT ReadBlock;       
FROM L2Errors IMPORT Status;

CONST   maxcache = 61;  (* largest possible on this machine *)

TYPE    CCB = RECORD Count, Stamp : ADDRESS;
                     Vol, Sec : CARDINAL;
              END;
        Block = ARRAY [0..511] OF CHAR;

VAR     Table : ARRAY [0..maxcache] OF CCB;
        Cache : ARRAY [0..maxcache] OF Block;
        Hits, Ops : ADDRESS;

PROCEDURE  ReadCache( VAR dvn : ARRAY OF CHAR; vol, sec : CARDINAL;
                      VAR block : ARRAY OF CHAR) : Status;

(* assumes that dvn and vol are consistent - no attempt is made to
   validate this assertion.
*)

VAR     i, m : CARDINAL;
        s : ADDRESS;
        rs : Status;

BEGIN   s := Table[0].Stamp;
        i := 0; m := 0;
        INC(Ops);
        LOOP IF i > maxcache THEN 
               i := m;
               WITH Table[i] DO
                 Count := 0;
                 Stamp := 0;
                 Vol := vol;
                 Sec := sec;
                 END;
               rs := ReadBlock(dvn,sec,Cache[i]);
               IF rs <> Ok THEN RETURN rs END;
               EXIT;
               END;
             WITH Table[i] DO
               IF (Sec = sec) AND (Vol = vol) AND (Stamp <> 0) THEN
                 INC(Hits);
                 EXIT 
                 END;
               IF Stamp < s THEN
                 m := i;
                 s := Stamp;
                 END;
               END;
             INC(i);
             END;
        INC(Table[i].Count);
        Table[i].Stamp := Ops;
        FastCopy(ADR(Cache[i]),ADR(block));  (* copies 512 bytes *)
        RETURN Ok;
END ReadCache;

PROCEDURE  PurgeBlock( vol, sec : CARDINAL);

VAR     i : CARDINAL;

BEGIN   FOR i := 0 TO maxcache DO
          WITH Table[i] DO
            IF (Sec = sec) AND (Vol = vol) AND (Stamp <> 0) THEN
              Count := 0;
              Stamp := 0;
              END;
            END;
          END;
END PurgeBlock;

PROCEDURE  PurgeVolume( vol : CARDINAL);

VAR     i : CARDINAL;

BEGIN   FOR i := 0 TO maxcache DO
          WITH Table[i] DO
            IF (Vol = vol) AND (Stamp <> 0) THEN
              Count := 0;
              Stamp := 0;
              END;
            END;
          END;
END PurgeVolume;           

PROCEDURE  ViewCache( i : CARDINAL; VAR vol, sec : CARDINAL;
                      VAR count, stamp : ADDRESS) : Status;

BEGIN   IF i > maxcache THEN RETURN NoSuch END;
        WITH Table[i] DO
          vol := Vol;
          sec := Sec;
          count := Count;
          stamp := Stamp;
          END;
        RETURN Ok;
END ViewCache;

PROCEDURE  CacheStats( VAR ops, hits : ADDRESS);

BEGIN   ops := Ops;
        hits := Hits;
END CacheStats;

PROCEDURE  InitCache;

VAR     i : CARDINAL;

BEGIN   FOR i := 0 TO maxcache DO 
          Table[i].Count := 0;
          Table[i].Stamp := 0;
          END;
        Hits := 0; Ops := 0;
END InitCache;

BEGIN   InitCache;
END L2Cache.
