IMPLEMENTATION MODULE L2Calendar[4];

(* Supports the NS58167 real time clock *)

(*  Implemented independently of L2Clock so that the driver can be excluded
    if necessary without impacting on any other definition modules.  (Like 
    the other device drivers.) *)

FROM L2Clock IMPORT Time;
FROM L2Exceptions IMPORT Establish;

CONST   Nmon = 7;       (* slots in Ndate vector *)
        Nday = 6;
        Ndow = 5;
        Nhour = 4;
        Nmin = 3;
        Nsec = 2;
        Ncsec = 1;
        CMSize = 0EEH;
        Bias = 1;       (* Magic number to make day of week work until 2099 *)

VAR     Ndate : ARRAY [1..7] OF CARDINAL;
        CMOS [0FF8000H] : ARRAY [0..CMSize+1] OF CHAR;
        N58167 [0FFC280H] : ARRAY [0..31] OF CARDINAL;
        CMonth, CYear : CARDINAL;
        RTCFlag : BOOLEAN;
        Cml : ARRAY [1..12] OF CARDINAL;

PROCEDURE  ReadClock( VAR t : Time);

BEGIN   ReadCMOS;
        ReadChip;
        CMonth := Ndate[Nmon];
        IF Ndate[Nmon] < CMonth 
          THEN INC(CYear);
               RTCFlag := FALSE;
          ELSE IF (Ndate[Nday] = 28) AND (CMonth = 2)
                 THEN IF (CYear MOD 4) = 0
                        THEN IF RTCFlag THEN INC(Ndate[Nday]) END;
                        ELSE RTCFlag := FALSE;
                        END;
                 ELSE RTCFlag := FALSE;
                 END;
          END;
        SetCMOS;
        WITH t DO
          csec := Ndate[Ncsec] + 100*Ndate[Nsec];
          min := Ndate[Nmin] + 60*Ndate[Nhour];
          day := (CYear-1980)*512 + (Ndate[Nmon]-1)*32 + Ndate[Nday];
          END;
END ReadClock;

PROCEDURE  SetClock( VAR t : Time);

BEGIN   WITH t DO
          CYear := (day DIV 512) + 1980;
          Ndate[Nmon] := ((day DIV 32) MOD 16) + 1;
          Ndate[Nday] := day MOD 32;
          Ndate[Nhour] := min DIV 60;
          Ndate[Nmin] := min MOD 60;
          Ndate[Nsec] := csec DIV 100;
          Ndate[Ncsec] := csec MOD 100;
          END;
        SetDOW;
        RTCFlag := FALSE;
        IF (Ndate[Nday] = 29) AND (Ndate[Nmon] = 2) 
          THEN Ndate[Nday] := 28;
               RTCFlag := (CYear MOD 4) = 0;
          END;
        WriteChip;
        CMonth := Ndate[Nmon];
        SetCMOS;
END SetClock;

PROCEDURE  SetDOW;

VAR     x, y : CARDINAL;

BEGIN   y := CYear - 1980;
        x := y*365 + ((y+3) DIV 4);
        IF ((y MOD 4) = 0) AND (Ndate[Nmon] > 2) THEN INC(x) END;
        INC(x,Cml[Ndate[Nmon]]+Ndate[Nday]+Bias);
        Ndate[Ndow] := (x MOD 7) + 1;
END SetDOW;

PROCEDURE  CalInt( junk : CARDINAL);  (* just in case *)

BEGIN   N58167[11H] := 0;     (* Clear interrupt enable mask *)
        junk := N58167[10H];  (* reset interrupt *)
END CalInt;

PROCEDURE  ReadChip;

VAR     i : CARDINAL;

BEGIN   FOR i := 1 TO 7 DO
          WHILE ODD(N58167[20]) DO END;
          Ndate[i] := FromBCD(N58167[i]);
          END;
END ReadChip;

PROCEDURE  WriteChip;

VAR     i : CARDINAL;

BEGIN   FOR i := 1 TO 7 DO
          N58167[i] := ToBCD(Ndate[i]);
          END;
END WriteChip;

PROCEDURE  ToBCD( c : CARDINAL) : CARDINAL;

BEGIN   c := c MOD 100;
        RETURN (c MOD 10) + 16*(c DIV 10);
END ToBCD;

PROCEDURE  FromBCD( b : CARDINAL) : CARDINAL;

BEGIN   b := b MOD 256;
        RETURN (b MOD 16) + 10*(b DIV 16);
END FromBCD;

PROCEDURE  ReadCMOS;

BEGIN   CYear := ORD(CMOS[42])*256 + ORD(CMOS[43]);
        CMonth := ORD(CMOS[44]);
        RTCFlag := ORD(CMOS[45]) = 1;
END ReadCMOS;

PROCEDURE  SetCMOS;

VAR     i, sum : CARDINAL;

BEGIN   CMOS[42] := CHR(CYear DIV 256);
        CMOS[43] := CHR(CYear MOD 256);
        CMOS[44] := CHR(CMonth);
        CMOS[45] := CHR(ORD(RTCFlag));
        sum := 0;
        FOR i := 0 TO CMSize DO INC(sum,ORD(CMOS[i])) END;
        CMOS[CMSize+1] := CHR(sum MOD 256);
END SetCMOS;

BEGIN   Cml[1]  :=   0;
        Cml[2]  :=  31;
        Cml[3]  :=  59;
        Cml[4]  :=  90;
        Cml[5]  := 120;
        Cml[6]  := 151;
        Cml[7]  := 181;
        Cml[8]  := 212;
        Cml[9]  := 243;
        Cml[10] := 273;
        Cml[11] := 304;
        Cml[12] := 334;
        Establish(44H,CalInt,0);
END L2Calendar.
