IMPLEMENTATION MODULE L2Console;

FROM SYSTEM IMPORT ADDRESS;

(*  A special synchronous driver for Serial port 0.  The driver 
    is designed to be used for debugging and panics.  The routine
    Init must be called before any of the other routines.  The
    driver sets a default baud rate. *)

VAR     Port0 [0FFC300H] : RECORD MRA, CSRA, CRA, TRBA, 
                                  ACR, IMR, CTUR, CTLR,
                                  MRB, CSRB, CRB, TRBB,
                                  IVR, OPCR, OPRBS, OPRBC : CARDINAL;
                           END;

PROCEDURE  Init;

VAR     i : CARDINAL;

BEGIN   FOR i := 1 TO 5000 DO END;
        WITH Port0 DO
          ACR := 80H;   (* Baud set 1        *)
          IMR := 0; 
          OPCR := 0;    (* No interrupts     *)
          OPRBS := 0FH; 
          OPRBC := 0F0H; 
          CRA := 10H;   (* MR1               *)
          MRA := 13H;   (* 8 bit, no parity  *)
          MRA := 7;     (* 1 stop bit        *)
          CSRA := 0CCH; (* 19.2 kBaud        *)
          CRA := 5;     (* Enable port       *)
        END;
END Init;

PROCEDURE  Write( ch : CHAR );

BEGIN   WHILE NOT (3 IN BITSET(Port0.CSRA)) DO END;
        Port0.TRBA := ORD(ch);
END Write;

PROCEDURE  WriteString( VAR s : ARRAY OF CHAR );

VAR     i : CARDINAL;

BEGIN   i := 0;
        WHILE (i <= HIGH(s)) AND (s[i] <> 0C) DO
          Write(s[i]);
          INC(i);
        END;
END WriteString;

PROCEDURE  WriteLn;

BEGIN   Write(CHR(13)); Write(CHR(10));
END WriteLn;

PROCEDURE  Read( VAR ch : CHAR );  (* Not exported *)

BEGIN   WHILE NOT (0 IN BITSET(Port0.CSRA)) DO END;
        ch := CHR(Port0.TRBA);
        IF ch >= ' ' THEN Write(ch) END;
END Read;

PROCEDURE  ReadString( VAR s : ARRAY OF CHAR );

(*  Reads up to a <RETURN>, processes DEL and ignores control chars *)

VAR     i : CARDINAL;

BEGIN   i := 0;
        LOOP Read(s[i]);
             IF s[i] = 177C THEN
               IF i > 0 THEN 
                 Write(10C); Write(' '); Write(10C);
                 DEC(i); 
                 END
             ELSIF s[i] = 15C THEN
               s[i] := 0C;
               EXIT;
             ELSIF s[i] < ' ' THEN (* ignore it *)
             ELSE INC(i) END;
             IF i > HIGH(s) THEN EXIT END;
             END;
        WriteLn;
END ReadString;

PROCEDURE  WriteCard( c, w : CARDINAL);      (* in Decimal *)

BEGIN   WriteAddr(ADDRESS(c),10,w);
END WriteCard;

PROCEDURE  WriteAddr( a : ADDRESS; r, w : CARDINAL);

(* Handles up to base 36 *)

VAR     s : ARRAY [1..100] OF CHAR;
        i, c : CARDINAL;

BEGIN   i := 1;
        WHILE (( a > 0 ) OR (w > 0)) AND (i < 100)  DO
          c := CARDINAL(a MOD ADDRESS(r));
          IF c < 10 THEN c := c + 48 ELSE c := c + 55 END;
          s[i] := CHR(c);
          IF (a = 0) AND (i > 1) THEN s[i] := ' ' END;
          a := a DIV ADDRESS(r);
          INC(i); 
          IF w > 0 THEN DEC(w) END;
          END;
        REPEAT DEC(i);
               Write(s[i]);
        UNTIL i = 1;
END WriteAddr;

END L2Console.
