IMPLEMENTATION MODULE L2Device;      (* The device independent i/o switch *)

FROM L2Errors IMPORT Status;
IMPORT L2Terminal;                (* T *)
IMPORT L2Floppy;                  (* F *)
IMPORT L2Printer;                 (* P *)
IMPORT L2Message;                 (* M *)
IMPORT L2Winchester;              (* W *)
IMPORT L2Cartridge;               (* C *)
(*IMPORT L2Omninet;                 (* O *)*)
IMPORT L2RamDisk;                 (* R *)

PROCEDURE  ReadBlock( VAR dn : ARRAY OF CHAR; blocknum : CARDINAL; 
                      VAR b : ARRAY OF CHAR ) : Status;

VAR     unit : CARDINAL;
        rs : Status;

BEGIN   IF dn[0] <> '#' THEN RETURN BadName END;
        rs := GetUnit(dn,unit);
        IF rs <> Ok THEN RETURN rs END;
        CASE dn[1] OF
          'T' : RETURN L2Terminal.ReadString(unit,b)                 |
          'F' : RETURN L2Floppy.ReadBlock(unit,blocknum,b)           |
          'N' : RETURN EndFile                                       |
          'M' : IF unit = 0 THEN RETURN L2Message.Receive(b);
                  ELSE RETURN NoFunction END                         |
          'W' : RETURN L2Winchester.ReadBlock(unit,blocknum,b);      |
          'R' : IF unit = 0 
                  THEN RETURN L2RamDisk.ReadBlock(blocknum,b)
                  ELSE RETURN NoUnit END;                            |
          'C' : IF unit = 0 
                  THEN RETURN L2Cartridge.ReadBlock(b);
                  ELSE RETURN NoUnit END;                            |
          'P' : RETURN NoFunction                                    
       (*   'O' : IF unit = 1
                 THEN RETURN L2Omninet.ReadBlock(b)
                 ELSE RETURN NoUnit
                 END;*)
          ELSE RETURN NoDevice
          END;
        RETURN Ok
END ReadBlock;

PROCEDURE  WriteBlock( VAR dn : ARRAY OF CHAR; blocknum : CARDINAL; 
                      VAR b : ARRAY OF CHAR ) : Status;

VAR     unit : CARDINAL;
        rs : Status;

BEGIN   IF dn[0] <> '#' THEN RETURN BadName END;
        rs := GetUnit(dn,unit);
        IF rs <> Ok THEN RETURN rs END;
        CASE dn[1] OF
          'T' : RETURN L2Terminal.WriteString(unit,b)                |
          'F' : RETURN L2Floppy.WriteBlock(unit,blocknum,b)          |
          'N' : RETURN Ok                                            |
          'M' : IF unit = 0 THEN RETURN NoFunction
                  ELSE RETURN L2Message.Send(unit,b) END             |
          'W' : RETURN L2Winchester.WriteBlock(unit,blocknum,b)      |
          'R' : IF unit = 0 
                  THEN RETURN L2RamDisk.WriteBlock(blocknum,b);
                  ELSE RETURN NoUnit END;                            |
          'C' : IF unit = 0
                  THEN RETURN L2Cartridge.WriteBlock(b);
                  ELSE RETURN NoUnit END;                            |
          'P' : IF unit = 0 THEN L2Printer.PrintString(b) 
                  ELSE RETURN NoUnit END                             
        (*  'O' : IF unit = 1
                 THEN RETURN L2Omninet.WriteBlock(b)
                 ELSE RETURN NoUnit
                 END; *)
          ELSE RETURN NoDevice
          END;
        RETURN Ok
END WriteBlock;

PROCEDURE  Flush( VAR dn : ARRAY OF CHAR ) : Status;

VAR     unit : CARDINAL;
        rs : Status;

BEGIN   IF dn[0] <> '#' THEN RETURN BadName END;
        rs := GetUnit(dn,unit);
        IF rs <> Ok THEN RETURN rs END;
        CASE dn[1] OF
          'T' : RETURN NoFunction                                    |
          'F' : RETURN L2Floppy.Flush(unit)                          |
          'N' : RETURN Ok                                            |
          'M' : IF unit = 0 THEN RETURN NoFunction 
                  ELSE RETURN L2Message.Flush(unit) END              |
          'W' : RETURN L2Winchester.Flush(unit);                     |
          'R' : RETURN NoFunction                                    |
          'C' : IF unit = 0 THEN RETURN L2Cartridge.Flush()
                  ELSE RETURN NoUnit END;                            |
          'P' : RETURN NoFunction                                    
        (*  'O' : RETURN NoFunction  *)
          ELSE RETURN NoDevice
          END;
        RETURN Ok
END Flush;

PROCEDURE  FormatTrack( VAR dn : ARRAY OF CHAR; track : CARDINAL) : Status;

VAR     unit : CARDINAL;
        rs : Status;

BEGIN   IF dn[0] <> '#' THEN RETURN BadName END;
        rs := GetUnit(dn,unit);
        IF rs <> Ok THEN RETURN rs END;
        CASE dn[1] OF 
          'T' : RETURN NoFunction                                     |
          'F' : RETURN L2Floppy.FormatTrack(unit,track)               |
          'N' : RETURN NoFunction                                     |
          'M' : RETURN NoFunction                                     |
          'W' : RETURN NoFunction                                     |
          'R' : RETURN NoFunction                                     |
          'C' : RETURN L2Cartridge.Format()                           |
          'P' : RETURN NoFunction                                     
        (*'O' : RETURN NoFunction *)
          ELSE RETURN NoDevice
          END;
        RETURN Ok
END FormatTrack;

PROCEDURE  GetDevAttr( VAR dn : ARRAY OF CHAR; VAR devsize : CARDINAL;
                       VAR attr : DevAttrSet ) : Status;

VAR     unit : CARDINAL;
        rs : Status;

BEGIN   IF dn[0] <> '#' THEN RETURN BadName END;
        rs := GetUnit(dn,unit);
        IF rs <> Ok THEN RETURN rs END;
        devsize := 0;
        CASE dn[1] OF 
          'T' : attr := DevAttrSet {RD,WR}                            |
          'F' : attr := DevAttrSet {FMT,RD,WR,FLSH,BLK};
                devsize := 1280                                       |
          'N' : attr := DevAttrSet {RD,WR,FLSH}                       |
          'M' : IF unit = 0 
                  THEN attr := DevAttrSet {RD}
                  ELSE attr := DevAttrSet {WR,FLSH} END               |
          'W' : attr := DevAttrSet {RD,WR,FLSH,BLK};
                RETURN L2Winchester.SectorCount(unit,devsize);        |
          'R' : attr := DevAttrSet {RD,WR,BLK};
                devsize := L2RamDisk.SectorCount;                     |
          'C' : attr := DevAttrSet {RD,WR,FLSH,FMT}                   |
          'P' : attr := DevAttrSet {WR}                               
         (* 'O' : attr := DevAttrSet {RD,WR}  *)
          ELSE RETURN NoDevice
          END;
        RETURN Ok
END GetDevAttr;

PROCEDURE  GetDevParams( VAR dn : ARRAY OF CHAR; p : CARDINAL;
                         VAR v : CARDINAL ) : Status;

VAR     unit : CARDINAL;
        rs : Status;

BEGIN   IF dn[0] <> '#' THEN RETURN BadName END;
        rs := GetUnit(dn,unit);
        IF rs <> Ok THEN RETURN rs END;
        IF (dn[1] = 'T') AND (p = 1) 
          THEN RETURN L2Terminal.GetBaud(unit,v)
  (*      ELSIF (dn[1] = 'O') AND (unit=1) AND (p=1)
          THEN v := L2Omninet.NodeNumber; RETURN Ok;  *)
          ELSE RETURN NoFunction; (* for now *)
          END;
END GetDevParams;

PROCEDURE  SetDevParams( VAR dn : ARRAY OF CHAR; p, v : CARDINAL) : Status;

VAR     unit : CARDINAL;
        rs : Status;

BEGIN   IF dn[0] <> '#' THEN RETURN BadName END;
        rs := GetUnit(dn,unit);
        IF rs <> Ok THEN RETURN rs END;
        IF (dn[1] = 'T') AND (p = 1)
          THEN RETURN L2Terminal.SetBaud(unit,v)
          ELSE RETURN NoFunction; (* for now *)
          END;
END SetDevParams;

PROCEDURE  GetUnit( VAR dn : ARRAY OF CHAR; VAR un : CARDINAL) : Status;

VAR     i, j : CARDINAL;          (* devices must have unique name strings *)

BEGIN   i := 2;
        j := 0;
        WHILE (j <= HIGH(dn)) AND (dn[j] <> 0C) AND (dn[j] <> '-') 
          DO INC(j) END;
        DEC(j);
        IF (j < 2) OR (j > 3) THEN RETURN BadName END;
        un := 0;
        WHILE (i <= j) AND (dn[i] >= '0') AND (dn[i] <= '9') DO
          un := 10*un + ORD(dn[i]) - 48;
          INC(i);
          END;
        IF i <> j+1 THEN RETURN BadName END;
        IF (j = 3) AND (un < 10) THEN RETURN BadName END;
        RETURN Ok;
END GetUnit;

END L2Device.
