IMPLEMENTATION MODULE L2Exceptions[4];

FROM L2Console IMPORT Init, WriteString, WriteLn, ReadString,
                      WriteAddr, WriteCard;
FROM L2Errors IMPORT Status, GetMessage;
FROM L2Machine IMPORT MoveToSR, S7, JSR, JMP, MoveStA0, RefWA0D0, IncSp,
                      SaveRegs, SaveAll, RestRegs, RestAll, RTE, Link, UnLink;
FROM SYSTEM IMPORT CODE, ADDRESS, REGISTER, SETREG, ADR;

(*$S-*)  (* If the stack is the problem, we don't want recursive errors ! *)

TYPE    VCB = RECORD  j : CARDINAL;
                      r : PROC;
                      v : CARDINAL;
                      H : Handler;
                      P : CARDINAL;
              END;

VAR     Vc : ARRAY [0..255] OF VCB;

PROCEDURE  Establish( vec : CARDINAL; h : Handler; param : CARDINAL);

BEGIN   WITH Vc[vec] DO
          j := JSR;
          r := Route;
          v := vec;
          H := h;
          P := param;
          Vectors[vec] := PROC(ADR(Vc[vec]));
          END;
END Establish;

(*$P-*)  (* A very special routine ! *)

PROCEDURE  Route;  (* The primary interrupt handler, it saves regs, and 
                      revectors to a specified handler. *)

VAR     v : CARDINAL;                   (* We must LINK to make this space   *)

BEGIN   CODE(SaveRegs,SaveAll);
        SETREG(13,KernelStackLimit);
        CODE(MoveStA0,56);              (* A0 now points to the v of the VCB *)
        CODE(RefWA0D0);                 (* D0 now contains v                 *)
        CODE(Link,-2);                  (* Create v                          *)
        v := CARDINAL(REGISTER(0));
        Vc[v].H(Vc[v].P);               (* Mike Herman eat your heart out    *)
        CODE(UnLink);
        CODE(RestRegs,RestAll);
        CODE(IncSp);                    (* Throw away the Return address     *)
        CODE(RTE);
END Route;

PROCEDURE  Assert( e : BOOLEAN; num : CARDINAL);

BEGIN   IF NOT e THEN
          Init;
          WriteLn;
          WriteString("PANIC -- Assertion #");
          WriteCard(num,1);
          WriteString(" failed");
          WriteLn;
          ReBoot;
          END;
END Assert;

PROCEDURE  AssertOk( st : Status; num : CARDINAL);

VAR     s : ARRAY [0..39] OF CHAR;

BEGIN   IF st <> Ok THEN
          Init;
          WriteLn;
          WriteString("PANIC -- Assertion #");
          WriteCard(num,1);
          WriteString(" failed, ");
          GetMessage(st,s);
          WriteString(s);
          WriteLn;
          ReBoot;
          END;
END AssertOk;

PROCEDURE  ReBoot;

BEGIN   Init;
        WriteLn;
        WriteString("Rebooting ... ");
        WriteLn;
        WriteLn;
        CODE(MoveToSR,S7);
        CODE(JMP,00FEH,0044H);  (* ReBoot the Stride via ROM (always mapped) *)
END ReBoot;

PROCEDURE  Nonsense( v : CARDINAL);

BEGIN   CODE(MoveToSR,S7);
        SETREG(13,ADDRESS(0));  (* Avoid further stack errors *)
        Init;
        WriteLn;
        WriteString('PANIC -- Unexpected interrupt # ');
        WriteCard(v,1);
        WriteLn;
        ReBoot;
END Nonsense;

VAR     i : CARDINAL;

BEGIN   FOR i := 2 TO 255 DO Establish(i,Nonsense,i) END;
        IntControl := { };
        KernelStackLimit := REGISTER(13);  (* Remember A5 for interrupts *)
END L2Exceptions.
