IMPLEMENTATION MODULE L2Kernel[4];

FROM L2Machine IMPORT MoveToSR, S0, MoveUSPA6, MoveA6USP, PushA6, PopA6,
                      PushW, PushA4, U0, RTE, MoveA6A7, MoveA7A6, JsrA0,
                      MoveUSPA0, MoveUSPA7, S4;
FROM L2Errors IMPORT Status;
FROM SYSTEM IMPORT ADDRESS, REGISTER, SETREG, ADR, CODE;
FROM L2Storage IMPORT ALLOCATE, DEALLOCATE;
FROM L2Clock IMPORT RequestAlarm, Ticks, Event;
IMPORT L2Memory;
FROM L2Memory IMPORT AddressSpace, NewSpace, DisposeSpace, MapSpace,
                     UserFrames, StackMax;
FROM L2Exceptions IMPORT Handler, KernelStackLimit, Assert, AssertOk;

(* Maintains the Process abstraction, context switching etc. *)

CONST   Quantum = 10;
        MaxGuards = 8;
        StackNeed = 800H;

TYPE    Stack = ARRAY [0..StackMax] OF ADDRESS;
        StackPointer = POINTER TO Stack;
        ProcessStates = (Free, Starting, Running, Ready, Blocked, Suspended);
        Semaphore = POINTER TO SemRec; 
        ProcessPointer = POINTER TO PCB;
        PCB = RECORD  State : ProcessStates;
                      KernelStack : StackPointer;
                      SP, KSLimit, USLimit : ADDRESS;
                      CodeFrames, StackFrames : CARDINAL;
                      User : AddressSpace;
                      Next, Prev : ProcessPointer;
                      Sem : Semaphore;
                      GuardStack : ARRAY [1..MaxGuards] OF Semaphore;
                      Gsp : CARDINAL;
                      Resumed : BOOLEAN;
                      ProcNum : Process;
                      Priority : CARDINAL;
                      Start : PROC;
              END;
        ProcessQueue = RECORD  Head, Tail : ProcessPointer;
                       END;
        SemRec = RECORD Count : INTEGER;
                        Waiting : ProcessQueue;
                 END;
        aptr = POINTER TO ADDRESS;
        cptr = POINTER TO CARDINAL;

VAR     ProcessTable : ARRAY Process OF PCB;
        FreeSlots : CARDINAL;
        ReadyProcs : ARRAY [1..MaxPriority] OF ProcessQueue;
        Current, IdleProcess : ProcessPointer;
        NoSwitch : BOOLEAN;

PROCEDURE  PushGuard( p : Process; s : Semaphore);

BEGIN   WITH ProcessTable[p] DO
          Assert(Gsp <> MaxGuards,108);
          INC(Gsp);
          GuardStack[Gsp] := s;
          END;
END PushGuard;

PROCEDURE  PopGuard( p : Process; VAR s : Semaphore);

BEGIN   WITH ProcessTable[p] DO
          Assert(Gsp <> 0,107);
          s := GuardStack[Gsp];
          DEC(Gsp);
          END;
END PopGuard;

PROCEDURE  Guard( s : Semaphore);

BEGIN   PushGuard(CurrentProcess,s);
        SemP(s);
END Guard;

PROCEDURE  UnGuard( s : Semaphore);

BEGIN   SemV(s);
        PopGuard(CurrentProcess,s);
END UnGuard;
                        
PROCEDURE  ReleaseGuards( p : Process);

VAR     s : Semaphore;

BEGIN   NoSwitch := TRUE;
        WHILE ProcessTable[p].Gsp > 0 DO
          PopGuard(p,s);
          SemV(s);
          END;          
        NoSwitch := FALSE;
END ReleaseGuards;

PROCEDURE  InitQueue( VAR q : ProcessQueue);

BEGIN   WITH q DO
          Head := NIL;
          Tail := NIL;
        END;
END InitQueue;

PROCEDURE  Remove( VAR q : ProcessQueue; p : ProcessPointer);

BEGIN   IF p^.Prev = NIL 
          THEN q.Head := p^.Next 
          ELSE p^.Prev^.Next := p^.Next 
          END;
        IF p^.Next = NIL
          THEN q.Tail := p^.Prev
          ELSE p^.Next^.Prev := p^.Prev
          END;
        p^.Next := NIL;
        p^.Prev := NIL;
END Remove;

PROCEDURE  Get( VAR q : ProcessQueue; VAR p : ProcessPointer);

BEGIN   p := q.Head;
        Assert(p <> NIL,101);
        Remove(q,p);
END Get;

PROCEDURE  Put( VAR q : ProcessQueue; p : ProcessPointer);

BEGIN   p^.Prev := q.Tail;
        p^.Next := NIL;
        IF q.Tail = NIL
          THEN q.Head := p
          ELSE  q.Tail^.Next := p
          END;
        q.Tail := p;
END Put;

PROCEDURE  Empty( VAR q : ProcessQueue) : BOOLEAN;

BEGIN   RETURN q.Head = NIL;
END Empty;

VAR     o, n : ProcessPointer;   (* Vars for switch cannot be local *)
        SwSt : Stack;            (* Switch's own stack *)
        ip   : CARDINAL;

PROCEDURE  Switch; 

BEGIN   IF NoSwitch THEN RETURN END;
        ip := MaxPriority;
        LOOP
          IF ReadyProcs[ip].Head <> NIL THEN
            n := ReadyProcs[ip].Head;
            
            IF ( Current^.Priority > n^.Priority ) AND
               ( Current^.State = Running ) THEN RETURN; END;
            Remove( ReadyProcs[ip], n );
            EXIT;
          END;
          IF ip = 1 THEN
            IF Current^.State = Running THEN RETURN; END;
            n := IdleProcess;
            EXIT;
          END;
          DEC(ip);
        END;
        o := Current;
        CODE(PushA6);
        CODE(MoveUSPA6);
        CODE(PushA6);
        o^.SP := REGISTER(15);
        KernelStackLimit := 0;
        SETREG(15,ADR(SwSt[StackMax-1]));
        SETREG(13,ADR(SwSt));
        KernelStackLimit := ADR(SwSt);
        IF o^.State = Free THEN
          DISPOSE(o^.KernelStack);
          AssertOk(DisposeSpace(o^.User),106);
        ELSIF o^.State = Running THEN 
          IF o <> IdleProcess THEN Put(ReadyProcs[o^.Priority],o) END;
          o^.State := Ready;
          END;
        CurrentProcess := n^.ProcNum;
        Current := n;
        Assert((MapSpace(n^.User) = Ok) OR (CurrentProcess = 1),103);
        KernelStackLimit := 0;
        SETREG(15,n^.SP);
        SETREG(13,n^.KSLimit);
        KernelStackLimit := n^.KSLimit;
        SETREG(7,n^.USLimit);
        CODE(PopA6);
        CODE(MoveA6USP);
        CODE(PopA6);
        IF n^.State = Starting
          THEN n^.State := Running;
               CODE(MoveToSR,S0);
               n^.Start;           (* This should never return *)
               Assert(FALSE,104);  (* We should NEVER get here *)
          ELSE n^.State := Running; 
          END;
END Switch;

PROCEDURE  SetNoSwitch( v : BOOLEAN);

BEGIN   NoSwitch := v;
        IF NOT v THEN Switch END;
END SetNoSwitch;
 
PROCEDURE  CreateProcess( VAR p : Process; cs, ss, priority : CARDINAL; 
                          code : PROC ) : Status;

VAR     s : Status;

BEGIN   IF FreeSlots = 0 THEN RETURN NoSlot END;
        p := 1; 
        WHILE ProcessTable[p].State <> Free DO INC(p) END;
        WITH ProcessTable[p] DO
          Sem := NIL;
          Gsp := 0;                    
          Resumed := FALSE;
          NEW(KernelStack);
          IF KernelStack = NIL THEN RETURN NoHeap END;
          s := NewSpace(User,cs,ss);
          IF s <> Ok THEN 
            DISPOSE(KernelStack);
            RETURN s 
            END;
          CodeFrames := cs;
          StackFrames := ss;
          KernelStack^[StackMax-1] := ADDRESS(UserFrames)*ADDRESS(1000H)
                                      + ADDRESS(0FFEH);  (* Initial USP *)
          KernelStack^[StackMax] := ADR(KernelStack^[StackMax]);  (* A6 *)
          SP := ADR(KernelStack^[StackMax-1]);
          USLimit := ADDRESS(UserFrames-ss+1)*1000H;
          KSLimit := ADDRESS(KernelStack);                
          Priority := priority;
          Start := code;
          Put(ReadyProcs[Priority],ADR(ProcessTable[p]));
          State := Starting;
          DEC(FreeSlots);
          END;
        RETURN Ok;
END CreateProcess;

PROCEDURE  DeleteProcess( p : Process ) : Status;

VAR     cp : ProcessPointer;

BEGIN   cp := ADR(ProcessTable[p]);
        WITH cp^ DO
          IF State = Free THEN RETURN NoSuch END;
          ReleaseGuards(p);
          IF State = Blocked THEN 
            INC(Sem^.Count);
            Remove(Sem^.Waiting,cp);
            END; 
          IF State = Ready THEN
            Remove(ReadyProcs[Priority],cp)
            END;
          INC(FreeSlots);
          IF State = Running
            THEN  State := Free;
                  Switch;
            ELSE  State := Free;
                  DISPOSE(KernelStack);
                  AssertOk(DisposeSpace(User),105);
            END;
          END;
        RETURN Ok;
END DeleteProcess;

PROCEDURE  SemP( s : Semaphore);

BEGIN   NoSwitch := FALSE;
        WITH s^ DO 
          DEC(Count);
          IF Count < 0 THEN
            Put(Waiting,Current);
            Current^.Sem := s;
            Current^.State := Blocked;
            Switch;
            END;
          END;
END SemP;

PROCEDURE  SemV( s : Semaphore);

VAR     t : ProcessPointer;

BEGIN   WITH s^ DO
          INC(Count);
          IF Count <= 0 THEN
            Get(Waiting,t);
            t^.State := Ready;
            Put(ReadyProcs[t^.Priority],t);
            Switch;
            END;
          END;
END SemV;

PROCEDURE  SemQ( s : Semaphore) : INTEGER; (* Returns the count *)

BEGIN   RETURN s^.Count;
END SemQ;

PROCEDURE  SemInit( VAR s : Semaphore; count : INTEGER);

BEGIN   NEW(s);
        WITH s^ DO
          InitQueue(Waiting);
          Count := count;
          END;
END SemInit;              

PROCEDURE  Suspend;

BEGIN   WITH Current^ DO
          IF Resumed THEN
            Resumed := FALSE;
            RETURN;
            END;
          NoSwitch := FALSE;
          State := Suspended;
          Switch;
          END;
END Suspend;

PROCEDURE  Resume( p : Process) : Status;

VAR     rp : ProcessPointer;

BEGIN   rp := ADR(ProcessTable[p]);
        WITH rp ^ DO
          IF State = Free THEN RETURN NoSuch END;
          IF State = Suspended
            THEN State := Ready;
                 Resumed := FALSE;
                 Put(ReadyProcs[Priority],rp);
            ELSE Resumed := TRUE;
            END;
          RETURN Ok;
          END;
END Resume;

PROCEDURE  SetPriority( priority : CARDINAL ) : Status;

BEGIN   IF (priority = 0) OR (priority > MaxPriority ) 
          THEN RETURN IllegalOp;
          END;
        Current^.Priority := priority;
        RETURN Ok;
END SetPriority;

PROCEDURE  GetPageCount( p : Process; VAR pages : CARDINAL) : Status;

BEGIN   WITH ProcessTable[p] DO
          IF State = Free THEN RETURN NoSuch END;
          pages := CodeFrames+StackFrames;
          RETURN Ok;
          END;
END GetPageCount;

PROCEDURE  ExpandCode( frames : CARDINAL) : Status;

VAR     rs : Status;

BEGIN   rs := L2Memory.ExpandCode(Current^.User,frames);
        IF rs <> Ok THEN RETURN rs END;
        INC(Current^.CodeFrames,frames);
        RETURN Ok;
END ExpandCode;

PROCEDURE  ExpandStack( frames : CARDINAL) : Status;

VAR     rs : Status;

BEGIN   WITH Current^ DO
          rs := L2Memory.ExpandStack(Current^.User,frames);
          USLimit := USLimit - ADDRESS(frames)*1000H;
          INC(StackFrames,frames);
          SETREG(7,USLimit);
          END;
        RETURN rs;
END ExpandStack;

PROCEDURE  ExecuteOnUserStack( p : PROC) : BOOLEAN;

(* Execute a super mode routine on the user stack after ensuring that 
   enough stack is available. *)

VAR     a0 : ADDRESS;
        rs : Status;

BEGIN   CODE(MoveUSPA0);
        a0 := REGISTER(8);
        rs := Ok;
        WITH Current^ DO
          IF (a0 <= USLimit) OR (a0-USLimit < StackNeed) 
            THEN rs := ExpandStack(1) END;
          END;
        IF rs <> Ok THEN RETURN FALSE END;
        SETREG(8,ADDRESS(p));
        KernelStackLimit := 0;
        Current^.KSLimit := Current^.USLimit;
        SETREG(13,Current^.KSLimit);
        CODE(PushA6);
        CODE(MoveA7A6);
        CODE(MoveUSPA7);
        KernelStackLimit := Current^.KSLimit;
        CODE(MoveToSR,S0);
        CODE(JsrA0);
        CODE(MoveToSR,S4);
        KernelStackLimit := 0;
        Current^.KSLimit := ADDRESS(Current^.KernelStack);
        SETREG(13,Current^.KSLimit);
        CODE(MoveA6A7);
        KernelStackLimit := Current^.KSLimit;
        CODE(PopA6);
        RETURN TRUE;
END ExecuteOnUserStack;

PROCEDURE  StartUserProgram( pc, ff : ADDRESS);

(* set the standard registers to start up the user program at pc *)

BEGIN   SETREG(6,ff);
        SETREG(7,Current^.USLimit);
        SETREG(15,ADR(Current^.KernelStack^[StackMax-1]));
        SETREG(12,pc);
        CODE(PushA4);
        CODE(PushW,U0);
        CODE(RTE);
END StartUserProgram;

PROCEDURE  PeekA( VAR a : ADDRESS ) : ADDRESS;

VAR     ap : aptr;
        t : BOOLEAN;

BEGIN   t := TRUE;
        Check(a,3,FALSE,t);
        ap := aptr(a);
        INC(a,4);
        IF t THEN RETURN ap^ ELSE RETURN 0FF0000H END;
END PeekA;

PROCEDURE  PeekC( VAR a : ADDRESS ) : CARDINAL;  (* For string lengths *)

VAR     cp : cptr;
        t : BOOLEAN;
        crd : CARDINAL;

BEGIN   t := TRUE;
        Check(a,1,FALSE,t);
        cp := cptr(a);
        INC(a,2);
        IF t THEN crd := cp^ ELSE crd := 0 END;
        IF crd > 0 THEN DEC(crd) END;
        RETURN crd;
END PeekC;

PROCEDURE  Check( adr : ADDRESS; len : CARDINAL; byte : BOOLEAN;
                  VAR flag : BOOLEAN);

VAR     ct, sb, at, alen : ADDRESS;

BEGIN   WITH Current^ DO
          ct := ADDRESS(CodeFrames + 1) * 1000H;
          sb := ADDRESS(UserFrames - StackFrames + 1) * 1000H;
          at := ADDRESS(UserFrames + 1) * 1000H;
          alen := ADDRESS(len);
          IF (NOT byte) AND ODD(adr) THEN flag := FALSE END;
          IF ((adr+alen >= ct) AND (adr < sb)) OR
             (adr+alen >= at) OR
             (adr < 1000H) OR (adr >= at)
               THEN flag := FALSE END;
          END;
END Check;

PROCEDURE  Valid( VAR pat : ARRAY OF CHAR) : BOOLEAN;

(* This procedure is designed to be called before switching to the user
   stack to service a system call.  A string coding the pattern of the
   parameters is passed in.  This routine should not cause any faults
   on its own, and its objective is to ensure that any bad parameter
   addresses cause the call to be aborted (and the program). *)

VAR     sp : ADDRESS;
        pp : CARDINAL;
        b : BOOLEAN;

BEGIN   CODE(MoveUSPA0);
        sp := REGISTER(8); (* USP *)
        b := TRUE;
        Check(sp,0,FALSE,b);
        pp := HIGH(pat);
        LOOP CASE pat[pp] OF
               'W' : Check(sp,1,FALSE,b); INC(sp,2)        |
               'L' : Check(sp,3,FALSE,b); INC(sp,4)        |
               'C' : Check(PeekA(sp),0,TRUE,b);            |
               'V' : Check(PeekA(sp),1,FALSE,b)            |
               'U' : Check(PeekA(sp),3,FALSE,b)            |
               'T' : Check(PeekA(sp),5,FALSE,b)            |
               'O' : Check(PeekA(sp),PeekC(sp),TRUE,b)   |
               'F' : Check(sp,1,FALSE,b)
               ELSE ;
               END;
             IF (pp = 0) OR NOT b THEN EXIT END;
             DEC(pp);
             END;
        RETURN b;
END Valid;

VAR     i : CARDINAL;
        e : Event;

BEGIN   FOR i := 1 TO ProcMax DO 
          WITH ProcessTable[i] DO
            ProcNum := i;
            State := Free;
            END;
          END;
        FreeSlots := ProcMax - 1;
        WITH ProcessTable[1] DO
          State := Running;
          KernelStack := StackPointer(KernelStackLimit);
          User := AddressSpace(NIL);
          Priority := 1;
          END;
        CurrentProcess := 1;
        Current := ADR(ProcessTable[1]);
        IdleProcess := Current;
        NoSwitch := FALSE;
        FOR i := 1 TO MaxPriority DO InitQueue(ReadyProcs[i]); END;
        AssertOk(RequestAlarm(Handler(Switch),0,Ticks(Quantum),TRUE,e),106);
END  L2Kernel.
