IMPLEMENTATION MODULE L2Loader;

FROM L2Exceptions IMPORT Handler, AssertOk;
IMPORT L2Exceptions;
FROM SYSTEM IMPORT ADDRESS;
FROM L2Terminal IMPORT SetCCTrap, MaxTerm, Read, Detach, BTWrite, VTDeallocate,
                       VTHangUp, VTClean;
FROM L2TTIO IMPORT Terminals, WriteString, WriteLn, Write;
FROM L2TTNumber IMPORT WriteAddr;
FROM L2Errors IMPORT Status;                              
IMPORT L2Kernel;
FROM L2Kernel IMPORT MaxPriority, Process, ProcMax, CurrentProcess, ExpandCode,
                     DeleteProcess, CreateProcess, StartUserProgram,
                     Semaphore, SemV, SemP, SemQ, SemInit;
FROM L2Clock IMPORT Ticks, Event, RequestAlarm, CancelAlarm, Time;
IMPORT L2Clock;
IMPORT L2Device;
FROM L2Calendar IMPORT SetClock, ReadClock;
FROM L2Files IMPORT Handle, Users, Open, ReadBlock, Close,
                    FileMode, ModeBits, GetMode,
                    PurgeFiles, Stall, UnStall, Mount;
FROM L2Logicals IMPORT PurgeLogicals, CreateLogical, Inherit, Translate,
                       CreateSystemLogical, DeleteSystemLogical;
FROM L2Memory IMPORT UserFrames;
FROM L2Message IMPORT Enable, Disable;
FROM L2Strings IMPORT Copy, Compare, Find;
FROM L2Locks IMPORT ClearLock;
FROM N2ICMP IMPORT LockICMP, UnLockICMP;
FROM N2UDP IMPORT LockUDPPort, ClearUDPPort;
FROM N2TCPUser IMPORT TCPSync, TCPCleanUp;

CONST   CCQMax = 29;

TYPE    JCB = RECORD Image, Command, Home, Name : ARRAY [0..79] OF CHAR;
                     Terminal, User, Child : CARDINAL;
                     Priv, CCFatal, CCFlag, Sleeping,
                     Detached, Active, Wait : BOOLEAN;
                     LoadResult, Result : Status;
                     LastPC : ADDRESS;
                     Alarm : Event;
                     Parent : Process;
                     Sync, Cont : Semaphore;
              END;

VAR     Jobs : ARRAY Process OF JCB;
        CCScan, Guard : Semaphore;
        JobCount : ARRAY [0..MaxTerm] OF CARDINAL;
        CCQ : ARRAY [0..CCQMax] OF CARDINAL;
        CCGet, CCPut : CARDINAL;         
        SystemDisk : ARRAY [0..5] OF CHAR;

PROCEDURE  Execute( VAR image, command, home : ARRAY OF CHAR;
                    terminal, user : CARDINAL; CCfatal, wait, priv : BOOLEAN;
                    VAR proc : Process; VAR result : Status; 
                    VAR pc : ADDRESS) : Status;

VAR     rs : Status;
        det : BOOLEAN;
        ImTemp : ARRAY [0..79] OF CHAR;

BEGIN   IF NOT Privilege(CurrentProcess) THEN
          IF priv THEN RETURN NoPriv END;
          user := Jobs[CurrentProcess].User;
          terminal := Jobs[CurrentProcess].Terminal;
          END;                    
        det := FALSE;
        IF terminal = 65535 THEN   (* Detached processes signaled by -1 *)
          det := TRUE;
          terminal := 0;        (* they run on #T0 *)
          END;
        IF terminal > MaxTerm THEN RETURN NoUnit END;
        rs := Translate(image,ImTemp);
        IF rs <> Ok THEN RETURN rs END;
        SemP(Guard);
        rs := CreateProcess(proc,1,1,3,Startup);
        IF rs <> Ok THEN SemV(Guard); RETURN rs; END;
        Jobs[CurrentProcess].Child := proc;
        Users[proc] := user;                 (* in Files *)
        Terminals[proc] := terminal;         (* in TTIO  *)
        WITH Jobs[proc] DO
          Copy(command,Command);
          Copy(ImTemp,Image);
          Copy(home,Home);
          Copy(Jobs[CurrentProcess].Name,Name);   (* Inherit user name *)
          Terminal := terminal;
          Detached := det;
          User := user;
          CCFatal := CCfatal;
          CCFlag := FALSE;
          Child := 0;
          Parent := CurrentProcess;
          Priv := priv;
          Wait := wait;
          Sleeping := FALSE;
          LastPC := 0;
          SemV(Cont);
          SemP(Sync);
          IF LoadResult = Ok
            THEN IF Wait THEN 
                   SemV(Guard);
                   SemP(Sync);
                   SemP(Guard);
                   result := Result;
                   pc := LastPC;
                   Cleanup(proc);
                   END;
            ELSE 
                 Cleanup(proc);
            END;
          Jobs[CurrentProcess].Child := 0;
          SemV(Guard);
          RETURN LoadResult;
          END;
END Execute;

PROCEDURE  Exit( rs : Status );

BEGIN   AssertOk(Kill(CurrentProcess,rs),301);
        SemP(Jobs[CurrentProcess].Cont);  (* wait to be deleted *)
END Exit;

PROCEDURE  Kill ( proc : Process; rs : Status ) : Status;

BEGIN   IF (proc = 0) OR (proc > ProcMax) THEN RETURN NoSuch END;
        IF NOT ((proc = CurrentProcess) OR Privilege(CurrentProcess) OR
          (Jobs[proc].User = Jobs[CurrentProcess].User)) 
          THEN RETURN NoPriv END;
        WITH Jobs[proc] DO
          SemP(Guard);
          IF NOT Active THEN SemV(Guard); RETURN NotActive END;
          IF Wait 
            THEN Result := rs;
                 SemV(Sync);
            ELSE 
                 Cleanup(proc);
            END;
          END;
        SemV(Guard);
        RETURN Ok;
END Kill;

PROCEDURE  SetLastPC( proc : Process; pc : ADDRESS) : Status;

BEGIN   IF (proc = 0) OR (proc > ProcMax) THEN RETURN NoSuch END;
        SemP(Guard);
        IF NOT Jobs[proc].Active THEN SemV(Guard); RETURN NotActive END;
        Jobs[proc].LastPC := pc;
        SemV(Guard);
        RETURN Ok;
END SetLastPC;

PROCEDURE  Sleep( interval : Ticks ) : Status;  (* New version *)

VAR     rs : Status;

BEGIN   SemP(Guard);
        rs := RequestAlarm(Handler(L2Kernel.Resume),CurrentProcess,
                           interval,FALSE,Jobs[CurrentProcess].Alarm);
        IF rs <> Ok THEN SemV(Guard); RETURN rs END;
        Jobs[CurrentProcess].Sleeping := TRUE;
        SemV(Guard);
        L2Kernel.Suspend; (* Until some signal, probably timeout *)
        rs := CancelAlarm(Jobs[CurrentProcess].Alarm);
        Jobs[CurrentProcess].Sleeping := FALSE;
        RETURN Ok;
END Sleep;

PROCEDURE  GetPID( VAR proc : Process) : Status;

BEGIN   proc := CurrentProcess;
        RETURN Ok;
END GetPID;

PROCEDURE  GetUser( proc : Process; VAR user : CARDINAL ) : Status;

BEGIN   IF (proc = 0) OR (proc > ProcMax) THEN RETURN NoSuch END;
        SemP(Guard);
        IF NOT Jobs[proc].Active THEN SemV(Guard); RETURN NotActive END;
        IF Privilege(CurrentProcess) 
          OR (Jobs[proc].User = Jobs[CurrentProcess].User)
          THEN user := Jobs[proc].User;
               SemV(Guard);
               RETURN Ok;
          ELSE SemV(Guard);
               RETURN NoPriv;
          END;
END GetUser;

PROCEDURE  SetUser( proc : Process; user : CARDINAL) : Status;

BEGIN   IF (proc = 0) OR (proc > ProcMax) THEN RETURN NoSuch END;
        SemP(Guard);
        IF NOT Jobs[proc].Active THEN SemV(Guard); RETURN NotActive END;
        IF Privilege(CurrentProcess)
          THEN Jobs[proc].User := user;
               Users[proc] := user; (* in Files *)
               SemV(Guard);
               RETURN Ok;
          ELSE SemV(Guard);
               RETURN NoPriv;
          END;
END SetUser;

PROCEDURE  GetCommand( VAR s : ARRAY OF CHAR);

BEGIN   Copy(Jobs[CurrentProcess].Command,s);
END GetCommand;

PROCEDURE  TestCC() : BOOLEAN;

VAR     cc : BOOLEAN;

BEGIN   WITH Jobs[CurrentProcess] DO
          cc := CCFlag;
          CCFlag := FALSE;
          END;
        RETURN cc;
END TestCC;

PROCEDURE  SetCCMode( fatal : BOOLEAN ); 

BEGIN   WITH Jobs[CurrentProcess] DO
          CCFatal := fatal;
          CCFlag := FALSE;
          END;
END SetCCMode;

PROCEDURE  CreateGlobal( VAR log, tran : ARRAY OF CHAR) : Status;

BEGIN   IF NOT Privilege(CurrentProcess) THEN RETURN NoPriv END;
        RETURN CreateSystemLogical(log,tran);
END CreateGlobal;

PROCEDURE  DeleteGlobal( VAR log : ARRAY OF CHAR) : Status;

BEGIN   IF NOT Privilege(CurrentProcess) THEN RETURN NoPriv END;
        RETURN DeleteSystemLogical(log);
END DeleteGlobal;

PROCEDURE  SetTime( VAR t : Time) : Status;

BEGIN   IF NOT Privilege(CurrentProcess) THEN RETURN NoPriv END;
        SetClock(t); 
        L2Clock.SetTime(t);
        RETURN Ok;
END SetTime;

PROCEDURE  GetTime( VAR t : Time) : Status;

BEGIN   ReadClock(t);
        L2Clock.SetTime(t);   (* In case it has strayed *)
        RETURN Ok;
END GetTime;

PROCEDURE  ReBoot() : Status;  (* Close files and reboot *)

VAR     p : Process;

BEGIN   IF NOT Privilege(CurrentProcess) THEN RETURN NoPriv END;
        Stall; (* Wait till current file i/o completes *)
        FOR p := 1 TO ProcMax DO PurgeFiles(p) END;
        L2Exceptions.ReBoot;  (* No life after this ! *)
END ReBoot;

PROCEDURE  Scanner;  (* This routine must be executed by its own process *)
                     (* sort of like the UNIX init process *)

CONST   log = "$SysDisk-LOGIN.LOD";
        sy  = "$SysDisk";
        syd = "SysDisk";

VAR     p : Process;
        tt : CARDINAL;
        rs : Status;
        a : ADDRESS;
        ch : CHAR;
        BootDev [0] : ARRAY [0..5] OF CHAR;
        MemTop [0FF8300H] : ADDRESS;

BEGIN   Users[CurrentProcess] := 0;
        Terminals[CurrentProcess] := 0;
        WriteLn;
        WriteLn;
        WriteString("LUMOS-2 V1.4  ");
        WriteAddr(MemTop DIV 400H,10,1);
        WriteString("K  booted from ");
        WriteString(BootDev);
        WriteLn;
        WriteLn;
        Copy(BootDev,SystemDisk);  (* secure copy for priv check *)
        AssertOk(CreateGlobal(syd,BootDev),3100);                    
        AssertOk(Mount(sy,"",FALSE),3101);
        WHILE Read(0,ch,FALSE) = Ok DO END;
        AssertOk(Execute(log,"StartUp",sy,0,0,FALSE,TRUE,TRUE,p,rs,a),3110);
        LOOP SemP(CCScan);
             tt := CCQ[CCGet];
             CCGet := (CCGet+1) MOD (CCQMax+1);
             IF JobCount[tt] = 0 
               THEN WHILE Read(tt,ch,FALSE) = Ok DO END; (* eat typeahead *)
                    rs := Execute(log,"",sy,tt,0,FALSE,FALSE,TRUE,p,rs,a);
               ELSE FOR p := 1 TO ProcMax DO
                      WITH Jobs[p] DO
                        IF Active AND (Terminal = tt)
                          THEN CCFlag := TRUE END;
                        IF Active AND CCFlag AND CCFatal 
                          THEN rs := Kill(p,Break) END;
                        END;
                      END;
               END;
             END;
END Scanner;

PROCEDURE  CCTrap( unit : CARDINAL);

BEGIN   IF SemQ(CCScan) > CCQMax THEN RETURN END;
        CCQ[CCPut] := unit;
        CCPut := (CCPut + 1) MOD (CCQMax+1);
        SemV(CCScan);
END CCTrap;

PROCEDURE  Cleanup( proc : Process);

VAR     rs : Status;
        kid : CARDINAL;

BEGIN   IF Jobs[proc].Active AND NOT Jobs[proc].Detached THEN 
          DEC(JobCount[Jobs[proc].Terminal]); END;
        Jobs[proc].Active := FALSE;
        Jobs[proc].Name[0] := 0C;           (* No user name *)
        rs := Detach(Jobs[proc].Terminal);
        ClearLock(proc);  (* cancel any locks and lock waits ... *)
        Stall;
        LockUDPPort(proc);  
        TCPSync(proc);
        LockICMP;
          IF proc <> CurrentProcess 
            THEN AssertOk(DeleteProcess(proc),302);
                 AssertOk(Disable(proc),303);
            END;
          VTClean( proc );
          PurgeLogicals(proc);
          PurgeFiles(proc);
          UnStall;
          UnLockICMP;
          ClearUDPPort(proc);
          TCPCleanUp();
        kid := Jobs[proc].Child;
        IF kid <> 0 THEN Jobs[kid].Wait := FALSE END;
        IF Jobs[proc].Sleeping THEN rs := CancelAlarm(Jobs[proc].Alarm) END;
        IF proc = CurrentProcess THEN 
          AssertOk(Disable(proc),305);
          SemV(Guard);  (* Last chance to do this *)
          AssertOk(DeleteProcess(proc),304);
          END;
END Cleanup;

PROCEDURE  Startup;

TYPE    memptr = POINTER TO ARRAY [0..511] OF CHAR;

VAR     LCB : RECORD CASE BOOLEAN OF
                TRUE  :  Flag, Size, Version : CARDINAL;
                         EntryPoint : ADDRESS;
                         LoadKeyOld, LoadKeyNew : CARDINAL;
                         FirstFreeOld, FirstFreeNew : ADDRESS;
                         LayerNo : CARDINAL;
                         FirstCode, CodePtr, RelocPtr, LinkerTablePtr,
                         DebuggerPtr : ADDRESS;
                         CheckSum : CARDINAL |
                FALSE :  Block : ARRAY [0..45] OF CHAR;
                END;
              END;
        h : Handle;
        sz, i, k : CARDINAL;
        mem : memptr;
        tiname : ARRAY [0..4] OF CHAR;
        lastblock : ARRAY [0..511] OF CHAR;

BEGIN   WITH Jobs[CurrentProcess] DO
          SemP(Cont);  (* Wait 'till jcb is ready *)
          Check(Open(Image,FALSE,h));
          Check(TestMode(h,Priv,Image));
          Check(ReadBlock(h,LCB.Block));
          WITH LCB DO
            IF (Flag <> 8) OR (CodePtr <> 512) OR (FirstCode < 1000H)
               OR ODD(EntryPoint) THEN Check(BadImage); END;
            sz := CARDINAL((FirstFreeNew + 0FFFH) DIV 1000H);
            IF sz > 2 THEN Check(ExpandCode(sz-2)) END;
            sz := CARDINAL((FirstFreeNew-FirstCode+511) DIV 512);
            FOR i := 1 TO sz-1 DO 
              mem := memptr( ADDRESS(i-1)*512 + FirstCode );
              Check(ReadBlock(h,mem^)); 
              END;
            Check(ReadBlock(h,lastblock));
            mem := memptr( ADDRESS(sz-1)*512 + FirstCode );
            k := CARDINAL( FirstFreeNew - ADDRESS(sz-1)*512 - FirstCode );
            FOR i := 0 TO k-1 DO mem^[i] := lastblock[i] END;
            Check(Close(h,TRUE));
            Check(Inherit(Parent)); (* inherit logicals *)
            Check(CreateLogical("CURRENT",Home));
        (*    Check(CreateLogical("HOME",Home));  Obsolete *)
            Check(CreateLogical("Image",Image));
            tiname := "#T0";
            IF Terminal > 9 THEN
              tiname[2] := CHR( Terminal DIV 10 + 48 );
              tiname[3] := CHR( Terminal MOD 10 + 48 );
              tiname[4] := 0C;
            ELSE
              tiname[2] := CHR( Terminal + 48 );
              tiname[3] := 0C;
            END; (* IF *)
            Check(CreateLogical("TI",tiname));
            Check(Enable(CurrentProcess));
            LoadResult := Ok;
            Active := TRUE;
            IF NOT Detached THEN INC(JobCount[Terminal]); END;
            SemV(Sync); (* Execute can now complete *)
            StartUserProgram(EntryPoint,FirstFreeNew);
            END;
          END;
END Startup;

PROCEDURE  GetJobInfo( p : Process; VAR pages, ti : CARDINAL;
              VAR par : Process; VAR im : ARRAY OF CHAR ) : Status;

VAR     rs : Status;
       
BEGIN   IF NOT (Privilege(CurrentProcess) OR 
                (Jobs[p].User = Jobs[CurrentProcess].User)) THEN
              RETURN NoPriv END;
        rs := L2Kernel.GetPageCount(p,pages);
        IF rs <> Ok THEN RETURN rs END;
        WITH Jobs[p] DO
          par := Parent;
          Copy(Image,im);
          ti := Terminal;
          END;
        RETURN Ok;
END GetJobInfo;

PROCEDURE  SetDevParams( VAR dn : ARRAY OF CHAR; p, v : CARDINAL) : Status;

BEGIN   IF NOT Privilege(CurrentProcess) THEN RETURN NoPriv END;
        RETURN L2Device.SetDevParams(dn,p,v);
END SetDevParams;

PROCEDURE  UpTime( VAR t : Ticks);

BEGIN   t := L2Clock.TickCount;
END UpTime;

PROCEDURE  Resume( p : Process) : Status;

BEGIN   IF NOT (Privilege(CurrentProcess) OR 
                (Jobs[p].User = Jobs[CurrentProcess].User)) THEN
              RETURN NoPriv END;
        RETURN L2Kernel.Resume(p);
END Resume;
                           
PROCEDURE SetPriority( priority : CARDINAL ) : Status;
BEGIN   IF NOT Privilege(CurrentProcess) THEN RETURN NoPriv; END;
        RETURN L2Kernel.SetPriority(priority);
END SetPriority;

PROCEDURE  SetName( proc : Process; VAR name : ARRAY OF CHAR) : Status;

BEGIN   IF (proc = 0) OR (proc > ProcMax) THEN RETURN NoSuch END;
        SemP(Guard);
        IF NOT Jobs[proc].Active THEN SemV(Guard); RETURN NotActive END;
        IF Privilege(CurrentProcess)
          THEN Copy(name,Jobs[proc].Name);
               SemV(Guard);
               RETURN Ok;
          ELSE SemV(Guard);
               RETURN NoPriv;
          END;
END SetName;

PROCEDURE  GetName( proc : Process; VAR name : ARRAY OF CHAR) : Status;

BEGIN   IF (proc = 0) OR (proc > ProcMax) THEN RETURN NoSuch END;
        SemP(Guard);
        IF NOT Jobs[proc].Active THEN SemV(Guard); RETURN NotActive END;
        Copy(Jobs[proc].Name,name);
        SemV(Guard);
        RETURN Ok;
END GetName;

PROCEDURE HangUp( unit : CARDINAL ) : Status;
  BEGIN     
    IF NOT Privilege( CurrentProcess ) THEN RETURN NoAccess; END;
    RETURN VTHangUp( unit );
  END HangUp;

PROCEDURE  WriteMessage( VAR name, message : ARRAY OF CHAR) : Status;

(* Try to force message on to some terminal at which name is logged in *)

VAR     terms : ARRAY [0..MaxTerm] OF BOOLEAN;
        i : CARDINAL;
        found, wrote : BOOLEAN;

BEGIN   found := FALSE;
        wrote := FALSE;
        IF name[0] = 0C THEN RETURN NoSuch END;
        FOR i := 0 TO MaxTerm DO terms[i] := FALSE END;   
        SemP(Guard);
        FOR i := 1 TO ProcMax DO
          WITH Jobs[i] DO
            IF (Compare(Name,name)=0) AND (NOT Detached) THEN
              found := TRUE;
              terms[Terminal] := TRUE;
              END;
            END;
          END;     
        SemV(Guard);
        IF NOT found THEN RETURN NoSuch END;
        FOR i := 0 TO MaxTerm DO
          IF terms[i] AND (BTWrite(i,message) = Ok) THEN wrote := TRUE END;
          END;
        IF wrote THEN RETURN Ok ELSE RETURN DeviceBusy END;
END WriteMessage;

PROCEDURE  TestMode( h : Handle; VAR privmode : BOOLEAN;
                     VAR im : ARRAY OF CHAR) : Status;

(* Here we define bit 12 of the Mode word on a LOD file to mean that
   the file is to execute with privilege if certain other conditions
   are met - it must not be writable by world and must belong to the
   system, i.e. it must have owner number 0 and live on the boot
   volume.  (We don't have the notion of privileged mounts.)

   User 0 - the system - can execute any file that its owner can execute.
*)

CONST   privbit = 12;     (* the 1000 bit in hex *)

VAR     d : Time;
        m : FileMode;
        o : CARDINAL;
        rs : Status;
        pi : BOOLEAN;

BEGIN   rs := GetMode(h,d,m,o);
        IF rs <> Ok THEN RETURN rs END;
        IF (Users[CurrentProcess] = o) OR (o=0)
          THEN IF NOT (OEx IN m) THEN RETURN NoAccess END;
          ELSE IF NOT (WEx IN m) THEN RETURN NoAccess END;
          END;
        pi := (privbit IN BITSET(m)) AND (NOT (WWr IN m)) AND (o=0)
              AND (Find(SystemDisk,im)=0);
        privmode := privmode OR pi;  (* Turn on privs if priv'd image *)
        RETURN Ok;
END TestMode;

PROCEDURE  Check( rs : Status);

BEGIN   IF rs <> Ok THEN
          WITH Jobs[CurrentProcess] DO
            LoadResult := rs;
            SemV(Sync);
            SemP(Cont);  (* wait for Cleanup *)
            END;
          END;
END Check;

PROCEDURE  Privilege( proc : Process) : BOOLEAN;

BEGIN   RETURN Jobs[proc].Priv;
END Privilege;

VAR     i : Process;
        j : CARDINAL;
        rs : Status;
        t : Time;
 
BEGIN   FOR i := 1 TO ProcMax DO
          WITH Jobs[i] DO
            SemInit(Cont,0);
            SemInit(Sync,0);
            Active := FALSE;
            Priv := TRUE;
            User := 0;
            Users[i] := 0;
            Terminals[i] := 0;
            Image[0] := 0C;
            Name[0] := 0C;         (* No user name *)
            END;
          END;
        FOR j := 0 TO MaxTerm DO 
          JobCount[j] := 0;
          AssertOk(SetCCTrap(CCTrap,j),305);
          END;
        CCGet := 0;
        CCPut := 0;
        SemInit(Guard,1);
        SemInit(CCScan,0);
        ReadClock(t);       (* get the real time *)
        L2Clock.SetTime(t);
        rs := CreateProcess(i,0,0,MaxPriority-2,Scanner);(*Init & CC symbiont*)
END L2Loader.
