IMPLEMENTATION MODULE L2Locks;    (* Application lock manager *)

FROM L2Kernel IMPORT Process, ProcMax, CurrentProcess, SetNoSwitch, Resume;
FROM L2Errors IMPORT Status;
FROM L2Strings IMPORT Copy, Compare;
FROM L2Storage IMPORT ALLOCATE, DEALLOCATE;
FROM L2Files IMPORT Users;

CONST   resmax = 19;

TYPE    LockPtr = POINTER TO LockRec;
        LockID = RECORD res : ARRAY [0..resmax] OF CHAR;
                        own : CARDINAL;
                 END;
        LockRec = RECORD id : LockID;
                         pr : Process;  (* The locker! *)
                         nx : LockPtr;
                  END;

VAR     LockList : LockPtr;
        Attachments : ARRAY Process OF LockID;
              
(* Resources are qualified with the user number of the caller.  Resources
   are also unlocked when the locking process exits.  Resource names
   are significant to 20 characters. *)

PROCEDURE Lock( VAR resource : ARRAY OF CHAR) : Status;

VAR     p, q : LockPtr;

BEGIN   IF resource[0] = 0C THEN RETURN BadName END;  (* Humpf *)
        SetNoSwitch(TRUE);
        p := LockList;
        q := NIL;
        WHILE p <> NIL DO
          IF ThisLock(p^.id,resource) THEN
            SetNoSwitch(FALSE);
            RETURN NoAccess;
            END;
          q := p;
          p := p^.nx;
          END;
        NEW(p);
        IF p = NIL THEN 
          SetNoSwitch(FALSE);
          RETURN NoHeap;
          END;
        IF q = NIL THEN LockList := p ELSE q^.nx := p END;
        p^.nx := NIL;
        p^.pr := CurrentProcess;
        p^.id.own := Users[CurrentProcess];
        Copy(resource,p^.id.res);
        SetNoSwitch(FALSE);
        RETURN Ok;
END Lock;

PROCEDURE UnLock( VAR resource : ARRAY OF CHAR) : Status;

VAR     p, q : LockPtr;
        i : Process;
        rs : Status;

BEGIN   SetNoSwitch(TRUE);
        p := LockList;
        q := NIL;
        WHILE p <> NIL DO
          WITH p^ DO
            IF ThisLock(id,resource) AND (pr=CurrentProcess) THEN
              IF q = NIL THEN LockList := nx ELSE q^.nx := nx END;
              DISPOSE(p);
              FOR i := 1 TO ProcMax DO
                IF ThisLock(Attachments[i],resource) THEN 
                  rs := Resume(i);
                  END;
                END;
              SetNoSwitch(FALSE);
              RETURN Ok;
              END;
            END; 
          q := p;
          p := p^.nx;
          END;
        SetNoSwitch(FALSE);
        RETURN NoSuch;
END UnLock;

PROCEDURE  ThisLock( VAR k : LockID; VAR res : ARRAY OF CHAR) : BOOLEAN;

BEGIN   RETURN (Compare(k.res,res) = 0) AND (k.own = Users[CurrentProcess]);
END ThisLock;

PROCEDURE  AttachLock( VAR resource : ARRAY OF CHAR);

(* The caller will be resumed when resource is next unlocked *)

BEGIN   SetNoSwitch(TRUE);
        WITH Attachments[CurrentProcess] DO
          own := Users[CurrentProcess];
          Copy(resource,res);
          END;
        SetNoSwitch(FALSE);
END AttachLock;

PROCEDURE  DetachLock;

BEGIN   Attachments[CurrentProcess].res[0] := 0C;
END DetachLock;

PROCEDURE  ClearLock( proc : Process);

VAR     p, q, tm : LockPtr;
        i : Process;
        rs : Status;

BEGIN   SetNoSwitch(TRUE);
        Attachments[proc].res[0] := 0C;
        p := LockList;
        q := NIL;
        WHILE p <> NIL DO
          WITH p^ DO
            IF pr = proc
              THEN FOR i := 1 TO ProcMax DO
                     IF (Compare(Attachments[i].res,p^.id.res) = 0) AND
                        (Attachments[i].own = p^.id.own) 
                       THEN rs := Resume(i) END;
                     END; 
                   tm := p;
                   IF q = NIL THEN LockList := nx ELSE q^.nx := nx END;
                   p := nx;
                   DISPOSE(tm);
              ELSE q := p;
                   p := nx;
              END;
            END;
          END;
        SetNoSwitch(FALSE);
END ClearLock;

PROCEDURE  ShowLock( n : CARDINAL; VAR resource : ARRAY OF CHAR;
                     VAR owner : CARDINAL; VAR proc : Process) : Status;

(* For displaying the lock list *)

VAR     p : LockPtr;

BEGIN   SetNoSwitch(TRUE);
        p := LockList;
        WHILE (p <> NIL) AND (n > 0) DO
          p := p^.nx;
          DEC(n);
          END;
        IF p = NIL THEN
          SetNoSwitch(FALSE);
          RETURN NoSuch;
          END;
        Copy(p^.id.res,resource);
        owner := p^.id.own;
        proc := p^.pr;
        SetNoSwitch(FALSE);
        RETURN Ok;
END ShowLock;

VAR     i : Process;

BEGIN   LockList := NIL;
        FOR i := 1 TO ProcMax DO
          Attachments[i].res[0] := 0C;
          END;
END L2Locks.
