IMPLEMENTATION MODULE L2Logicals;

(*  Maintains logical name tables for the system and the processes.  *)

FROM L2Errors IMPORT Status;
FROM L2Strings IMPORT Copy, Insert, Remove, Find, Compare, Length, Substring;
FROM L2Kernel IMPORT Process, ProcMax, CurrentProcess, 
                     Semaphore, SemInit, Guard, UnGuard, SetNoSwitch;
FROM L2Storage IMPORT ALLOCATE, DEALLOCATE;

CONST   LogLen = 8;
        TransLen = 80;

TYPE    Logical = ARRAY [0..LogLen-1] OF CHAR;
        Translation = ARRAY [0..TransLen-1] OF CHAR;
        LogPointer = POINTER TO LogNode;
        LogNode = RECORD  LogName : Logical;
                          Trans : Translation;
                          Next : LogPointer;
                  END;

VAR     Lists : ARRAY Process OF LogPointer;
        Hyphen : ARRAY [0..1] OF CHAR;
        Lguard : Semaphore;

PROCEDURE  Translate( VAR sn, tn : ARRAY OF CHAR) : Status;

VAR     p : LogPointer;
        ln : Logical;
        hp, xc, kp : CARDINAL;

BEGIN   Guard(Lguard);
        Copy(sn,tn);
        xc := 0;
        LOOP IF (tn[0] > '$') OR (tn[0] <= ' ')
               THEN Insert("$CURRENT-",tn,0) END;
             IF tn[0] <> '$' THEN EXIT END;
             hp := Find(Hyphen,tn);
             IF hp-1 > LogLen THEN UnGuard(Lguard); RETURN BadName END;
             Substring(tn,ln,1,hp-1);
             Remove(tn,0,hp); 
             p := Lists[CurrentProcess];
             WHILE (p <> NIL) AND (Compare(p^.LogName,ln) <> 0) 
               DO p := p^.Next;
               END;
             IF p = NIL THEN
               p := Lists[1];
               WHILE (p <> NIL) AND (Compare(p^.LogName,ln) <> 0) 
                 DO p := p^.Next;
                 END;
               END;
             IF p = NIL THEN UnGuard(Lguard); RETURN NoLogical END;
             IF Length(p^.Trans) + Length(tn) > HIGH(tn)+1 
               THEN UnGuard(Lguard); RETURN BadName END;
             Insert(p^.Trans,tn,0);
             INC(xc);
             IF xc = 20 THEN UnGuard(Lguard); RETURN BadName END;
             END;
        LOOP kp := 0;  (* Process carets now *)
             WHILE (kp <= HIGH(tn)) AND (tn[kp] <> 0C) AND (tn[kp] <> '^')
               DO INC(kp) END;
             IF (kp > HIGH(tn)) OR (tn[kp] = 0C) THEN EXIT END;
             IF kp < 4 THEN UnGuard(Lguard); RETURN BadName END;
             hp := kp - 2;
             WHILE (hp > 0) AND (tn[hp] <> '-') DO DEC(hp) END;
             IF hp = 0 THEN UnGuard(Lguard); RETURN BadName END;
             Remove(tn,hp+1,kp-hp);
             END;
        UnGuard(Lguard);
        RETURN Ok;
END Translate;

PROCEDURE  CreateLogical( VAR log, tran : ARRAY OF CHAR) : Status;

VAR     rs : Status;

BEGIN   Guard(Lguard);
        rs := Create(CurrentProcess,log,tran);
        UnGuard(Lguard); 
        RETURN rs;
END CreateLogical;

PROCEDURE  DeleteLogical( VAR log : ARRAY OF CHAR) : Status;

VAR     rs : Status;

BEGIN   Guard(Lguard);
        rs := Delete(CurrentProcess,log);
        UnGuard(Lguard); 
        RETURN rs;
END DeleteLogical;

PROCEDURE  PurgeLogicals( proc : Process);

VAR     p, q : LogPointer;

BEGIN   IF (proc = 0) OR (proc > ProcMax) THEN RETURN END;
        Guard(Lguard);    
        SetNoSwitch(TRUE);
        p := Lists[proc];
        WHILE p <> NIL DO
          q := p;
          p := p^.Next;
          DISPOSE(q);
          END;
        Lists[proc] := NIL;
        SetNoSwitch(FALSE);
        UnGuard(Lguard); 
END PurgeLogicals;

PROCEDURE  CreateSystemLogical( VAR log, tran : ARRAY OF CHAR) : Status;

VAR     rs : Status;

BEGIN   Guard(Lguard);
        rs := Create(1,log,tran);
        UnGuard(Lguard); 
        RETURN rs;
END CreateSystemLogical;

PROCEDURE  DeleteSystemLogical( VAR log : ARRAY OF CHAR) : Status;

VAR     rs : Status;

BEGIN   Guard(Lguard);
        rs := Delete(1,log);
        UnGuard(Lguard); 
        RETURN rs;
END DeleteSystemLogical;

PROCEDURE  Inherit( parent : Process) : Status;

VAR     p : LogPointer;
        rs : Status;

BEGIN   IF (parent = 0) OR (parent > ProcMax) THEN RETURN NoSuch END;
        IF parent = CurrentProcess THEN RETURN BadFunction END;
        Guard(Lguard);
        p := Lists[parent];
        rs := Ok;
        WHILE (p <> NIL) AND (rs = Ok) DO
          WITH p^ DO 
            rs := Create(CurrentProcess,LogName,Trans);
            p := Next;
            END;
          END;
        UnGuard(Lguard); 
        RETURN rs;
END Inherit;

PROCEDURE  GetLogical( n : CARDINAL; VAR log, tran : ARRAY OF CHAR) : Status;

VAR     p : LogPointer;
        rs : Status;

BEGIN   Guard(Lguard);
        p := Lists[CurrentProcess];
        WHILE (p <> NIL) AND (n > 1) DO
          p := p^.Next;
          DEC(n);
          END;
        IF p = NIL THEN rs := NoSuch ELSE
          Copy(p^.LogName,log);
          Copy(p^.Trans,tran);
          rs := Ok;
          END;
        UnGuard(Lguard);
        RETURN rs;
END GetLogical;

PROCEDURE  GetSystemLogical( n : CARDINAL; VAR log, tran : ARRAY OF CHAR)
                          : Status;

VAR     p : LogPointer;
        rs : Status;

BEGIN   Guard(Lguard);
        p := Lists[1];
        WHILE (p <> NIL) AND (n > 1) DO
          p := p^.Next;
          DEC(n);
          END;
        IF p = NIL THEN rs := NoSuch ELSE
          Copy(p^.LogName,log);
          Copy(p^.Trans,tran);
          rs := Ok;
          END;
        UnGuard(Lguard);
        RETURN rs;
END GetSystemLogical;

PROCEDURE  Create( pr : Process; VAR log, tran : ARRAY OF CHAR) : Status;

VAR     p : LogPointer;
        i, ll : CARDINAL;

BEGIN   ll := Length(log);
        IF ll > LogLen THEN RETURN BadLogical END;
        FOR i := 0 TO ll-1 DO 
          IF NOT (((CAP(log[i]) <= 'Z') AND (CAP(log[i]) >= 'A')) OR
                 ((log[i] <= '9') AND (log[i] >= '0')))
            THEN RETURN BadLogical END;
          END;
        IF Length(tran) > TransLen THEN RETURN BadName END;
        p := Lists[pr];   
        SetNoSwitch(TRUE);
        WHILE (p <> NIL) AND (Compare(p^.LogName,log) <> 0) 
          DO p := p^.Next END;
        IF p = NIL THEN
          NEW(p);
          IF p = NIL THEN RETURN NoHeap END;
          p^.Next := Lists[pr];
          Lists[pr] := p;
          Copy(log,p^.LogName);
          END;
        Copy(tran,p^.Trans);
        SetNoSwitch(FALSE);
        RETURN Ok;
END Create;

PROCEDURE  Delete( pr : CARDINAL; VAR log : ARRAY OF CHAR) : Status;

VAR     p, q : LogPointer;

BEGIN   p := Lists[pr];
        q := NIL;
        SetNoSwitch(TRUE);
        WHILE (p <> NIL) AND (Compare(p^.LogName,log) <> 0) DO
          q := p;
          p := p^.Next;
          END;
        IF p = NIL THEN RETURN NoLogical END;
        IF q = NIL 
          THEN Lists[pr] := p^.Next
          ELSE q^.Next := p^.Next
          END;           
        SetNoSwitch(FALSE);
        DISPOSE(p);
        RETURN Ok;
END Delete;

VAR     i : Process;

BEGIN   FOR i := 1 TO ProcMax DO Lists[i] := NIL END;
        Hyphen[0] := "-"; Hyphen[1] := 0C;
        SemInit(Lguard,1); (* One at a time, please *)
END L2Logicals.
