IMPLEMENTATION MODULE L2Message;  (* The interprocess message driver *)

FROM L2Kernel IMPORT Process, ProcMax, CurrentProcess, SetNoSwitch,
                     Semaphore, SemV, SemP, SemInit, Resume;
FROM L2Errors IMPORT Status;
FROM L2Storage IMPORT ALLOCATE, DEALLOCATE;
FROM L2Exceptions IMPORT Assert;

(* No array can have more than 32767 components with this compiler *)

TYPE    Buffer = ARRAY [0..32767] OF CHAR;
        BufPtr = POINTER TO Buffer;
        MCB = RECORD Buf : BufPtr;
                     Size, Proc : CARDINAL;
                     Msgs, Slots : Semaphore;
                     Active, Attached : BOOLEAN;
              END;

VAR     Messages : ARRAY Process OF MCB;

PROCEDURE  Send( dest : CARDINAL; VAR s : ARRAY OF CHAR) : Status;

VAR     i : CARDINAL;           
        rs : Status;

BEGIN   IF dest > ProcMax THEN RETURN NoUnit END;
        IF dest = 0 THEN RETURN NoFunction END;
        IF dest = CurrentProcess THEN RETURN BadFunction END;
        WITH Messages[dest] DO
          SemP(Slots);
          SetNoSwitch(TRUE);
          IF NOT Active THEN 
            SemV(Slots); 
            SetNoSwitch(FALSE); 
            RETURN NoUnit;
            END;
          Size := HIGH(s)+1;
          ALLOCATE(Buf,Size);
          IF Buf = NIL THEN 
            SemV(Slots); 
            SetNoSwitch(FALSE); 
            RETURN NoHeap;
            END;
          Proc := CurrentProcess;
          FOR i := 0 TO HIGH(s) DO Buf^[i] := s[i] END;
          SemV(Msgs);     
          SetNoSwitch(FALSE);
          IF Attached THEN rs := Resume(dest) END;
          END;
        RETURN Ok;
END Send;

PROCEDURE  Receive( VAR s : ARRAY OF CHAR) : Status;

VAR     rs : Status;
        k, i : CARDINAL;

BEGIN   WITH Messages[CurrentProcess] DO
          SemP(Msgs);
          SetNoSwitch(TRUE);
          IF Buf = NIL 
            THEN rs := EndFile; 
            ELSE rs := Ok;
                 k := Size - 1;
                 IF k > HIGH(s) THEN k := HIGH(s) END;
                 FOR i := 0 TO k DO s[i] := Buf^[i] END;
                 IF k < HIGH(s) THEN s[k+1] := 0C END;
                 DEALLOCATE(Buf,Size);
            END;
          IF Proc = 65535 THEN rs := NoSuch END;
          Proc := 0;
          SemV(Slots);
          END;
        SetNoSwitch(FALSE);
        RETURN rs;
END Receive;

PROCEDURE TestMessage( wait : BOOLEAN ) : CARDINAL;

BEGIN   WITH Messages[CurrentProcess] DO
          IF wait THEN 
            SemP(Msgs); 
            SemV(Msgs); 
            END;
          RETURN Proc;
          END;
END TestMessage;

PROCEDURE  Flush( unit : CARDINAL) : Status;  (* send an EOF message *)

BEGIN   IF unit > ProcMax THEN RETURN NoUnit END;
        IF unit = CurrentProcess THEN RETURN BadFunction END;
        WITH Messages[unit] DO
          SetNoSwitch(TRUE);
          IF NOT Active THEN 
            SetNoSwitch(FALSE);
            RETURN NoUnit;
            END;
          SemP(Slots);
          SetNoSwitch(TRUE);
          Proc := CurrentProcess;
          Assert(Buf=NIL,210);
          SemV(Msgs);
          END;
        SetNoSwitch(FALSE);
        RETURN Ok;
END Flush;

PROCEDURE  Enable( unit : CARDINAL) : Status;

BEGIN   IF unit > ProcMax THEN RETURN NoUnit END;
        IF unit = 0 THEN unit := CurrentProcess END;
        Messages[unit].Active := TRUE;
        RETURN Ok;
END Enable;

PROCEDURE  Disable( unit : CARDINAL) : Status;

VAR     p : CARDINAL;

BEGIN   IF unit > ProcMax THEN RETURN NoUnit END;
        IF unit = 0 THEN unit := CurrentProcess END;
        SetNoSwitch(TRUE);
        FOR p := 1 TO ProcMax DO
          WITH Messages[p] DO
            IF Proc = unit THEN Proc := 65535 END;
            END;
          END;
        WITH Messages[unit] DO
          IF Active THEN 
            Attached := FALSE;
            Active := FALSE;
            IF Proc <> 0 THEN
              IF Buf <> NIL THEN DEALLOCATE(Buf,Size) END;
              Proc := 0;
              SemP(Msgs);
              SemV(Slots);
              END;
            END;
          END;
        SetNoSwitch(FALSE);
        RETURN Ok;
END Disable;

PROCEDURE  Attach;

BEGIN   Messages[CurrentProcess].Attached := TRUE;
END Attach;

PROCEDURE  Detach;

BEGIN   Messages[CurrentProcess].Attached := FALSE;
END Detach;

VAR     i : Process;

BEGIN   FOR i := 1 TO ProcMax DO
          WITH Messages[i] DO
            SemInit(Msgs,0);
            SemInit(Slots,1);
            Proc := 0;
            Buf := NIL;
            Active := FALSE;
            Attached := FALSE;
            END;
          END;
END L2Message.
