IMPLEMENTATION MODULE L2Omninet;

FROM L2Errors     IMPORT Status;
FROM L2Kernel     IMPORT Semaphore, SemV, SemP, SemInit, SemQ,
                         SetNoSwitch, Guard, UnGuard;
FROM L2Exceptions IMPORT Establish, IntControl;

CONST   NETWORK = 6;
        NETINT  = 5;
        maxlength = 1280;
        delayCount = 10;

VAR     WHADDR [0FFC380H] : CARDINAL;
        WLADDR [0FFC384H] : CARDINAL;
        WDATA  [0FFC386H] : CARDINAL;
        NADDR  [0FFC388H] : CARDINAL;
        WCMDAR [0FFC382H] : CARDINAL;
        OSTAT  [0FFC380H] : BITSET;
        CMNode [0FF8026H] : CHAR;    (* CMOS stored node number *)
        Oguard, Tsync, Rsync : Semaphore;
        Initialised : BOOLEAN;
        Result : CARDINAL;

PROCEDURE Delay;

VAR     i : CARDINAL;

BEGIN FOR i := 1 TO delayCount DO END;
END Delay;

PROCEDURE  Ready;

BEGIN   WHILE NOT (7 IN OSTAT) DO END;
END Ready;

PROCEDURE  SetCmdAddr( a : CARDINAL);

BEGIN   Ready;
        WCMDAR := 0;
        Ready;
        WCMDAR := a DIV 256;
        Ready;
        WCMDAR := a MOD 256;
        Delay;
END SetCmdAddr;
                                   
PROCEDURE  PointRam( a : CARDINAL);

BEGIN   WLADDR := a MOD 256;
        Delay;
        WHADDR := a DIV 256;
        Delay;
END PointRam;

PROCEDURE  WriteRamChar( c : CHAR);

BEGIN   WDATA := ORD(c);
        Delay;
END WriteRamChar;

PROCEDURE  WriteRamByte( c : CARDINAL);

BEGIN   WDATA := c MOD 256;
        Delay;
END WriteRamByte;

PROCEDURE ReadRamChar( VAR c : CHAR );

BEGIN   c := CHR( WDATA MOD 256 );
        Delay;
END ReadRamChar;

PROCEDURE  ReadRamByte( VAR c : CARDINAL);

BEGIN   c := WDATA MOD 256;
        Delay;
END ReadRamByte;

PROCEDURE Command( a, r : CARDINAL);

BEGIN         
   SetNoSwitch(TRUE);
   PointRam(r);
   WriteRamByte(0FFH);
   SetCmdAddr(a);
   LOOP SemP(Tsync); 
        SetNoSwitch(TRUE);
        PointRam(r);
        ReadRamByte(Result);
        IF Result <> 0FFH THEN EXIT END;
        END;
   SetNoSwitch(FALSE);
END Command;

PROCEDURE SetUpReceive;

BEGIN   Command(4,27);
END SetUpReceive;

PROCEDURE WriteBlock( VAR b : ARRAY OF CHAR) : Status;

VAR     i, k : CARDINAL;
        rs : Status;

BEGIN
   IF HIGH(b) < 40 THEN RETURN BoundsError END;
   Guard(Oguard);
   IF NOT Initialised THEN 
     TransportInit;
     IF Result <> 0FEH THEN 
       UnGuard(Oguard);
       RETURN DevError;
       END;
     END;
   SetNoSwitch(TRUE);
   PointRam(26);
   WriteRamChar(b[7]);  (* extract destination from DIX address *)
   k := 256*ORD(b[0])+ORD(b[1]);
   IF k > maxlength THEN k := maxlength END;
   IF k > HIGH(b) - 3 THEN k := HIGH(b) - 3 END;
   PointRam(23);
   WriteRamChar(b[0]);
   WriteRamChar(b[1]);
   PointRam(600H);
   FOR i := 0 TO k-1 DO
     WriteRamChar(b[i+2]);
   END;      
   Command(15,31);
   IF Result = 0 
     THEN rs := Ok
     ELSE rs := DevError
     END;
   UnGuard(Oguard);
   RETURN rs;
END WriteBlock;

PROCEDURE ReadBlock( VAR b : ARRAY OF CHAR) : Status;

VAR    i, k : CARDINAL;

BEGIN        
   IF HIGH(b) < 40 THEN RETURN BoundsError END;
   Guard(Oguard);
   IF NOT Initialised THEN 
     TransportInit;
     IF Result <> 0FEH THEN 
       UnGuard(Oguard);
       RETURN DevError;
       END;
     END;
   LOOP SetNoSwitch(TRUE);
        PointRam(27);
        ReadRamByte(Result);
        IF Result = 0 THEN EXIT END;
        UnGuard(Oguard);  (* Allow other ops while we wait *)
        SemP(Rsync);    
        Guard(Oguard);    (* One operation at a time *)
        END;
   SetNoSwitch(TRUE);
   PointRam(29);
   ReadRamChar(b[0]); ReadRamChar(b[1]);
   k := 256*ORD(b[0])+ORD(b[1]);
   IF k > HIGH(b) - 3 THEN k := HIGH(b) - 3 END;
   PointRam(100H);
   FOR i := 0 TO k-1 DO
     ReadRamChar(b[i+2]);
     END;
   SetUpReceive;
   UnGuard(Oguard);
   RETURN Ok;
END ReadBlock;    

PROCEDURE Init;

BEGIN
   SetNoSwitch(TRUE);
   PointRam(0);              (* Layout commands *)

   WriteRamByte(20H);   (* init *)
   WriteRamByte(0);
   WriteRamByte(0);
   WriteRamByte(31);

   WriteRamByte(0F0H);  (* setup receive *)
   WriteRamByte(0);
   WriteRamByte(0);
   WriteRamByte(27);
   WriteRamByte(90H);
   WriteRamByte(0);
   WriteRamByte(1);
   WriteRamByte(0);
   WriteRamByte(5);
   WriteRamByte(0);
   WriteRamByte(0);

   WriteRamByte(40H);   (* send *)
   WriteRamByte(0);
   WriteRamByte(0);
   WriteRamByte(31);
   WriteRamByte(90H);
   WriteRamByte(0);
   WriteRamByte(6);
   WriteRamByte(0);
   WriteRamByte(0);
   WriteRamByte(0);
   WriteRamByte(0);
   WriteRamByte(0);

   Command(0,31);         (* Do init *)
END Init;

MODULE Interrupts[4];

IMPORT Establish, Tsync, Rsync, SemV, SemQ, NETINT, IntControl;

PROCEDURE NetInt( e : CARDINAL);

BEGIN
   EXCL(IntControl,NETINT);
   INCL(IntControl,NETINT);
   IF SemQ(Tsync) < 1 THEN SemV(Tsync) END;
   IF SemQ(Rsync) < 1 THEN SemV(Rsync) END;
END NetInt;

BEGIN
   Establish(47H,NetInt,0);
END Interrupts;

PROCEDURE  TransportInit;

BEGIN
   INCL(IntControl,NETINT);
   Init;  
   IF Result <> NodeNumber THEN RETURN END;
   SetUpReceive;  
   IF Result <> 0FEH THEN RETURN END;
   Initialised := TRUE;
END TransportInit;

BEGIN 
   SemInit(Tsync,0);
   SemInit(Rsync,0);
   SemInit(Oguard,1);
   NodeNumber := ORD(CMNode) MOD 256;
   NADDR := NodeNumber;
   INCL(IntControl,NETWORK);
   Initialised := FALSE;
END L2Omninet.
