IMPLEMENTATION MODULE L2Printer;

FROM L2Exceptions IMPORT Establish, Handler;
FROM L2Kernel IMPORT Semaphore, SemV, SemP, SemInit, Guard, UnGuard;

VAR     PrGrd : Semaphore;

PROCEDURE  PrintLn;

BEGIN   Guard(PrGrd);
        Print(15C);
        Print(12C); 
        UnGuard(PrGrd);
END PrintLn;

PROCEDURE  PrintString( VAR s : ARRAY OF CHAR );

VAR     i : CARDINAL;

BEGIN   Guard(PrGrd);
        i := 0;
        WHILE ( i <= HIGH(s)) AND (s[i] <> 0C) DO
          Print(s[i]);
          INC(i);
          END;      
        UnGuard(PrGrd);
END PrintString;

MODULE  PrinterMonitor[4];

IMPORT Establish, Handler, Semaphore, SemV, SemP, SemInit;

EXPORT Print;

CONST   PrintBufferSize = 1024;

VAR     PrintData    [0FFC1C0H] : CARDINAL;
        PrintControl [0FFC1C4H] : BITSET;   (* Write only *)

        PControl : BITSET;
        Busy : BOOLEAN;
        Buffer : ARRAY [0..PrintBufferSize-1] OF CHAR;
        GetP, PutP, Count : CARDINAL;
        Holes : Semaphore;

PROCEDURE  Print( ch : CHAR );

BEGIN   IF Busy 
          THEN SemP(Holes);
               Put(ch)
          ELSE Send(ch)
          END
END Print;

PROCEDURE  Get( VAR ch : CHAR);

BEGIN   ch := Buffer[GetP];
        GetP := (GetP + 1) MOD PrintBufferSize;
        DEC(Count);
END Get;

PROCEDURE  Put( ch : CHAR);

BEGIN   Buffer[PutP] := ch;
        PutP := (PutP + 1) MOD PrintBufferSize;
        INC(Count);
END Put;

PROCEDURE  Send( ch : CHAR);

BEGIN   PrintData := ORD(ch);
        INCL(PControl,0);         (* Int enable *)
        EXCL(PControl,4);         (* Strobe line *)
        PrintControl := PControl;
        Busy := TRUE;
        INCL(PControl,4);
        PrintControl := PControl;
END Send;

PROCEDURE  PrinterInt;

VAR     ch : CHAR;

BEGIN   EXCL(PControl,0);
        PrintControl := PControl;
        Busy := FALSE;
        IF Count > 0 THEN
          Get(ch);
          Send(ch);
          SemV(Holes);
          END;
END PrinterInt;

BEGIN   GetP := 0;
        PutP := 0;
        Count := 0;
        PControl := {7,6,5,4};  (* There is other junk on this port *)
        PrintControl := PControl;
        SemInit(Holes,PrintBufferSize);
        Establish(43H,Handler(PrinterInt),0);
        Busy := FALSE;
END PrinterMonitor;

BEGIN   SemInit(PrGrd,1);
END L2Printer.
