IMPLEMENTATION MODULE L2System; 

(*$T-*)
(*$S-*)

(*  L2System.MOD -- Minimal support for M2-68K -- LUMOS kernel version
 
A simplified startup module derived from Smiler-2. Real arithmetic and 32 
bit signed arithmetic are not supported, neither are the process management
primitives.  All unsupported features cause an illegal instruction trap.
The module supports priorities (monitors) for supermode routines only.
A7 and A5 are set by the bootstrap loader.  Stack overflow causes
a TRAP #0 and all compiler generated halts (HALTX) cause TRAP #1. 
The LUMOS kernel does not use nested procedures and thus a display is
not needed.  Here, A5 is used as the kernel stack limit register. *)


FROM SYSTEM IMPORT CODE, REGISTER, ADDRESS, PROCESS;

(*$P-*)

PROCEDURE CASEX;
  BEGIN
    CODE(020137B);           (*       MOVEA.L (SP)+,A0     *)
    CODE(110220B);           (*       SUB.L   (A0),D0      *)
    CODE(130250B,4);         (*       CMP.L   4(A0),D0     *)
    CODE(061014B);           (*       BHI     ELSE         *)
    CODE(161500B);           (*       ASL.L   #1,D0        *)
    CODE(150250B,8);         (*       ADD.L   8(A0),D0     *)
    CODE(150360B,4000B);     (*       ADDA.W  (A0,D0.L),A0 *)
    CODE(047320B);           (*       JMP     (A0)         *)
    CODE(150750B,8);         (* ELSE: ADDA.L  8(A0),A0     *)
    CODE(054610B);           (*       SUBQ.L  #4,A0        *)
    CODE(047320B);           (*       JMP     (A0)         *)
  END CASEX;

(*$P-*)

PROCEDURE HALTX;
BEGIN   CODE(047101B);       (*       TRAP    #1           *)
END HALTX;

(*$P-*)

PROCEDURE STACKTEST;

BEGIN   CODE(020117B);       (*       MOVE.L  SP,A0        *)
        CODE(040750B,-128);  (*       LEA     -128(A0),A0  *)
        CODE(110700B);       (*       SUB.L   D0,A0        *)
        CODE(130715B);       (*       CMP.L   A5,A0        *)
        CODE(061002B);       (*       BHI     OK           *)
        CODE(047100B);       (*       TRAP    #0           *)
        CODE(047165B);       (* OK:   RTS                  *)
END STACKTEST;

PROCEDURE TERMINATE;
BEGIN   CODE(04AFCH);
END TERMINATE;

PROCEDURE NEWPROCESS (PROCESSCODE : PROC;
                      FWAWSP : ADDRESS; LENGTHWSP : ADDRESS;
                      VAR PROCESSDESCRIPTOR : PROCESS;
                      INITIALPRIO : ADDRESS);
BEGIN   CODE(04AFCH);
END NEWPROCESS;

PROCEDURE TRANSFER;
BEGIN   CODE(04AFCH);
END TRANSFER;

PROCEDURE IOTRANSFER;
BEGIN   CODE(04AFCH);
END IOTRANSFER;

PROCEDURE MODCALL;
BEGIN   CODE(04AFCH);
END MODCALL;

(*$P-*)
PROCEDURE SYSCALL;      (* Handles priority changes *)
                        (* Will give privilege fault in user mode *)

BEGIN   IF REGISTER(0) > 1 THEN CODE(04AFCH)        (* ILLEGAL *)
          ELSE IF REGISTER(0) = 0
                 THEN CODE(150374B,2000H) END;      (* ADDA.W  #2000H,A0 *)
               CODE(030010B);                       (* MOVE.W  A0,D0     *)
               CODE(043300B);                       (* MOVE    D0,SR     *)
               CODE(047165B);                       (* RTS               *)
          END;
END SYSCALL;

PROCEDURE MULU32     (MULTIPLICAN, MULTIPLIER : ADDRESS);
BEGIN
    CODE(044347B,07C00H);
    CODE(021056B,8); 
    CODE(022056B,12); 
    CODE(02601H,02801H,02A02H,04844H,04845H);
    CODE(0C2C2H,0C4C4H,0C6C5H,0C8C5H,04841H);
    CODE(0D242H,04285H,0D985H,0D243H,0D985H);
    CODE(04841H,04242H,04243H,04842H,04843H);
    CODE(0D483H,0D484H);
    CODE(026501B,12);
    CODE(026502B,8);
    CODE(046337B,03EH);
END MULU32;

PROCEDURE DIVU32     (DIVIDEND, DIVISOR       : ADDRESS);
BEGIN
    CODE(044347B,0F000H);
    CODE(021056B,12);
    CODE(020056B,8);
    CODE(04283H,082C0H,06800H,00010H,02401H);
    CODE(04241H,04841H,082C0H,03601H,03202H);
    CODE(082C0H,02001H,04841H,03203H,04841H);
    CODE(04240H,04840H);
    CODE(026501B,12);
    CODE(026500B,8);
    CODE(046337B,0FH);
END DIVU32;

PROCEDURE MULS32     (MULTIPLICAN, MULTIPLIER : ADDRESS);
BEGIN   CODE(04AFCH);
END MULS32;

PROCEDURE DIVS32     (DIVIDEND, DIVISOR       : ADDRESS);
BEGIN   CODE(04AFCH);
END DIVS32;

PROCEDURE FADD       (ADDER,       ADDEND     : REAL);
BEGIN   CODE(04AFCH);
END FADD;

PROCEDURE FSUB       (MINUEND,     SUBTRAHEND : REAL);
BEGIN   CODE(04AFCH);
END FSUB;

PROCEDURE FMUL       (MULTIPLICAN, MULTIPLIER : REAL);
BEGIN   CODE(04AFCH);
END FMUL;

PROCEDURE FDIV       (DIVIDEND,    DIVISOR    : REAL);
BEGIN   CODE(04AFCH);
END FDIV;

PROCEDURE FCMP       (DESTINATION, SOURCE     : REAL);
BEGIN   CODE(04AFCH);
END FCMP;

PROCEDURE FTST       (TOTEST                  : REAL);
BEGIN   CODE(04AFCH);
END FTST;

PROCEDURE FLOATX     (TOCONVERT               : ADDRESS);
BEGIN   CODE(04AFCH);
END FLOATX;

PROCEDURE TRUNCX     (TOCONVERT               : REAL);
BEGIN   CODE(04AFCH);
END TRUNCX;

(*$P-*)

BEGIN 
END L2System.
