IMPLEMENTATION MODULE L2Terminal;

FROM L2Kernel IMPORT  Semaphore, SemInit, SemV, SemP, Guard, UnGuard,
                      Process, CurrentProcess, Resume;
FROM L2Exceptions IMPORT AssertOk, Assert, Establish, Handler;
FROM L2Errors IMPORT Status;
FROM L2Clock IMPORT RequestAlarm, Event;
FROM SYSTEM IMPORT ADDRESS;

VAR     i : CARDINAL;
        UnitGuard, OutGuard : ARRAY [0..MaxTerm] OF Semaphore;

PROCEDURE  SetCCTrap( a : Handler; unit : CARDINAL) : Status;

(*  Arranges that a(unit) will be called if a <CTRL-C> is received on a
    particular unit.  If this is not called, <CTRL-C> is processed as an
    ordinary character (it acts as <CTRL-Z> on ReadString). *)

VAR     e : Status;

BEGIN   e := Check(unit);
        IF e <> Ok THEN RETURN e END;
        Terminals[unit].Killer := a;
        RETURN Ok;
END SetCCTrap;

PROCEDURE  PutStr( unit : CARDINAL; VAR s : ARRAY OF CHAR);

VAR     i : CARDINAL;

BEGIN   i := 0;
        WHILE (i <= HIGH(s)) AND (s[i] <> 0C) DO
          PutCh(unit,s[i]);
          INC(i);
        END;
END PutStr;

PROCEDURE  Check( unit : CARDINAL) : Status;  (* Validate access *)

BEGIN   IF unit > MaxTerm THEN RETURN NoUnit END;
        RETURN Ok;
END Check;

PROCEDURE  Read( unit : CARDINAL; VAR ch : CHAR; wait : BOOLEAN ) : Status;

(*  Unformatted, single character input without interpretation or
    echo.  The only special characters are <CTRL-C>, which causes a
    call to a CCTrap handler (if requested), and flow control  *)

VAR     s : Status;

BEGIN   s := Check(unit);
        IF s <> Ok THEN RETURN s END;
        s := GetCh(unit,ch,wait);
        RETURN s;
END Read;

PROCEDURE  RubOut( unit : CARDINAL; VAR i : CARDINAL);

BEGIN   IF i > 0 THEN 
          PutCh(unit,10C); 
          PutCh(unit,' '); 
          PutCh(unit,10C);
          DEC(i); 
          END
END RubOut;

PROCEDURE  ReadString( unit : CARDINAL; VAR s : ARRAY OF CHAR ) : Status;

(*  Formatted line input, processes <DEL>, <CTRL-U>, <CTRL-Z>, etc.
    The procedure blocks the process until a <RETURN> is seen.  Echoing
    is done for the characters read.  If too many characters are available for
    the string s, they are discarded. *)

VAR     i, j : CARDINAL;
        e : Status;

BEGIN   e := Check(unit);
        IF e <> Ok THEN RETURN e END;
        Guard(UnitGuard[unit]);
        i := 0;  e := Ok;
        LOOP AssertOk(GetCh(unit,s[i],TRUE),111);
             IF s[i] = 177C THEN 
               RubOut(unit,i) 
             ELSIF (s[i] = 3C) OR (s[i] = 32C) THEN
               s[i] := 0C;
               PutStr(unit,"<EOF>"); 
               PutCh(unit,15C); PutCh(unit,12C);
               e := EndFile;
               EXIT;
             ELSIF s[i] = 15C THEN
               s[i] := 0C;
               PutCh(unit,15C);
               PutCh(unit,12C);
               EXIT;
             ELSIF s[i] = 33C THEN
               s[i] := 0C;
               e := EscKey;
               EXIT;
             ELSIF s[i] = 25C THEN
               WHILE i > 0 DO RubOut(unit,i) END;
             ELSIF s[i] = 22C THEN
               PutCh(unit,15C); PutCh(unit,12C);
               s[i] := 0C;
               PutStr(unit,s);
             ELSIF s[i] < ' ' THEN (* ignore it *)
             ELSIF i < HIGH(s) THEN
               PutCh(unit,s[i]);
               INC(i);
               END;
             END;
        UnGuard(UnitGuard[unit]);
        RETURN e;
END ReadString;

PROCEDURE  Write( unit : CARDINAL; ch : CHAR) : Status;

VAR     s : Status;

BEGIN   s := Check(unit);
        IF s <> Ok THEN RETURN s END;
        Guard(UnitGuard[unit]);
        PutCh(unit,ch);
        UnGuard(UnitGuard[unit]);
        RETURN Ok;
END Write;

PROCEDURE  WriteString( unit : CARDINAL; VAR s : ARRAY OF CHAR) : Status;

VAR     e : Status;

BEGIN   e := Check(unit);
        IF e <> Ok THEN RETURN e END;
        Guard(UnitGuard[unit]);
        PutStr(unit,s);
        UnGuard(UnitGuard[unit]);
        RETURN Ok;
END WriteString;

PROCEDURE  WriteLn( unit : CARDINAL) : Status;

VAR     e : Status;

BEGIN   e := Check(unit);
        IF e <> Ok THEN RETURN e END;
        Guard(UnitGuard[unit]);
        PutCh(unit,15C); PutCh(unit,12C);
        UnGuard(UnitGuard[unit]);
        RETURN Ok;
END WriteLn;

PROCEDURE  BTWrite( unit : CARDINAL; VAR msg : ARRAY OF CHAR) : Status;

(* Non blocking breakthrough write *)

VAR     e : Status;
        i : CARDINAL;

BEGIN   e := Check(unit);
        IF e <> Ok THEN RETURN e END;
        Guard(OutGuard[unit]);
        i := 0;
        WHILE ( i <= HIGH(msg)) AND (msg[i] <> 0C) DO
          e := BTPutCh(unit,msg[i]);
          IF e <> Ok THEN 
            UnGuard(OutGuard[unit]);
            RETURN e;
            END;
          INC(i);
          END;
        UnGuard(OutGuard[unit]);
        RETURN Ok;
END BTWrite;

PROCEDURE  PutCh( unit : CARDINAL; ch : CHAR);

BEGIN   Guard(OutGuard[unit]);  (* don't mix with BTWrite *)
        NPutCh(unit,ch);
        UnGuard(OutGuard[unit]);
END PutCh; 

MODULE TerminalMonitor[4];

IMPORT Semaphore, SemInit, SemV, SemP, AssertOk, Assert, Establish, Handler,
       Status, ADDRESS, MaxTerm, RealTerms, RequestAlarm, Event, 
       CurrentProcess, Resume, Process;

EXPORT NPutCh, GetCh, Terminals, Attach, Detach, SetBaud, GetBaud, BTPutCh,
       VTIn, VTOut, VTAllocate, VTDeallocate, VTHangUp, VTClean;

CONST   TerminalBufferSize = 256;

TYPE    Duart = RECORD MRA, CSRA, CRA, TRBA, 
                       ACR, IMR, CTUR, CTLR,
                       MRB, CSRB, CRB, TRBB,
                       IVR, OPCR, OPRBS, OPRBC : CARDINAL;
                END;
        PortPointer = POINTER TO Duart;
        Buffer = RECORD  buf : ARRAY [0..TerminalBufferSize-1] OF CHAR;
                         GetP, PutP, Count : CARDINAL;
                 END;
        TCB = RECORD rBuff, tBuff : Buffer;
                     Chars, Holes : Semaphore;
                     Done, Frozen, Virtual : BOOLEAN;
                     SpeedMask : CARDINAL;
                     Killer : Handler;
                     Port : PortPointer;
                     Pid, Attached : Process;
              END;

VAR     Terminals : ARRAY [0..MaxTerm] OF TCB;

PROCEDURE  Init( unit, tv, rv : CARDINAL; port : ADDRESS; speed : CARDINAL );

BEGIN   WITH Terminals[unit] DO
          InitBuffer(rBuff);
          InitBuffer(tBuff);
          SemInit(Chars,0);
          SemInit(Holes,TerminalBufferSize);
          Frozen := FALSE;          
          SpeedMask := speed;
          Attached := 0;
          Killer := Handler(NIL);
          Port := PortPointer(port);
          Virtual := FALSE;
          Establish(tv,OutputInt,unit);
          Establish(rv,InputInt,unit);
          WITH Port^ DO
            IF NOT ODD(unit) THEN  (* Program the chip *)
              ACR := 80H;          (* Baud set 1       *)
              IMR := 0;
              OPCR := 0F0H;        (* Connect interrupts to Rdy bits *)
              OPRBS := 0FH;
              OPRBC := 0F0H;
              END;
            CRA := 10H;   (* MR1              *)
            MRA := 13H;   (* 8 bit, no parity *)
            MRA := 7;     (* 1 stop bit       *)
            CSRA := SpeedMask;
            CRA := 1;     (* Enable receiver  *)
            END;
          END;
END Init;                                    

PROCEDURE  Attach( unit : CARDINAL) : Status;

BEGIN   IF unit > MaxTerm THEN RETURN NoUnit END;
        Terminals[unit].Attached := CurrentProcess;
        RETURN Ok;
END Attach;

PROCEDURE  Detach( unit : CARDINAL) : Status;

BEGIN   IF unit > MaxTerm THEN RETURN NoUnit END;
        Terminals[unit].Attached := 0;
        RETURN Ok;
END Detach;

PROCEDURE  SetBaud( unit, speedmask : CARDINAL) : Status;

BEGIN   IF unit > MaxTerm THEN RETURN NoUnit END;
        WITH Terminals[unit] DO
        IF Virtual THEN RETURN NoFunction END;
          SpeedMask := speedmask;
          Port^.CSRA := SpeedMask;
          END;
        RETURN Ok;
END SetBaud;

PROCEDURE  GetBaud( unit : CARDINAL; VAR speedmask : CARDINAL) : Status;

BEGIN   IF unit > MaxTerm THEN RETURN NoUnit END;
        IF Terminals[unit].Virtual THEN RETURN NoFunction END;
        speedmask := Terminals[unit].SpeedMask;
        RETURN Ok;
END GetBaud;

PROCEDURE  InitBuffer( VAR b : Buffer );

BEGIN   WITH b DO
          GetP := 0;
          PutP := 0;
          Count := 0;
          END;
END InitBuffer;

PROCEDURE  Get( VAR b : Buffer; VAR ch : CHAR);

BEGIN   WITH b DO
          ch := buf[GetP];
          GetP := (GetP + 1) MOD TerminalBufferSize;
          DEC(Count);
          END;
END Get;

PROCEDURE  Put( VAR b : Buffer; ch : CHAR); 

BEGIN   WITH b DO
          buf[PutP] := ch;
          PutP := (PutP + 1) MOD TerminalBufferSize;
          INC(Count);
          END;
END Put;

PROCEDURE  Empty ( VAR b : Buffer ) : BOOLEAN;

BEGIN   RETURN b.Count = 0
END Empty;

PROCEDURE  Full ( VAR b : Buffer ) : BOOLEAN;

BEGIN   RETURN b.Count = TerminalBufferSize;
END Full;

PROCEDURE  NPutCh ( unit : CARDINAL; ch : CHAR );

(* Put a character into the unit buffer, waiting if necessary *)
                               
VAR
  rs:Status;

BEGIN   WITH Terminals[unit] DO
          SemP(Holes);
          Put(tBuff,ch);
          IF NOT Frozen THEN
            IF Virtual THEN
              IF Pid<>0 THEN rs:=Resume(Pid) END;
            ELSE
              Port^.CRA := 4 END; (* Enable Transmitter & Interrupts *)
            END(*IF*);
          END;
END NPutCh;

PROCEDURE  BTPutCh ( unit : CARDINAL; ch : CHAR ) : Status;

(* Put a character into the unit buffer, return error if no room *)
VAR
  rs : Status;

BEGIN   WITH Terminals[unit] DO
          IF tBuff.Count + 2 >= TerminalBufferSize THEN RETURN DeviceBusy END;
          SemP(Holes);
          Put(tBuff,ch);
          IF NOT Frozen THEN 
            IF Virtual THEN
               IF Pid<>0 THEN rs:=Resume(Pid) END;
               ELSE
                 Port^.CRA := 4 END; (* Enable Transmitter & Interrupts *)
               END(*IF*);
          END;
        RETURN Ok;
END BTPutCh;

PROCEDURE  OutputInt( unit : CARDINAL );

(*  Called when an output interrupt occurs on a unit *)

VAR     ch : CHAR;
        i : CARDINAL;         
        s : Status;   
        e : Event;

BEGIN   WITH Terminals[unit] DO
          IF Empty(tBuff) OR Frozen 
            THEN Port^.CRA := 8  (* Switch off the transmitter *)
            ELSE Get(tBuff,ch);
               Port^.TRBA := ORD(ch);
               SemV(Holes);
            END;
          END;
END OutputInt;   

PROCEDURE  GetCh( unit : CARDINAL; VAR ch : CHAR; wait : BOOLEAN) : Status;

BEGIN   WITH Terminals[unit] DO
          IF Empty(rBuff) AND NOT wait
            THEN RETURN EmptyBuff;
            ELSE SemP(Chars);
                 Get(rBuff,ch);
            END;
        END;
        RETURN Ok;
END GetCh;

PROCEDURE  InputInt( unit : CARDINAL );

(* Called when an input interrupt occurs on a unit *)

VAR     ch : CARDINAL;
        rs : Status;

BEGIN   WITH Terminals[unit] DO
          ch := Port^.TRBA MOD 256;
          IF ch = 17 
            THEN Frozen := FALSE;
                 Port^.CRA := 4;  (* Enable interrupts *)
                 RETURN;
          ELSIF ch = 19 
            THEN Frozen := TRUE;
                 RETURN;  
          ELSIF (ch = 3) AND (ADDRESS(Killer) <> NIL) 
            THEN Killer(unit) 
          ELSIF NOT Full(rBuff)
            THEN Put(rBuff,CHR(ch));
                 SemV(Chars);
            END; 
          IF Attached <> 0 THEN rs := Resume(Attached) END;
          END;
END InputInt;

PROCEDURE  VTInit( unit : CARDINAL );
  BEGIN
    WITH Terminals[unit] DO
      InitBuffer(rBuff);
      InitBuffer(tBuff);
      SemInit(Chars,0);
      SemInit(Holes,TerminalBufferSize);
      Frozen := FALSE;
      Attached := 0;
      Killer:=Handler(NIL);
      Pid:=0;
      Virtual := TRUE;
      Done := TRUE;
    END; (* WITH *)
  END VTInit;

PROCEDURE VTAllocate( VAR unit : CARDINAL ) : Status;
  VAR
    i : CARDINAL;
  BEGIN
    i := RealTerms;
    LOOP
      IF Terminals[i].Pid = 0 THEN 
        Terminals[i].Pid := CurrentProcess;
        Terminals[i].Done := FALSE;
        unit := i;
        RETURN Ok;
      END; (* IF *)
      IF i = MaxTerm THEN RETURN NoSlot; END;
      INC( i );
    END; (* LOOP *)
  END VTAllocate;

PROCEDURE VTDeallocate( unit : CARDINAL );
  BEGIN         
    IF (unit > MaxTerm) OR ( Terminals[unit].Pid = 0 ) OR
       ( NOT Terminals[unit].Virtual) OR 
       ( Terminals[unit].Pid <> CurrentProcess  )THEN 
      RETURN;
    END; (* IF *)
    Terminals[unit].Pid := 0;
    Terminals[unit].Attached := 0;
  END VTDeallocate;
 
PROCEDURE VTHangUp( unit : CARDINAL ) : Status;
  VAR
    rs : Status;
  BEGIN
    IF (unit > MaxTerm) OR ( NOT Terminals[unit].Virtual) OR
       ( Terminals[unit].Pid = 0 ) THEN
      RETURN IllegalOp; 
    END; (* IF *)
    rs := Resume( Terminals[unit].Pid ); 
    Terminals[unit].Done := TRUE;             
    RETURN Ok;
  END VTHangUp;
              
PROCEDURE VTClean( pid : CARDINAL );
  VAR
    i : CARDINAL;
  BEGIN
    FOR i := RealTerms TO MaxTerm DO
      IF Terminals[i].Pid = pid THEN
        Terminals[i].Pid := 0;
      END;
    END;
  END VTClean;
           
PROCEDURE VTOut( unit : CARDINAL; VAR ch : CHAR ) : Status;
  BEGIN
    IF (unit > MaxTerm) OR ( NOT Terminals[unit].Virtual) THEN 
      RETURN IllegalOp; 
    END; (* IF *)
    WITH Terminals[unit] DO
      IF CurrentProcess <> Pid THEN RETURN NoPriv; END;
      IF Empty( tBuff ) OR Frozen THEN 
        IF Done THEN RETURN NotActive; END;
        RETURN EmptyBuff;
      ELSE
        Get( tBuff, ch );
        SemV( Holes );
        RETURN Ok;
      END; (* IF *)    
    END; (* WITH *)
  END VTOut;

PROCEDURE VTIn( unit : CARDINAL; ch : CHAR ) : Status;
  VAR
    rs : Status;
  BEGIN
    IF (unit > MaxTerm) OR ( NOT Terminals[unit].Virtual) THEN 
      RETURN IllegalOp; 
    END; (* IF *)
    WITH Terminals[unit] DO                     
      IF CurrentProcess <> Pid THEN RETURN NoPriv; END;
      IF Done THEN RETURN NotActive; END;
      IF ch = CHR(17) THEN
        Frozen := FALSE;
        RETURN Ok;
      ELSIF ch = CHR(19) THEN
        Frozen := TRUE;
        RETURN Ok;
      ELSIF (ch=CHR(3)) AND (ADDRESS(Killer)<>NIL) THEN
        Killer(unit);
      ELSIF NOT Full(rBuff) THEN
        Put( rBuff, ch );
        SemV( Chars );
      END; (* IF *)
      IF Attached <> 0 THEN rs := Resume(Attached); END;
      RETURN Ok;
    END; (* WITH *)
  END VTIn;

VAR
  i : CARDINAL;                             
BEGIN   
  Init(0,4BH,4FH,0FFC300H,0CCH);  (* 19.2 k *)
  Init(1,4AH,4EH,0FFC310H,0CCH);
  Init(2,49H,4DH,0FFC340H,0CCH);
  Init(3,48H,4CH,0FFC350H,0CCH);
  Init(4,69H,6FH,0FFC600H,0CCH);
  Init(5,68H,6EH,0FFC610H,0CCH);
  Init(6,67H,6DH,0FFC640H,0CCH);
  Init(7,66H,6CH,0FFC650H,0CCH);
  Init(8,65H,6BH,0FFC680H,0CCH);
  Init(9,64H,6AH,0FFC690H,066H);   (* 1200 *)
  FOR i := RealTerms TO MaxTerm DO VTInit( i ); END;
END TerminalMonitor;

BEGIN   FOR i := 0 TO MaxTerm DO 
          SemInit(UnitGuard[i],1);
          SemInit(OutGuard[i],1);
          END;
END L2Terminal.
