IMPLEMENTATION MODULE L2Transport;

FROM L2Errors     IMPORT Status;
FROM L2Kernel     IMPORT Semaphore, SemV, SemP, SemInit, SemQ,
                         SetNoSwitch, Guard, UnGuard;
FROM L2Exceptions IMPORT Establish, IntControl;

CONST   NETWORK = 6;
        NETINT  = 5;
        maxlength = 1023;

VAR     WHADDR [0FFC380H] : CARDINAL;
        WLADDR [0FFC384H] : CARDINAL;
        WDATA  [0FFC386H] : CARDINAL;
        NADDR  [0FFC388H] : CARDINAL;
        WCMDAR [0FFC382H] : CARDINAL;
        OSTAT  [0FFC380H] : BITSET;
        CMNode [0FF8026H] : CHAR;    (* CMOS stored node number *)
        Oguard, Tsync, Rsync : Semaphore;
        Initialised : BOOLEAN;

PROCEDURE Delay;

VAR     i : CARDINAL;

BEGIN FOR i := 1 TO 10 DO END;
END Delay;

PROCEDURE  Ready;

BEGIN   WHILE NOT (7 IN OSTAT) DO END;
END Ready;

PROCEDURE  SetCmdAddr( a : CARDINAL);

BEGIN   Ready;
        WCMDAR := 0;
        Ready;
        WCMDAR := a DIV 256;
        Ready;
        WCMDAR := a MOD 256;
        Delay;
END SetCmdAddr;
                                   
PROCEDURE  PointRam( a : CARDINAL);

BEGIN   WLADDR := a MOD 256;
        Delay;
        WHADDR := a DIV 256;
        Delay;
END PointRam;

PROCEDURE  WriteRamChar( c : CHAR);

BEGIN   WDATA := ORD(c);
        Delay;
END WriteRamChar;

PROCEDURE  WriteRamByte( c : CARDINAL);

BEGIN   WDATA := c MOD 256;
        Delay;
END WriteRamByte;

PROCEDURE ReadRamChar( VAR c : CHAR );

BEGIN   c := CHR( WDATA MOD 256 );
        Delay;
END ReadRamChar;

PROCEDURE  ReadRamByte( VAR c : CARDINAL);

BEGIN   c := WDATA MOD 256;
        Delay;
END ReadRamByte;

PROCEDURE Command( a, r : CARDINAL);

BEGIN         
   SetNoSwitch(TRUE);
   PointRam(r);
   WriteRamByte(0FFH);
   SetCmdAddr(a);
   LOOP SemP(Tsync); 
        SetNoSwitch(TRUE);
        PointRam(r);
        ReadRamByte(Result);
        IF Result <> 0FFH THEN EXIT END;
        END;
   SetNoSwitch(FALSE);
END Command;

PROCEDURE SetUpReceive;

BEGIN   Command(09H,3F7H);
END SetUpReceive;

PROCEDURE WriteBlock( node : CARDINAL; VAR b : ARRAY OF CHAR) : Status;

VAR     i, k : CARDINAL;
        rs : Status;

BEGIN
   Guard(Oguard);
   IF NOT Initialised THEN 
     TransportInit;
     IF Result <> 0FEH THEN 
       UnGuard(Oguard);
       RETURN DevError;
       END;
     END;
   SetNoSwitch(TRUE);
   PointRam(29H);
   WriteRamByte(node);
   k := HIGH(b);
   IF k > maxlength THEN k := maxlength END;
   PointRam(26H);
   WriteRamByte((k+1) DIV 256);
   WriteRamByte((k+1) MOD 256);
   PointRam(800H);
   FOR i := 0 TO k DO
     WriteRamChar(b[i]);
   END;      
   Command(1EH,15H);
   IF Result = 0 
     THEN rs := Ok
     ELSE CASE Result OF
            80H : IF Echo(node) 
                    THEN rs := TimedOut 
                    ELSE rs := NoSuch 
                    END                  |
            81H : rs := OverRun          |
            82H : rs := NoSuch           |
            83H : rs := NoSuch
            ELSE  rs := DevError
            END;
     END;   
   UnGuard(Oguard);
   RETURN rs;
END WriteBlock;

PROCEDURE ReadBlock( VAR b : ARRAY OF CHAR) : Status;

VAR    i, k : CARDINAL;

BEGIN
   Guard(Oguard);
   IF NOT Initialised THEN 
     TransportInit;
     IF Result <> 0FEH THEN 
       UnGuard(Oguard);
       RETURN DevError;
       END;
     END;
   LOOP SetNoSwitch(TRUE);
        PointRam(3F7H);
        ReadRamByte(Result);
        IF Result = 0 THEN EXIT END;
        UnGuard(Oguard);  (* Allow other ops while we wait *)
        SemP(Rsync);    
        Guard(Oguard);    (* One operation at a time *)
        END;
   SetNoSwitch(TRUE);
   PointRam(3F8H);
   k := HIGH(b);
   IF k > maxlength+8 THEN k := maxlength+8 END;
   FOR i := 0 TO k DO
     ReadRamChar(b[i]);
     END;
   SetUpReceive;
   UnGuard(Oguard);
   RETURN Ok;
END ReadBlock;    

PROCEDURE  Echo( n : CARDINAL) : BOOLEAN;

BEGIN   SetNoSwitch(TRUE);
        PointRam(34H);
        WriteRamByte(n);
        Command(30H,2FH);
        RETURN Result = 0C0H;
END Echo;

PROCEDURE Init;

BEGIN
   SetNoSwitch(TRUE);
   PointRam(9);              (* Layout setup receive command *)
   WriteRamByte(0F0H);
   WriteRamByte(0);
   WriteRamByte(3);
   WriteRamByte(0F7H);
   WriteRamByte(0A0H);
   WriteRamByte(0);
   WriteRamByte(4);
   WriteRamByte(0);
   WriteRamByte(4);
   WriteRamByte(0);
   WriteRamByte(5);
   PointRam(19H);            (* control text *)
   WriteRamChar('L');
   WriteRamChar('U');
   WriteRamChar('M');
   WriteRamChar('O');
   WriteRamChar('S');
   PointRam(1EH);            (* Layout send command *)
   WriteRamByte(40H);
   WriteRamByte(0);
   WriteRamByte(0);
   WriteRamByte(15H);
   WriteRamByte(0A0H);
   WriteRamByte(0);
   WriteRamByte(8);
   WriteRamByte(0);
   WriteRamByte(4);
   WriteRamByte(0);
   WriteRamByte(5);
   PointRam(2BH);            (* Layout init command *)
   WriteRamByte(20H);
   WriteRamByte(0);
   WriteRamByte(0);
   WriteRamByte(2AH);
   PointRam(30H);            (* Layout Echo command *)
   WriteRamByte(2);
   WriteRamByte(0);
   WriteRamByte(0);
   WriteRamByte(2FH);
   Command(2BH,2AH);         (* Do init *)
END Init;

MODULE Interrupts[4];

IMPORT Establish, Tsync, Rsync, SemV, SemQ, NETINT, IntControl;

PROCEDURE NetInt( e : CARDINAL);

BEGIN
   EXCL(IntControl,NETINT);
   INCL(IntControl,NETINT);
   IF SemQ(Tsync) < 1 THEN SemV(Tsync) END;
   IF SemQ(Rsync) < 1 THEN SemV(Rsync) END;
END NetInt;

BEGIN
   Establish(47H,NetInt,0);
END Interrupts;

PROCEDURE  TransportInit;

BEGIN
   INCL(IntControl,NETINT);
   Init;  
   IF Result <> NodeNumber THEN RETURN END;
   SetUpReceive;  
   IF Result <> 0FEH THEN RETURN END;
   Initialised := TRUE;
END TransportInit;

BEGIN 
   SemInit(Tsync,0);
   SemInit(Rsync,0);
   SemInit(Oguard,1);
   NodeNumber := ORD(CMNode) MOD 256;
   NADDR := NodeNumber;
   INCL(IntControl,NETWORK);
   Initialised := FALSE;
END L2Transport.
