IMPLEMENTATION MODULE L2Traps[4];

FROM SYSTEM IMPORT REGISTER, ADDRESS, CODE, SETREG;
FROM L2Machine IMPORT MoveToSR, S0;
FROM L2Console IMPORT Init, WriteAddr, WriteString, WriteLn, WriteCard;
FROM L2Errors IMPORT Status, GetMessage;
FROM L2Kernel IMPORT Process, ProcMax, CurrentProcess;
FROM L2Exceptions IMPORT ReBoot, Establish, AssertOk;
FROM L2Loader IMPORT SetLastPC, Exit;

TYPE    TrapRec = RECORD Last : Status;
                         Fatal : BOOLEAN;
                  END;

VAR     Traps : ARRAY Process OF TrapRec;

PROCEDURE  SetTrapMode( fatal : BOOLEAN);

BEGIN   Traps[CurrentProcess].Fatal := fatal;
END SetTrapMode;

PROCEDURE  TestTrap() : Status;

VAR     t : Status;

BEGIN   t := Traps[CurrentProcess].Last;
        Traps[CurrentProcess].Last := Ok;
        RETURN t;
END TestTrap;

(*$S-*)  (* No more stack errors please *)

PROCEDURE  PeekAddr( a : ADDRESS ) : ADDRESS;

TYPE    aptr = POINTER TO ADDRESS;

VAR     ap : aptr;

BEGIN   ap := aptr(a);
        RETURN ap^;
END PeekAddr;

PROCEDURE  PeekCard( a : ADDRESS ) : CARDINAL;

TYPE    cptr = POINTER TO CARDINAL;

VAR     cp : cptr;

BEGIN   cp := cptr(a);
        RETURN cp^;
END PeekCard;

PROCEDURE  Trap( r : CARDINAL);

VAR     rs : Status;
        pc, sp, a6, d0, a5 : ADDRESS;
        sr, sro, pco : CARDINAL;
        str : ARRAY [0..39] OF CHAR;

BEGIN   rs := Status(r);
        pco := 0; sro := 0;
        IF (rs = BusError) OR (rs = AddrError)
          THEN pco := 8; sro := 8; END;
        IF (rs = Halt) OR (rs = StackOver) 
          THEN pco := 4; END;
        a5 := REGISTER(13);
        a6 := REGISTER(14);       (* here we know about stack frame *)
        a6 := PeekAddr(a6); 
        sr := PeekCard(a6+ADDRESS(64+sro));
        pc := PeekAddr(a6+ADDRESS(66+pco));
        d0 := PeekAddr(a6+4);
        sp := a6+ADDRESS(70+pco);
        IF 13 IN BITSET(sr) (* SuperMode *)
          THEN SETREG(13,0);  (* Kill the stack checks *)
               IF (rs = Halt) AND (d0 <> 0) THEN
                 IF d0 = 1 
                   THEN rs := NoReturn 
                   ELSE rs := RangeError 
                   END;
                 END;
               Init;
               WriteLn;
               WriteString("PANIC -- ");
               GetMessage(rs,str);
               WriteString(str);
               WriteLn;
               WriteString("PC = ");
               WriteAddr(pc,16,1);
               WriteString("  SR = ");
               WriteAddr(ADDRESS(sr),16,1);
               WriteString("  SP = ");
               WriteAddr(sp,16,1);
               WriteString("  SL = ");
               WriteAddr(a5,16,1);
               WriteString(" Current Process = ");
               WriteCard(CARDINAL(CurrentProcess),1);
               WriteLn;
               ReBoot;
          ELSE IF (rs = StackOver) OR (rs = Halt) THEN rs := BadTrap END;
               WITH Traps[CurrentProcess] DO
                 IF Fatal OR (Last <> Ok) OR (rs = BusError)
                    OR (rs = AddrError) THEN
                   CODE(MoveToSR,S0);  (* Drop priority *)
                   AssertOk(SetLastPC(CurrentProcess,pc),401);
                   Exit(rs);
                   END;
                 Last := rs;
                 END;
          END;
END Trap;

MODULE Parity[7];

  IMPORT Init, WriteString, WriteLn, ReBoot, Establish, SETREG;

  PROCEDURE  ParCrash( bd : CARDINAL);

  BEGIN   SETREG(13,0);
          Init;
          WriteString('PANIC -- Memory parity fault on ');
          CASE bd OF
            0 : WriteString('System board') |
            1 : WriteString('Winchester board')
            END;
          WriteLn;
          ReBoot;
  END ParCrash;

BEGIN   Establish(50H,ParCrash,0);
        Establish(51H,ParCrash,1);
END Parity;

VAR     i : Process;
        k : CARDINAL;

BEGIN   FOR i := 1 TO ProcMax DO 
          WITH Traps[i] DO
            Fatal := TRUE;
            Last := Ok;
            END;
          END;
        Establish(2,Trap,ORD(BusError));
        Establish(3,Trap,ORD(AddrError));
        Establish(4,Trap,ORD(IllInst));
        Establish(5,Trap,ORD(DivZero));
        Establish(6,Trap,ORD(BoundsError));
        Establish(7,Trap,ORD(ArithOver));
        Establish(8,Trap,ORD(PrivViol));
        Establish(9,Trap,ORD(TraceTrap));
        Establish(32,Trap,ORD(StackOver));  (* LUMOS kernel only *)
        Establish(33,Trap,ORD(Halt)); 
        FOR k := 34 TO 46 DO Establish(k,Trap,ORD(BadTrap)) END;
END L2Traps.
