IMPLEMENTATION MODULE N2Arp;
FROM N2Packets IMPORT PacketPtr,NewPacket,DisposePacket;
FROM L2Clock   IMPORT Ticks,TickCount;
FROM N2Base    IMPORT HostIP;
FROM L2Kernel  IMPORT Semaphore,SemInit,SemP,SemV,SetNoSwitch;
FROM L2Storage IMPORT ALLOCATE,DEALLOCATE;
FROM N2Errors  IMPORT Status;
FROM N2Driver  IMPORT WriteBlock,HostMac;
FROM SYSTEM    IMPORT ADDRESS;
             
CONST                   
   maxtime=3000;              (*3000 ticks or 30 seconds for tmo*)
   ReQuest=1C;                (*ARP request number*)
   RePly=2C;                  (*ARP response number*)
   HToffset=14;               (*Offsets in the packet*)
   PToffset=16;
   HLenoffset=18;
   PLenoffset=19;
   SHoffset=22;
   SIPoffset=28;
   TIPoffset=38;
   OPoffset=20;
   THoffset=32;
   soffset=6;TFoffset=12;
   Hlen=6C;                   (*Physical address length is 6 bytes*)
   Plen=4C;                   (*Protocol address length is 4 bytes *)
   PTnum=0800H;               (*Indicates IP addresses*)
   TFnum=0806H;               (*An ARP type Packet*)
   HType=1C;                  (*Ethernet Type *)

TYPE
   PhyAddr = ARRAY[0..5] OF CHAR;
   ArpPointer = POINTER TO ArCache;
   ArCache = RECORD
               Ipnumber : ARRAY[0..3] OF CHAR;      (*A cache node*)
               Physaddr : PhyAddr;
               TStamp : Ticks;
               next : ArpPointer;
               datagram : PacketPtr;
               resolving : BOOLEAN;
             END;

   ArpCache = RECORD
                head:ArpPointer;
              END;                                              

   GStats = RECORD                            (*Stats variables*)
              ArpResponse,ArpRequest,TArpPackets:CARDINAL;
              numcache,PacketsDisc:CARDINAL;
              PacketsSent:ADDRESS;
            END;

VAR           
   cache:ArpCache;
   maxsize:CARDINAL;                  
   guard:Semaphore;                      
   Astats:GStats;
                                                          
(*This procedure intializes all the statstics variables to zero*)

PROCEDURE InitGlobals();
BEGIN
  Astats.ArpResponse:=0;          
  Astats.ArpRequest:=0;
  Astats.PacketsSent:=0;
  Astats.TArpPackets:=0;
  Astats.numcache:=0;
  Astats.PacketsDisc:=0;
  maxsize:=10;
END InitGlobals;
         
(*This procedure returns all the statistics to the caller*)

PROCEDURE ArpStats(VAR ArpResp,ArpReq,Cachenum,Pdisc,TArpPack:CARDINAL;
                          VAR Psent:ADDRESS);
BEGIN
  ArpResp:=Astats.ArpResponse;
  ArpReq:=Astats.ArpRequest;
  Cachenum:=Astats.numcache;
  Psent:=Astats.PacketsSent;
  Pdisc:=Astats.PacketsDisc;
  TArpPack:=Astats.TArpPackets;
END ArpStats;

              
(*This procedure returns the information for the Arp node num*)
PROCEDURE StatsCache(VAR Ipnum,Hnum:ARRAY OF CHAR;VAR resolved:BOOLEAN;
                         num:CARDINAL):BOOLEAN;
VAR
   p:ArpPointer;
   i:CARDINAL;
BEGIN
   i:=0;     
   SetNoSwitch(TRUE);                        (*protects critical section*)
   p:=cache.head;
   WHILE (p<>NIL) AND (i<>num) DO            (*locates the node number*)
     p:=p^.next;
     i:=i+1;
   END(*WHILE*);
   IF p=NIL THEN   
     SetNoSwitch(FALSE);
     RETURN FALSE;
   ELSE
     FOR i:=0 TO 3 DO                         (*inserts IP number*)
       Ipnum[i]:=p^.Ipnumber[i];
     END(*FOR*);
     FOR i:=0 TO 5 DO                         (*inserts Phys. address*)
       Hnum[i]:=p^.Physaddr[i];
     END(*FOR*);     
     IF p^.resolving THEN                     (*Set resolved field*)
        resolved:=FALSE;                      
     ELSE
        resolved:=TRUE;
     END;
     SetNoSwitch(FALSE);
     RETURN TRUE;
   END(*IF*);  
END StatsCache;
                                                                              
(*This procedure searches the ARP cache for an entry that matches the IP 
  address given. If found a pointer is assigned to it*)

PROCEDURE ArpFind(NumIp:ARRAY OF CHAR;VAR p:ArpPointer):BOOLEAN;
BEGIN
  p:=cache.head;
  WHILE p <> NIL DO
    IF  ((p^.Ipnumber[0]=NumIp[0]) AND (p^.Ipnumber[1]=NumIp[1])
         AND (p^.Ipnumber[2]=NumIp[2]) AND (p^.Ipnumber[3]=NumIp[3])) THEN 
       RETURN TRUE;
    ELSE
       p:=p^.next;
    END(*IF*);
  END(*WHILE*);
  RETURN FALSE;
END ArpFind;
                                                                              
(*This procedure adds an entry to the cache with the proper IP number 
  inserted, but with the phys. address unresolved*)

PROCEDURE ArpAdd(NumIp:ARRAY OF CHAR;VAR r:PacketPtr);
VAR
  p,q:ArpPointer;               
  i:CARDINAL;
BEGIN          
  IF (Astats.numcache=maxsize) THEN
    CleanCache();                (*tries to make room in the cache*)
  END(*IF*);
  p:=cache.head;
  NEW(q);                        (*makes a new cache entry at start of list*)
  q^.next:=cache.head;
  cache.head:=q;
  INC(Astats.numcache);
  FOR i:=0 TO 3 DO
    q^.Ipnumber[i]:=NumIp[i];    (*inserts IP number in cache entry*)
  END(*FOR*);
  q^.datagram:=r;                (*queues the packet waiting for phys. addr.*)
  q^.TStamp:=TickCount;          (*sets time stamp*)
  q^.resolving:=TRUE;            (*sets resolving as true*)
  r:=NIL;     
  q:=NIL;
END ArpAdd;

(*This procedure tries to remove all stale entries from the cache*)
PROCEDURE CleanCache();
VAR
  p,q,d:ArpPointer;
  lapsed:Ticks;
BEGIN
  q:=cache.head;p:=NIL;     
  WHILE q<>NIL DO
    lapsed:=TickCount-q^.TStamp;
    IF (lapsed>maxtime) AND (NOT q^.resolving) THEN
      IF p=NIL THEN                    (*deleting 1st cache entry*)
        d:=q;
        q:=q^.next;
        cache.head:=cache.head^.next;
      ELSE                         (*general case for deleting cache entry*)
        d:=q;
        q:=q^.next;
        p^.next:=q;
      END(*IF*);
      IF d^.datagram<>NIL THEN          (*removes a queued packet*)
        DisposePacket(d^.datagram);
      END;                                     
      DEC(Astats.numcache);
      DISPOSE(d);
    ELSE
     p:=q;
     q:=q^.next;
    END(*IF*);
  END(*WHILE*);
END CleanCache;
                 
(*This procedure updates cache entries with the info. given in the ARP packet 
  and sends out an ARP response if an ARP request was given*)

PROCEDURE ArpIn(VAR r:PacketPtr);
VAR
   SIpnum:ARRAY[0..3] OF CHAR;
   p:ArpPointer;
   q:PacketPtr;
   i:CARDINAL;          
   rs:Status;
BEGIN     
   INC(Astats.TArpPackets);
   IF ((r^.data[HToffset]=0C) AND (r^.data[HToffset+1]=HType)) THEN (*Ethernet type?*)
     IF ((r^.data[PToffset]=CHR(PTnum DIV 100H)) AND 
        (r^.data[PToffset+1]=CHR(PTnum MOD 100H))) THEN    (*Check IP TYPE*)
       FOR i:=0 TO 3 DO                        (*Extracts Sender IP num*)
          SIpnum[i]:=r^.data[SIPoffset+i];
       END(*FOR*);
       SemP(guard);      

     (*updates ARP table if an entry exists for that IP number and
       sends any queued packet*)                 
  
       IF ArpFind(SIpnum,p) THEN              
         p^.resolving:=FALSE;
         FOR i:=0 TO 5 DO
           p^.Physaddr[i]:=r^.data[SHoffset+i]; (*Updates Cache*)
         END(*FOR*);
         p^.TStamp:=TickCount;
         IF p^.datagram<>NIL THEN              (*a packet is waiting*)
           q:=p^.datagram;         
           p^.datagram:=NIL;
           FOR i:=0 TO 5 DO
             q^.data[i]:=r^.data[SHoffset+i];  (*inserts dest. addr. in MAC*)
           END(*FOR*);
           INC(Astats.PacketsSent);
           rs:=WriteBlock(q^.data,q^.plen); (*sends pkt q'd in cache*)
           DisposePacket(q);
         END(*IF*);
       END(*IF*); 
       SemV(guard);                        

      (*checks if we are the target and the packet is an ARP request*)

       IF ((r^.data[TIPoffset]=HostIP[0]) AND (r^.data[TIPoffset+1]=HostIP[1]) AND
          (r^.data[TIPoffset+2]=HostIP[2]) AND (r^.data[TIPoffset+3]=HostIP[3]))
                                         THEN (*We are the target*)
          IF ((r^.data[OPoffset]=0C) AND (r^.data[OPoffset+1]=ReQuest)) THEN 
         (*Inserting required information for an ARP resolution*)

             FOR i:=0 TO 3 DO                  
                r^.data[TIPoffset+i]:=SIpnum[i]; (*inserting dest. IPnum*)
                r^.data[SIPoffset+i]:=HostIP[i]; (*inserting source IPnum*)
             END(*FOR*);
             FOR i:=0 TO 5 DO
               r^.data[THoffset+i]:=r^.data[SHoffset+i];(*insert. dest. Phy.*)
               r^.data[SHoffset+i]:=HostMac[i];  (*insert. host addr. in ARP*)
               r^.data[i]:=r^.data[soffset+i];   (*insert. dest. into MAC*)
               r^.data[soffset+i]:=HostMac[i];   (*insert. host addr. in MAC*)
             END(*FOR*);
             r^.data[OPoffset]:=0C;              (*inserting reply operation*)
             r^.data[OPoffset+1]:=RePly;    
             INC(Astats.ArpResponse);
             rs:=WriteBlock(r^.data,42); (*sends the ARP response*)
          END(*IF*);
        END(*IF*);   
     ELSE
       INC(Astats.PacketsDisc);
     END(*IF*);
   ELSE
     INC(Astats.PacketsDisc);
   END(*IF*);          
   DisposePacket(r);
END ArpIn;
                                                                          
(*This procedure makes an ARP request packet and then broadcasts it. The
  broadcast number is all ones*)

PROCEDURE ArpSend(Ipnumber:ARRAY OF CHAR);
VAR
  p:PacketPtr;
  i:CARDINAL;
  rs:Status;
BEGIN                       
  NewPacket(p);                                                          

 (*Inserting required information for an ARP broadcast*)

  FOR i:=0 TO 5 DO
    p^.data[i]:=377C;                   (*inserts Broadcast address*)
    p^.data[soffset+i]:=HostMac[i];     (*inserts hard. addr. in MAC*)
    p^.data[SHoffset+i]:=HostMac[i];    (*inserts hard. addr. in ARP frame*)
  END(*FOR*);                                                              
  p^.data[TFoffset]:=CHR(TFnum DIV 100H);  
  p^.data[TFoffset+1]:=CHR(TFnum MOD 100H);
  p^.data[HToffset]:=0C;               
  p^.data[HToffset+1]:=HType;                               
  p^.data[PToffset]:=CHR(PTnum DIV 100H);
  p^.data[PToffset+1]:=CHR(PTnum MOD 100H);
  p^.data[HLenoffset]:=Hlen;
  p^.data[PLenoffset]:=Plen;
  p^.data[OPoffset]:=0C;
  p^.data[OPoffset+1]:=ReQuest;
  FOR i:=0 TO 3 DO
    p^.data[SIPoffset+i]:=HostIP[i];        (*placing in host IP number*)
    p^.data[TIPoffset+i]:=Ipnumber[i];      (*placing in dest. IP number*)
  END(*FOR*);
  p^.plen:=42;      
  INC(Astats.ArpRequest);               
  rs:=WriteBlock(p^.data,p^.plen);     (*sends the ARP broadcast*)
  DisposePacket(p);    
END ArpSend;
                                                                             
(*This procedure sends a packet onto the network if the IPnum given could
  be resolved. Otherwise a cache entry is made and an ARP request is made*)

PROCEDURE PacketSend(p:PacketPtr;Ipnumber:ARRAY OF CHAR):Status;
VAR
  r:ArpPointer;
  lapsed:Ticks;  
  i:CARDINAL;
  rs:Status;             
BEGIN                                                   
   FOR i:=0 TO 5 DO                 
     p^.data[soffset+i]:=HostMac[i];  (*inserts hard. addr. in MAC*)
   END(*FOR*);
   SemP(guard);
   IF ArpFind(Ipnumber,r) THEN
      lapsed:=TickCount-r^.TStamp;
      IF (lapsed<maxtime) AND (NOT r^.resolving) THEN
       (*The destination address exists in the cache*)
         FOR i:=0 TO 5 DO
           p^.data[i]:=r^.Physaddr[i];      (*inserts dest. addr. in MAC*)
         END(*FOR*);
         INC(Astats.PacketsSent);
         rs:=WriteBlock(p^.data,p^.plen);   (*sends the packet*)
         SemV(guard);
         DisposePacket(p);
         RETURN rs;
      ELSE                             (*Queue packet for address resolution*)
         IF r^.datagram<>NIL THEN
           DisposePacket(r^.datagram);  (*removes an old packet queued*)
         END(*IF*);
         r^.datagram:=p;
         ArpSend(Ipnumber);
      END(*IF*);
    ELSE
      ArpAdd(Ipnumber,p);               (*Phys. address not resolved*)
      ArpSend(Ipnumber);
   END(*IF*);
   SemV(guard);                
   RETURN Ok;
END PacketSend;  

BEGIN                 
  InitGlobals();
  cache.head := NIL;
  SemInit(guard,1);
END N2Arp.
