IMPLEMENTATION MODULE N2IP;

FROM N2Errors     IMPORT Status;
FROM N2Base       IMPORT IP1, IPhdrlen, IPtlen, IPproto, IPsaddr, IPdaddr,
                     HostIP, ChkSum, HostNet2, NetHost2, GateIP, NetMaskH,
                     NetMaskL;
FROM N2Packets    IMPORT Packet, PacketPtr, DisposePacket, PacketHandler,
                                            NewPacket;
FROM L2Strings    IMPORT Substring;
FROM L2Kernel     IMPORT MaxPriority, Process,CreateProcess;
FROM L2Exceptions IMPORT Assert;
FROM N2Arp        IMPORT PacketSend , ArpIn;
FROM N2Driver     IMPORT ReadBlock;
FROM SYSTEM       IMPORT ADDRESS;
IMPORT L2Errors;

CONST
   IPtos    = IP1 + 1;
   IPid     = IP1 + 4;
   IPflags  = IP1 + 6;
   IPFrgoff = IP1 + 6;
   IPttl    = IP1 + 8;
   IPchksum = IP1 + 10;                         
   FrameType= 12;
   IPopts   = IP1 + 20;
   Version  = 4;
   TOS      = 0C;         (*Routine, Normal : Delay, Throughput, Reliability *)
   ID1      = 0C;
   ID2      = 0C;
   TTL      = 40C;        (*32 seconds/hops*)
   Frag     = 0C;                 
   HdrLen   = 5;
   maxreg   = 4;
   IpType1  = 10C;      (*IP type number is 0800H*)
   IpType2  = 0C;
   ArpType1 = 10C;      (*Arp type number is 0806H*)
   ArpType2 = 6C;

TYPE
   Data = RECORD
              protonum:CHAR;                (*Structure of register table*)
              procname:PacketHandler;
          END;

   RegTable = ARRAY[0..maxreg] OF Data;

   globalstats = RECORD
                   BadPktsChksum,  (*Global variables for debugging*)
                   PktsNotReg,
                   WrongType,
                   WrongDest,
                   BadVersionNum:CARDINAL;  (*purposes and to keep statistics*)
                   PktsSentToLoc,
                   PktsSentToGate,
                   PktsRcvd:ADDRESS;       
                 END;
                                  
(*A variant record to hold the source and destination address*)

   MaskRec = RECORD
              CASE BOOLEAN OF
               TRUE  : S:ARRAY[0..3] OF CHAR; |  (*hold as characters*)
               FALSE : A,B : BITSET;             (*take as type Bitset*)
              END(*CASE*);
             END;

VAR
   rtable : RegTable;  
   globals : globalstats;
   pid : Process;

(*This procedure will initialize the register table*)

PROCEDURE InitRtable();
VAR
  p:CARDINAL;
BEGIN
  FOR p:=0 TO maxreg-1 DO
    rtable[p].protonum:=0C;
  END(*FOR*);
END InitRtable;

(*This procedure will take an address and will 'and' it with the netmask
  if flag is true else it will 'or' it with the netmask. In either case the
  result is passed back.*)
                        
PROCEDURE Mask(VAR addr:MaskRec;flag:BOOLEAN);
BEGIN
  IF flag THEN                              (*If flag then 'and' them*)
     addr.A:=addr.A*NetMaskH;
     addr.B:=addr.B*NetMaskL;
  ELSE                                      (*else 'or' them together*)
     addr.A:=addr.A+NetMaskH;
     addr.B:=addr.B+NetMaskL;
  END(*IF*);
END Mask;

(*This procedure will check the register table for the protocol number*)

PROCEDURE CheckRegister(protocol:CHAR;VAR proc:PacketHandler) : Status;
VAR
   p:CARDINAL;
BEGIN
  p:=0;
  WHILE (rtable[p].protonum<>0C) AND (p<maxreg) DO
    IF protocol=rtable[p].protonum THEN
        proc:=rtable[p].procname;          (*Extract procedure name*)
        RETURN Ok;
    ELSE
        INC(p);
    END(*IF*);
  END(*WHILE*);
  INC(globals.PktsNotReg);
  RETURN NotRegistered;                    (*Not found in table*)
END CheckRegister;
                                                                
(*This procedure will register a higher level protocol with IP. Thus,
  if a packet comes in for this protocol it can be sent to it*)

PROCEDURE IPRegister(proc : PacketHandler; protocol:CHAR) : Status;  
VAR
  p:CARDINAL;
BEGIN                                                      
  p:=0;                       
  WHILE (rtable[p].protonum<>0C) AND (p<maxreg) DO
     INC(p);
  END(*WHILE*);
  IF p<maxreg THEN                         (*If table not full, register*)
     rtable[p].protonum:=protocol;
     rtable[p].procname:=proc;
     RETURN Ok;
  ELSE                                     (*Else return status full*)
     RETURN RTableFull;
  END(*IF*);
END IPRegister;

(*This procedure will return all the statistics that are kept*)

PROCEDURE IPStats(VAR BadPktsChksum, PktsNotReg, WrongType, WrongDest,
                      BadVersionNum: CARDINAL;VAR PktsSentToLoc, PktsSentToGate,
                      PktsRcvd  : ADDRESS);
BEGIN
  PktsRcvd       := globals.PktsRcvd;
  PktsSentToLoc  := globals.PktsSentToLoc;
  PktsSentToGate := globals.PktsSentToGate;
  BadPktsChksum  := globals.BadPktsChksum;   (*Copy over statistics*)   
  BadVersionNum  := globals.BadVersionNum;
  WrongType      := globals.WrongType;
  PktsNotReg     := globals.PktsNotReg;
  WrongDest      := globals.WrongDest;
END IPStats;
                  
(*This procedure will take a packet, fill in the required fields and will
  pass it off to arp, with the right destination. (ie either give it the
  destination address if it is on this network or give it the gateway 
  address if it is not on this network*)

PROCEDURE IPSend( VAR p : PacketPtr ) : Status;
VAR
  chk : CARDINAL;
  rs : Status;
  daddr, saddr, daddr2 : MaskRec;  
  j:CARDINAL;
  proc:PacketHandler;
BEGIN                                                  

    (*The following code fills in the necessary fields*)
  p^.data[FrameType]:=IpType1;
  p^.data[FrameType+1]:=IpType2;
  p^.data[IP1]:=CHR(Version*10H + HdrLen); (*Fill in version number and *)
  p^.data[IPtos] := TOS;                   (*header length*)
  p^.data[IPid] := ID1;                    (*Fill in the identification fields*)
  p^.data[IPid+1] := ID2;
  p^.data[IPFrgoff] := Frag;               (*Fill in the flag field and the*)
  p^.data[IPFrgoff+1] := Frag;             (*fragment offset field*)
  p^.data[IPttl] := TTL;                   (*Fill in the time to live*)
  p^.data[IPchksum] :=0C;                  (*To calculate checksum we must *)
  p^.data[IPchksum+1]:=0C;                 (*store 0 in the checksum*)
  HostNet2(p^.plen-IP1,IPtlen,p^.data);    (*Insert total length field*)
  ChkSum(chk,IP1,HdrLen*4,p^.data,FALSE);  (*Calculate header checksum*)
  HostNet2(chk,IPchksum,p^.data);          (*Insert checksum*)

   (*Extract source and destination IP numbers*)

  FOR j:=0 TO 3 DO
    saddr.S[j]:=p^.data[IPsaddr+j];
    daddr.S[j]:=p^.data[IPdaddr+j];
    daddr2.S[j]:=p^.data[IPdaddr+j];
  END(*FOR*);

  (*Check to see if the packets for us, if so physically turn it around*)
  (* the class A net address 127.x.x.x is loop back! *)

  IF ((saddr.S[3]=daddr.S[3]) AND (saddr.S[2]=daddr.S[2]) AND
         (saddr.S[1]=daddr.S[1]) AND (saddr.S[0]=daddr.S[0])) OR
         (daddr.S[0]=CHR(127)) THEN
    rs:=CheckRegister(p^.data[IPproto],proc);
    IF rs=Ok THEN                                            
      IF daddr.S[0] = CHR(127) THEN
        p^.data[IPsaddr] := daddr.S[0];
        p^.data[IPsaddr+1] := daddr.S[1];
        p^.data[IPsaddr+2] := daddr.S[2];
        p^.data[IPsaddr+3] := daddr.S[3];
        p^.data[IPdaddr] := saddr.S[0];
        p^.data[IPdaddr+1] := saddr.S[1];
        p^.data[IPdaddr+2] := saddr.S[2];
        p^.data[IPdaddr+3] := saddr.S[3];
      END; (* IF *)
      proc(p);
      RETURN rs;
    ELSE       
      DisposePacket(p);
      RETURN NotRegistered;
    END(*IF*);
  END(*IF*);
                                                                   
  (*Check to see if the destination is on this network. If so then send
    Arp the destination address else send it the Gateway IP.*)

  Mask(saddr,TRUE);
  Mask(daddr,TRUE);
  IF ((saddr.S[3]=daddr.S[3]) AND (saddr.S[2]=daddr.S[2]) AND
       (saddr.S[1]=daddr.S[1]) AND (saddr.S[0]=daddr.S[0])) THEN
    rs:=PacketSend(p,daddr2.S);        
    INC(globals.PktsSentToLoc);
  ELSE  
    rs:=PacketSend(p,GateIP);
    INC(globals.PktsSentToGate);
  END(*IF*);
  p:=NIL;    
  RETURN rs;
END IPSend;

(*This procedure will be executing as a process. It is a blocking process. That
  is when ever there are no datagrams available to be read, then it will block.
  The only time it will awaken is when ever there is an indication that there
  is data available. Thus, no cpu time is wasted.*)

PROCEDURE NetIn; (* Process to handle incoming data *)                
VAR
  p : PacketPtr;             
  type, headerLen : CARDINAL;
  destIp, destIp2 : MaskRec;
  proc : PacketHandler;        
  rs:Status;
  chk1,chk2,j,vers:CARDINAL;                           
BEGIN     
  NewPacket(p);
  LOOP                                        (*infinite loop*)

   (*Read the data from the Driver, if any.*)

   rs:=ReadBlock(p^.data,p^.plen); 
   IF (p<>NIL) THEN                        (*Was there heap left*)
      IF p^.plen<>0 THEN                   (*Was data read *)
   (* We have a packet from the network, we must calculate the checksum*)
   (* , check the version number and check the frame type.              *)

        p^.next:=NIL;p^.prev:=NIL;
        INC(globals.PktsRcvd);
        IF (p^.data[FrameType]=ArpType1) AND         (*Check frame type*)
            (p^.data[FrameType+1]=ArpType2) THEN 
          ArpIn(p);                                  (*IF Arp, then hand off*)
          NewPacket(p);                            
        ELSE
          NetHost2(chk1,IPchksum,p^.data);             (*Extract checksum*)
          headerLen:=CARDINAL(p^.data[IP1]) MOD 10H;
          p^.data[IPchksum]:=0C;                       (*Insert 0 into checksum*)
          p^.data[IPchksum+1]:=0C;
          ChkSum(chk2,IP1,headerLen*4,p^.data,FALSE);  (*Recalculate checksum*)
          IF (chk1=chk2) THEN  
           vers:=CARDINAL(p^.data[IP1]) DIV 10H;     (*check version number*)
           IF Version=vers THEN  
             IF (p^.data[FrameType]=IpType1) AND  (*IP type*)
                       (p^.data[FrameType+1]=IpType2) THEN

 (*Check register table to see if protocol is registered, if not throw packet*)

               rs:=CheckRegister(p^.data[IPproto],proc);   
               IF rs=Ok THEN
                 FOR j:=0 TO 3 DO
                   destIp.S[0+j]:=p^.data[IPdaddr+j];
                   destIp2.S[0+j]:=p^.data[IPdaddr+j];
                 END(*FOR*);
                 Mask(destIp2,FALSE); 

  (*Check to see that packet was for us or was a broadcast address*)

                 IF ((destIp.S[3]=HostIP[3]) AND (destIp.S[2]=HostIP[2]) 
                   AND (destIp.S[1]=HostIP[1]) AND (destIp.S[0]=HostIP[0])) THEN 
                   proc(p);
                   NewPacket(p);
                 ELSIF ((destIp2.S[3]=377C) AND (destIp2.S[2]=377C) 
                       AND (destIp2.S[1]=377C) AND (destIp2.S[0]=377C)) THEN 
                   proc(p);    
                   NewPacket(p);
                 ELSE                              (*Datagram wasn't for us*)
                   INC(globals.WrongDest);
                 END(*IF*);
               END(*IF*);
             ELSE
               INC(globals.WrongType);             (*Wasn't for IP or Arp*)
             END(*IF*);            
           ELSE
            INC(globals.BadVersionNum);            (*Bad version number*)
           END(*IF*);
          ELSE
           INC(globals.BadPktsChksum);               (*bad checksum*)
          END(*IF*);
        END(*IF*);
      END(*IF*);
   END(*IF*)
  END(*LOOP*);
END NetIn;
                       
PROCEDURE Init();
BEGIN
   globals.PktsRcvd:=0;
   globals.BadPktsChksum:=0;
   globals.PktsNotReg:=0;
   globals.PktsSentToLoc:=0;
   globals.PktsSentToGate:=0;
   globals.WrongType:=0;
   globals.WrongDest:=0;
   globals.BadVersionNum:=0;
END Init;
  
BEGIN                                        
   Init();
   InitRtable();
   (*Create a light weight process, NetIn*)
   Assert( CreateProcess( pid, 0, 0, MaxPriority, NetIn ) = L2Errors.Ok, 500 );
END N2IP.
