IMPLEMENTATION MODULE N2TCPIn;

FROM L2Exceptions IMPORT Assert;
FROM L2Kernel     IMPORT CreateProcess, Process, MaxPriority, SemP, SemV;
FROM L2Storage    IMPORT ALLOCATE, DEALLOCATE;
FROM L2Clock      IMPORT TickCount;
FROM N2Packets    IMPORT PacketLen, Packet, PacketPtr, GetPacket, PutPacket,
                         PacketQ, NewPacket, InitPacketQ, DisposePacket;
FROM N2Base       IMPORT IP1, IPsaddr, IPhdrlen, IPtlen, IPdaddr, IPAddr,
                         NetHost2, NetHost4, HostNet2, HostNet4, ChkSum, 
                         HostIP, IPproto, IPMinLen;
FROM N2IP         IMPORT IPRegister,IPSend;
FROM N2TCPMain    IMPORT TCPProtocol, TCBPtr, TCB, BuffSz, MaxSocs, 
                         TCBState, Header, MaxSegSz, FragPtr, activeSocs,
                         Code, OutState, socList, SocPtr, Sleeper,
                         TCPInSegs, TCPPassOpen, TCPOuts, TCPEstabResets,
                         TCPCurrEstab, TCPAttemptFails, Ack, Syn;
FROM N2TCPTimer   IMPORT SetTimerEvent, ClearTimerEvents, PutMessage,Event;
FROM N2Errors     IMPORT Status;
                  IMPORT L2Errors;
FROM N2TCPLib     IMPORT TCPAbort, TCPWait, TCPRWindow, WakeUp, TCPRtt,
                         TCBDeallocate;
FROM SYSTEM       IMPORT ADDRESS;            
                                      
CONST                        
  MaxRcvrPortPkts = MaxSocs;
  HalfSeq         = 80000000H;        (*used for sequence # comparisons*) 
  HalfCard        = 32768;
  destoffset = 0;
  sIpoffset = 4;
  TCPoffset = IP1 + IPMinLen;         (*1st octet of TCP header*)
  tcpMinHLen = 5H;

 (*TCP header offsets*)

  seqoffset = 4;
  ackoffset = 8;
  tcpHLenoff = 12;
  codeBoffset = 13;
  windowoff = 14;
  chkoffset = 16;
  urgoffset = 18;
  optoffset = 20;

TYPE
  TCBStateProc = PROCEDURE( TCBPtr, PacketPtr );

VAR
  rcvPort     : PacketQ;           (* TCPIn (ie IP) / TCPRcvr interface *)
  stateProc   : ARRAY TCBState  OF TCBStateProc;
  tcpChkoff,ipHLen : CARDINAL;
  rs : Status;
  ipLen : CARDINAL;


(*This procedure places the data received from the network into the receive
  buffer if it was in our window*)

PROCEDURE TCPData(tcb:TCBPtr;p:PacketPtr);
VAR
  wLast,first,last:ADDRESS;
  rwind,pacPos,buffPos,i,dataLen:CARDINAL;
BEGIN                                     
  IF Header.synBit THEN
    INC(tcb^.rcvNext);
    tcb^.flags.needOut:=TRUE;
    INC(Header.seqNum);
  END(*IF*);           
  dataLen:=Header.tcpLen-Header.TCPHLen;
  rwind:=BuffSz-tcb^.rcvBuffCnt;       
  wLast:=tcb^.rcvNext+ADDRESS(rwind)-1;
  first:=Header.seqNum;
  last:=first+ADDRESS(dataLen)-1;

(*This will chop of duplicate data received in front of our window*)
(*and also looks after wrap around*)

  IF ((tcb^.rcvNext-first)-1<HalfSeq) THEN
    dataLen:=dataLen-CARDINAL(tcb^.rcvNext-first);
    first:=tcb^.rcvNext;
  END(*IF*);

(*This will chop of data received past our receive window*)
(*and also looks after wrap around*)

  IF ((last-wLast)-1<HalfSeq) THEN
    dataLen:=CARDINAL(ADDRESS(dataLen)-(last-wLast));
    Header.finBit:=FALSE;
  END(*IF*);
  buffPos:=tcb^.rcvBuffPos+tcb^.rcvBuffCnt;       (*next byte to write to*)
  buffPos:=buffPos+CARDINAL(first-tcb^.rcvNext);  (*looks after out of order*)
  buffPos:=buffPos MOD BuffSz;                    (*may wrap arround*)
  pacPos:=CARDINAL(first-Header.seqNum)+Header.tcp1+Header.TCPHLen;       
  IF tcb^.rcvBuffer<>NIL THEN
    FOR i:=1 TO dataLen DO
      tcb^.rcvBuffer^[buffPos]:=p^.data[pacPos];
      INC(pacPos);
      buffPos:=(buffPos + 1) MOD BuffSz;           (*looks after wrap around*)
    END(*FOR*);
  END(*IF*);
  TCPDoData(tcb,p,first,dataLen);
  IF tcb^.flags.needOut THEN  
    SetTimerEvent( tcb, SENDPKT, 20 );             (*sets timer event*)
  END(*IF*);          
END TCPData;
             
(*This procedure changes the output state according to the ACKs received*)

PROCEDURE TCPOState(tcb:TCBPtr;numAcked:INTEGER);
BEGIN    
  IF numAcked<=0 THEN
    RETURN;
  END(*IF*);
  IF tcb^.outState=RETRANSMIT THEN
    tcb^.retransCnt:=0;          
    tcb^.outState:=TRANSMIT;     
  END(*IF*);
  IF tcb^.sendBuffCnt=0 THEN
    tcb^.outState:=IDLE;
    RETURN;
  END(*IF*);
  PutMessage(tcb^.socPtr,SENDPKT); 
END TCPOState;

                                 
(* This procedure looks after the acknowledgements in received packets
   and recalculates the round trip estimates*)       

PROCEDURE TCPAcked(tcb:TCBPtr;p:PacketPtr):Status;
VAR
  dataAck, tcbNum, codeAck:INTEGER;
BEGIN    
  IF (NOT Header.ackBit) THEN
    RETURN SysErr;
  END(*IF*);
  dataAck:=INTEGER(Header.ackNum-tcb^.sendUnAck);
  codeAck:=0;
  IF dataAck<=0 THEN
    RETURN DuplicateAck;                     (*duplicate ack*);
  END(*IF*);
  IF (Header.ackNum-tcb^.sendNext)-1<HalfSeq THEN
    IF tcb^.tcbState=SYNRCVD THEN
       RETURN Reset(p);
    ELSE         
       RETURN AckIt(tcb,p);
    END(*IF*);
  END(*IF*);
  TCPRtt(tcb);
  tcb^.sendUnAck:=Header.ackNum;
  IF (dataAck>0) AND (tcb^.tcbCode.syn) THEN
    DEC(dataAck);
    INC(codeAck);
    tcb^.tcbCode.syn:=FALSE;
    tcb^.flags.firstSend:=FALSE;
  END(*IF*);
  IF (tcb^.tcbCode.fin) AND (Header.ackNum-tcb^.sendNext=0) THEN
     DEC(dataAck);
     INC(codeAck);
     tcb^.tcbCode.fin:=FALSE;
     tcb^.flags.sendFin:=FALSE;
  END(*IF*);
  tcb^.sendBuffPos:=(tcb^.sendBuffPos+CARDINAL(dataAck)) MOD BuffSz;
  tcb^.sendBuffCnt:=tcb^.sendBuffCnt-CARDINAL(dataAck);
  IF dataAck>0 THEN
    WakeUp(tcb,WRITER);
  END;
  TCPOState(tcb,CARDINAL(dataAck+codeAck));
  RETURN Ok;
END TCPAcked;

                                             
(*This procedure updates the number of contiguous octets received from
  the foreign host if any of the fragments can be used*)

PROCEDURE TCPFrags( VAR tcb : TCBPtr; dataLen : CARDINAL; p : PacketPtr );
VAR
  s : FragPtr;
  new : CARDINAL;
BEGIN               
  WITH tcb^ DO
    rcvNext  := rcvNext + ADDRESS( dataLen );
    rcvBuffCnt := rcvBuffCnt + dataLen;
    IF rcvNext = finSeq THEN
      Header.finBit := TRUE;
      RETURN;
    END; (* IF *)
    IF fragHead = NIL THEN RETURN; END;
    s := fragHead;
    fragHead := fragHead^.next;
    WHILE (( s^.seqNum - rcvNext)-1 > HalfSeq ) DO
      new := s^.length - CARDINAL( rcvNext - s^.seqNum );  (*deletes excess*)
      IF new < HalfCard THEN
        rcvNext := rcvNext + ADDRESS( new );
        rcvBuffCnt := rcvBuffCnt + new;
      END; (* IF *)
      IF rcvNext = finSeq THEN
        Header.finBit := TRUE;
        DISPOSE( s );
        RETURN;
      END; (* IF *)
      DISPOSE( s );
      s := fragHead;
      IF s = NIL THEN RETURN; END;
      fragHead := fragHead^.next;
    END; (* WHILE *)
    fragHead := s;                            (*took one too many frags off*)
  END; (* WITH *)
END TCPFrags;

(*This procedure adds another fragment of sequence numbers to the fragment
  queue. This indicates a packet of information was received out of order*) 

PROCEDURE AddSeg( VAR head : FragPtr; seq : ADDRESS; dataLen : CARDINAL );
VAR
  r, q, frag : FragPtr;
BEGIN 
  IF dataLen = 0 THEN RETURN; END;
  NEW(frag);
  frag^.seqNum := seq;
  frag^.length := dataLen;
  r := NIL;
  q := head;
  WHILE ( q <> NIL ) AND ( seq-q^.seqNum<HalfSeq ) DO
    r := q;
    q := q^.next;
  END (*WHILE*);
  IF r = NIL THEN
    head := frag;
  ELSE
    r^.next := frag;
  END;
  frag^.next := q;
END AddSeg;
         
(*Queues fragment if out of order or updates acknowledgments from data
  received and if a fragment hole was filled.*)

PROCEDURE TCPDoData(tcb:TCBPtr;p:PacketPtr;first:ADDRESS;dataLen:CARDINAL);
VAR
  wakeUp:BOOLEAN;
BEGIN
  wakeUp:=FALSE;
  IF tcb^.rcvNext=first THEN               (*Can update rcvNext*)
    IF dataLen>0 THEN
      TCPFrags(tcb,dataLen,p);
      tcb^.flags.needOut:=TRUE;
      wakeUp:=TRUE;
    END(*IF*);
    IF Header.finBit THEN
      tcb^.flags.needOut:=TRUE;
      tcb^.flags.rcvDone:=TRUE;
      INC(tcb^.rcvNext);
      wakeUp:=TRUE;
    END(*IF*);
    IF wakeUp THEN
      WakeUp(tcb,READER);
    END(*IF*);
  ELSE                                       (*got a fragment*)
    IF Header.finBit THEN
      tcb^.finSeq:=Header.seqNum+ADDRESS(dataLen);
    END(*IF*);
    Header.finBit:=FALSE;
    AddSeg(tcb^.fragHead,first,dataLen);
  END(*IF*);
END TCPDoData;
                                                       

(*This procedure is used to compute the sender's window size and thus
  indicates how much data the foreign host will accept*)

PROCEDURE TCPSWindow(tcb:TCBPtr; p:PacketPtr);
VAR
  olastseg,lastseg:ADDRESS;
BEGIN    
  WITH tcb^ DO
    IF (Header.seqNum-lastWinSeq)>HalfSeq THEN
      RETURN;
    END(*IF*);
    IF ((Header.seqNum-lastWinSeq)=0) AND ((Header.ackNum-lastWinAck)>HalfSeq) THEN
      RETURN;
    END(*IF*);

    (*There is a sender window update *)
    (*Compute the last sequences of the new and old windows*)

    olastseg:=lastWinAck+ADDRESS(sendWinSz);
    lastseg:=Header.ackNum+ADDRESS(Header.winSz);
    sendWinSz:=Header.winSz;
    lastWinSeq:=Header.seqNum;
    lastWinAck:=Header.ackNum;
    IF ((lastseg-olastseg)-1>HalfSeq) THEN
      RETURN;
    END(*IF*);
    IF (outState=PERSIST) THEN
      ClearTimerEvents(tcb);
      outState:=TRANSMIT;
    END(*IF*);
    PutMessage(tcb^.socPtr,SENDPKT); 
  END(*WITH*);
END TCPSWindow;            

(************STATE PROCEDURES FOLLOW****************)

(*This procedure looks after the Time-Wait state processing*)

PROCEDURE TimeWait(tcb:TCBPtr;p:PacketPtr);
BEGIN    
  IF Header.resetBit THEN
    TCBDeallocate(tcb);
    RETURN;
  END(*IF*);
  IF Header.synBit THEN
    rs:=Reset(p);
    TCBDeallocate(tcb);
    RETURN;
  END(*IF*);
  rs:=TCPAcked(tcb,p);
  TCPData(tcb,p);              (*just ack any packets*)
  TCPWait(tcb);
END TimeWait;
                                                            
(*This procedure looks after the closed state processing*)

PROCEDURE TCPClosed(tcb:TCBPtr;p:PacketPtr);
BEGIN   
  rs:=Reset(p);                                          
END TCPClosed;

(*This procedure looks after the closing state processing*)

PROCEDURE  TCPClosing(tcb:TCBPtr;p:PacketPtr);
BEGIN
  IF Header.resetBit THEN
    TCPAbort(tcb,FReset);
    RETURN;
  END(*IF*);
  IF Header.synBit THEN
    rs:=Reset(p);
    TCPAbort(tcb,LReset);
    RETURN;
  END(*IF*);
  rs:=TCPAcked(tcb,p);
  IF NOT tcb^.tcbCode.fin THEN           (*Received an ack*)
    WakeUp(tcb,CLOSER);                  (*Wake up closer*)
    tcb^.tcbState:=TIMEWAIT;
    TCPWait(tcb);
  END(*IF*);
END TCPClosing;
                                                          
(*This procedure implements the state processing for the FinWait2 state*)

PROCEDURE FinWait2(tcb:TCBPtr;p:PacketPtr);
BEGIN    
  IF Header.resetBit THEN
    TCBDeallocate(tcb);
    RETURN;
  END(*IF*);
  IF Header.synBit THEN
    rs:=Reset(p);
    TCBDeallocate(tcb);
    RETURN;
  END(*IF*);
  IF TCPAcked(tcb,p)=SysErr THEN
     RETURN;
  END(*IF*);
  TCPData(tcb,p);                           (*for data & FIN acking*)
  IF tcb^.flags.rcvDone THEN
    tcb^.tcbState:=TIMEWAIT;
    TCPWait(tcb);
  END(*IF*);
END FinWait2;
                                                                        
(*This procedure implements the state processing for the FinWait1 state*)

PROCEDURE FinWait1(tcb:TCBPtr;p:PacketPtr);
BEGIN    
  IF Header.resetBit THEN
    TCPAbort(tcb,FReset);
    RETURN;
  END(*IF*);
  IF Header.synBit THEN
    rs:=Reset(p);
    TCPAbort(tcb,LReset);
    RETURN;
  END(*IF*);
  IF TCPAcked(tcb,p)=SysErr THEN
    RETURN;
  END(*IF*);
  TCPData(tcb,p);
  TCPSWindow(tcb,p);
  IF tcb^.flags.rcvDone THEN
    IF tcb^.tcbCode.fin THEN            (*Fin NOT acked*)
      tcb^.tcbState:=CLOSING;
    ELSE
      tcb^.tcbState:=TIMEWAIT;
      WakeUp(tcb,CLOSER);               (*Wake Closer*)
      TCPWait(tcb);
    END(*IF*);
  ELSE
    IF NOT tcb^.tcbCode.fin THEN
      WakeUp(tcb,CLOSER);               (*Wake closer*)
      tcb^.tcbState:=FINWAIT2;
      TCPWait(tcb);
    END(*F*);
  END(*IF*);
END FinWait1;


(*This procedure implements the state processing for the Listen state*)

PROCEDURE TCPListen(tcb:TCBPtr;p:PacketPtr);
BEGIN    
  IF Header.resetBit THEN
    RETURN
  END(*IF*);
  IF (Header.ackBit) OR NOT (Header.synBit) THEN
    rs:=Reset(p);
    RETURN;
  END(*IF*);
  tcb^.sendWinSz:=Header.winSz;             (*Init. some variables*)
  tcb^.lastWinAck:=tcb^.initSendSeq;
  tcb^.winSeq:=tcb^.rcvNext+BuffSz;
  tcb^.lastWinSeq:=Header.seqNum;
  tcb^.rcvNext := Header.seqNum;
  tcb^.tcbState:=SYNRCVD;
  tcb^.outState:=IDLE;
  tcb^.forHost:=Header.forHost;
  tcb^.forPort:=Header.forPort;
  tcb^.flags.needOut:=TRUE;
  INC(TCPPassOpen);
  Header.finBit:=FALSE;       (*don't process fins in listen*)
  TCPData(tcb,p);
END TCPListen;

(*This procedure implements the state processing for the SynSent state*)

PROCEDURE TCPSynSent(tcb:TCBPtr; p:PacketPtr);
BEGIN    
  IF Header.ackBit AND (((Header.ackNum-tcb^.initSendSeq)-1>HalfSeq) OR
                                ((Header.ackNum-tcb^.sendNext)-1<HalfSeq)) THEN
    rs:=Reset(p);
    RETURN;
  END(*IF*);
  IF Header.resetBit THEN
    tcb^.tcbState:=CLOSED;
    tcb^.status:=FReset;
    INC(TCPAttemptFails);
    ClearTimerEvents(tcb);
    WakeUp(tcb,OPENER);
    RETURN;
  END(*IF*);
  IF NOT Header.synBit THEN
    RETURN
  END(*IF*);
  tcb^.sendWinSz:=Header.winSz;           (*ack and syn received*)
  tcb^.lastWinSeq:=Header.seqNum;
  tcb^.rcvNext:=Header.seqNum;
  tcb^.winSeq:=tcb^.rcvNext+BuffSz;
  rs:=TCPAcked(tcb,p);
  TCPData(tcb,p);
  Header.finBit:=FALSE;
  IF tcb^.tcbCode.syn THEN                (*Our syn. not acked*)
    tcb^.tcbState:=SYNRCVD;
  ELSE
    INC(TCPCurrEstab);
    tcb^.tcbState:=ESTABLISHED;
    WakeUp(tcb,OPENER);
  END(*IF*);
END TCPSynSent;
                                                               
(*This procedure implements the state processing for the SynRcvd state*)

PROCEDURE TCPSynRcvd(tcb:TCBPtr; p:PacketPtr);
BEGIN    
  IF Header.resetBit THEN
    INC(TCPAttemptFails);
    TCPAbort(tcb,FRefusal);
  END(*IF*);
  IF Header.synBit THEN
    INC(TCPAttemptFails);
    rs:=Reset(p);
    TCPAbort(tcb,LReset);
    RETURN;
  END(*IF*);
  IF (TCPAcked(tcb,p)=SysErr) THEN         (*Wait for a good ack*)
    RETURN;
  END(*IF*);
  WakeUp(tcb,OPENER);                      (*Wake up the OPENER*)
  INC(TCPCurrEstab);
  tcb^.tcbState:=ESTABLISHED;
  TCPData(tcb,p);
  IF tcb^.flags.rcvDone THEN
    tcb^.tcbState:=CLOSEWAIT;
  END(*IF*);
END TCPSynRcvd;
                                                               
(*This procedure implements the state processing for the Established state*)

PROCEDURE Established(tcb:TCBPtr; p:PacketPtr);
BEGIN    
  IF Header.resetBit THEN
    INC(TCPEstabResets);
    DEC(TCPCurrEstab);
    TCPAbort(tcb,FReset);
    RETURN;
  END(*IF*);
  IF Header.synBit THEN
    INC(TCPEstabResets);
    DEC(TCPCurrEstab);
    rs:=Reset(p);
    TCPAbort(tcb,LReset);
    RETURN;
  END(*IF*);
  IF (TCPAcked(tcb,p) = SysErr) THEN
    RETURN;
  END(*IF*);
  TCPData(tcb,p);
  TCPSWindow(tcb,p);
  IF tcb^.flags.rcvDone THEN
    tcb^.tcbState:=CLOSEWAIT;
  END(*IF*);
END Established;
                                                                   
(*This procedure implements the state processing for the LastAck state*)

PROCEDURE LastAck(tcb:TCBPtr; p:PacketPtr);
BEGIN    
  IF Header.resetBit THEN
    TCPAbort(tcb,FReset);
    RETURN;
  END(*IF*);
  IF Header.synBit THEN
    rs:=Reset(p);
    TCPAbort(tcb,LReset);
    RETURN;
  END(*IF*);
  rs:=TCPAcked(tcb,p);
  IF NOT tcb^.tcbCode.fin THEN
    WakeUp(tcb,CLOSER);
  END(*IF*);
END LastAck;
                                                               
(*This procedure implements the state processing for the CloseWait state*)

PROCEDURE TCPCloseWait(tcb:TCBPtr; p:PacketPtr);
BEGIN     
  IF Header.resetBit THEN
    INC(TCPEstabResets);
    DEC(TCPCurrEstab);
    TCPAbort(tcb,FReset);            (*Aborts the connection*)
    RETURN;
  END(*IF*);
  IF Header.synBit THEN
    INC(TCPEstabResets);
    DEC(TCPCurrEstab);
    rs:=Reset(p);
    TCPAbort(tcb,LReset);           (*Aborts the connection*)
    RETURN;
  END(*IF*);
  rs:=TCPAcked(tcb,p);
  TCPSWindow(tcb,p);               (*Updates the sender window*)
END TCPCloseWait;

(****************END OF STATE PROCEDURES*******************)
                                                                
(*This procedure generates an acknowledgement for an acknowledgement
  received outside our window*)

PROCEDURE AckIt(tcb:TCBPtr;p:PacketPtr):Status;                  
VAR
  r : PacketPtr;
  rs : Status;
  rcvWinSz,chkValue : CARDINAL;
BEGIN     
  IF Header.resetBit THEN     (*Do not respond to RESETS*)
     RETURN Ok;
  END(*IF*);
  IF  (ipLen<=ipHLen+Header.TCPHLen) AND 
             NOT (Header.synBit OR Header.finBit) THEN
    RETURN Ok;                         (*duplicate ACK*)
  END(*IF*);
  NewPacket(r);
  IF r=NIL THEN
     RETURN SysErr;
  END(*IF*);
   (*Set IP fields*)
   r^.data[IPsaddr]:=HostIP[0];
   r^.data[IPsaddr+1]:=HostIP[1];
   r^.data[IPsaddr+2]:=HostIP[2];
   r^.data[IPsaddr+3]:=HostIP[3];
   r^.data[IPdaddr]:=p^.data[IPsaddr];
   r^.data[IPdaddr+1]:=p^.data[IPsaddr+1];
   r^.data[IPdaddr+2]:=p^.data[IPsaddr+2];
   r^.data[IPdaddr+3]:=p^.data[IPsaddr+3];
   r^.data[IPproto]:=TCPProtocol;

  (*Set TCP fields*)
   r^.data[TCPoffset]:=p^.data[Header.tcp1+2];    (*sets source port*)
   r^.data[TCPoffset+1]:=p^.data[Header.tcp1+3];
   r^.data[TCPoffset+2]:=p^.data[Header.tcp1];    (*sets dest. port*)
   r^.data[TCPoffset+3]:=p^.data[Header.tcp1+1];
   HostNet4(tcb^.sendNext,TCPoffset+seqoffset,r^.data);
   HostNet4(tcb^.rcvNext,TCPoffset+ackoffset,r^.data);
   IF tcb^.tcbState = SYNRCVD THEN
     r^.data[TCPoffset+codeBoffset]:=CHR(Ack + Syn);
   ELSE
     r^.data[TCPoffset+codeBoffset]:=CHR(Ack);            (*sets ack BIT*)
   END; (* IF *)
   r^.data[tcpHLenoff+TCPoffset]:=CHR(tcpMinHLen*10H);
   rcvWinSz:=TCPRWindow(tcb);  
   HostNet2(rcvWinSz,TCPoffset+windowoff,r^.data);
   r^.data[TCPoffset+urgoffset]:=0C;
   r^.data[TCPoffset+urgoffset+1]:=0C;
   r^.data[TCPoffset+chkoffset]:=0C;
   r^.data[TCPoffset+chkoffset+1]:=0C;
   ChkSum(chkValue,TCPoffset,tcpMinHLen*4,r^.data,TRUE);
   HostNet2(chkValue,TCPoffset+chkoffset,r^.data);
   r^.plen:=TCPoffset+20;              (*Min length being sent out*)
   INC(TCPOuts);                       
   RETURN IPSend(r);
END AckIt;


(*This procedure sends a reset when segments arrive unexpectedly. ie when
  no connection exists*)

PROCEDURE Reset(p:PacketPtr):Status;
VAR       
  r:PacketPtr;                     
  ackNum:ADDRESS;
  chkValue,TCPseqoff,TCPackoff,dataLen:CARDINAL;
BEGIN     
  IF Header.resetBit THEN          (*Don't send a reset on a reset*)
    RETURN Ok;
  END;
  NewPacket(r);
  IF r=NIL THEN
    RETURN SysErr;
  END(*IF*);

     (*Set IP Fields*)

  r^.data[IPproto]:=TCPProtocol;
  r^.data[IPsaddr]:=HostIP[0];
  r^.data[IPsaddr+1]:=HostIP[1];
  r^.data[IPsaddr+2]:=HostIP[2];
  r^.data[IPsaddr+3]:=HostIP[3];
  r^.data[IPdaddr]:=p^.data[IPsaddr];
  r^.data[IPdaddr+1]:=p^.data[IPsaddr+1];
  r^.data[IPdaddr+2]:=p^.data[IPsaddr+2];
  r^.data[IPdaddr+3]:=p^.data[IPsaddr+3];

     (*Sets TCP Header Fields*)

  r^.data[TCPoffset]:=p^.data[Header.tcp1+2];      (*inserting source port*)
  r^.data[TCPoffset+1]:=p^.data[Header.tcp1+3];
  r^.data[TCPoffset+2]:=p^.data[Header.tcp1];      (*inserting dest. port*)
  r^.data[TCPoffset+3]:=p^.data[Header.tcp1+1];
  TCPseqoff:=TCPoffset+seqoffset;
  TCPackoff:=Header.tcp1+ackoffset;
  IF Header.ackBit THEN
    r^.data[TCPseqoff]:=p^.data[TCPackoff];
    r^.data[TCPseqoff+1]:=p^.data[TCPackoff+1];
    r^.data[TCPseqoff+2]:=p^.data[TCPackoff+2];
    r^.data[TCPseqoff+3]:=p^.data[TCPackoff+3];
    r^.data[TCPoffset+codeBoffset]:=4C;        (*sets reset bit*)
  ELSE
    r^.data[TCPseqoff]:=0C;                    (*sets seq. num as zero*)
    r^.data[TCPseqoff+1]:=0C;
    r^.data[TCPseqoff+2]:=0C;
    r^.data[TCPseqoff+3]:=0C;
    r^.data[TCPoffset+codeBoffset]:=24C;       (*sets ack. & reset bits*)
  END(*IF*);
  dataLen:=Header.tcpLen-Header.TCPHLen;
  IF Header.finBit THEN
     INC(dataLen);
  END(*IF*);
  ackNum:=Header.seqNum+ADDRESS(dataLen);
  HostNet4(ackNum,TCPoffset+ackoffset,r^.data);
  r^.data[tcpHLenoff+TCPoffset]:=CHR(tcpMinHLen*10H);
  r^.data[TCPoffset+windowoff]:=0C;
  r^.data[TCPoffset+windowoff+1]:=0C;
  r^.data[TCPoffset+urgoffset]:=0C;
  r^.data[TCPoffset+urgoffset+1]:=0C;
  r^.data[TCPoffset+chkoffset]:=0C;
  r^.data[TCPoffset+chkoffset+1]:=0C;
  ChkSum(chkValue,TCPoffset,tcpMinHLen*4,r^.data,TRUE);
  r^.data[TCPoffset+chkoffset]:=CHR(chkValue DIV 100H);
  r^.data[TCPoffset+chkoffset+1]:=CHR(chkValue MOD 100H);
  r^.plen:=TCPoffset+20;          (*length to end of TCP Header*)
  INC(TCPOuts);
  RETURN IPSend(r);
END Reset;


(*This procedure makes sure the data received is atleast partially in
  our receiving window. If it is not FALSE is returned*)

PROCEDURE CheckWind(tcb:TCBPtr):BOOLEAN;
VAR
  rWindow,dataLen:CARDINAL;
  seqLast,winLast:ADDRESS;
  rv:BOOLEAN;
BEGIN    
  IF tcb^.tcbState<=SYNSENT THEN
     RETURN TRUE;
  END;
  dataLen:=Header.tcpLen-Header.TCPHLen;
  IF Header.synBit THEN                    (*acknowledge another octet*)
    INC(dataLen);
  END;
  IF Header.finBit THEN
    INC(dataLen);
  END(*IF*);
  rWindow:=BuffSz-tcb^.rcvBuffCnt;
  IF (rWindow=0) AND (dataLen=0) THEN
    RETURN (Header.seqNum=tcb^.rcvNext);
  END;                               
  winLast:=tcb^.rcvNext+ADDRESS(rWindow)-1;
  rv:= ((Header.seqNum-tcb^.rcvNext)<HalfSeq) AND   (*check negative*)
                ((Header.seqNum-winLast)-1>HalfSeq);
  IF dataLen=0 THEN                      
    RETURN rv;
  END;
  seqLast:=Header.seqNum+ADDRESS(dataLen)-1;
  rv:=rv OR ((seqLast-tcb^.rcvNext<HalfSeq) AND ((seqLast-winLast)-1>HalfSeq));
  IF rWindow=0 THEN
    Header.tcpLen:=Header.TCPHLen;
  END(*IF*);    
  RETURN rv;
END CheckWind;

                                            
(*This procedure tries to find the proper TCB for the packet of information
  received off the network*)

PROCEDURE FindSoc(VAR soc:SocPtr;p:PacketPtr);
VAR
  r,ListenTcb,tcb:TCBPtr;
BEGIN     
  soc:=NIL;
  tcb:=NIL;
  ListenTcb:=NIL;
  SemP(activeSocs.guard);
  r:=activeSocs.head;
  LOOP
    IF r=NIL THEN
      EXIT;
    END;
    IF (r^.forPort=Header.forPort) AND (r^.locPort=Header.locPort) AND
       (r^.forHost=Header.forHost) THEN
       tcb:=r;
       EXIT;
    END;
    IF (r^.tcbState=LISTEN) AND (r^.locPort=Header.locPort) THEN
      ListenTcb:=r;
    END;
    r:=r^.next;
  END(*LOOP*);
  IF tcb=NIL THEN
    IF Header.synBit THEN
      tcb:=ListenTcb;
    END;
  END;
  IF tcb<>NIL THEN
    soc:=tcb^.socPtr;
  END(*IF*);
  SemV(activeSocs.guard);
END FindSoc;
                           
(*This procedure extracts all the relevant information from a packet  
  received off the network*)                                    

PROCEDURE ExtractHeader(VAR p:PacketPtr);
VAR
  code:CARDINAL;
BEGIN     
  NetHost2(ipLen, IPtlen, p^.data);
  ipHLen:=CARDINAL(p^.data[IPhdrlen]) MOD 10H;
  Header.tcp1:= IP1 + 4*ipHLen;
  Header.tcpLen:= IP1 + ipLen - Header.tcp1;
  tcpChkoff := Header.tcp1 + chkoffset;
  WITH Header DO
    NetHost2( chkSum,tcpChkoff, p^.data);
    NetHost4(seqNum,Header.tcp1+seqoffset,p^.data);
    NetHost4(ackNum,Header.tcp1+ackoffset,p^.data);
    NetHost2(forPort,Header.tcp1,p^.data);
    NetHost2(locPort,Header.tcp1+2,p^.data);
    NetHost4(forHost,IPsaddr,p^.data);
    TCPHLen:=4 * CARDINAL(p^.data[tcp1+tcpHLenoff]) DIV 10H;
    NetHost2(winSz,Header.tcp1+windowoff,p^.data);
    code:=CARDINAL(p^.data[codeBoffset+tcp1]) MOD 40H;
    finBit:=(code MOD 2H)=1;
    code:=code DIV 2H;
    synBit:=(code MOD 2H)=1;
    code:=code DIV 2H;
    resetBit:=(code MOD 2H)=1;
    code:=code DIV 4H;
    ackBit:=(code MOD 2H)=1;
  END(*WITH*);
END ExtractHeader;

CONST       
  Indefinitely = 0;
                           
(*This procedure is used by the TCP input process. The process also loops
  for ever in this procedure*)

PROCEDURE TCPRcvr;
 (* TCP receiving process *)
  VAR
    p : PacketPtr;
    chk2,i : CARDINAL;
    tcb : TCBPtr;
    socPtr:SocPtr;
    j,a,b:CARDINAL;
  BEGIN
    InitPacketQ( rcvPort, MaxRcvrPortPkts );      (*queue shared between*)
    LOOP                                           (*IP and TCP*)
       rs:=GetPacket( p,rcvPort,Indefinitely );    (* Blocking call *)
       ExtractHeader(p);
       p^.data[tcpChkoff] := 0C;
       p^.data[tcpChkoff+1] := 0C;
       ChkSum(chk2, Header.tcp1, Header.tcpLen, p^.data, TRUE);
       IF Header.chkSum=chk2 THEN
         FindSoc(socPtr,p);
         tcb:=NIL;
         IF socPtr<>NIL THEN
            SemP(socPtr^.tcbGuard);
            tcb:=socPtr^.tcb;
         END(*IF*);
         IF (tcb = NIL) OR (tcb^.status <> Ok) THEN
           rs:=Reset(p);
           IF socPtr<>NIL THEN            (*Release the guard*)
             SemV(socPtr^.tcbGuard);
           END(*IF*);
         ELSE
           IF CheckWind(tcb) THEN
             stateProc[tcb^.tcbState](tcb,p);
           ELSE
             rs:=AckIt(tcb,p);            (*If a packet is outside our window*)
           END(*IF*);
           SemV(socPtr^.tcbGuard);
         END(*IF*);
       END(*IF*);
       DisposePacket(p);
    END; (* LOOP *)
  END TCPRcvr;

(*This procedure is used by the IP input process only. The packets for
 TCP are placed on the queue in this procedure*)

PROCEDURE TCPIn( p : PacketPtr );
 (* Packet delivery from IP  ->> TCPRcvr process *)
  BEGIN
    INC(TCPInSegs);
    PutPacket( p, rcvPort );
  END TCPIn;

VAR
  proc : Process;
BEGIN                                  
(*Starts the TCP input process*)
  Assert(CreateProcess(proc,0,0,MaxPriority-1,TCPRcvr ) = L2Errors.Ok, 590 );

(*Register the TCPIn procedure with IP*)
  Assert(IPRegister(TCPIn,TCPProtocol) = Ok, 506 );
  stateProc[CLOSED]:=TCPClosed; 

(*initializes the array of procedure type*)             
  stateProc[LISTEN]:=TCPListen;
  stateProc[SYNSENT]:=TCPSynSent;
  stateProc[SYNRCVD]:=TCPSynRcvd;
  stateProc[ESTABLISHED]:=Established;
  stateProc[FINWAIT1]:=FinWait1;
  stateProc[FINWAIT2]:=FinWait2;
  stateProc[CLOSEWAIT]:=TCPCloseWait;
  stateProc[LASTACK]:=LastAck;
  stateProc[CLOSING]:=TCPClosing;
  stateProc[TIMEWAIT]:=TimeWait;
END N2TCPIn.
