IMPLEMENTATION MODULE N2TCPOut;

FROM L2Exceptions IMPORT Assert;
FROM L2Kernel     IMPORT CreateProcess, Process, SetNoSwitch, Suspend, Resume,
                         MaxPriority, SemP, SemV, Semaphore, SemInit;
FROM N2Base       IMPORT IP1, IPMinLen, IPsaddr, IPdaddr, IPAddr, HostIP,
                         HostNet2, HostNet4, ChkSum, IPproto;
FROM N2IP         IMPORT IPSend;
FROM N2Packets    IMPORT PacketLen, PacketPtr, Packet, NewPacket;
FROM N2TCPMain    IMPORT TCPProtocol, TCBPtr, TCB, OutState, TCPMAXRXT,
                         TCPMAXPRS, Ack, Push, Reset, Syn, Fin, BuffSz,
                         TCBState, activeSocs, MaxSegSz, TCPMAXRETRIES,
                         SocPtr, TCPRetransSegs, TCPOutSegs;
FROM N2TCPLib     IMPORT TCPAbort, TCPWait, TCPRWindow, TCPRtt, TCBDeallocate;
FROM SYSTEM       IMPORT ADDRESS;
FROM N2Errors     IMPORT Status;
FROM N2TCPTimer   IMPORT ClearTimerEvents, DeleteTimerEvent, SetTimerEvent,
                           Event, GetMessage, PutMessage, TimerEventExists;
FROM L2Clock      IMPORT Ticks, TickCount;
                  IMPORT L2Errors;                  

CONST
  HalfSeq = 80000000H;
  RXMIT = TRUE;
  NEWDATA = FALSE;

TYPE
  TCBStateProc = PROCEDURE(TCBPtr,Event);


VAR
  first       : BOOLEAN;
  outProc     : Process;
  tcpOSwitch  : ARRAY OutState OF TCBStateProc;

(**********************************************************)
(*                                                        *)
(*      TCP output process and supporting routines        *)
(*                                                        *)
(**********************************************************)

CONST
  TCP1        = IP1 + IPMinLen;
  TCPsport    = TCP1;
  TCPfport    = TCP1 + 2;
  TCPseq      = TCP1 + 4;
  TCPack      = TCP1 + 8;
  TCPhlen     = TCP1 + 12;
  TCPcode     = TCP1 + 13;
  TCPwin      = TCP1 + 14;
  TCPchksum   = TCP1 + 16;
  TCPurg      = TCP1 + 18;
  TCPopt      = TCP1 + 20;
  TCPHdrLen   = 20;
  TCPHdr      = 5;

(*This procedure calculates the offset and the data length*)

PROCEDURE TCPSndLen( tcb : TCBPtr; retransmit : BOOLEAN;      (*pg 235*)
                            VAR offset : CARDINAL):ADDRESS;
VAR
  temp:ADDRESS;
BEGIN
  IF retransmit OR tcb^.tcbCode.syn THEN
    offset:=0;
  ELSE
    offset:=CARDINAL(tcb^.sendNext-tcb^.sendUnAck);
  END(*IF*);
  temp := ADDRESS(tcb^.sendBuffCnt - offset);
  IF temp>MaxSegSz THEN 
     RETURN MaxSegSz;
  ELSE        
     RETURN temp;
  END(*IF*);
END TCPSndLen;

(*This procedure will calculate the amount of data to be sent*)

PROCEDURE HowMuch( tcb : TCBPtr) : ADDRESS;       
VAR
  num:ADDRESS;
BEGIN
  num:=tcb^.sendUnAck+ADDRESS(tcb^.sendBuffCnt)-tcb^.sendNext;
  IF tcb^.tcbCode.syn THEN
    INC(num);
  END(*IF*);
  IF tcb^.flags.sendFin THEN
    INC(num);
  END(*IF*);
  RETURN num;
END HowMuch;

                                                      
(* This procedure handles the case when we are in state retransmit and
   an event has occurred. *)

PROCEDURE ReTransmit(tcb:TCBPtr; event:Event);           
VAR
  temp : Ticks;
  j:CARDINAL;
BEGIN        
  IF event<>RETRANSMITSEG THEN
    RETURN
  END(*IF*);
  INC(tcb^.retransCnt);
  IF (tcb^.retransCnt)>TCPMAXRETRIES THEN
    TCPAbort(tcb,TimedOut);
    RETURN;
  END(*IF*); 
  TCPSend(tcb,RXMIT);
  temp:=tcb^.retransTmoVal;
  FOR j:=0 TO tcb^.retransCnt DO
    temp:=temp*2H;
  END(*FOR*);
  IF temp>TCPMAXRXT THEN                            (* set timer events *)
    SetTimerEvent(tcb,RETRANSMITSEG,TCPMAXRXT);
  ELSE
    SetTimerEvent(tcb,RETRANSMITSEG,temp);
  END(*IF*);
  IF tcb^.outState<>RETRANSMIT THEN
    tcb^.slowStartThresh:=tcb^.congWinSz;
  END(*IF*);
  IF tcb^.sendWinSz>tcb^.slowStartThresh THEN          (* flow control *)
     tcb^.slowStartThresh:=tcb^.slowStartThresh DIV 2;
  ELSE
     tcb^.slowStartThresh:=tcb^.sendWinSz DIV 2;
  END(*IF*);
  IF tcb^.slowStartThresh<MaxSegSz THEN
     tcb^.slowStartThresh:=MaxSegSz;
  END(*IF*);
  tcb^.congWinSz:=MaxSegSz;
END ReTransmit;

(*This procedure handles the case where were in the idle state in the finite
  state machine*)

PROCEDURE TCPIdle( tcb:TCBPtr; event:Event);           
BEGIN
  IF event=SENDPKT THEN
    Transmit(tcb,event);
  END(*IF*);
END TCPIdle;

(*This procedure handles the case where were in the persist state in the
  finite state machine*)

PROCEDURE Persist( tcb:TCBPtr; event:Event);     
BEGIN
  IF event<>PROBERCVR THEN
    RETURN;
  END(*IF*);
  TCPSend(tcb,RXMIT);
  tcb^.persistTimeVal:=tcb^.persistTimeVal*2H;
  IF tcb^.persistTimeVal>TCPMAXPRS THEN
    tcb^.persistTimeVal:=TCPMAXPRS;
  END(*IF*);
  SetTimerEvent(tcb,PROBERCVR,tcb^.persistTimeVal);
END Persist;

(*If we are currently in the transmit state do the following*)

PROCEDURE Transmit(tcb:TCBPtr; event:Event);        
VAR
  winSz:CARDINAL;
  tosend, waiting:ADDRESS;      
  rtt:Ticks;
BEGIN    
  IF event=RETRANSMITSEG THEN
    DeleteTimerEvent(tcb,SENDPKT,rtt);
    ReTransmit(tcb,event);
    tcb^.outState:=RETRANSMIT;
    tcb^.flags.needOut := TRUE;
    RETURN;
  END(*IF*);
  tosend:=HowMuch(tcb);                     (* How much data to send? *)
  IF tosend=0 THEN
    IF tcb^.flags.needOut THEN
      TCPSend(tcb,NEWDATA);
      IF tcb^.tcbState < ESTABLISHED THEN              (* COMER missed this! *)
        SetTimerEvent(tcb,RETRANSMITSEG,tcb^.retransTmoVal);
      END; (* IF *)
    END(*IF*);
    RETURN;
  END(*IF*);            
  IF tcb^.sendWinSz=0 THEN
    tcb^.outState:=PERSIST;
    tcb^.persistTimeVal:=tcb^.retransTmoVal;
    TCPSend(tcb,NEWDATA);
    SetTimerEvent(tcb,PROBERCVR,tcb^.persistTimeVal);
    RETURN;
  END(*IF*);
  tcb^.outState:=TRANSMIT;
  IF tcb^.sendWinSz>tcb^.congWinSz THEN
     winSz:=tcb^.congWinSz;
  ELSE    
     winSz:=tcb^.sendWinSz;
  END(*IF*);
  waiting:=tcb^.sendNext-tcb^.sendUnAck;
  WHILE (HowMuch(tcb)-1 < HalfSeq) AND (waiting<=ADDRESS(winSz)) DO
    TCPSend(tcb,NEWDATA);
    waiting:=tcb^.sendNext-tcb^.sendUnAck;
  END(*WHILE*);
  IF (NOT TimerEventExists(tcb,RETRANSMITSEG)) THEN
    SetTimerEvent(tcb,RETRANSMITSEG,tcb^.retransTmoVal);
  END(*IF*);
END Transmit;    


(*Compute and send a TCP segment for the given TCB*)

PROCEDURE TCPSend(tcb:TCBPtr; retransmit:BOOLEAN);       
VAR
  code, offset, i, j, chk, Data1, hdrLength:CARDINAL;
  p:PacketPtr;
  Seqnum, newdata, dataLen:ADDRESS;
  rs:Status;
BEGIN     
  code:=0;
  NewPacket(p);
  IF p=NIL THEN
    RETURN ;
  END(*IF*);
  p^.data[IPproto]:=TCPProtocol;                 (* Fill in packet fields *)
  p^.data[IPsaddr]:=HostIP[0];
  p^.data[IPsaddr+1]:=HostIP[1];
  p^.data[IPsaddr+2]:=HostIP[2];
  p^.data[IPsaddr+3]:=HostIP[3];
  HostNet4(tcb^.forHost,IPdaddr,p^.data);
  HostNet2(tcb^.locPort,TCPsport,p^.data);
  HostNet2(tcb^.forPort,TCPfport,p^.data);
  dataLen:=TCPSndLen(tcb,retransmit,offset);
  IF NOT(retransmit) THEN
    HostNet4(tcb^.sendNext,TCPseq,p^.data);
    Seqnum:=tcb^.sendNext;
  ELSE
    HostNet4(tcb^.sendUnAck,TCPseq,p^.data);
    Seqnum:=tcb^.sendUnAck;
  END(*IF*);             
  HostNet4(tcb^.rcvNext,TCPack,p^.data);
  p^.data[TCPcode]:=0C;
  IF (tcb^.flags.sendFin) AND (((Seqnum+dataLen)-tcb^.sendLast)=0) THEN
    tcb^.tcbCode.fin:=TRUE;
  END(*IF*);
  IF NOT tcb^.flags.firstSend THEN
    code:=code+Ack;
  END(*IF*);
  IF tcb^.tcbCode.reset THEN
    code:=code+Reset;
  END(*IF*);
  IF tcb^.tcbCode.syn THEN
    code:=code+Syn;
   (*send max seg size as option*)

    p^.data[TCPopt]:=CHR(2);
    p^.data[TCPopt+1]:=CHR(4);
    HostNet2(MaxSegSz,TCPopt+2,p^.data);

    p^.data[TCPhlen]:=CHR((TCPHdr+1)*10H);    (*Look out for options*)
    Data1:=TCP1+TCPHdrLen+4;                                       
    hdrLength:=TCPHdrLen+4;
  ELSE
    p^.data[TCPhlen]:=CHR(TCPHdr*10H);
    Data1:=TCP1+TCPHdrLen;                        (*No options*)
    hdrLength:=TCPHdrLen;
  END(*IF*);
  IF tcb^.tcbCode.fin THEN
    code:=code+Fin;
  END(*IF*);         
  IF (dataLen>0) THEN
    code:=code+Push;
  END(*IF*);
  p^.data[TCPcode]:=CHR(code MOD 100H);
  HostNet2(TCPRWindow(tcb),TCPwin,p^.data);
  i:=(tcb^.sendBuffPos+offset) MOD BuffSz;
  FOR j:= Data1 TO (CARDINAL(dataLen)+Data1-1) DO     (*send all unacked data*)
   p^.data[j]:=tcb^.sendBuffer^[i];
   i:=(i+1) MOD BuffSz;
  END(*FOR*);        
  tcb^.flags.needOut:=FALSE;
  IF retransmit THEN                                  (*We're sending it out*)
    newdata:=tcb^.sendUnAck+dataLen-tcb^.sendNext;
    IF (newdata>HalfSeq) THEN
      newdata:=0;
    END;
    INC(TCPRetransSegs);
  ELSE
    newdata:=dataLen;
    IF tcb^.tcbCode.syn THEN
      INC(newdata);
    END(*IF*);
    IF tcb^.tcbCode.fin THEN
      INC(newdata);
    END(*IF*);
  END(*IF*);
  tcb^.sendNext:=tcb^.sendNext+newdata;
  IF newdata>=0 THEN
    INC(TCPOutSegs);
  END(*IF*);
  IF tcb^.tcbState=TIMEWAIT THEN             (*Final ack*)
    TCPWait(tcb);
  END(*IF*);
  p^.plen:=CARDINAL(dataLen)+Data1;
  p^.data[TCPchksum]:=0C;
  p^.data[TCPchksum+1]:=0C;
  ChkSum(chk, TCP1, CARDINAL(dataLen)+hdrLength, p^.data,TRUE);
  HostNet2(chk,TCPchksum,p^.data);
  rs:=IPSend(p);
END TCPSend;


(*Given a message it will call the correct procedure for the current
  output*)

PROCEDURE TCPOut;                     
 (* TCP output process *)
  VAR
    tcb    : TCBPtr;
    event  : Event;            
    socPtr : SocPtr;
  BEGIN
    LOOP
      GetMessage( socPtr, event );  (* Blocking call *)
      WITH socPtr^ DO
        SemP(tcbGuard);
        IF (tcb<>NIL) AND (tcb^.tcbState<>CLOSED) THEN
          IF event=DELETETCB THEN
            TCBDeallocate(tcb);
          ELSE  
            tcpOSwitch[tcb^.outState](tcb,event);
          END(*IF*);
        END(*IF*);
        SemV(tcbGuard);
      END(*WITH*);
    END; (* LOOP *)
  END TCPOut;

BEGIN 
  first:=TRUE;
  tcpOSwitch[IDLE]:=TCPIdle;
  tcpOSwitch[PERSIST]:=Persist;
  tcpOSwitch[TRANSMIT]:=Transmit;
  tcpOSwitch[RETRANSMIT]:=ReTransmit;
  Assert( CreateProcess( outProc, 0, 0, MaxPriority-1, TCPOut ) 
                                                   = L2Errors.Ok, 591 );
END N2TCPOut.
