IMPLEMENTATION MODULE N2TCPUser;

                IMPORT L2Errors;
FROM L2Clock    IMPORT TickCount;
FROM L2Files    IMPORT Users;
FROM L2Kernel   IMPORT Process, CurrentProcess, SemP, SemV, SemQ, 
                       SetNoSwitch, Resume;
FROM L2Storage  IMPORT ALLOCATE, DEALLOCATE;
FROM N2Base     IMPORT IPAddr, NetHost4, HostIP;
FROM N2Errors   IMPORT Status;  
FROM N2TCPLib   IMPORT TCBDeallocate, TCPRWindow, WakeUp;
                IMPORT N2TCPLib;
FROM N2TCPMain  IMPORT AnyForPort, BuffSz, Port, SocHandle, TCBState, 
                       Sleeper, TCBPtr, socList, activeSocs, MaxSocs, 
                       TCPCurrEstab, OutState, TCPTWOMSL, MaxSegSz;
FROM N2TCPTimer IMPORT Event, PutMessage, ClearTimerEvents, SetTimerEvent,
                       DeleteTimerEvent;
FROM SYSTEM     IMPORT ADDRESS, ADR;

CONST
  SeqIncrement = 904;  
  noPrivFloorPort = 256;

VAR
  nextPortNum     : Port;
  newSeq          : ADDRESS;

PROCEDURE AddSoc( port, forPort : Port; forAddr : IPAddr;
                  VAR soc : SocHandle ) : Status;
  VAR
    m : TCBPtr;
    ourAddr : IPAddr;
    i : SocHandle;
  BEGIN

   (* ...if request is for a reserved port then check for privilege *)

    IF ( port < noPrivFloorPort ) AND ( Users[CurrentProcess] <> 0 ) THEN
      RETURN PrivViol;
    END; (* IF *)
    NetHost4( ourAddr, 0, HostIP );
    IF ( port = forPort ) AND ( forAddr = ourAddr ) THEN
      RETURN IllegalOp;
    END; (* IF *)

   (* ...make sure request is not for a duplicate connection *)

    SemP( activeSocs.guard );
    IF forPort <> AnyForPort THEN  (* don't check if passive open! *)
      m := activeSocs.head;
      WHILE m <> NIL DO
        IF ( port = m^.locPort ) AND ( forPort = m^.forPort ) AND
           ( forAddr = m^.forHost ) THEN
          SemV( activeSocs.guard );
          RETURN ConInUse;
        END; (* IF *)
        m := m^.next;
      END; (* WHILE *)
    END; (* IF *)

   (* ...find a free socket *)

    FOR i := 1 TO MaxSocs DO
      IF socList[i].tcb = NIL THEN  (* found our free soc! *)
        SemV( activeSocs.guard );
        SemP( socList[i].tcbGuard );
        IF socList[i].tcb = NIL THEN (* we still have it *)
          SemP( activeSocs.guard );
          NEW( socList[i].tcb );
          IF socList[i].tcb = NIL THEN 
            SemV( activeSocs.guard );
            SemV( socList[i].tcbGuard );
            RETURN NoHeap; 
          ELSE
            NEW( socList[i].tcb^.sendBuffer );
            IF socList[i].tcb^.sendBuffer = NIL THEN 
              DISPOSE( socList[i].tcb );
              SemV( activeSocs.guard );
              SemV( socList[i].tcbGuard );
              RETURN NoHeap;           
            ELSE
              NEW( socList[i].tcb^.rcvBuffer );
              IF socList[i].tcb^.rcvBuffer = NIL THEN 
                DISPOSE( socList[i].tcb^.sendBuffer );
                DISPOSE( socList[i].tcb );
                SemV( activeSocs.guard );
                SemV( socList[i].tcbGuard );
                RETURN NoHeap; 
              END; (* IF *)
            END; (* IF *)
          END; (* IF *)
          soc := i;
          InitTCB( soc, port, forPort, forAddr );
          socList[soc].tcb^.prev := NIL;
          socList[soc].tcb^.next := activeSocs.head;
          IF activeSocs.head <> NIL THEN 
            activeSocs.head^.prev := socList[soc].tcb;
          END; (* IF *)
          activeSocs.head := socList[soc].tcb;
          SemV( activeSocs.guard );
          SemV( socList[soc].tcbGuard );
          RETURN Ok;
        ELSE                                        
         (* we lost it !!, look for another empty soc... *)
          SemV( socList[soc].tcbGuard );
          SemP( activeSocs.guard );
        END; (* IF *)
      END; (* IF *)
    END; (* FOR *)

   (* ...only get here if all the socs are in use. *)   

    SemV( activeSocs.guard );
    RETURN NoSocs
  END AddSoc;

PROCEDURE InitTCB( thisSoc : SocHandle; lPort, fPort : Port; fHost : IPAddr );
  BEGIN               
    newSeq := newSeq + SeqIncrement;
    WITH socList[thisSoc].tcb^ DO         
      pid := CurrentProcess;
      tcbState := CLOSED;
      soc := thisSoc;
      outState := IDLE;
      locPort := lPort;
      forPort := fPort;
      forHost := fHost;
      WITH tcbCode DO
        ack := FALSE;
        reset := FALSE;
        syn := TRUE;
        fin := FALSE;
      END; (* WITH *)
      WITH flags DO
        firstSend := FALSE;
        rcvdFin := FALSE;
        rcvDone := FALSE;
        sendDone := FALSE;
        sendFin := FALSE;
      END; (* WITH *)
      rdBlocked := FALSE;
      wrBlocked := FALSE;
      attached := FALSE;
      status := Ok;        
      initSendSeq := newSeq;
      sendUnAck := newSeq;
      sendNext := newSeq;
      lastWinAck := newSeq;
      congWinSz := 536;
      slowStartThresh := 65535;
      smoothRndTripTime := 0;
      rndTripDevEst := 0;
      retransTmoVal := 500;
      retransCnt := 0;
      sendBuffPos := 0;
      sendBuffCnt := 0;
      rcvBuffPos := 0;
      rcvBuffCnt := 0;
      fragHead := NIL;
      finSeq := 0;
      socList[thisSoc].tcb^.socPtr := ADR( socList[thisSoc] );
    END; (* WITH *)
  END InitTCB;

PROCEDURE TCPListen( port : Port; VAR soc : SocHandle ) : Status;
 (* Passive *)
  VAR
    n2res : Status;
  BEGIN
    n2res := AddSoc( port, 0, 0, soc );
    IF n2res <> Ok THEN RETURN n2res; END;
    SemP( socList[soc].tcbGuard );
    LOOP
      WITH socList[soc] DO
        tcb^.tcbState := LISTEN;
        tcb^.flags.needOut := FALSE;
        tcb^.ocBlocked := TRUE;
        SemV( tcbGuard );
        SemP( ioGuard );
        SemP( tcbGuard );
        IF tcb^.status = Ok THEN SemV( tcbGuard ); RETURN tcb^.status; END;
        InitTCB( soc, port, 0, 0 );
      END; (* WITH *)
    END; (* LOOP *)
  END TCPListen;

PROCEDURE TCPOpen( forPort : Port; forHost : IPAddr;
                                   VAR soc : SocHandle ) : Status;
 (* Active *)
  VAR
    n2res : Status;
    port  : Port;
  BEGIN
    SetNoSwitch( TRUE );
    port := nextPortNum;
    IF nextPortNum = 0FFFFH THEN
      nextPortNum := noPrivFloorPort;
    ELSE
      INC( nextPortNum );
    END; (* IF *) 
    SetNoSwitch( FALSE );
    n2res := AddSoc( port, forPort, forHost, soc );
    IF n2res <> Ok THEN RETURN n2res; END;
    SemP( socList[soc].tcbGuard );
    PutMessage( socList[soc].tcb^.socPtr, SENDPKT );
    WITH socList[soc] DO      
      tcb^.tcbState := SYNSENT;
      tcb^.flags.needOut := TRUE;
      tcb^.flags.firstSend := TRUE;
      tcb^.ocBlocked := TRUE;
      SemV( tcbGuard );
      SemP( ioGuard );
      SemP( tcbGuard );
      n2res := tcb^.status;
      IF n2res <> Ok THEN TCBDeallocate( tcb ); END;
      SemV( tcbGuard );
      RETURN n2res;
    END; (* WITH *)
  END TCPOpen;

PROCEDURE TCPWrite( soc : SocHandle; VAR data : ARRAY OF CHAR;
                len : CARDINAL ) : Status;
  VAR
    i, j, pos : CARDINAL;                       
    time      : ADDRESS;
  BEGIN
    IF ( 1 > soc ) OR ( soc > MaxSocs ) THEN RETURN BadSoc; END;
    WITH socList[soc] DO
      SemP( tcbGuard );

     (* ...check for error conditions and privilege. *)

      IF tcb = NIL THEN SemV( tcbGuard ); RETURN NoTCB; END;
      IF tcb^.pid <> CurrentProcess THEN 
        SemV( tcbGuard ); 
        RETURN PrivViol; 
      END; (* IF *)
      IF len > ( HIGH( data ) + 1 ) THEN 
        SemV( tcbGuard ); 
        RETURN BuffOverFlow; 
      END; (* IF *)
      i := 0;
      LOOP
        IF tcb^.tcbState <> ESTABLISHED THEN
          SemV( tcbGuard );
          RETURN BadState;
        END; (* IF *)
        IF tcb^.status <> Ok THEN SemV( tcbGuard ); RETURN tcb^.status; END;

        j := i;
        pos := (tcb^.sendBuffPos + tcb^.sendBuffCnt) MOD BuffSz;
        WHILE ( i < len ) AND ( i < BuffSz - tcb^.sendBuffCnt ) DO
          tcb^.sendBuffer^[pos] := data[i];
          pos := ( pos + 1 ) MOD BuffSz;
          INC( i );
        END; (* WHILE *)
        tcb^.sendBuffCnt := tcb^.sendBuffCnt + i - j;

        tcb^.flags.needOut := TRUE; 

        IF i = len THEN 
          IF ( tcb^.sendNext = tcb^.sendUnAck ) OR 
             ( tcb^.sendBuffCnt > MaxSegSz )  THEN
            PutMessage( tcb^.socPtr, SENDPKT ); 
          END; (* IF *)
          SemV( tcbGuard ); 
          RETURN Ok; 
        END;

        tcb^.wrBlocked := TRUE;
        SemV( tcbGuard );
        SemP( ioGuard );
        SemP( tcbGuard );
      END; (* LOOP *)
    END; (* WITH *)
  END TCPWrite;

PROCEDURE TCPRead( soc : SocHandle; VAR data : ARRAY OF CHAR;
                   VAR len : CARDINAL; wait : BOOLEAN ) : Status;
  BEGIN
    IF ( 1 > soc ) OR ( soc > MaxSocs ) THEN RETURN BadSoc; END;
    WITH socList[soc] DO
      SemP( tcbGuard );

     (* ...check for error conditions and privilege. *)

      IF tcb = NIL THEN 
        SemV( tcbGuard ); 
        RETURN NoTCB; 
      END; (* IF *)
      IF tcb^.pid <> CurrentProcess THEN 
        SemV( tcbGuard ); 
        RETURN PrivViol; 
      END; (* IF *)
      IF ( tcb^.tcbState <> ESTABLISHED ) 
         AND ( tcb^.tcbState <> CLOSEWAIT ) THEN
        SemV( tcbGuard );
        RETURN BadState;
      END; (* IF *)
      IF tcb^.status <> Ok THEN 
        SemV( tcbGuard );
        RETURN tcb^.status; 
      END; (* IF *)
                  
     (* need to let user know if other side of connection is closing *)

      IF tcb^.flags.rcvDone AND ( tcb^.rcvBuffCnt = 0 ) THEN
        SemV( tcbGuard );
        RETURN RcvDone;
      END; (* IF *)

     (* ...block till data arrives if required *)

      IF wait AND ( tcb^.rcvBuffCnt = 0 ) THEN
        tcb^.rdBlocked := TRUE;
        SemV( tcbGuard );
        SemP( ioGuard );     (* waiting for data from for host *)
        SemP( tcbGuard );
        IF ( tcb^.tcbState <> ESTABLISHED ) 
           AND ( tcb^.tcbState <> CLOSEWAIT ) THEN
          SemV( tcbGuard );
          RETURN BadState;
        END; (* IF *)
        IF tcb^.status <> Ok THEN 
          SemV( tcbGuard ); 
          RETURN tcb^.status; 
        END; (* IF *)
        IF tcb^.flags.rcvDone AND ( tcb^.rcvBuffCnt = 0 ) THEN
          SemV( tcbGuard );
          RETURN RcvDone;
        END; (* IF *)
      END; (* IF *)

     (* ...read data into user buffer from tcb buffer *)

      len := 0;
      WHILE ( tcb^.rcvBuffCnt > len ) AND ( len < HIGH( data ) + 1 ) DO
        data[len] := tcb^.rcvBuffer^[tcb^.rcvBuffPos];
        tcb^.rcvBuffPos := ( tcb^.rcvBuffPos + 1 ) MOD BuffSz;
        INC( len );
      END; (* WHILE *)
      tcb^.rcvBuffCnt := tcb^.rcvBuffCnt - len;

     (* ...see if we need to update from a zero window *)

      IF ( tcb^.winSeq <= tcb^.rcvNext ) AND
         ( TCPRWindow( tcb ) > 0 ) THEN
        tcb^.flags.needOut := TRUE;
        PutMessage( tcb^.socPtr, SENDPKT );
      END; (* IF *)
      SemV( tcbGuard );
      RETURN Ok;
    END; (* WITH *)
  END TCPRead;

PROCEDURE TCPClose( soc : SocHandle ) : Status;
  VAR
    n2res : Status;
  BEGIN
    IF ( 1 > soc ) OR ( soc > MaxSocs ) THEN RETURN BadSoc; END;
    WITH socList[soc] DO
      SemP( tcbGuard );
      IF tcb = NIL THEN 
        SemV( tcbGuard ); 
        RETURN NoTCB; 
      END; (* IF *)
      IF tcb^.pid <> CurrentProcess THEN 
        SemV( tcbGuard ); 
        RETURN PrivViol; 
      END; (* IF *)
      IF (tcb^.tcbState <> ESTABLISHED ) 
         AND ( tcb^.tcbState <> CLOSEWAIT ) THEN
        SemV( tcbGuard );
        RETURN BadState;
      END; (* IF *)
      IF tcb^.status <> Ok THEN
        TCBDeallocate( tcb );
        SemV( tcbGuard );
        RETURN Ok;
      END; (* IF *)

      DEC( TCPCurrEstab );
      tcb^.flags.sendFin := TRUE;
      tcb^.sendLast := tcb^.sendUnAck + ADDRESS( tcb^.sendBuffCnt );
      IF tcb^.tcbState = ESTABLISHED THEN
        tcb^.tcbState := FINWAIT1;
      ELSE  (* tcb^.tcbState = CLOSEWAIT *)
        tcb^.tcbState := LASTACK;
      END; (* IF *)
      tcb^.flags.needOut := TRUE;
      PutMessage( tcb^.socPtr, SENDPKT );
      tcb^.ocBlocked := TRUE;
      SemV( tcbGuard );
      SemP( ioGuard );        (* block till receive ACK of FIN or aborted *)
      SemP( tcbGuard );
      n2res := tcb^.status;
      IF ( n2res <> Ok ) OR ( tcb^.tcbState = LASTACK ) THEN
        TCBDeallocate( tcb );
      ELSE
        IF tcb^.sendBuffer <> NIL THEN DISPOSE( tcb^.sendBuffer ); END;
        IF tcb^.rcvBuffer <> NIL THEN DISPOSE( tcb^.rcvBuffer ); END;
        tcb^.pid := 0;
      END; (* IF *)
      SemV( tcbGuard );
      RETURN n2res;
    END; (* WITH *)
  END TCPClose;

PROCEDURE TCPAbort( soc : SocHandle ) : Status;
  BEGIN       
    IF ( 1 > soc ) OR ( soc > MaxSocs ) THEN RETURN BadSoc; END;
    WITH socList[soc] DO
      SemP( tcbGuard );
      IF tcb = NIL THEN SemV( tcbGuard ); RETURN NoTCB; END;
      IF ( tcb^.pid <> CurrentProcess ) AND
         ( Users[CurrentProcess] <> 0 ) THEN
        SemV( tcbGuard );
        RETURN PrivViol;
      END; (* IF *)
      IF ( tcb^.tcbState = ESTABLISHED ) AND ( tcb^.status = Ok ) THEN 
        DEC( TCPCurrEstab );
      END; (* IF *)
      N2TCPLib.TCPAbort( tcb, NotActive );
      tcb^.pid := 0;
      tcb^.tcbState := TIMEWAIT;
      SetTimerEvent( tcb, DELETETCB, 100 );
      SemV( tcbGuard );
      RETURN Ok;
    END; (* WITH *)
  END TCPAbort;

PROCEDURE TCPAttach( soc : SocHandle ) : Status;
  VAR
   n2res : Status;
  BEGIN
    IF ( 1 > soc ) OR ( soc > MaxSocs ) THEN RETURN BadSoc; END;
    WITH socList[soc] DO
      SemP( tcbGuard );
      IF tcb = NIL THEN
        n2res := NoTCB;
      ELSIF tcb^.pid <> CurrentProcess THEN
        n2res := PrivViol;
      ELSIF tcb^.status <> Ok THEN
        n2res := BadTCB;
      ELSIF (tcb^.tcbState<>ESTABLISHED) AND (tcb^.tcbState<>CLOSEWAIT) THEN
        n2res := NotActive;
      ELSE
        n2res := Ok;
        tcb^.attached := TRUE;
      END; (* IF *)
      SemV( tcbGuard );
      RETURN n2res;
    END; (* WITH *)
  END TCPAttach;

PROCEDURE TCPDetach( soc : SocHandle ) : Status;
  VAR
    n2res : Status;
  BEGIN
    IF ( 1 > soc ) OR ( soc > MaxSocs ) THEN RETURN BadSoc; END;
    WITH socList[soc] DO
      SemP( tcbGuard );
      IF tcb = NIL THEN
        n2res := NoTCB;
      ELSIF tcb^.pid <> CurrentProcess THEN
        n2res := PrivViol;
      ELSIF tcb^.status <> Ok THEN
        n2res := BadTCB;
      ELSE
        n2res := Ok;
        tcb^.attached := FALSE;
      END; (* IF *)
      SemV( tcbGuard );
      RETURN n2res;
    END; (* WITH *)
  END TCPDetach;

PROCEDURE TCPSync( pid : Process );
  VAR
    m : TCBPtr;
  BEGIN
    SemP( activeSocs.guard );
    m := activeSocs.head;
    WHILE ( m <> NIL ) DO
      IF m^.pid = pid THEN         
        SemV( activeSocs.guard );
        SemP( m^.socPtr^.tcbGuard );
        IF ( m^.tcbState = ESTABLISHED ) OR
           (( m^.tcbState > ESTABLISHED ) AND m^.ocBlocked AND
                                              ( m^.status = Ok )) THEN
          DEC( TCPCurrEstab );
        END; (* IF *)
        WakeUp( m, ALL );
        m^.socPtr^.flagged := TRUE;
        SemP( activeSocs.guard );
      END; (* IF *)
      m := m^.next;
    END; (* WHILE *)
    SemV( activeSocs.guard );
  END TCPSync;

PROCEDURE TCPCleanUp();
  VAR
    i  : SocHandle;
  BEGIN                        
    FOR i := 1 TO MaxSocs DO
      IF socList[i].flagged THEN 
        TCBDeallocate( socList[i].tcb );
        socList[i].flagged := FALSE;
        SemV( socList[i].tcbGuard );
      END; (* IF *)
    END; (* FOR *)
  END TCPCleanUp;

BEGIN
  newSeq := TickCount;
  nextPortNum := noPrivFloorPort;
END N2TCPUser.
