IMPLEMENTATION MODULE Exec;  (* Execute system command *)

(* G.R. Guenther  1989 01 03  LUMOS-2 version *)

(* This version executes EXTERNAL commands only - we do not invoke
   the CLI. (So the DEFINE tables are not available either.)
*)

FROM SYSTEM IMPORT ADDRESS;
FROM R2SysCalls IMPORT Status, GetMessage, Translate, Exit, 
                       WriteLn, WriteString;
IMPORT R2SysCalls;
FROM Errors IMPORT Error;
FROM R2Strings IMPORT Insert, Remove, Substring, Concatenate;

PROCEDURE  Execute( VAR command : ARRAY OF CHAR; echo, do : BOOLEAN);

(* If echo is TRUE, the command is printed on the user's terminal.
   IF do is TRUE, the command is executed by the system. If
   the command fails, an error message will be displayed (and the
   caller will be terminated).
*)

VAR     i, c, h, t, x : ARRAY [0..79] OF CHAR;
        r, s : Status;
        a : ADDRESS;
        p : CARDINAL;

BEGIN   IF echo THEN 
          WriteString("** ");
          WriteString(command);
          WriteLn;
          END;
        IF do THEN
          s := Translate("$CURRENT",h);
          c[0] := 0C; Insert(command,c,0);
          Component(c,t);
          Concatenate(t,".LOD",i);
          s := R2SysCalls.Execute(i,c,h,0,0,TRUE,TRUE,FALSE,p,r,a);
          IF s <> Ok THEN
            Concatenate("$LODLIB-",i,x);
            s := R2SysCalls.Execute(x,c,h,0,0,TRUE,TRUE,FALSE,p,r,a);
            END;
          IF s <> Ok THEN Error("Cannot execute",t) END;
          IF r <> Ok THEN
            IF r <> Status(-1) THEN 
              GetMessage(r,x);
              Error("Exit status",x);
              END;
            Exit(Status(-1));
            END;
          END;
END Execute;

PROCEDURE  Component( VAR s, t : ARRAY OF CHAR);

VAR     a : CARDINAL;

BEGIN   a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] = " ") DO INC(a) END;
        Remove(s,0,a);
        a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] > " ") DO INC(a) END;
        Substring(s,t,0,a);
        IF t[0] = ";" THEN t[0] := 0C; RETURN END;  (* Comments *)
        Remove(s,0,a);
END Component;

END Exec.
