IMPLEMENTATION MODULE IdLists;   (* manages simple ordered sets of IdPtrs *)

(* G.R. Guenther 1988 05 26  *)
(* Modified: GRG 1989 01 03  LUMOS-2 version *)

(* IdLists defines the IdList type and provides the basic operations.
   An IdList is an ordered list of IdPtrs in which each pointer has 
   an associate boolean attribute. Further, each IdPtr may occur only
   once in any list.
*)
            
FROM UsesLex IMPORT IdPtr;
FROM R2Storage IMPORT ALLOCATE;
FROM Errors IMPORT Assert;

TYPE    IdList = POINTER TO ICB;   
        Ilp = POINTER TO Ilink;
        ICB = RECORD get, head : Ilp;
              END;
        Ilink = RECORD ip : IdPtr;
                       ia : BOOLEAN;
                       nx : Ilp;
                END;

PROCEDURE  IdInit( VAR il : IdList);

(* Sets il to be an empty IdList. *)

BEGIN   NEW(il);
        WITH il^ DO
          get := NIL;
          head := NIL;
          END;
END IdInit;

PROCEDURE  IdPut( VAR il : IdList; idp : IdPtr; iat : BOOLEAN);

(* il is searched to see if idp occurs.  If it is found, IdPut 
   does nothing more, otherwise idp is added to the end of the
   list and its attribute is set to iat.
*)

VAR     p : Ilp;

  PROCEDURE  Init( VAR p : Ilp);

  BEGIN    NEW(p);
           WITH p^ DO
             nx := NIL;
             ip := idp;
             ia := iat;
             END;
           IF il^.get = NIL THEN il^.get := p END;
  END Init;

BEGIN   Assert(il <> NIL,"Uninitialised IdList");
        WITH il^ DO
          IF head = NIL 
            THEN Init(head); 
            ELSE p := head;
                 LOOP IF p^.ip = idp THEN RETURN END; 
                      IF p^.nx = NIL THEN EXIT END; 
                      p := p^.nx;
                      END;
                 Init(p^.nx);
            END;
          END;
END IdPut;

PROCEDURE  IdGet( VAR il : IdList; VAR idp : IdPtr; 
                  VAR iat : BOOLEAN) : BOOLEAN;
                          
(* IdGet iterates along the list.  Each time it is called, it yields
   the next IdPtr and its attribute.  It does not remove items,
   so even if an item is added again, IdGet will only see it once. 
   The return value indicates whether there was any item to get.
*)

BEGIN   Assert(il <> NIL,"Uninitialised IdList");
        WITH il^ DO
          IF get = NIL THEN RETURN FALSE END;
          idp := get^.ip;
          iat := get^.ia;
          get := get^.nx;
          END;
        RETURN TRUE;
END IdGet;

END IdLists.
