MODULE Manual;  (* extracts manuals from source files *)

(* G.R. Guenther  1988 06 07 *)

(* This program extracts manuals from source files (so that only
   one version of documentation needs to be maintained).

   Manual is invoked as in:

        Manual <source_file> [ <manual_file> ]

   <source_file> has a default extension of ".MOD". <manual_file>
   defaults to <source_file>.MAN.

   Manual honours the following embedded commands in source files:

        #START           -- begin copying to manual file
        #STOP            -- stop copying
        #INCLUDE ...     -- open another source file

   Commands must begin in column 1.  Other text may follow on #START
   and #STOP commands, but not on an #INCLUDE.  If you #INCLUDE while
   #START is in effect, the included file will begin with #START still
   in effect.  #START and #STOP are not stacked, #INCLUDE is.
*)

FROM TextIO IMPORT OpenInput, CloseInput, Eof, ReadLn,
                   CreateOutput, CloseOutput, PutString, PutLn;
FROM Commands IMPORT GetArg;
FROM Strings IMPORT Assign, Pos, Delete;
FROM Errors IMPORT SetName;

VAR     sf, mf, ln : ARRAY [0..133] OF CHAR;
        copy : BOOLEAN;

BEGIN   SetName("Manual");
        copy := FALSE;
        sf := "";
        GetArg("Source file",sf,ln,TRUE,FALSE);
        Assign(sf,mf);
        GetArg("Manual file",mf,ln,FALSE,FALSE);
        OpenInput(sf,".MOD","Cannot open source file",FALSE,TRUE);
        CreateOutput(mf,".MAN","Cannot create manual file");
        LOOP ReadLn(ln);
             IF Eof THEN 
               CloseInput; 
               IF NOT Eof THEN ReadLn(ln) END;
               END;
             IF Eof THEN EXIT END;
             IF Pos("#START",ln) = 0 THEN
                 copy := TRUE;
               ELSIF Pos("#STOP",ln) = 0 THEN
                 copy := FALSE;
               ELSIF Pos("#INCLUDE",ln) = 0 THEN
                 Delete(ln,0,8);
                 OpenInput(ln,".MOD","Cannot open #INCLUDE file",
                           FALSE,TRUE);
               ELSIF copy THEN
                 PutString(ln,0,0);
                 PutLn;
               END;
             END;
        CloseOutput;
END Manual.
