IMPLEMENTATION MODULE StrTable;  (* String table manager *)

(* G.R. Guenther 1988 05 28  *)
(* Modified:     1989 01 03  LUMOS-2 version *)

(* StrTable supports the storage of string constants.  It
   provides the procedure Enter which caches away a string
   and returns a pointer.  It guarantees that two identical
   strings will have identical pointers.

   CAUTION:  Do not trust the value of HIGH for any string
   pointer returned by this module.  StrTable does not
   guarantee that unused characters have actually been allocated.

   This implementation uses Binary Search Trees.
*)
                                                
FROM SYSTEM IMPORT ADDRESS;
FROM R2Strings IMPORT Length, Compare;
FROM R2Storage IMPORT ALLOCATE, DEALLOCATE;
 
TYPE    StrPtr = POINTER TO String;
        String = ARRAY [0..32767] OF CHAR; (* never fully allocated *)
        TreePtr = POINTER TO TreeNode;
        TreeNode = RECORD t : StrPtr;
                          l, r : TreePtr;
                   END;

VAR     Root : TreePtr;

PROCEDURE  Init( VAR p : TreePtr; VAR s : ARRAY OF CHAR ) : ADDRESS;

(* Creates a new node in the tree and returns a pointer to the
   string stored there.
*)

VAR     i, c : CARDINAL;

BEGIN   NEW(p);
        WITH p^ DO 
          c := Length(s);
          ALLOCATE(t,c+1);
          FOR i := 1 TO c DO t^[i-1] := s[i-1] END;
          t^[c] := 0C;
          l := NIL; r := NIL;        
          RETURN t;
          END;
END Init;

PROCEDURE  Enter( VAR str : ARRAY OF CHAR) : ADDRESS;

(* Enters str into the string table and returns its address. *)

VAR     p : TreePtr;
        d : INTEGER;

BEGIN   IF Root = NIL THEN RETURN Init(Root,str) END;
        p := Root;
        LOOP d := Compare(str,p^.t^);
             IF d = 0 THEN RETURN p^.t END;
             IF d < 0
               THEN IF p^.r = NIL 
                      THEN RETURN Init(p^.r,str) 
                      ELSE p := p^.r 
                      END;
               ELSE IF p^.l = NIL 
                      THEN RETURN Init(p^.l,str) 
                      ELSE p := p^.l 
                      END;
               END;
             END;
END Enter;

BEGIN   Root := NIL;
END StrTable.
