IMPLEMENTATION MODULE Symbols; (* Symbol tables for Update *)

(* G.R. Guenther  1988 05 30 *)
(* Modified:      1989 01 03  LUMOS-2 version *)

(* This module manages a simple kind of symbol table.  Both
   identifers and values are assumed to be StrPtrs.
*)                          

FROM UpdLex IMPORT StrPtr;
FROM R2Storage IMPORT ALLOCATE, DEALLOCATE;
IMPORT StrTable;

TYPE    SymTab = POINTER TO Symbol;
        Symbol = RECORD var, val : StrPtr;
                        next : SymTab;
                        marked : BOOLEAN;
                 END;

(* VAR     Null : StrPtr;  (* The string "" *)  *)

PROCEDURE  InitSymTab( VAR s : SymTab);

(* Initialisation *)

BEGIN   s := NIL;
END InitSymTab;

PROCEDURE  Enter( VAR s : SymTab; l, r : StrPtr);

(* Sets the value in s of ident l to string r. *)

VAR     p : SymTab;

BEGIN   p := s;
        WHILE (p <> NIL) AND NOT p^.marked DO
          IF p^.var = l THEN
            p^.val := r;
            RETURN;
            END;
          p := p^.next;
          END;
        NEW(p);
        WITH p^ DO
          var := l;
          val := r;
          next := s;
          marked := FALSE;
          END;
        s := p;
END Enter;

PROCEDURE  LookUp( VAR s : SymTab; l : StrPtr) : StrPtr;

(* Returns the value of l in s, or Null if not found. *)

VAR     p : SymTab;

BEGIN   p := s;
        WHILE (p <> NIL) AND NOT p^.marked DO
          IF p^.var = l THEN RETURN p^.val END;
          p := p^.next;
          END;
        RETURN Null;
END LookUp;

PROCEDURE  Mark( VAR s : SymTab);

(* Marks an activation record boundary in SymTab.  The next
   call to Release will delete all values entered since this Mark.
   Marks are stacked.
*)

BEGIN   IF s <> NIL THEN s^.marked := TRUE END;
END Mark;

PROCEDURE  Release( VAR s : SymTab);

(* Discard all entries since the last Mark.
*)

VAR     g : SymTab;

BEGIN   WHILE (s <> NIL) AND NOT s^.marked DO
          g := s;
          s := s^.next;
          DISPOSE(g);
          END;
        IF s <> NIL THEN s^.marked := FALSE END;
END Release;

BEGIN   Null := StrTable.Enter("");
END Symbols.
