IMPLEMENTATION MODULE UsesLex;  (* Lexical analyser for Uses *)

(* G.R. Guenther 1988 05 29  *)

(* This module is the lexical analyser for the Uses program.  It
   recognises a small fragment of the lexical rules for Modula-2
   programs.

   The lexer keeps only one copy of each identifier, so if
   it returns two equal pointers in different contexts, you
   may assume that the identifiers were equal without comparing
   the actual strings.

   Be cautious when using IdPtrs.  Do not assume that all the
   MaxIdLen characters have actually been allocated.  Never
   write on LexText^.

CONST   MaxIdLen = 133;  (* Identifiers cannot cross line boundaries ... *)

TYPE    Symbol = ( semiSY,   (* ;              *)
                   commaSY,  (* ,              *)
                   defnSY,   (* DEFINITION     *)
                   fromSY,   (* FROM           *)
                   implemSY, (* IMPLEMENTATION *)
                   importSY, (* IMPORT         *)
                   moduleSY, (* MODULE         *)
                   identSY,  (* identifier     *) 
                   otherSY   (* none of these  *) );

        IdType = ARRAY [0..MaxIdLen] OF CHAR;
        IdPtr  = POINTER TO IdType;

VAR     LexText : IdPtr;
        LexSym  : Symbol;
*)

FROM TextIO IMPORT Eof, GetChar, PutBack;
FROM StrTable IMPORT Enter;

VAR     defnID, fromID, implemID, importID, moduleID : IdPtr;

PROCEDURE  IsLetter( c : CHAR) : BOOLEAN;

BEGIN   c := CAP(c);
        RETURN ('A' <= c) AND (c <= 'Z');
END IsLetter;

PROCEDURE  IsIdentChar( c : CHAR): BOOLEAN;

BEGIN   RETURN IsLetter(c) OR (('0' <= c) AND (c <= '9'));
END IsIdentChar;

PROCEDURE  Lex;

(* Lex skips blanks, tabs and comments to find the next token in the
   current input file (in TextIO).  The type of the token is stored
   in LexSym. If LexSym = identSY then LexText^ will contain the 
   identifier. 
*)

VAR     text : IdType;
        ch : CHAR;
        i : CARDINAL;

BEGIN   LexSym := otherSY;
        i := 0; (* comment nesting depth *)
        LOOP IF Eof THEN RETURN END;
             ch := GetChar();
             IF ch = '(' THEN
               ch := GetChar();
               IF ch = '*' 
                 THEN INC(i)
                 ELSE IF i = 0 THEN PutBack; RETURN END;
                 END;
             ELSIF ch = '*' THEN
               IF i = 0 THEN RETURN END;
               ch := GetChar();
               IF ch = ')' THEN DEC(i) ELSE PutBack END;
             ELSIF (ch > ' ') AND (i = 0) THEN EXIT;
               END;
             END;
        IF ch = ';' THEN LexSym := semiSY; RETURN END;
        IF ch = ',' THEN LexSym := commaSY; RETURN END;
        IF IsLetter(ch) THEN
          i := 0;
          REPEAT text[i] := ch;
                 IF i < MaxIdLen THEN INC(i) END;
                 ch := GetChar();
            UNTIL NOT IsIdentChar(ch);
          PutBack;
          text[i] := 0C;
          LexText := Enter(text);
          IF LexText = implemID THEN LexSym := implemSY; RETURN END;
          IF LexText = defnID THEN LexSym := defnSY; RETURN END;
          IF LexText = moduleID THEN LexSym := moduleSY; RETURN END;
          IF LexText = fromID THEN LexSym := fromSY; RETURN END;
          IF LexText = importID THEN LexSym := importSY; RETURN END;
          LexSym := identSY;
          RETURN;
          END;
END Lex;

BEGIN   (* shuffled - why ? *)
        implemID := Enter("IMPLEMENTATION");
        defnID   := Enter("DEFINITION");
        moduleID := Enter("MODULE");
        fromID   := Enter("FROM");
        importID := Enter("IMPORT");
END UsesLex.
