IMPLEMENTATION MODULE UsesParse;  (* Parser for Uses *)

(* G.R. Guenther 1988 05 29  *)
(* Modified: GRG 1989 01 03  LUMOS-2 version *)
(*           GRG 1990 01 18  Ignore module priority *)

(* UsesParse contains a procedure to parse an initial fragment
   of a Modula-2 program. The parser is implemented as a nasty 
   little state machine.

   The parser accepts the following grammar:

   <strt> ::= [ 'DEFINITION' | 'IMPLEMENTATION' ] 'MODULE' 
                 <module_ident> <mdnm>
   <mdnm> ::= <anything else> ';' <imcl>
   <imcl> ::= ( 'FROM' <module_ident> <from> )   |
              ( 'IMPORT' <module_ident> <impo> ) |
              <anything_else> 
   <from> ::= 'IMPORT' <other_ident> (',' <other_ident>)* ';' <imcl>
   <impo> ::= <mdnm> | (',' <module_ident> <impo>)
*)

FROM UsesLex IMPORT Lex, LexSym, Symbol;
FROM Errors IMPORT Error, Assert;
FROM R2Strings IMPORT Concatenate;
FROM TextIO IMPORT ErrorDisplay;

TYPE    states = (strt, mdnm, imcl, from, impo);

VAR     st : states;

PROCEDURE  Expect( sy : Symbol);   (* syntax error messages *)

VAR     m : ARRAY [0..29] OF CHAR;

BEGIN   IF LexSym = sy THEN RETURN END;
        CASE sy OF
          semiSY   : m := '";"';
        | commaSY  : m := '","';
        | importSY : m := '"IMPORT"';
        | moduleSY : m := '"MODULE"';
        | identSY  : m := 'Identifier';
          ELSE Assert(FALSE,"Strange syntax error");
          END;
        Concatenate(m," expected",m);
        ErrorDisplay;
        Error("Modula-2 syntax error",m);  (* bye for now *)
END Expect;

PROCEDURE  Parse() : BOOLEAN;

(* Parse is an iterator.  Each time it is called, it looks for a
   new module name (in MODULE, FROM and IMPORT clauses).  If a module
   name is found Parse returns TRUE, otherwise FALSE.  The actual name
   is available in the LexText^ variable (in the UsesLex module).
   
   Some syntax errors can be detected.  These will generate 
   a fatal error.  Parse ignores all text beyond the last import,
   however.

        WARNING:  Use of LOOP & CASE to simulate GOTO !!!
*)

BEGIN   LOOP  Lex;
              CASE st OF
                 strt : IF (LexSym = defnSY) OR (LexSym = implemSY)
                          THEN Lex (* ignore it *) END;
                        Expect(moduleSY);
                        Lex; Expect(identSY);
                        st := mdnm;
                        EXIT;
               | mdnm : WHILE LexSym <> semiSY DO Lex END; (* ignore prio *)
                        st := imcl;
               | imcl : IF LexSym = fromSY THEN
                          Lex; Expect(identSY);
                          st := from;
                          EXIT;
                        ELSIF LexSym = importSY THEN
                          Lex; Expect(identSY);
                          st := impo;
                          EXIT;
                        ELSE
                          st := strt;  (* end of this module *)
                          RETURN FALSE;
                          END;
               | from : Expect(importSY);
                        Lex;
                        IF LexSym <> semiSY THEN
                          LOOP Expect(identSY); Lex;
                               IF LexSym = semiSY 
                                 THEN EXIT
                                 ELSE Expect(commaSY); Lex;
                                 END;
                               END;
                          END;
                        st := imcl;
               | impo : IF LexSym = semiSY 
                          THEN st := imcl
                          ELSE Expect(commaSY); Lex;
                               Expect(identSY);
                               EXIT;  (* state still impo *)
                          END;
                 ELSE Assert(FALSE,"Parser confused");
                 END;
              END;
        RETURN TRUE;
END Parse;

BEGIN   st := strt;  (* First file starts at the beginning *)
END UsesParse.
