MODULE  Account;  (* Manage the PASSWORDS.SYS file and directories *)

FROM R2SysCalls IMPORT ReadString, WriteString, Read, WriteLn, Status,
                       SetUser, Rename, GetPID, GetMessage, Exit, 
                       SetCCMode, Delete, GetCommand, ModeBits, SetMode,
                       Handle, FileMode, Open, Close, Create, Translate;
FROM R2Format IMPORT ScanCard, FormatCard;
FROM R2Streams IMPORT Stream, Done, EndStream, OpenStream, ReadLine,
                      CreateStream, WriteLine, CloseStream, 
                      SetPos, ReleaseStream;
FROM R2Strings IMPORT Remove, Substring, Copy, Compare, Insert, 
                      Length, Concatenate;
FROM R2Directories IMPORT NMode, FullName, SetWild, NextWild, NOwner;
FROM LockSync IMPORT LockWait, LockClear;

CONST   uaf = "$SysDisk-PASSWORDS.SYS";
        tmp = "$SysDisk-passwords.new";
        jnk = "$SysDisk-passwords.old";
        lck = "passwords.lock";
        usr = "$UserDisk-";
        sys = "$SysDisk-";

TYPE    String = ARRAY [0..79] OF CHAR;
                                             
PROCEDURE  Error( VAR s : ARRAY OF CHAR);

BEGIN   WriteString("Account -- ");
        WriteString(s);
        WriteLn;
        Exit(Status(-1));
END Error;

PROCEDURE  Check( rs : Status);

VAR     s : ARRAY [0..79] OF CHAR;

BEGIN   IF rs = Ok THEN RETURN END;
        GetMessage(rs,s);
        Error(s);
END Check;
   
PROCEDURE  Drop( VAR tree : ARRAY OF CHAR);

VAR     h : Handle;
        rs : Status; 
        wc, df : String;
        oo : CARDINAL;

BEGIN   Check(SetWild(tree));
        WHILE NextWild() = Ok DO
          IF DIR IN FileMode(NMode)
            THEN Copy(FullName,df);
                 oo := NOwner;
                 Concatenate(df,"-*",wc);
                 Drop(wc);  
                 Check(Open(df,TRUE,h));
                 Check(SetMode(h,FileMode(0FFH),oo));
                 Check(Close(h,TRUE));
                 Check(SetWild(tree));  (* Restart *)
            ELSE Check(Delete(FullName));
            END;
          END;
END Drop;

PROCEDURE  Component( VAR s, t : ARRAY OF CHAR);

VAR     a : CARDINAL;

BEGIN   a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] = " ") DO INC(a) END;
        Remove(s,0,a);
        a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] > " ") DO INC(a) END;
        Substring(s,t,0,a);
        IF t[0] = ";" THEN t[0] := 0C; RETURN END;  (* Comments *)
        Remove(s,0,a);
END Component;

PROCEDURE  AddAcc( VAR name : ARRAY OF CHAR);

VAR     ps : Stream;
        d, r, s, t : String;
        h : Handle;
        i, max, n : CARDINAL;
        x : Status;

BEGIN   Copy(name,t);
        Concatenate(usr,t,s);
        x := Translate(s,d);
        IF x = NoLogical THEN
          Concatenate(sys,t,s);
          x := Translate(s,d);
          END;
        IF x <> Ok THEN Error("Cannot construct directory name") END;
        max := 0;
        OpenStream(ps,uaf,TRUE);
        IF NOT Done THEN Error("Cannot access password file") END;
        WHILE NOT EndStream(ps) DO
          ReadLine(ps,s);
          Component(s,t);
          IF Compare(name,t) = 0 THEN
            CloseStream(ps);
            Error("Username already in use");
            END;
          FOR i := 1 TO 3 DO Component(s,t) END;
          IF ScanCard(t,10,n) AND (n > max) THEN max := n END;
          END;
        FormatCard(max+1,10,1,r);
        Concatenate(name,"  ",s);
        Concatenate(s,s,t);
        Concatenate(t,d,s);
        Concatenate(s,"  ",t);
        Concatenate(t,r,s);
        Concatenate(s,"  N  N",t);
        SetPos(ps,0FFFFFFFFH);     (* Append *)
        WriteLine(ps,t);
        CloseStream(ps);
        Check(Create(d,FileMode(01F0H),h));
        Check(SetMode(h,FileMode(01F0H),max+1));
        Check(Close(h,TRUE));
        WriteString(t);
        WriteLn;
END AddAcc;

PROCEDURE  DelAcc( VAR name : ARRAY OF CHAR);

VAR     ps, np : Stream;
        found : BOOLEAN;
        s, t, n, d : String;

BEGIN   WriteString("Are you sure ? ");
        ReadString(s);
        IF CAP(s[0]) <> 'Y' THEN RETURN END;
        OpenStream(ps,uaf,FALSE);
        IF NOT Done THEN Error("Cannot access password file") END;
        CreateStream(np,tmp);
        IF NOT Done THEN Error("Cannot create temporary file") END;
        found := FALSE;
        WHILE NOT EndStream(ps) DO
          ReadLine(ps,s);
          Copy(s,t);
          Component(t,n);
          IF Compare(n,name) = 0
            THEN found := TRUE;
                 Component(t,d);
                 Component(t,d);  (* dir *)
            ELSE WriteLine(np,s);
            END;
          END;
        CloseStream(ps);
        CloseStream(np);
        IF found
          THEN Check(Rename(uaf,jnk));
               Check(Rename(tmp,uaf));
               Check(Delete(jnk));
               WriteString("Deleting directory and files ...");
               WriteLn;
               LockClear(lck);  (* don't make everybody wait for this *)
               Drop(d);
          ELSE Check(Delete(tmp));
               Error("Password record not found");
          END; 
END DelAcc;

VAR     gs, cs, ns : String;
        c : CHAR;
        k, i : CARDINAL;

BEGIN   IF NOT LockWait(lck,30)
          THEN Error("Waited too long for password file") END;
        SetCCMode(FALSE);
        GetCommand(gs);
        Component(gs,cs);
        Component(gs,ns);
        k := Length(ns);
        IF (k < 1) OR (k > 8) THEN Error("Username syntax error") END;
        FOR i := 0 TO k-1 DO 
          c := CAP(ns[i]);
          IF NOT (((c >= 'A') AND (c <= 'Z')) OR ((c >= '0') AND (c <= '0')))
            THEN Error("Username syntax error") END;
          END;
        c := CAP(cs[0]);
        IF (c= 'C') OR (c = 'A') THEN AddAcc(ns)
          ELSIF  (c = 'R') OR (c = 'D') THEN DelAcc(ns);
          ELSE Error("Unknown command")
          END;
END Account.
