MODULE Compare; (* Compare text files *)

FROM R2SysCalls IMPORT GetCommand, GetMessage, Status, Translate, WriteString,
                       WriteLn, Exit;
FROM R2Streams IMPORT Stream, OpenStream, EndStream, ReadLine, CreateStream,
                      Done, CloseStream, WriteLine, GetPos, SetPos;
FROM R2Strings IMPORT Length, Compare, Insert, Remove, Substring, Copy;
FROM R2Format IMPORT FormatCard;
FROM SYSTEM IMPORT ADDRESS;

CONST   MaxDiffs = 150;  (* Maximum number of lines in a difference section *)
       
TYPE    Line = ARRAY [0..79] OF CHAR;  (* Diffs beyond col 80 not seen ! *)
        DiffRec = RECORD Lin : Line;
                         Sum, Num : CARDINAL;
                         FilAdr : ADDRESS;
                  END;

VAR     fs : ARRAY [0..1] OF Stream;
        os : Stream;
        dr : ARRAY [0..1] OF ARRAY [0..MaxDiffs] OF DiffRec;
        lc, ma, pt : ARRAY [0..1] OF CARDINAL;
        es : ARRAY [0..1] OF BOOLEAN;
        dc : CARDINAL;

PROCEDURE  Err( VAR s : ARRAY OF CHAR);

BEGIN   WriteString("Compare -- ");
        WriteString(s);
        WriteLn;
        Exit(Status(-1));
END Err;

PROCEDURE  Component( VAR s, t : ARRAY OF CHAR);

VAR     a : CARDINAL;

BEGIN   a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] = " ") DO INC(a) END;
        Remove(s,0,a);
        a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] > " ") DO INC(a) END;
        Substring(s,t,0,a);
        IF t[0] = ";" THEN t[0] := 0C; RETURN END;  (* Comments *)
        Remove(s,0,a);
END Component;

PROCEDURE  CheckSum( VAR s : ARRAY OF CHAR ) : CARDINAL;

VAR     cs, i : CARDINAL;

BEGIN   i := 0;
        cs := 0;
        WHILE (i <= HIGH(s)) AND (s[i] <> 0C) DO
          INC(cs,ORD(s[i]));
          INC(i);
          END;
        RETURN cs;
END CheckSum;       

PROCEDURE  StartFiles;

VAR     st, fn0, fn1, fn2 : Line;
        rs : Status;

BEGIN   GetCommand(st);
        Component(st,fn0);
        Component(st,fn1);
        IF fn1[0] = 0C THEN Err("Two filenames required") END;
        Component(st,fn2);
        IF fn2[0] = 0C THEN Copy("$TI",fn2) END;
        OpenStream(fs[0],fn0,FALSE);
        IF NOT Done THEN Err("Cannot open first input file") END;
        OpenStream(fs[1],fn1,FALSE);
        IF NOT Done THEN Err("Cannot open second input file") END;
        CreateStream(os,fn2);
        IF NOT Done THEN Err("Cannot open output file") END;
        rs := Translate(fn0,st);
        Insert("0 = ",st,0);     
        WriteLine(os,st);
        rs := Translate(fn1,st);
        Insert("1 = ",st,0);
        WriteLine(os,st);
END StartFiles;
               
PROCEDURE  DumpList( n : CARDINAL);

VAR     i : CARDINAL;
        st, tm : Line;

BEGIN   i := 0;
        IF (ma[n] > 0) OR (ma[1-n] > 0) THEN
          WriteLine(os,"");
          INC(dc);
          WHILE i <= ma[n] DO
            Copy(dr[n][i].Lin,st);
            FormatCard(dr[n][i].Num,10,5,tm);
            Insert(": ",st,0);
            Insert(tm,st,0);
            IF n = 0 
              THEN Insert("0 ",st,0) 
              ELSE Insert("1 ",st,0) 
              END;
            WriteLine(os,st);
            INC(i);
            END;
          END;
        IF ma[n] <> pt[n] THEN 
          SetPos(fs[n],dr[n][ma[n]].FilAdr);
          lc[n] := dr[n][ma[n]].Num+1;
          END;
END DumpList;

PROCEDURE  ReadRec( n : CARDINAL) : BOOLEAN;

(* Returns TRUE if the record read is also on the other list.  Sets ma[] *)

VAR     s, i : CARDINAL;

BEGIN   IF pt[n] > MaxDiffs THEN Err("Difference table overflow") END;
        IF es[n] THEN RETURN FALSE END;
        WITH dr[n][pt[n]] DO
          IF EndStream(fs[n]) 
            THEN Copy("<EOF>",Lin);
                 es[n] := TRUE;
            ELSE ReadLine(fs[n],Lin);
            END;
          GetPos(fs[n],FilAdr);
          Sum := CheckSum(Lin);
          s := Sum;
          Num := lc[n];
          INC(lc[n]);
          END;
        INC(pt[n]);      
        i := 0;
        WHILE i < pt[1-n] DO
          WITH dr[1-n][i] DO
            IF s = Sum THEN
              IF Compare(dr[n][pt[n]-1].Lin,Lin) = 0 THEN
                ma[n] := pt[n]-1;
                ma[1-n] := i;
                RETURN TRUE;
                END;
              END;
            END;
          INC(i);
          END;
        RETURN FALSE;
END ReadRec;

PROCEDURE  NextSection;

BEGIN   pt[0] := 0;
        pt[1] := 0;
        WHILE NOT(ReadRec(0) OR ReadRec(1)) DO END;  (* Quits on Match *)
        DumpList(0);
        DumpList(1);
END NextSection;

VAR     st : Line;

BEGIN   StartFiles;
        dc := 0;
        lc[0] := 1;
        lc[1] := 1;
        es[0] := FALSE;
        es[1] := FALSE;
        REPEAT NextSection;
        UNTIL es[0] OR es[1];
        CloseStream(fs[0]);
        CloseStream(fs[1]);
        FormatCard(dc DIV 2,10,1,st);
        Insert(" difference sections found",st,Length(st));
        WriteLine(os,"");
        WriteLine(os,st);
        CloseStream(os);
END Compare.
