MODULE Compile;  (* Compiles an index file for a help database *)

FROM R2Strings IMPORT Concatenate, Compare, Copy, Length, Remove;
FROM R2SysCalls IMPORT Status, Create, Close, WriteBlock, Handle, 
                       FileMode, GetCommand, WriteString, WriteLn;
FROM R2NumbIO IMPORT WriteCard;
FROM Errors IMPORT SetName, Error;
FROM R2Streams IMPORT Stream, OpenStream, EndStream, CloseStream, ReadLine,
                      Done, GetPos;
FROM SYSTEM IMPORT ADDRESS;

CONST   maxrecs = 320;

TYPE    toptyp = ARRAY [0..11] OF CHAR;
        block = ARRAY [0..511] OF CHAR;
        indrec = RECORD nam : toptyp;
                        pos : ADDRESS;
                 END;

VAR     i, j, k : CARDINAL;
        ts, hfn, ifn, line : ARRAY [0..79] OF CHAR;
        hs : Stream;
        ih : Handle;
        index : RECORD
                  CASE BOOLEAN OF
                       TRUE : blocks : ARRAY [0..(maxrecs DIV 32)-1] OF block;
                    | FALSE : recs : ARRAY [0..maxrecs-1] OF indrec;
                    END;
                  END;

PROCEDURE  Swap( i, j : CARDINAL);  (* for bubble sort *)

VAR     t : indrec;

BEGIN   WITH index DO
          t := recs[i];
          recs[i] := recs[j];
          recs[j] := t;
          END;
END Swap;

BEGIN   SetName("Compile");
        GetCommand(ts);
        Concatenate(ts,".HLP",hfn);
        Concatenate(ts,".IND",ifn);
        OpenStream(hs,hfn,FALSE);
        IF NOT Done THEN Error("Cannot open",hfn) END;
        IF Create(ifn,FileMode(0B0H),ih) <> Ok 
          THEN Error("Cannot create",ifn) END;
        WITH index DO
          k := 0;
          WHILE NOT EndStream(hs) DO
            ReadLine(hs,line);
            IF (line[0] = '*') AND (Length(line) > 1) THEN
              Remove(line,0,1);
              Concatenate(line,"            ",recs[k].nam);
              GetPos(hs,recs[k].pos);
              INC(k);
              IF k >= maxrecs-1 THEN
                Error("Too many help texts in",hfn) END;
              END;
            END;
          CloseStream(hs);
          recs[k].nam[0] := 0C;
          FOR i := 1 TO k-1 DO
            FOR j := 1 TO k-i DO
              IF Compare(recs[j-1].nam,recs[j].nam) > 0 
                THEN Swap(j-1,j) 
                END;
              END;
            END;
          j := (k+31) DIV 32;
          FOR i := 1 TO j DO 
            IF WriteBlock(ih,blocks[i-1],TRUE) <> Ok
              THEN Error("Error writing index file",""); END;
            END;
          END;
        IF Close(ih,TRUE) <> Ok 
          THEN Error("Error closing index file","") END;
        WriteCard(k,1);
        WriteString(" help records processed.");
        WriteLn;
END Compile.
