MODULE Document;

FROM SYSTEM IMPORT
  ADDRESS ;
FROM R2SysCalls IMPORT 
  ReadString, Write, WriteString, WriteLn, Status, 
  GetCommand, Exit, GetMessage, Result, Translate;
FROM R2Format IMPORT
  FormatCard;
FROM R2Strings IMPORT
  Remove, Copy, Length, Insert, Substring, Compress;
FROM R2Streams IMPORT 
  Stream, OpenStream,CreateStream,Done,WriteChar,ReadLine,
  WriteLine, CloseStream, EndStream;

CONST
  FF       = 14C;
  MAXLINES = 59;
  LINESIZE = 79;               
  NAMELEN  = 22;

TYPE
 
  BufferType = ARRAY[0..131] OF CHAR;
  LineType   = ARRAY[0..LINESIZE]  OF CHAR;

VAR
  curfilename : LineType;
  instream,outstream,tocstream,curstream : Stream;
  pagenum,toclines : CARDINAL;

PROCEDURE Check(
                   r : Status);
 
VAR
  str : LineType;

BEGIN
  IF r = Ok THEN RETURN END;
  GetMessage(r,str);
  Err(str);
END Check;

PROCEDURE  Err( VAR s : ARRAY OF CHAR);

BEGIN
  WriteString("Document -- ");
  WriteString(s);
  WriteLn;
  Exit(Status(-1));
END Err;

PROCEDURE  Component( VAR s, t : ARRAY OF CHAR);

VAR     a : CARDINAL;

BEGIN   a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] = " ") DO INC(a) END;
        Remove(s,0,a);
        a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] > " ") DO INC(a) END;
        Substring(s,t,0,a);
        IF t[0] = ";" THEN t[0] := 0C; RETURN END;  (* Comments *)
        Remove(s,0,a);
END Component;

PROCEDURE CenterText(VAR s,t : ARRAY OF CHAR);

VAR
  i, len, pad : CARDINAL;

BEGIN
  len := Length(s);     
  IF len <= HIGH(t) + 1 THEN
    pad := (HIGH(t) + 1 - len) DIV 2;
    FOR i := 0 TO pad-1 DO
      t[i] := " ";
    END;             
    FOR i := 0 TO len-1 DO
      t[pad + i] := s[i];
    END;              
    IF len + pad <= HIGH(t) THEN
      t[len+pad] := 0C;
    END;
  END;
END CenterText;
          
PROCEDURE UpdateTOC;
                                
VAR
  line,buf : LineType;
  i, j, k : CARDINAL;

BEGIN
          
  IF toclines MOD MAXLINES = 0 THEN
    toclines := 0;
    WriteChar(tocstream,FF);
    FOR i := 0 TO 1 DO
      WriteLine(tocstream,"");
      INC(toclines);
    END;  
    CenterText("Table of Contents",line);
    WriteLine(tocstream,line);
    INC(toclines);
    FOR i := 0 TO 1 DO
      WriteLine(tocstream,"");
      INC(toclines);
    END;  
  END;   

  FOR i := 0 TO NAMELEN+2 DO
    buf[i] := " ";
  END;
  k := Length(curfilename);
  FOR i := 0 TO k-1 DO
    buf[i] := curfilename[i];
  END;

  FormatCard(pagenum,10,3,line);
  k := NAMELEN + 3;
  FOR i := 0 TO 2 DO
    buf[k] := line[i];
    INC(k);
  END;    
  buf[k] := 0C;
 
  CenterText(buf,line);
  WriteLine(tocstream,line);
  INC(toclines);
  WriteString(line); WriteLn;

END UpdateTOC;
  
PROCEDURE PrintFile;
                              
VAR
  line,temp : LineType;
  buf : BufferType;
  lc, i, j, k  : CARDINAL;

BEGIN 
                  
  UpdateTOC;
  lc := 0;
   
  WHILE NOT EndStream(curstream) DO

    IF lc MOD MAXLINES = 0 THEN 
      WriteChar(outstream,FF);
      lc := 0;
      FOR i := 0 TO 1 DO
        WriteLine(outstream,"");
        INC(lc);
      END;          
      FOR i := 0 TO HIGH(line) DO
        line[i] := " ";
      END;
      Insert(curfilename,line,21);
      FormatCard(pagenum,10,3,temp);
      j := 0;
      FOR i := HIGH(line)-4 TO HIGH(line)-1 DO
        line[i] := temp[j];
        INC(j);
      END;                                
      line[HIGH(line)] := 0C;
      WriteLine(outstream,line);
      INC(lc);
      FOR i := 0 TO 1 DO
        WriteLine(outstream,"");
        INC(lc);
      END;          
      INC(pagenum);
    END;

    ReadLine(curstream,buf);
    WriteLine(outstream,buf);
    INC(lc);

  END;(*WHILE*)

END PrintFile;  
    
VAR
  line,temp,buf,infilename,outfilename,tocfilename : LineType;    
  i, j : CARDINAL;

BEGIN
  WriteString("Enter file name of source documents: ");
  ReadString(infilename);
  WriteString("Enter name of output file: ");
  ReadString(outfilename);
  WriteString("Enter name of Table of Contents file: ");
  ReadString(tocfilename);
  OpenStream(instream,infilename,FALSE);
  IF NOT Done THEN
    Err("Error opening input file");
  END;
  CreateStream(outstream,outfilename);
  IF NOT Done THEN
    Err("Error opening output file");
  END;                   
  CreateStream(tocstream,tocfilename);
  IF NOT Done THEN
    Err("Error opening table of contents file");
  END;
                   
  toclines := 0;
  pagenum := 1;

  WHILE NOT EndStream(instream) DO

    ReadLine(instream,line);
    Component(line,curfilename);
    Check(Translate(curfilename,buf));
    i := 0; j := 65535;
    WHILE (i <= HIGH(curfilename)) AND (curfilename[i] <> 0C) DO
      IF curfilename[i] = '-' THEN j := i END;
      INC(i);
      END;
    Remove(curfilename,0,j+1);
    OpenStream(curstream,buf,FALSE);
    IF NOT Done THEN
      Insert("Error opening input file ",buf,0);
      WriteString(buf);WriteLn;
    ELSE
      IF NOT EndStream(curstream) THEN
          PrintFile;
      END;
      CloseStream(curstream);
    END;
  END;

  CloseStream(instream);
  CloseStream(outstream);
  CloseStream(tocstream);


END Document.
