MODULE Dump;

FROM R2SysCalls IMPORT 
  ReadString, WriteString, WriteLn, ReadBlock, Translate, Status, 
  Seek, Size, Handle, Open, Close, GetCommand, Exit, GetMessage, Result;
FROM R2Format IMPORT
  FormatCard, ScanCard;
FROM R2Strings IMPORT
  Remove, Copy, Length, Insert, Substring;                               
FROM R2Streams IMPORT
  Stream, Done, CreateStream, CloseStream, WriteLine, WriteChar;


TYPE
 
  LineType = ARRAY[0..79] OF CHAR;
  Three = [0..2];
  Block = RECORD 
            CASE Three OF
              0 : a : ARRAY [0..511] OF CHAR|
              1 : n : ARRAY [0..31] OF ARRAY [0..7] OF CARDINAL|
              2 : c : ARRAY [0..31] OF ARRAY [0..15] OF CHAR;
            END;
          END;

PROCEDURE Check(
                   r : Status);
 
VAR
  str : LineType;

BEGIN
  IF r = Ok THEN RETURN END;
  GetMessage(r,str);
  Err(str);
END Check;
    

PROCEDURE  Err(
               VAR s : ARRAY OF CHAR);

BEGIN
  WriteString("Dump -- ");
  WriteString(s);
  WriteLn;
  Exit(Status(-1));
END Err;

PROCEDURE  Component( VAR s, t : ARRAY OF CHAR);

VAR     a : CARDINAL;

BEGIN   a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] = " ") DO INC(a) END;
        Remove(s,0,a);
        a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] > " ") DO INC(a) END;
        Substring(s,t,0,a);
        IF t[0] = ";" THEN t[0] := 0C; RETURN END;  (* Comments *)
        Remove(s,0,a);
END Component;
            
VAR
  Line, infilename, outfilename, temp, buffer : LineType;
  rs : Status;                 
  infile : Handle; 
  outstream : Stream;          
  blk : Block;
  filesize, bc,row,col,blknum,address,i,charval : CARDINAL;

BEGIN                                                         
  GetCommand(Line);
  Component(Line,temp);
  IF temp[0] = 0C THEN
    LOOP
      WriteString("Dump> ");
      ReadString(temp);
      CASE Result() OF
      
        EndFile,EscKey : Exit(Status(-1)); |
        
        Ok             : IF temp[0] <> 0C THEN EXIT END;
  
      ELSE
      END;
    END;
  END;
  Check(Translate(temp,infilename));
  Check(Open(infilename,FALSE,infile));
  rs := Size(infile,filesize);
  IF rs <> Ok THEN
    filesize := 65535;  (* Must be NFS Open ? *)
  END;
  Component(Line,temp);
  IF temp[0] = 0C THEN  (* No command line options *)
    Copy("$TI",temp);
    blknum := 0;
  ELSE
    IF NOT ScanCard(temp,10,blknum) THEN
      Err("No starting block number indicated");
    ELSE
      Component(Line,temp);
      IF temp[0] = 0C THEN  (* Output to terminal *)
        Copy("$TI",temp); 
      END;
    END;
  END;
  Check(Translate(temp,outfilename));
  CreateStream(outstream,outfilename);
  IF NOT Done THEN
    Err("Unable to open output file");
  END;                 
  IF blknum <> 0 THEN (* Seek to proper block *)
    IF blknum < filesize THEN 
      Check(Seek(infile,blknum));
      address := blknum*512;
    ELSE
      Err("Block number out of range");
    END;
  ELSE
    address := 0;
  END;

  LOOP
    rs := ReadBlock(infile,blk.a);
    IF rs <> Ok THEN EXIT END;
    WriteChar(outstream,14C);
    WriteLine(outstream,infilename);
    Copy("Block Number = ",buffer);
    FormatCard(blknum,10,1,temp);
    Insert(temp,buffer,15);
    WriteLine(outstream,buffer);
    FOR row := 0 TO 31 DO
      bc := 0;
      buffer[0] := 0C;
      FormatCard(address,16,8,temp);
      INC(address,16);
      FOR i := 1 TO 7 DO
        IF temp[i] = " " THEN
          temp[i] := "0";
        END;
      END;
      Insert(temp,buffer,bc);
      INC(bc,8);
      Insert("  ",buffer,bc);
      INC(bc,2);
      FormatCard(16*row,16,6,temp);
      FOR i := 1 TO 5 DO
        IF temp[i] = " " THEN
          temp[i] := "0";
        END;
      END;
      Insert(temp,buffer,bc);
      INC(bc,6);
      Insert("  ",buffer,bc);
      INC(bc,2);
      FOR col := 0 TO 7 DO
        FormatCard(blk.n[row,col],16,5,temp);
        FOR i := 1 TO 3 DO
          IF temp[i] = " " THEN
            temp[i] := "0";
          END;
        END;
        Insert(temp,buffer,bc);
        INC(bc,5);
      END;      
      Insert("  ",buffer,bc);
      INC(bc,2);
      FOR col := 0 TO 15 DO
        charval := ORD(blk.c[row,col]) MOD 128; 
        IF (charval >= 32) AND (charval < 127)  THEN
          buffer[bc] := blk.c[row,col];
        ELSE
          buffer[bc] := ".";
        END;
        INC(bc);
      END;(*FOR*)
      buffer[bc] := 0C;
      WriteLine(outstream,buffer);
    END;(*FOR*)
    INC(blknum);
  END;(*LOOP*)
  IF rs <> EndFile THEN
    Check(rs);
  END;
  CloseStream(outstream);
  Check(Close(infile,FALSE));

END Dump.
