MODULE Err;

FROM R2NumbIO IMPORT WriteCard;
FROM SYSTEM IMPORT ADDRESS;
FROM R2SysCalls IMPORT
  Write, ReadString, WriteString, WriteLn, Translate, Status,
  Read, GetCommand, Exit, GetMessage, Result;
FROM R2Strings IMPORT
  Remove, Copy, Length, Insert, Substring, Compare, Compress, Concatenate;
FROM R2Streams IMPORT
  Stream, Done, CloseStream, ResetStream, CreateStream,
  OpenStream, WriteLine, ReadLine, EndStream;

CONST   MaxErrs = 181;
        Hat     = '^';
        Comma   = ',';

TYPE    LineType = ARRAY[0..79] OF CHAR;
        ErrTable = ARRAY[0..MaxErrs] OF LineType;
        Buffer4  = ARRAY[0..3] OF CHAR;
        Buffer   = ARRAY[0..1] OF LineType;

VAR     ErrorTable : ErrTable;
        Stars, buf4   : Buffer4;
        buffer : Buffer;
        Line, infile, outfile, errfile, temp, 
        temp1, temp2, ending, ErrorLine : LineType;
        Stats : Status;
        instream, outstream, errstream : Stream;
        ErrorLinesRead, ErrorNum, Counter, I, BufferCount : CARDINAL;
        
PROCEDURE Check(r : Status);

VAR     str : LineType;

BEGIN   IF r = Ok THEN RETURN END;
        GetMessage(r,str);
        Error(str);
END Check;

PROCEDURE Error(VAR s : ARRAY OF CHAR);

BEGIN   WriteString("Err -- ");
        WriteString(s);
        WriteLn;
        Exit(Status(-1));
END Error;

PROCEDURE Component(VAR s,t : ARRAY OF CHAR);

VAR     a : CARDINAL;

BEGIN   a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] = " ") DO INC(a) END;
        Remove(s,0,a);
        a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] > " ") DO INC(a) END;
        Substring(s,t,0,a);
        IF t[0] = ";" THEN t[0] := 0C; RETURN END;
        Remove(s,0,a);
END Component;

PROCEDURE Find(pat , pat1 : CHAR; VAR s : LineType) : CARDINAL;

VAR     I : CARDINAL;

BEGIN   I := 0;
        WHILE (I <= HIGH(s)) AND (s[I] <> 0C) DO
              IF (s[I] = pat) OR (s[I] = pat1) THEN RETURN I; END;
              INC(I);
        END;
        RETURN I;
END Find;

PROCEDURE FindNext(VAR TempLine : LineType;
                   VAR ErrNum : CARDINAL) : BOOLEAN;

VAR     pos, I : CARDINAL;

BEGIN   ErrNum := 0;
        pos := Find(Hat,Comma,TempLine);
        IF (pos = HIGH(TempLine)) OR (TempLine[pos] = 0C) THEN
              RETURN FALSE 
        END;
        Remove(TempLine,0,pos+1);
        pos := 0;
        WHILE (TempLine[pos]>57C) AND (TempLine[pos]<72C) DO
              ErrNum := ErrNum*10 + ORD(TempLine[pos]) - 48;
              INC(pos);
        END;
        RETURN TRUE;
END FindNext;

PROCEDURE FindError(ErrNum : CARDINAL; VAR ErrorLine : LineType): BOOLEAN;

VAR     Subtract : CARDINAL;

BEGIN   CASE ErrNum OF
                 0..8     : Subtract :=   0;|
                 20..61   : Subtract :=  11;|
                 70..111  : Subtract :=  19;|
                 120..161 : Subtract :=  27;|
                 198..211 : Subtract :=  63;|
                 300..304 : Subtract := 151;|
                 400..405 : Subtract := 246;|
                 923,924  : Subtract := 763;|
                 941..946 : Subtract := 779;|
                 973,974  : Subtract := 805;|
                 981,982  : Subtract := 811;|
                 990..999 : Subtract := 818;
                 ELSE RETURN FALSE;
        END;
        ErrNum := ErrNum - Subtract;
        WHILE ErrorLinesRead <= ErrNum DO
                 ReadLine(errstream,ErrorTable[ErrorLinesRead]);
                 INC(ErrorLinesRead);
        END;
        ErrorLine := ErrorTable[ErrNum];
        RETURN TRUE;
END FindError;

BEGIN GetCommand(Line);
      Component(Line,temp);
      IF temp[0] = 0C THEN
         LOOP WriteString("Error Listing > ");
              ReadString(temp);
              CASE Result() OF
                   EndFile,EscKey : Exit(Status(-1));|
                   Ok             : IF temp[0] <> 0C THEN EXIT END;
                   ELSE
              END;
         END;
      END;
      Compress(temp);
      Substring(temp,ending,HIGH(temp)-4,4);
      IF (Compare(ending,'.LIS') <> 0) THEN
         Concatenate(temp,'.LIS',temp2);
         temp := temp2;
      ELSE Error("Cannot find LIS file");
      END;

      Check(Translate(temp,infile));
      OpenStream(instream,infile,FALSE);
      IF NOT Done THEN
         Error("Cannot open input file");
      END;

      Component(Line,temp1);
      IF temp1[0] = 0C THEN
         Check(Translate('$TI',outfile));
         OpenStream(outstream,outfile,TRUE);
         IF NOT Done THEN
            Error("Cannot open terminal for output");
         END;
      ELSE
         Check(Translate(temp1,outfile));
         OpenStream(outstream,outfile,TRUE);
         IF NOT Done THEN
            CreateStream(outstream,outfile);
            IF NOT Done THEN
               Error("Cannot open output file");
            END;
         END;
      END;

      Check(Translate('$ERRORS',errfile));
      OpenStream(errstream,errfile,FALSE);
      IF NOT Done THEN
         Error("Cannot open ERRS.DAT");
      END;

      ResetStream(errstream);
      ErrorLinesRead := 0;
      ResetStream(instream);      
      ResetStream(outstream);
      FOR I := 0 TO 3 DO Stars[I] := '*'; END;
      BufferCount := 0;
      WHILE NOT EndStream(instream) DO
         ReadLine(instream,buffer[BufferCount MOD 2]);
         Substring(buffer[BufferCount MOD 2],buf4,1,4);
         IF (Compare(buf4,Stars) = 0) THEN
            WriteLine(outstream,buffer[(BufferCount+1) MOD 2]);
            WriteLine(outstream,buffer[BufferCount MOD 2]);
            Compress(buffer[BufferCount MOD 2]);
            Remove(buffer[BufferCount MOD 2],0,4);
            WHILE FindNext(buffer[BufferCount MOD 2],ErrorNum) DO
                IF FindError(ErrorNum,ErrorLine) THEN
                   WriteLine(outstream,ErrorLine);
                ELSE
                   WriteLine(outstream,"Error number not in tables");
                END;
            END;
         END;
         INC(BufferCount);
      END;
      CloseStream(instream);
      CloseStream(outstream);
      CloseStream(errstream);
END Err.
