MODULE FixDisk;

FROM R2SysCalls IMPORT WriteString, WriteLn, Status, Open, WriteBlock,
                       Time, Exit, ReadString, Translate, Dismount,
                       Handle, ReadBlock, Close, Seek, Mount,
                       ModeBits, FileMode, GetMode;
FROM R2NumbIO IMPORT WriteCard;
FROM R2Strings IMPORT Find, Copy, Length, Insert;
FROM R2Storage IMPORT ALLOCATE;

TYPE    LineType = ARRAY [0..77] OF CHAR;
        DirRec = RECORD Name : ARRAY [0..19] OF CHAR;
                        First : CARDINAL;
                        Date : Time;
                        Owner, Mode : CARDINAL;
                 END;
        Home = RECORD Key : ARRAY [0..3] OF CHAR;
                      Label : ARRAY [0..19] OF CHAR;
                      Date : Time;
                      Size, Owner, Mode : CARDINAL;
               END;
        Four = [0..3];
        Block = RECORD CASE Four OF
                 0 : a : ARRAY [0..511] OF CHAR      | 
                 1 : d : ARRAY [0..15] OF DirRec     |
                 2 : f : ARRAY [0..255] OF CARDINAL  |
                 3 : h : Home;
                 END;
                END;               
        MapPtr = POINTER TO MapRec;
        MapRec = RECORD Fat : Block;
                        Use : ARRAY [0..255] OF BOOLEAN;
                 END; 

VAR     Hyphen : ARRAY [0..1] OF CHAR;
        Map : ARRAY [0..199] OF MapPtr;
        Dev, Device : ARRAY [0..9] OF CHAR;
        Dsize, FatCount : CARDINAL;

PROCEDURE  Mark( k : CARDINAL);

BEGIN   Map[k DIV 256]^.Use[k MOD 256] := TRUE;
END Mark;

PROCEDURE  Test( k : CARDINAL) : BOOLEAN;

BEGIN   RETURN Map[k DIV 256]^.Use[k MOD 256];
END Test;

PROCEDURE  MakeMap;

VAR     i, j : CARDINAL;

BEGIN   FOR i := 0 TO 199 DO
          NEW(Map[i]);
          WITH Map[i]^ DO
            FOR j := 0 TO 255 DO Use[j] := FALSE END;
            END;
          END;
END MakeMap;

PROCEDURE  Check( rs : Status);

BEGIN   IF rs <> Ok THEN Exit(rs) END;
END Check;

PROCEDURE  Scan( VAR fn : ARRAY OF CHAR; dFirst : CARDINAL);

VAR     h : Handle;                 
        rs : Status;
        db : Block;
        tim : Time;
        m : FileMode;
        i, j, k, p : CARDINAL;
        Spec, ffn : LineType;

BEGIN   Copy(fn,ffn);
        i := Find(Hyphen,ffn);
        j := Length(ffn);
        IF i = j THEN Insert("-DIRECTORY.SYS",ffn,j) END;
        Check(Open(ffn,FALSE,h));
        Check(GetMode(h,tim,m,i));
        IF NOT (DIR IN m) THEN Exit(BadDirect) END;
        LOOP rs := ReadBlock(h,db.a);
             IF rs <> Ok THEN EXIT END;
             FOR i := 0 TO 15 DO 
               WITH db.d[i] DO
                 k := Length(Name);
                 IF k > 0 THEN
                   p := First;
                   WHILE p <> 65535 DO
                     IF Test(p)
                       THEN WriteString('Bad ');
                            WriteString(Name);
                            WriteString('  ');
                            WriteCard(dFirst,1);
                            WriteString('  ');
                            WriteCard(i,1);
                            WriteLn;
                            p := 65535;
                       ELSE
                            Mark(p);
                            p := Fat(p);
                       END;
                     END;
                   IF (DIR IN FileMode(Mode)) AND 
                      NOT(SYS IN FileMode(Mode)) THEN
                     Copy(Name,Spec);
                     Insert(Hyphen,Spec,0);
                     Insert(fn,Spec,0);
                     Scan(Spec,First);      (* Recursively *)
                     END; 
                   END;
                 END;
               END;
               dFirst := Fat(dFirst);
             END;
        IF (rs <> Ok) AND (rs <> EndFile) THEN Check(rs) END;
        rs := Close(h,FALSE);
END Scan;

PROCEDURE  SetVolParams;

VAR     h : Handle;
        b : Block;
        spec : LineType;

BEGIN   Copy("-HOMEBLOCK.SYS",spec);
        Insert(Device,spec,0);
        Check(Open(spec,FALSE,h));
        Check(ReadBlock(h,b.a));
        Dsize := b.h.Size;
        Check(Close(h,FALSE));
        FatCount := (Dsize+255) DIV 256;
        IF FatCount > 200 THEN Exit(BadVolume) END;
        WriteString('Device size: '); WriteCard(Dsize,1); WriteLn;
        WriteString('Fat blocks: '); WriteCard(FatCount,1); WriteLn;
END SetVolParams;

PROCEDURE  LoadFat;

VAR     i, j, m : CARDINAL;
        h : Handle;    

BEGIN   Check(Dismount(Device));
        Check(Open(Device,FALSE,h));
        Check(Seek(h,3));
        FOR i := 0 TO FatCount-1 DO
          Check(ReadBlock(h,Map[i]^.Fat.a)) END;
        Check(Close(h,FALSE)); 
        Check(Mount(Device,"junk",FALSE));
END LoadFat;

PROCEDURE  Repair;

VAR     i, j, m : CARDINAL;
        h : Handle;    

BEGIN   Check(Dismount(Device));
        Check(Open(Device,TRUE,h));
        Check(Seek(h,3));
        FOR i := 0 TO FatCount-1 DO
          WITH Map[i]^ DO    
            FOR j := 0 TO 255 DO 
              IF NOT Use[j] THEN Fat.f[j] := 0 END;
              END;
            Check(WriteBlock(h,Fat.a,FALSE)) 
            END;
          END;
        Check(Close(h,TRUE)); 
        Check(Mount(Device,"junk",FALSE));
END Repair;

PROCEDURE  Fat( k : CARDINAL) : CARDINAL;

BEGIN   RETURN Map[k DIV 256]^.Fat.f[k MOD 256];
END Fat;
       
VAR     i, k : CARDINAL;
        yorn : ARRAY [0..1] OF CHAR;

BEGIN   Hyphen[0] := "-"; Hyphen[1] := 0C;
        WriteString('Device: ');
        ReadString(Dev);                 
        Check(Translate(Dev,Device));
        IF Find(Hyphen,Device) <> Length(Device) THEN Exit(BadName) END;
        SetVolParams;
        MakeMap;
        WriteString('Loading FAT'); WriteLn;
        LoadFat;   (* Dismounts and remounts Device ! *)
        WriteString('Scanning Directories'); WriteLn;
        Scan(Device,FatCount+3);
        WriteString('Unallocated block count: ');
        k := 0;
        FOR i := 0 TO Dsize-1 DO
          IF NOT Test(i) THEN INC(k) END;
          END;
        WriteCard(k,1);
        WriteLn;
        WriteString("Repair ? ");
        ReadString(yorn);
        IF CAP(yorn[0]) = "Y" THEN Repair END;
END FixDisk.
