MODULE Kermit;  (* Kermit server for LUMOS-2 *)

(* 1989 07 09 - GRG *)

(* This program implements a simple Kermit server as described in
   the Kermit protocol manual by Frank da Cruz.  It supports file
   transfer only and treats all files as binary.  8th bit prefixing
   is done.  No advanced features are supported.  Timeouts are not 
   implemented.
*)

FROM R2SysCalls IMPORT Read, Status, WriteString, Write, WriteLn,
                       Result, Handle, Open, Create, FileMode, Translate,
                       Close, WriteBlock, ReadBlock, CreateLogical;
FROM R2Strings IMPORT Length, Copy, Remove, Concatenate;
FROM R2Directories IMPORT SetWild, NextWild, NName;

CONST   MAXSP = 1000;
        MAXRP = 2000;
        MAXTRY = 5;

TYPE    states = (ssfil, ssdat, sseot, srfil, srdat, sserv, 
                  sdfil, sddat, sdeof);

VAR     (* Send-init parameters *)

        spsiz, rpsiz : INTEGER;
        seol, reol, rctlq, sctlq, ebq : CHAR;

        (* packet vars *)

        seq, size, osize, maxsiz, rln, rsn,
        sndpkl, rdatap : INTEGER;
        smark, rmark : CHAR;
        sndpkt : ARRAY [0..MAXSP+99] OF CHAR;
        rcvpkt : ARRAY [0..MAXRP+199] OF CHAR;
        data : ARRAY [0..MAXRP] OF CHAR;

        remain, parstr, tmpstr, filnam : ARRAY [0..133] OF CHAR;
        usestr, first : BOOLEAN;
        tmpptr : INTEGER;
        state : states;
        xflag : CHAR;

        fbuf : ARRAY [0..511] OF CHAR;
        fp : INTEGER;
        fh : Handle;
        rs : Status;

(* character functions *)

PROCEDURE  tochar( i : INTEGER) : CHAR;

BEGIN   RETURN CHR(i + 32);
END tochar;

PROCEDURE  unchar( ch : CHAR) : INTEGER;

BEGIN   RETURN ORD(ch) -32;
END unchar;

PROCEDURE  ctl( ch : CHAR) : CHAR;     (* toggle the 64 - ctl - bit *)

BEGIN   IF ODD(ORD(ch) DIV 64)
          THEN RETURN CHR(ORD(ch)-64)
          ELSE RETURN CHR(ORD(ch)+64)
          END;
END ctl;
        
(* system interface routines *)

PROCEDURE  ttinl( VAR s : ARRAY OF CHAR; eol : CHAR) : INTEGER;

(* reads from the control line until s is full or eol is matched *)

VAR     i : INTEGER;

BEGIN   i := 0;
        LOOP Read(s[i],TRUE);
             IF s[i] = eol THEN EXIT
               ELSIF i = INTEGER(HIGH(s)) THEN EXIT
               ELSE INC(i)
               END;
             END;
        IF s[i] = eol THEN s[i] := 0C END;
        RETURN i;
END ttinl;

PROCEDURE  ttflui;   (* flush input *)

VAR     ch : CHAR;

BEGIN   REPEAT Read(ch,FALSE)
        UNTIL Result() <> Ok;
END ttflui;

PROCEDURE  zopeni( VAR fn : ARRAY OF CHAR) : BOOLEAN;

BEGIN   rs := Open(fn,FALSE,fh);
        fp := 512;
        RETURN rs = Ok;
END zopeni;

PROCEDURE  zopeno( VAR fn : ARRAY OF CHAR) : BOOLEAN;

BEGIN   rs := Create(fn,FileMode(0F0H),fh);
        fp := 0;
        RETURN rs = Ok;
END zopeno;

PROCEDURE  zclosi;

BEGIN   rs := Close(fh,TRUE);
END zclosi;

PROCEDURE  zcloso;

BEGIN   IF fp <> 0 THEN rs := WriteBlock(fh,fbuf,TRUE) END;
        rs := Close(fh,TRUE);
END zcloso;

PROCEDURE  zgetc( VAR c : CHAR) : BOOLEAN;

BEGIN   IF fp > 511 THEN
          rs := ReadBlock(fh,fbuf);
          IF rs <> Ok THEN RETURN FALSE END;
          fp := 0;
          END;
        c := fbuf[fp];
        INC(fp);
        RETURN TRUE;
END zgetc;

PROCEDURE  zputc( c : CHAR);

BEGIN   fbuf[fp] := c;
        INC(fp);
        IF fp > 511 THEN
          rs := WriteBlock(fh,fbuf,TRUE);
          fp := 0;
          END;
END zputc;

PROCEDURE  zltor( VAR nl, nr : ARRAY OF CHAR);

VAR     i, j : INTEGER;
        ds : BOOLEAN;

BEGIN   ds := FALSE;
        i := 0; j := 0;
        WHILE (i <= INTEGER(HIGH(nl))) AND (nl[i] <> 0C) DO
          IF (nl[i] = '.')
            THEN IF NOT ds THEN
                   nr[j] := '.';
                   INC(j);
                   ds := TRUE;
                   END;
            ELSE nr[j] := CAP(nl[i]);
                 INC(j);
            END;
          INC(i);
          END;
        IF j <= INTEGER(HIGH(nr)) THEN nr[j] := 0C END;
END zltor;

(* transport *)

PROCEDURE  input() : CHAR;

(* read a packet, check sequence number and return type *)

VAR     type : CHAR;
        try : INTEGER;

BEGIN   try := 0;
        LOOP type := rpack();
             IF (rsn = seq) AND (type <> 'Q') AND (type <> 'N')
               THEN ttflui; RETURN type; END;
             IF try > MAXTRY THEN error("Too many retries"); RETURN 'E' END;
             IF (type = 'N') AND (rsn = (seq+1)MOD 64) 
               THEN RETURN 'Y' ELSE resend END;
             INC(try);
             END;
END input;

PROCEDURE  nxtpkt;

BEGIN   seq := (seq + 1) MOD 64;
END nxtpkt;

PROCEDURE  tinit;

BEGIN   seq := 0;
        sndpkt[0] := 0C;
        rcvpkt[0] := 0C;
END tinit;

PROCEDURE  error( VAR s : ARRAY OF CHAR);

BEGIN   spack('E',seq,Length(s),s);
END error;

PROCEDURE  ack;

BEGIN   spack('Y',seq,0,"");
        nxtpkt;
END ack;

PROCEDURE  ack1( VAR s : ARRAY OF CHAR);

BEGIN   spack('Y',seq,Length(s),s);
        nxtpkt;
END ack1;

PROCEDURE  nak;

BEGIN   spack('N',seq,0,"");
END nak;

PROCEDURE  sinit;

BEGIN   rpar;
        spack("S",seq,Length(parstr),parstr);
END sinit;

PROCEDURE  sfile() : BOOLEAN;

VAR     rnam : ARRAY [0..19] OF CHAR;
        x : INTEGER;

BEGIN   IF NOT zopeni(filnam) THEN RETURN FALSE END;
        zltor(filnam,rnam);
        x := encstr(rnam);
        maxsiz := spsiz - 4;
        nxtpkt;
        spack(xflag,seq,x,data);
        RETURN TRUE;
END sfile;

PROCEDURE  sdir;

BEGIN   maxsiz := spsiz - 4;
        nxtpkt;
        spack(xflag,seq,0,"");
END sdir;

PROCEDURE sdata() : BOOLEAN;

VAR     x : INTEGER;

BEGIN   x := getpkt(maxsiz);
        IF x = 0 THEN RETURN FALSE END;
        nxtpkt;
        spack('D',seq,x,data);
        RETURN TRUE;
END sdata;

PROCEDURE sdirent() : BOOLEAN;

VAR     x : INTEGER;
        crlf : ARRAY [0..1] OF CHAR;
        buf : ARRAY [0..19] OF CHAR;

BEGIN   rs := NextWild();
        IF rs <> Ok THEN RETURN FALSE END;
        crlf[0] := 15C; crlf[1] := 12C;
        FOR x := 0 TO 19 DO buf[x] := ' ' END;
        Copy(NName,filnam);
        rs := NextWild();
        IF rs = Ok THEN 
          Concatenate(filnam,buf,filnam);
          Remove(filnam,20,20);
          Concatenate(filnam,NName,filnam);
          rs := NextWild();
          IF rs = Ok THEN
            Concatenate(filnam,buf,filnam);
            Remove(filnam,40,20);
            Concatenate(filnam,NName,filnam);
            END;
          END;
        Concatenate(filnam,crlf,filnam);
        x := encstr(filnam);
        nxtpkt;
        spack("D",seq,x,data);
        RETURN TRUE;
END sdirent;

PROCEDURE  seof;

BEGIN   zclosi;
        nxtpkt;
        spack("Z",seq,0,"");
END seof;

PROCEDURE  seot;

BEGIN   nxtpkt;
        spack("B",seq,0,"");
END seot;

PROCEDURE  gnfile() : BOOLEAN;

BEGIN   rs := NextWild();
        IF rs = Ok
          THEN Copy(NName,filnam);
               RETURN TRUE
          ELSE RETURN FALSE
          END;
END gnfile;

PROCEDURE  rcvfil() : BOOLEAN;

BEGIN   decstr(filnam);
        RETURN zopeno(filnam);
END rcvfil;

(* datalink *)

PROCEDURE  spack( type : CHAR; sqn, len : INTEGER; VAR pkt : ARRAY OF CHAR);

VAR     i : INTEGER;

BEGIN   sndpkt[0] := smark;
        sndpkt[1] := tochar(len+3);
        sndpkt[2] := tochar(sqn);
        sndpkt[3] := type;     
        FOR i := 0 TO len-1 DO sndpkt[4+i] := pkt[i] END;
        sndpkt[4+len] := 0C;
        i := chksum(sndpkt,1);
        sndpkt[len+4] := tochar(i);
        sndpkt[len+5] := seol;
        sndpkt[len+6] := 0C;
        sndpkl := len+7;
        WriteString(sndpkt);
END spack;

PROCEDURE  resend;

BEGIN   IF sndpkt[0] <> 0C THEN WriteString(sndpkt) ELSE nak END;
END resend;

PROCEDURE  chksum( VAR s : ARRAY OF CHAR; off : INTEGER) : INTEGER;

VAR     c, i, t : INTEGER;

BEGIN   i := off; c := 0;
        WHILE s[i] <> 0C DO INC(c,ORD(s[i]) MOD 200B); INC(i) END;
        t := (((c MOD 256) DIV 64) + c ) MOD 64;
        RETURN t;
END chksum;

PROCEDURE  rpack() : CHAR;

VAR     j, i, k : INTEGER;
        type : CHAR;
          
BEGIN   rsn := -1;
        rln := -1;
        rcvpkt[0] := 0C;
        j := ttinl(rcvpkt,reol);
        i := 0;
        WHILE (i < j) AND (rcvpkt[i] <> rmark) DO INC(i) END;
        IF i = j THEN RETURN 'Q' END;
        rln := unchar(rcvpkt[i+1]) - 3;
        IF (rln < 0) OR (rln+i+4 > MAXRP) THEN RETURN 'Q' END;
        rsn := unchar(rcvpkt[i+2]);
        type := rcvpkt[i+3];
        rdatap := i+4;
        j := unchar(rcvpkt[i+4+rln]);
        rcvpkt[i+4+rln] := 0C;
        k := chksum(rcvpkt,i+1);
        IF k <> j THEN RETURN 'Q' ELSE RETURN type END;
END rpack;

(* presentation *)

PROCEDURE  getpkt( max : INTEGER) : INTEGER;

VAR     i : INTEGER;
        c : CHAR;

BEGIN   size := 0;    (* global !! *)   
        WHILE remain[size] <> 0C DO
          data[size] := remain[size];
          INC(size);
          END;
        remain[0] := 0C;
        data[size] := 0C;
        WHILE gnchar(c) DO
          osize := size;
          encode(c);
          IF size = max THEN RETURN size END;
          IF size > max THEN
            i := 0;
            WHILE data[osize+i] <> 0C DO
              remain[i] := data[osize+i];
              INC(i);
              END;
            remain[i] := 0C;
            size := osize;
            data[size] := 0C;
            RETURN size;
            END;
          END;
        RETURN size;   (* for short packets *)
END getpkt;

PROCEDURE  gnchar( VAR c : CHAR) : BOOLEAN;

BEGIN   IF usestr 
          THEN c := tmpstr[tmpptr];
               INC(tmpptr);
               RETURN c <> 0C;
          ELSE RETURN zgetc(c);
          END;
END gnchar;

PROCEDURE  encstr( VAR s : ARRAY OF CHAR) : INTEGER;

VAR     x : INTEGER;

BEGIN   Copy(s,tmpstr);
        tmpptr := 0;
        usestr := TRUE;
        x := getpkt(spsiz);
        usestr := FALSE;
        RETURN size;
END encstr;

PROCEDURE  decstr( VAR s : ARRAY OF CHAR);

BEGIN   usestr := TRUE;
        tmpptr := 0;
        decode;
        usestr := FALSE;
        tmpstr[tmpptr] := 0C;
        Copy(tmpstr,s);
END decstr;

PROCEDURE  pnchar( c : CHAR);

BEGIN   IF usestr
          THEN tmpstr[tmpptr] := c;
               INC(tmpptr);
          ELSE zputc(c)
          END;
END pnchar;

PROCEDURE  encode( a : CHAR);

VAR     b : INTEGER;

BEGIN   b := ORD(a);        
        IF b >= 128 THEN 
          data[size] := ebq;
          INC(size);
          b := b MOD 128;
          a := CHR(b);
          END;
        IF (b < 32) OR (b = 127) THEN
          data[size] := sctlq;
          INC(size);
          a := ctl(a);
        ELSIF (a = sctlq) OR (a = ebq) THEN
          data[size] := sctlq;
          INC(size);
          END;
        data[size] := a;
        INC(size);
        data[size] := 0C;
END encode;

PROCEDURE  decode;  (* from rcvpkt @ rdatap via pnchar *)

VAR     a : CHAR;
        b, b8 : INTEGER;

BEGIN   WHILE rcvpkt[rdatap] <> 0C DO
          a := rcvpkt[rdatap];  INC(rdatap);
          b8 := 0;
          IF a = ebq THEN b8 := 128; a := rcvpkt[rdatap]; INC(rdatap); END;
          IF a = rctlq THEN 
            a := rcvpkt[rdatap]; INC(rdatap);
            b := ORD(a) MOD 128;
            IF (b > 62) AND (b < 96) THEN a := ctl(a) END;
            END;
          b := ORD(a) MOD 128;
          pnchar(CHR(b+b8));
          END;
END decode;

(* negotiations *)

PROCEDURE  spar;   (* get parameters from rcvpkt at rdatap *)

BEGIN   IF rln >= 1 
          THEN spsiz := unchar(rcvpkt[rdatap+0])
          ELSE spsiz := 80
          END;
        IF rln >= 5 THEN seol := CHR(unchar(rcvpkt[rdatap+4])) END;
        IF (seol < 2C) OR (seol > 37C) THEN seol := 15C END;
        IF rln >= 6 THEN rctlq := rcvpkt[rdatap+5] END;
        IF (rctlq <= 40C) OR ((rctlq >= '@') AND (rctlq <= '^'))
           OR (rctlq = 177C) THEN rctlq := '#' END;
END spar;

PROCEDURE  rpar;

BEGIN   parstr[0] := tochar(rpsiz);
        parstr[1] := tochar(15);    (* timeout requested *)
        parstr[2] := tochar(0);
        parstr[3] := ctl(0C);
        parstr[4] := tochar(ORD(reol));
        parstr[5] := '#';
        parstr[6] := '&';
        parstr[7] := 0C;
END rpar;
        
(* protocol state machine *)        

PROCEDURE  restart;

BEGIN   state := sserv;
        tinit;
        xflag := 'F';
END restart;

PROCEDURE  err( s : ARRAY OF CHAR);

BEGIN   error(s);
        restart;
END err;

PROCEDURE  generic;  (* server commands *)

VAR     c : CHAR;
        i : INTEGER;

BEGIN   xflag := 'X';
        c := parstr[0];
        Remove(parstr,0,2);
        CASE c OF
          'C' : IF parstr[0] = 0C THEN Copy("$HOME",parstr) END;
                rs := Translate(parstr,filnam);
                i := Length(filnam);
                IF (i > 0) AND (filnam[i-1] = '-') THEN filnam[i-1] := 0C END;
                rs := CreateLogical("CURRENT",filnam);
                rs := Translate("$CURRENT",filnam);
                i := encstr(filnam);
                ack1(data); restart;
        | 'D' : IF parstr[0] = 0C THEN Copy("$CURRENT-*",parstr) END;
                rs := SetWild(parstr);
                IF rs <> Ok THEN err("Cannot open directory") END;
                sinit; first := TRUE;
                state := sdfil;
        | 'T' : rs := SetWild(parstr);
                sinit; first := TRUE;
                state := ssfil;
        ELSE err("Unimplemented generic server command");
        END;
END generic;

PROCEDURE  protocol;

VAR     next : CHAR;

BEGIN   restart;
        LOOP next := input();
             CASE state OF
               sserv : CASE next OF
                         'I' : spar; rpar; ack1(parstr); tinit;
                       | 'R' : decstr(filnam);
                               rs := SetWild(filnam);            
                               IF rs <> Ok THEN 
                                 err("Cannot access directory") END;
                               sinit; first := TRUE;
                               state := ssfil;
                       | 'S' : spar; rpar; ack1(parstr); state := srfil;
                       | 'G' : decstr(parstr);
                               IF (parstr[0] = 'F') OR (parstr[0] = 'L')
                                 THEN ack; EXIT
                                 ELSE generic
                                 END;
                       | 'E' : restart;
                       ELSE err('Unknown server command');
                       END;
             | ssfil : CASE next OF
                         'Y' : IF first THEN spar; first := FALSE END;
                               IF gnfile( )
                                 THEN IF sfile()
                                        THEN state := ssdat
                                        ELSE err("Cannot access file");
                                        END;
                                 ELSE seot; state := sseot;
                                 END;
                       | 'E' : restart;
                       ELSE err('Unexpected packet type');
                       END;
             | ssdat : CASE next OF
                         'Y' : IF NOT sdata() THEN seof; state := ssfil END;
                       | 'E' : restart;
                       ELSE err('Unexpected packet type');
                       END;
             | sseot : CASE next OF
                         'Y' : restart;
                       | 'E' : restart;
                       ELSE err('Unexpected packet type');
                       END;
             | srfil : CASE next OF
                         'B' : ack; restart;
                       | 'F' : IF rcvfil() 
                                 THEN ack; state := srdat;
                                 ELSE err("Cannot create file");
                                 END;
                       | 'E' : restart;
                       ELSE err('Unexpected packet type');
                       END;
             | srdat : CASE next OF
                         'D' : decode; ack;
                       | 'Z' : zcloso; ack; state := srfil;
                       | 'E' : restart;
                       ELSE err('Unexpected packet type');
                       END;
             | sdfil : CASE next OF
                         'Y' : spar; sdir; state := sddat;
                       | 'E' : restart;
                       ELSE err('Unexpected packet type');
                       END;
             | sddat : CASE next OF
                         'Y' : IF NOT sdirent()
                                 THEN seof; state := sdeof END;
                       | 'E' : restart;
                       ELSE err('Unexpected packet type');
                       END;
             | sdeof : CASE next OF
                         'Y' : seot; state := sseot;
                       | 'E' : restart;
                       ELSE err('Unexpected packet type');
                       END;
             ELSE err('Invalid state');
             END;
        END;
END protocol;

BEGIN   spsiz := 90;  rpsiz := 90;
        seol := 15C;  reol := 15C;
        sctlq := '#'; rctlq := '#';
        smark := 1C;  rmark := 1C;
        seq := 0; ebq := '&'; 
        usestr := FALSE; 
        remain[0] := 0C;
        xflag := 'F';

        WriteString("LUMOS-2 Kermit Server V1.0"); WriteLn;

        protocol;

END Kermit.
