MODULE Linda;  (* A Line editor for LUMOS-2 *)

FROM LineBuffer IMPORT  InsertLine, DeleteLine, ReplaceLine, Line,
                        Position, Top, Bottom, Up, Down, Number, EOB,
                        Cur, New, Count, Max;
FROM R2SysCalls IMPORT  Status, GetCommand, Result, ReadString, WriteString,
                        WriteLn, Translate;
FROM R2Strings IMPORT   Find, Insert, Length, Remove;
FROM R2Streams IMPORT   Stream, Done, CreateStream, OpenStream, CloseStream,
                        EndStream, ReadLine, WriteLine;
FROM R2NumbIO  IMPORT   WriteCard;
FROM Match     IMPORT   PatternMatch;
FROM Change IMPORT Edit;

VAR     FileName, null, Command, Save, Pattern : Line;
        cp : CARDINAL;

PROCEDURE ScanCard( VAR n : CARDINAL);

BEGIN   n := 0;
        WHILE Command[cp] = ' ' DO INC(cp) END;
        WHILE (Command[cp] >= '0') AND (Command[cp] <= '9') DO
          (*$T-*) n := n*10 + ORD(Command[cp]) - 48; (*$T=*)
          INC(cp);
          END;
END ScanCard;

PROCEDURE ScanString( VAR s : Line) : CARDINAL;

(* 0 = nothing, 1 = ok, 2 = error *)

VAR     b : CHAR;
        i : CARDINAL;

BEGIN   WHILE Command[cp] = ' ' DO INC(cp) END;
        b := Command[cp];
        s[0] := 0C;
        IF b = 0C THEN RETURN 0 END;
        i := 0;
        INC(cp);
        WHILE (Command[cp] <> b) AND (Command[cp] <> 0C) DO
          s[i] := Command[cp];
          INC(i);
          INC(cp);
          END; 
        IF Command[cp] = b THEN s[i] := 0C; INC(cp); RETURN 1 END;
        IF Command[cp] = 0C THEN RETURN 2 END;
END ScanString;

PROCEDURE Still() : BOOLEAN;

BEGIN   WHILE Command[cp] = ' ' DO INC(cp) END;
        RETURN Command[cp] <> 0C;
END Still;

PROCEDURE  Err( VAR s : ARRAY OF CHAR);

BEGIN   WriteString("?? ");
        WriteString(s);
        WriteLn;
END Err;
         
PROCEDURE  ReadIn( VAR fn : Line; b, e : CARDINAL; VAR pr : Line);

VAR     s : Stream;
        i, lc : CARDINAL;
        tn : Line;
        rs : Status;

BEGIN   OpenStream(s,fn,FALSE);
        IF NOT Done THEN Err("Cannot open input file"); RETURN; END;
        i := 0;
        lc := 0;
        LOOP IF EndStream(s) THEN EXIT END;
             ReadLine(s,New);
             INC(i);
             IF (i >= b) AND ((e = 0) OR (i <= e)) THEN
               IF pr[0] <> 0C THEN Insert(pr,New,0) END;
               InsertLine;
               INC(lc);
               END;
             IF (e <> 0) AND (i = e) THEN EXIT END;
             END;
        CloseStream(s);
        rs := Translate(fn,tn);
        WriteString(tn);
        WriteString("  ");
        WriteCard(lc,1);
        WriteString(" lines");
        WriteLn;
END ReadIn;

PROCEDURE  WriteOut( VAR fn : Line; n : CARDINAL) : BOOLEAN;

VAR     s : Stream;
        i, lc : CARDINAL;
        rs : Status;
        tn : Line;

BEGIN   CreateStream(s,fn);
        IF NOT Done THEN Err("Cannot create output file"); RETURN FALSE; END;
        i := 1;
        lc := 0;
        LOOP IF EOB THEN EXIT END;
             WriteLine(s,Cur);
             INC(lc);
             Down;
             IF i = n THEN EXIT END;
             INC(i);
             END;
        rs := Translate(fn,tn);
        WriteString(tn);
        WriteString('  ');
        WriteCard(lc,1);
        WriteString(' lines');
        WriteLn;
        CloseStream(s);
        RETURN TRUE;
END WriteOut;

PROCEDURE  Print;

BEGIN   IF EOB 
          THEN WriteString("[EOB]")
          ELSE WriteString(Cur);
          END;
        WriteLn;
END Print;             

PROCEDURE  SynErr;

BEGIN   Err("Syntax error");
END SynErr;

PROCEDURE  eAppend;

VAR     t : Line;
        n : CARDINAL;

BEGIN   IF ScanString(t) <> 1 THEN SynErr; RETURN END;
        ScanCard(n);
        IF Still() THEN SynErr; RETURN  END;
        IF n = 0 THEN n := 1 END;
        WHILE (n > 0) AND NOT EOB DO
          New := t;
          Insert(Cur,New,0);
          ReplaceLine;
          Print;
          Down; 
          DEC(n);
          END;
END eAppend;
             
PROCEDURE  ePrefix;

VAR     t : Line;
        n : CARDINAL;

BEGIN   IF ScanString(t) <> 1 THEN SynErr; RETURN END;
        ScanCard(n);
        IF Still() THEN SynErr; RETURN  END;
        IF n = 0 THEN n := 1 END;
        WHILE (n > 0) AND NOT EOB DO
          New := Cur;
          Insert(t,New,0);
          ReplaceLine;           
          Print;
          Down;
          DEC(n);
          END;
END ePrefix;
                 
PROCEDURE eBack;

VAR     n, i : CARDINAL;

BEGIN   ScanCard(n);
        IF Still() THEN SynErr; RETURN END;
        IF n = 0 THEN n := 1 END;
        IF n >= Number THEN n := Number - 1 END;
        Position(Number-n);
END eBack;

PROCEDURE eNext;

VAR     n, i : CARDINAL;

BEGIN   ScanCard(n);
        IF Still() THEN SynErr; RETURN END;
        IF n = 0 THEN n := 1 END;
        Position(Number+n);
END eNext;

PROCEDURE eHelp;   

VAR     h : Stream;
        l : Line;

BEGIN   OpenStream(h,"$LinHelp",FALSE);
        IF NOT Done THEN Err("Help is not available"); RETURN END;
        WHILE NOT EndStream(h) DO
          ReadLine(h,l);
          WriteString(l);
          WriteLn;
          END;
        CloseStream(h);
END eHelp;

PROCEDURE eWrite;

VAR     f : Line;
        n : CARDINAL;
        b : BOOLEAN;

BEGIN   IF ScanString(f) <> 1 THEN SynErr; RETURN END;
        ScanCard(n);
        IF Still() THEN SynErr; RETURN END;
        b := WriteOut(f,n);
END eWrite;

PROCEDURE eType;

VAR     n, p : CARDINAL;

BEGIN   ScanCard(n);
        IF Still() THEN SynErr; RETURN END;
        IF n = 0 THEN n := 20 END;
        p := Number;
        WHILE (n > 0) AND NOT EOB DO
          Print;
          Down;
          DEC(n);
          END;
        WriteLn;
        Position(p);
END eType;

PROCEDURE eSubs( glob : BOOLEAN);
     
VAR     p, s : Line;
        n, c, i, k : CARDINAL;

BEGIN   IF ScanString(p) <> 1 THEN SynErr; RETURN END;
        IF ScanString(s) <> 1 THEN SynErr; RETURN END;
        ScanCard(n);
        IF Still() THEN SynErr; RETURN END;
        IF n = 0 THEN n := 1 END;
        k := Length(p);
        WHILE (n > 0) AND NOT EOB DO
          New := Cur;                  
          c := 0;
          LOOP i := Find(p,New);
               IF i < Length(New)
                 THEN Remove(New,i,k);
                      Insert(s,New,i);
                      INC(c);
                 ELSE EXIT
                 END;    
               IF NOT glob THEN EXIT END;
               IF c > Max THEN Err("Recursive substitution"); RETURN END;
               END;
          IF c > 0 THEN 
            ReplaceLine;
            Print;
            END;
          Down;
          DEC(n);
          END;
END eSubs;

PROCEDURE  eChange;

BEGIN   IF EOB THEN Err("Nothing to change"); RETURN END;
        IF Still() THEN SynErr; RETURN END;
        New := Cur;
        Edit(New);
        ReplaceLine;
        Down;
END eChange;

PROCEDURE eReplace;

BEGIN   IF EOB THEN Err("Nothing to replace"); RETURN END;
        IF ScanString(New) <> 1 THEN SynErr; RETURN END;
        IF Still() THEN SynErr; RETURN END;
        ReplaceLine;
END eReplace;

PROCEDURE  eLoad;

VAR     f, p : Line;
        b, e : CARDINAL;

BEGIN   IF ScanString(f) <> 1 THEN SynErr; RETURN END;
        ScanCard(b);
        ScanCard(e);
        IF ScanString(p) = 2 THEN SynErr; RETURN END;
        IF Still() THEN SynErr; RETURN END;
        ReadIn(f,b,e,p);
END eLoad;                                

PROCEDURE eJump;

VAR     n : CARDINAL;

BEGIN   ScanCard(n);
        IF Still() THEN SynErr; RETURN END;
        IF n = 0 THEN n := 1 END;
        IF n > Count THEN n := Count + 1 END;
        Position(n);
END eJump;

PROCEDURE eFind;
       
VAR     p : Line;
        r : CARDINAL;

BEGIN   r := ScanString(p);
        IF r = 2 THEN SynErr; RETURN END;
        IF r = 1 THEN Pattern := p ELSE p := Pattern END;
        IF Still() THEN SynErr; RETURN END;
        Down;
        LOOP IF EOB THEN EXIT END;
             IF Find(p,Cur) < Length(Cur) THEN EXIT END;
             Down;
             END;
END eFind;

PROCEDURE eMatch;
       
VAR     p : Line;

BEGIN   IF ScanString(p) <> 1 THEN SynErr; RETURN END;
        IF Still() THEN SynErr; RETURN END;
        LOOP IF EOB THEN EXIT END;
             IF PatternMatch(p,Cur,FALSE) THEN EXIT END;
             Down;
             END;
END eMatch;

PROCEDURE eDelete;

VAR     n : CARDINAL;

BEGIN   ScanCard(n);
        IF Still() THEN SynErr; RETURN END;
        IF n = 0 THEN n := 1 END;
        WHILE (n > 0) AND NOT EOB DO
          Save := Cur;
          DeleteLine;
          DEC(n);
          END;
END eDelete;

PROCEDURE  eUndo;

BEGIN   New := Save;
        InsertLine;
END eUndo;

PROCEDURE  eInsert;

VAR     t : Line;
        e : CARDINAL;

BEGIN   e := ScanString(t);
        IF e = 2 THEN SynErr; RETURN END;
        IF e = 1 THEN New := t; InsertLine; RETURN END;
        LOOP ReadString(New);
             IF Result() <> Ok THEN EXIT END;
             InsertLine;
             END;
END eInsert;
            
PROCEDURE  CommandLoop;

VAR     com : CHAR;

BEGIN   LOOP WriteCard(Number,1);
             WriteString(": ");
             ReadString(Command);
             cp := 1;                 
             com := CAP(Command[0]);
             IF Result() = EscKey THEN com := 33C; WriteLn END;
             IF Result() = EndFile THEN com := 'X' END;
             IF (com >= '0') AND (com <= '9') THEN cp := 0; com := 'J' END;
             CASE com OF
                 "A" : eAppend;
              |  "B" : eBack;
              |  "C" : eChange;
              |  "D" : eDelete;
              |  "E" : Bottom;
              |  "F" : eFind;
              |  "G" : eSubs(TRUE);
              |  "H" : eHelp;
              |  "I" : eInsert;
              |  "J" : eJump;
              |  "L" : eLoad;
              |  "M" : eMatch;
              |  "N" : eNext;
              |  "P" : ePrefix;
              |  "Q" : EXIT;
              |  "R" : eReplace;
              |  "S" : eSubs(FALSE);
              |  "T" : Top;
              |  "U" : eUndo;
              |  "V" : eType;
              |  "W" : eWrite;
              |  "X" : Top;
                       IF WriteOut(FileName,Count) THEN EXIT END;
              |  0C  : Down;
              |  33C : Up;
              |  "?" : eHelp;
              |  "+" : eNext;
              |  "-" : eBack;
                 ELSE  Err("Unknown command");
                 END;
             Print;
             END;
END CommandLoop;

BEGIN   GetCommand(FileName);
        IF FileName[0] <> 0C THEN 
          null[0] := 0C;
          ReadIn(FileName,0,0,null);
          END;
        Position(1);
        Pattern[0] := 0C;
        Print;
        CommandLoop;
END Linda.
