IMPLEMENTATION MODULE LineBuffer;  (* For the Linda editor *)

FROM R2SysCalls IMPORT Write;
FROM R2Storage IMPORT ALLOCATE, DEALLOCATE;

TYPE    LinePtr = POINTER TO LineRec;
        LineRec = RECORD Next, Prev : LinePtr;
                         Len : CARDINAL;
                         Text : Line;    (* must be last, underallocated *)
                  END;

VAR     Head, Tail, Here : LinePtr;

PROCEDURE  Top;

BEGIN   Here := Head;
        Number := 1;
        SetAll;
END Top;

PROCEDURE  Bottom;

BEGIN   Here := NIL;
        Number := Count+1;
        SetAll;
END Bottom;

PROCEDURE  Up;

BEGIN   GoUp;
        SetAll;
END Up;

PROCEDURE  Down;

BEGIN   GoDown;
        SetAll;
END Down;

PROCEDURE  Position( n : CARDINAL);
        
BEGIN   IF n = 0 THEN n := 1 END;
        IF n > Count THEN n := Count+1 END;
        WHILE n <> Number DO
          IF n < Number THEN GoUp ELSE GoDown END;
          END;
        SetAll;
END Position;

PROCEDURE  InsertLine;

VAR     p, q : LinePtr; 
        c : CARDINAL;

BEGIN   c := 0;
        New[Max] := 0C;
        WHILE New[c] <> 0C DO INC(c) END;
        INC(c,11);
        ALLOCATE(p,c);
        WITH p^ DO
          Len := c;
          FOR c := 0 TO Len-11 DO Text[c] := New[c] END;
          IF Here = NIL
            THEN q := Tail;
                 Tail := p;
            ELSE q := Here^.Prev;
                 Here^.Prev := p;
            END;
          IF q = NIL 
            THEN Head := p;
            ELSE q^.Next := p
            END;
          Next := Here;
          Prev := q;
          END;
        INC(Count);
        INC(Number);
END InsertLine;

PROCEDURE  ReplaceLine;

VAR     p : LinePtr;
        c : CARDINAL;

BEGIN   IF Here = NIL THEN RETURN END;
        c := 0;
        New[Max] := 0C;
        WHILE New[c] <> 0C DO INC(c) END;
        INC(c,11);
        ALLOCATE(p,c);
        p^.Next := Here^.Next;
        p^.Prev := Here^.Prev;
        IF Here^.Prev = NIL 
          THEN Head := p 
          ELSE Here^.Prev^.Next := p
          END;
        IF Here^.Next = NIL
          THEN Tail := p 
          ELSE Here^.Next^.Prev := p
          END;
        p^.Len := c;
        FOR c := 0 TO p^.Len-11 DO p^.Text[c] := New[c] END;
        c := Here^.Len;
        DEALLOCATE(Here,c);
        Here := p;
        Cur := New;
END ReplaceLine;

PROCEDURE  DeleteLine;

VAR     p, q : LinePtr;
        c : CARDINAL;

BEGIN   IF Here <> NIL THEN
          p := Here;
          q := Here^.Prev;
          Here := Here^.Next;
          IF q = NIL 
            THEN Head := Here
            ELSE q^.Next := Here
            END;
          IF Here = NIL
            THEN Tail := q
            ELSE Here^.Prev := q;
            END;
          c := p^.Len;
          DEALLOCATE(p,c);
          DEC(Count);
          SetAll;
          END;
END DeleteLine;

PROCEDURE  SetAll; 

VAR     i : CARDINAL;

BEGIN   EOB := Here = NIL;
        IF EOB 
          THEN Cur[0] := 0C
          ELSE WITH Here^ DO
                FOR i := 0 TO Len-11 DO 
                  Cur[i] := Text[i]
                  END
                END
          END
END SetAll;

PROCEDURE  GoUp;

BEGIN   IF Number = 1 THEN RETURN END;
        IF Here = NIL
          THEN Here := Tail;
               IF Tail <> NIL THEN DEC(Number) END;
          ELSE Here := Here^.Prev;
               DEC(Number);
          END;
END GoUp;

PROCEDURE  GoDown;

BEGIN   IF Here <> NIL THEN
          INC(Number);
          Here := Here^.Next;
          END;
END GoDown;
                    

BEGIN   Head := NIL;
        Tail := NIL;
        Here := NIL;
        EOB := TRUE;
        Count := 0;
        Number := 1;
        Cur[0] := 0C;
END LineBuffer.
