MODULE LOGIN;  (* The Basic LUMOS command processor *)

FROM R2SysCalls IMPORT WriteString, WriteLn, ReadString, DeleteGlobal, Write,
                       CreateGlobal, DeleteLogical, CreateLogical, ReBoot,
                       Translate, GetTime, Time, GetMessage, Status, Open,
                       FastSize, Handle, ReadBlock, Close, ModeBits, FileMode,
                       GetMode, Delete, Rename, WriteBlock, Create, FreeSpace,
                       Mount, Dismount, Process, GetPID, GetUser, Execute,
                       SetUser, Read, Result, DevAttrSet, GetDevAttr, Seek,
                       TestCC, GetCommand, DevAttr, SetMode, FormatTrack, 
                       SetTime, Kill, Exit, Sleep;
FROM R2Streams IMPORT Stream, Done, CreateStream, OpenStream, CloseStream,
                      ReleaseStream, EndStream, ReadLine, WriteLine, WriteChar,
                      SysErr;
FROM R2Storage IMPORT ALLOCATE, DEALLOCATE;
FROM R2Format IMPORT ScanCard, FormatCard, FormatAddr;
FROM R2TimeSubs IMPORT ScanTime, FormatTime;
FROM R2NumbIO IMPORT WriteCard, WriteAddr, WriteTime;
FROM R2Strings IMPORT Find, Remove, Copy, Length, Substring, Insert, Compare;
FROM R2Directories IMPORT Path, NName, NFirst, NMode, NOwner, NDate,
                          SetWild, NextWild, FullName, TestWild;
FROM R2NewCalls IMPORT SetName, GetName, SetFileDate;
FROM LockSync IMPORT LockWait, LockClear;
FROM OneWay IMPORT Cipher;
FROM SYSTEM IMPORT ADDRESS, ADR;
FROM R2NewerCalls IMPORT HangUp;

(* ----  Utilities section               ---- *)

TYPE    ComName = ARRAY [0..15] OF CHAR;
        LineType = ARRAY [0..77] OF CHAR;
        DirRec = RECORD Name : ARRAY [0..19] OF CHAR;
                        First : CARDINAL;
                        Date : Time;
                        Owner, Mode : CARDINAL;
                 END;
        Home = RECORD Key : ARRAY [0..3] OF CHAR;
                      Label : ARRAY [0..19] OF CHAR;
                      Date : Time;
                      Size, Owner, Mode : CARDINAL;
               END;
        Four = [0..3];
        Block = RECORD CASE Four OF
                 0 : a : ARRAY [0..511] OF CHAR      | 
                 1 : d : ARRAY [0..15] OF DirRec     |
                 2 : f : ARRAY [0..255] OF CARDINAL  |
                 3 : h : Home;
                 END;
                END;

VAR     Hyphen : ARRAY [0..1] OF CHAR;
        ThisProc : Process;
        ThisUser : CARDINAL;
        ThisUserName : LineType;

PROCEDURE  Spawn0;

BEGIN   Spawn(0);
END Spawn0;

PROCEDURE  Spawn1;

BEGIN   Spawn(1);
END Spawn1;

PROCEDURE Spawn2;

BEGIN   Spawn(2);
END Spawn2;

PROCEDURE  Spawn( mode : CARDINAL);

(* mode :  0 - subprocess, wait             lodfile / util
           1 - subprocess, not waited for   RUN
           2 - detached process             DETACH  *)

VAR     im, imag, home : LineType;
        u, ti : CARDINAL;
        res, rs : Status;
        pc : ADDRESS;
        p : Process;

BEGIN   IF (mode = 2) AND NOT PrivFlag THEN
          Err("Privilege violation");
          RETURN;
          END;
        IF mode > 0
          THEN Component(Line,im);
               IF im[0] = 0C THEN Err("Image name required"); RETURN END;
          ELSE Copy(Command,im);
          END;
        u := 0;  (* remove leading blanks on command *)
        WHILE (u <= HIGH(Line)) AND (Line[u] = " ") DO INC(u) END;
        Remove(Line,0,u);
        Suffix(".LOD",im);
        rs := Translate(im,imag);
        rs := Translate("$TI",home);
        IF mode = 2
          THEN ti := 65535
          ELSE ti := ORD(home[2]) - 48;  (* kludge *)
               IF home[3] <> 0C THEN ti := ti*10 + ORD(home[3]) - 48; END; 
          END;
        p := ThisProc;
        rs := Translate("$CURRENT",home);
        rs := Execute(imag,Line,home,ti,ThisUser,BreakFlag,mode=0,
                      PrivSpawnFlag,p,res,pc);
        IF rs = NoFile THEN
          Insert("$LODLIB-",im,0);
          rs := Translate(im,imag);
          rs := Execute(imag,Line,home,ti,ThisUser,BreakFlag,mode=0,
                         PrivSpawnFlag,p,res,pc);
          END;
        IF rs <> Ok THEN Err("Cannot execute"); END;
        IF (rs = Ok) AND (mode <> 0) THEN
          WriteString("Process ");
          WriteCard(p,1);
          WriteString(" started");
          IF mode = 2 THEN WriteString(" (detached)") END;
          WriteLn;
          END;
        IF (rs = Ok) AND (res <> Ok) AND (mode = 0) THEN
          IF res <> Status(-1) THEN
            GetMessage(res,im);
            WriteLn;
            WriteString('CLI -- Exit status: "');
            WriteString(im); Write('"');
            IF pc <> 0 THEN 
              WriteString(' at PC = '); 
              WriteAddr(pc,16,1);
              END;
            WriteLn;
            END;
          IF FatalFlag THEN ReDirect END;
          END;
END Spawn;

PROCEDURE  WideFiles;  (* wide directory listings to terminal only *)

VAR     s : LineType;
        c, n, k : CARDINAL;
        rs : Status;
       
BEGIN   Component(Line,s);
        rs := SetWild(s);
        IF rs <> Ok THEN ShowStatus(rs); RETURN END;
        s[0] := 0C;
        c := 0; n := 0;
        LOOP rs := NextWild();
             IF rs <> Ok THEN EXIT END;
             IF TestCC() THEN
               rs := Break;
               EXIT;
               END;
             INC(n);
             k := Length(NName);
             IF c+k+2 > 78 THEN
               WriteString(s);
               WriteLn;
               s[0] := 0C;
               c := 0;
               END;
             IF c > 0 THEN 
               Insert("  ",s,c);
               INC(c,2);
               END;
             Insert(NName,s,c);
             INC(c,k);
             END;
        IF c > 0 THEN 
          WriteString(s);
          WriteLn;
          END;
        IF rs <> EndFile THEN ShowStatus(rs); RETURN END;
        IF n = 0 THEN 
          WriteString("Files -- No files found");
          WriteLn;
          END;
END WideFiles;

PROCEDURE  DirectoryT;

BEGIN   Directory(TRUE);
END DirectoryT;

PROCEDURE  DirectoryF;

BEGIN   Directory(FALSE);
END DirectoryF;

PROCEDURE  Directory( full : BOOLEAN);

CONST   b22 = "                      ";

VAR     rs : Status;
        fs, tot, fc : CARDINAL;
        spec, out, temp : LineType;
        os : Stream;

BEGIN   Component(Line,spec);
        rs := SetWild(spec);
        IF rs <> Ok THEN ShowStatus(rs); RETURN END;
        Component(Line,out);
        IF out[0] = 0C THEN Copy("$TI",out) END;
        CreateStream(os,out);
        IF NOT Done THEN
          Err("Cannot create output file"); 
          RETURN;
          END;
        IF full THEN 
          WriteLine(os,"");
          Copy(Path,out);
          Insert('Directory of ',out,0);
          WriteLine(os,out);
          WriteLine(os,"");
          END;
        tot := 0; fc := 0;
        LOOP rs := NextWild();
             IF rs <> Ok THEN EXIT END;
             IF TestCC() THEN
               rs := Break;
               EXIT;
               END;
             Copy(NName,out);
             IF full THEN
               Insert(b22,out,Length(out));
               Remove(out,22,22);
               rs := FastSize(Path,NFirst,fs);
               IF rs <> Ok THEN EXIT END;
               FormatCard(fs,10,6,temp);
               Insert(temp,out,22);
               Insert("   ",out,28);
               INC(tot,fs);
               INC(fc);
               FormatTime(NDate,temp);
               Insert(temp,out,31);
               FormatCard(NOwner,10,6,temp);
               Insert(temp,out,Length(out));
               FormatCard(NMode,16,5,temp);
               Insert(temp,out,Length(out));
               END;
             WriteLine(os,out);
             END;
        IF full THEN
          WriteLine(os,"");
          out := "Total of  blocks in  files";
          FormatCard(fc,10,1,temp);
          Insert(temp,out,20);
          FormatCard(tot,10,1,temp);
          Insert(temp,out,9);
          WriteLine(os,out); 
          WriteLine(os,"");
          END;
        CloseStream(os);
        IF (rs <> Ok) AND (rs <> EndFile) THEN ShowStatus(rs) END;
END Directory;

PROCEDURE  SetLogicalF;

BEGIN   SetLogical(FALSE);
END SetLogicalF;

PROCEDURE  SetLogicalT;

BEGIN   SetLogical(TRUE);
END SetLogicalT;

PROCEDURE  SetLogical( glob : BOOLEAN);

VAR     log, tran : LineType;
        rs : Status;

BEGIN   Component(Line,log);
        IF glob AND NOT PrivFlag THEN Err("Privilege violation"); RETURN END;
        IF log[0] = '$' THEN Remove(log,0,1) END;
        IF log[0] = 0C THEN Err('Logical name required'); RETURN END;
        Component(Line,tran);
        IF tran[0] = 0C 
          THEN IF glob 
                 THEN rs := DeleteGlobal(log);
                 ELSE rs := DeleteLogical(log);
                 END
          ELSE IF glob
                 THEN rs := CreateGlobal(log,tran);
                 ELSE rs := CreateLogical(log,tran);
                 END;
          END;
        ShowStatus(rs);
END SetLogical;

PROCEDURE  SetCurrent;

VAR     tran, curr : LineType;
        rs : Status;
        i, h : CARDINAL;

BEGIN   Component(Line,tran);
        IF tran[0] <> 0C THEN
          rs := Translate(tran,curr);
          h := Length(curr);
          IF (h > 0) AND (curr[h-1] = '-') THEN curr[h-1] := 0C END;
          IF rs <> Ok 
            THEN ShowStatus(rs);
            ELSE ShowStatus(CreateLogical("CURRENT",curr));
            END;
          END;
        rs := Translate("$CURRENT",tran);
        IF rs <> Ok 
          THEN ShowStatus(rs)
          ELSE WriteString("Current directory is ");
               WriteString(tran);
               WriteLn;
          END;
END SetCurrent;

PROCEDURE  ShowLogical;

VAR     log, tran : LineType;
        rs : Status;

BEGIN   Component(Line,log);
        IF log[0] = 0C THEN log := "$CURRENT" END;
        rs := Translate(log,tran);
        IF rs = Ok 
          THEN WriteString(log);
               WriteString(" translates to ");
               WriteString(tran); 
               WriteLn;
          END;
        ShowStatus(rs);
END ShowLogical;

PROCEDURE  DeleteFile;

VAR     victim, resp : LineType;
        rs : Status;
        mode : CHAR;
        n : CARDINAL;

BEGIN   Component(Line,victim);
        IF victim[0] = 0C THEN Err("File specification required"); RETURN END;
        IF NOT TestWild(victim) THEN
          ShowStatus(Delete(victim));
          RETURN;
          END;
        n := 0;
        rs := SetWild(victim);
        IF rs <> Ok THEN ShowStatus(rs); RETURN END;
        Component(Line,resp);
        IF resp[0] = 0C THEN
          WriteString("All, Fast, None or Individual ? ");
          ReadString(resp);
          IF Result() = EscKey THEN WriteLn END;
          END;
        mode := CAP(resp[0]);
        IF (mode <> 'A') AND (mode <> 'I') AND (mode <> 'F') THEN RETURN END;
        LOOP rs := NextWild();
             IF rs = EndFile THEN EXIT END;
             IF rs <> Ok THEN ShowStatus(rs); RETURN END;
             INC(n);
             IF TestCC() THEN ShowStatus(Break); RETURN END;
             IF mode <> 'F' THEN Copy(FullName,resp) END;
             IF mode = 'I' THEN Insert(" ? ",resp,Length(resp)) END;
             IF mode <> 'F' THEN WriteString(resp) END;
             IF mode = 'I' THEN
               ReadString(resp);                    
               IF Result() = EscKey THEN WriteLn; RETURN END;
               IF Result() = EndFile THEN RETURN END;
               END;
             IF mode = 'A' THEN WriteLn END;
             IF (mode <> 'I') OR (CAP(resp[0]) = 'Y') THEN
               rs := Delete(FullName);
               IF rs <> Ok THEN 
                 IF mode = 'F' THEN WriteString(FullName); WriteLn; END;
                 ShowStatus(rs); (* but keep going *)
                 END;
               END;
             END;
        IF n = 0 THEN Err("No files found"); END;
END DeleteFile;

PROCEDURE  CopyFile;

VAR     src, dest, temp : LineType;
        rs : Status;
        i, n : CARDINAL;
        listing, wildsrc, wilddest : BOOLEAN;
        c : CHAR;

BEGIN   Component(Line,src);          
        IF src[0] = 0C THEN Err("No files specified"); RETURN END;
        Component(Line,temp);
        Component(Line,dest);
        wildsrc := TestWild(src);
        c := CAP(dest[0]);
        listing := ((c = 'L') OR wildsrc) AND (c <> 'N');
        rs := Translate(temp,dest);
        IF rs <> Ok THEN ShowStatus(rs); RETURN END;
        i := Length(dest);
        wilddest := (i > 0) AND (dest[i-1] = '-');
        IF wildsrc AND NOT wilddest THEN Err("Wildcard error"); RETURN END;
        IF NOT (wildsrc OR wilddest) THEN
          ShowStatus(DoCopy(src,dest));
          RETURN;
          END;
        n := 0;
        rs := SetWild(src);
        IF rs <> Ok THEN ShowStatus(rs); RETURN END;
        LOOP rs := NextWild();
             IF rs = EndFile THEN EXIT END;
             IF rs <> Ok THEN ShowStatus(rs); RETURN END;
             INC(n);
             IF TestCC() THEN ShowStatus(Break); RETURN END;
             IF NOT (wildsrc AND (DIR IN FileMode(NMode))) THEN
               temp[0] := 0C;
               IF wilddest THEN Insert(NName,temp,0) END;
               Insert(dest,temp,0);
               rs := DoCopy(FullName,temp);
               IF (rs <> Ok) OR listing THEN
                 WriteString(FullName);
                 WriteString(" -> ");
                 WriteString(temp);
                 WriteLn;
                 END;
               IF rs <> Ok THEN ShowStatus(rs) END;  (* keep going *)
               END;
             END;
        IF n = 0 THEN Err("No files found"); END;
END CopyFile;

PROCEDURE  DoCopy( VAR src, dest : ARRAY OF CHAR ) : Status;

CONST   blkmax = 15;

VAR     b : ARRAY [0..blkmax] OF ARRAY [0..511] OF CHAR;
        rs, xs : Status;
        hs, hd : Handle;
        m : FileMode;
        t : Time;
        u, i, bc : CARDINAL;
        done : BOOLEAN;

BEGIN   rs := Open(src,FALSE,hs);
        IF rs <> Ok THEN RETURN rs END;
        rs := GetMode(hs,t,m,u);
        IF rs <> Ok THEN RETURN rs END;
        m := m - FileMode{SYS,DIR};
        rs := Create(dest,m,hd);
        IF rs = BadName THEN rs := Open(dest,TRUE,hd); (* record devs *)
          ELSIF rs = Ok THEN rs := SetFileDate(hd,t); (* copy gets same date *)
          END;
        IF rs <> Ok THEN 
          xs := Close(hs,FALSE); 
          RETURN rs;
          END;
        done := FALSE;
        LOOP bc := 0;
             LOOP rs := ReadBlock(hs,b[bc]);
                  IF rs = EndFile THEN 
                    rs := Ok; 
                    done := TRUE; 
                    EXIT;
                    END;
                  INC(bc);
                  IF (bc = blkmax+1) OR (rs <> Ok) THEN EXIT END;
                  END;
             IF rs <> Ok THEN EXIT END;
             i := 0;
             LOOP IF i = bc THEN EXIT END;
                  rs := WriteBlock(hd,b[i],TRUE);
                  IF rs <> Ok THEN EXIT END;
                  INC(i);
                  END;
             IF done OR (rs <> Ok) THEN EXIT END;
             END;
        IF rs = Ok 
          THEN rs := Close(hd,TRUE);
          ELSE xs := Close(hd,FALSE);
          END;
        xs := Close(hs,FALSE);
        RETURN rs;
END DoCopy;

PROCEDURE  RenameFile;

VAR     src, dest, temp : LineType;
        rs : Status;
        i, n : CARDINAL;
        listing, wildsrc, wilddest : BOOLEAN;
        c : CHAR;

BEGIN   Component(Line,src);          
        IF src[0] = 0C THEN Err("No files specified"); RETURN END;
        Component(Line,temp);
        Component(Line,dest);
        wildsrc := TestWild(src);
        c := CAP(dest[0]);
        listing := ((c = 'L') OR wildsrc) AND (c <> 'N');
        rs := Translate(temp,dest);
        IF rs <> Ok THEN ShowStatus(rs); RETURN END;
        i := Length(dest);
        wilddest := (i > 0) AND (dest[i-1] = '-');
        IF wildsrc AND NOT wilddest THEN Err("Wildcard error"); RETURN END;
        IF NOT (wildsrc OR wilddest) THEN
          ShowStatus(Rename(src,dest));
          RETURN;
          END;
        n := 0;
        rs := SetWild(src);
        IF rs <> Ok THEN ShowStatus(rs); RETURN END;
        LOOP rs := NextWild();
             IF rs = EndFile THEN EXIT END;
             IF rs <> Ok THEN ShowStatus(rs); RETURN END;
             INC(n);
             IF TestCC() THEN ShowStatus(Break); RETURN END;
             temp[0] := 0C;
             IF wilddest THEN Insert(NName,temp,0) END;
             Insert(dest,temp,0);
             rs := Rename(FullName,temp);
             IF (rs <> Ok) OR listing THEN
               WriteString(FullName);
               WriteString(" -> ");
               WriteString(temp);
               WriteLn;
               END;
             IF rs <> Ok THEN ShowStatus(rs) END;
             IF rs = OffVolume THEN RETURN END;
             END;
        IF n = 0 THEN Err("No files found"); END;
END RenameFile;

PROCEDURE  FormatVolume;

VAR     k, fe, ds : CARDINAL;
        a : DevAttrSet;
        temp, dev : LineType;
        vb : ARRAY [0..3] OF CHAR;
        rs : Status;
        h : Handle;

BEGIN   Component(Line,temp);
        IF temp[0] = 0C THEN Err("Device name required"); RETURN END;
        rs := Translate(temp,dev);
        IF rs <> Ok THEN ShowStatus(rs); RETURN END;
        k := Find(Hyphen,dev);
        Remove(dev,k,80);  (* device only *)
        rs := GetDevAttr(dev,ds,a);
        IF rs <> Ok THEN ShowStatus(rs); RETURN END;
        IF NOT (FMT IN a) THEN 
          Err("Device does not support formatting"); RETURN END;
        rs := Open(dev,TRUE,h);
        IF rs = BadName THEN rs := DeviceBusy END;
        IF rs <> Ok THEN ShowStatus(rs); RETURN END;
        k := 0; fe := 0;
        LOOP IF TestCC() THEN 
               IF fe > 0 THEN WriteLn END;
               rs := Close(h,FALSE);
               RETURN;
               END;
             rs := FormatTrack(h,k);
             IF rs = NoTrack THEN EXIT END;
             IF rs <> Ok THEN
               IF fe = 0 THEN 
                 WriteString("Error formatting tracks: "); 
                 END;
               IF (fe > 0) AND ((fe MOD 10) = 0) THEN
                 WriteLn;
                 WriteString("                         ");
                 END;
               WriteCard(k,4);
               INC(fe);
               END;
             INC(k);
             END;
        IF fe > 0 THEN WriteLn END;
        k := 0; fe := 0;
        REPEAT IF TestCC() THEN 
                 IF fe > 0 THEN WriteLn END;
                 rs := Close(h,FALSE);
                 RETURN;
                 END;
               rs := Seek(h,k);
               IF rs = Ok THEN rs := ReadBlock(h,vb); END;
               IF rs <> Ok THEN
                 IF fe = 0 THEN 
                   WriteString("Error verifying sectors: "); 
                   END;
                 IF (fe > 0) AND ((fe MOD 10) = 0) THEN
                   WriteLn;
                   WriteString("                         ");
                   END;
                 WriteCard(k,4);
                 INC(fe);
                 END;
               INC(k,8);
        UNTIL k >= ds;
        IF fe > 0 THEN WriteLn END;
        ShowStatus(Close(h,FALSE));
END FormatVolume;

PROCEDURE  InitialiseVolume;

CONST   EOF = 65535;

VAR     db : Block;
        da : DevAttrSet;
        ds, dm, fs, i : CARDINAL;
        rs : Status;
        h : Handle;
        label, dev, dvn : LineType;
        now : Time;

  PROCEDURE  EnterFile( recnum, first, mode : CARDINAL; 
                        VAR nam : ARRAY OF CHAR);

  BEGIN WITH db.d[recnum] DO
          Copy(nam,Name);
          Mode := mode;
          First := first;
          Owner := ThisUser;
          Date := now;
          END;
  END EnterFile;

BEGIN   rs := GetTime(now);
        Component(Line,dev);
        rs := Translate(dev,dvn); 
        IF rs <> Ok THEN ShowStatus(rs); RETURN END;
        i := Find(Hyphen,dvn);
        Remove(dvn,i,80);  (* device only for NFS open *)
        Component(Line,label);
        IF label[0] = 0C 
          THEN Err("Device name and label required"); RETURN END;
        Component(Line,dev);
        IF dev[0] = 0C 
          THEN dm := 0FFH;
          ELSE IF NOT ScanCard(dev,16,dm) 
                 THEN Err("Mode must be HEX numeral"); RETURN END;
          END;
        rs := GetDevAttr(dvn,ds,da);
        IF rs <> Ok THEN ShowStatus(rs); RETURN END;
        IF NOT (BLK IN da) 
          THEN Err("Must be block structured device"); RETURN END;
        rs := Open(dvn,TRUE,h);
        IF rs = BadName THEN rs := DeviceBusy END; 
        IF rs <> Ok THEN ShowStatus(rs); RETURN END;
        WITH db.h DO
          Key := "HOME";
          Copy(label,Label);
          Date := now;
          Size := ds;
          Mode := dm;
          Owner := ThisUser;
          END;
        rs := Seek(h,2);
        IF rs <> Ok THEN ShowStatus(rs); rs := Close(h,FALSE); RETURN END;
        rs := WriteBlock(h,db.a,FALSE);
        IF rs <> Ok THEN ShowStatus(rs); rs := Close(h,FALSE); RETURN END;
        fs := (ds-1) DIV 256;      (* FAT size - 1 *)
        WITH db DO
          f[0] := 1; f[1] := EOF;  (* BOOTBLOCKS.SYS *)
          f[2] := EOF;             (* HOMEBLOCK.SYS  *)
          FOR i := 1 TO fs DO f[i-1+3] := i-1+4 END;
          f[fs+3] := EOF;          (* ALLOCATION.SYS *)
          f[fs+4] := EOF;          (* DIRECTORY.SYS  *)
          FOR i := fs+5 TO 255 DO f[i] := 0 END; 
          END;
        rs := Seek(h,3);
        IF rs <> Ok THEN ShowStatus(rs); rs := Close(h,FALSE); RETURN END;
        rs := WriteBlock(h,db.a,FALSE);
        IF rs <> Ok THEN ShowStatus(rs); rs := Close(h,FALSE); RETURN END;
        FOR i := 0 TO 255 DO db.f[i] := 0 END;
        FOR i := 1 TO fs DO 
          rs := Seek(h,i+3);
          IF rs <> Ok THEN ShowStatus(rs); rs := Close(h,FALSE); RETURN END;
          rs := WriteBlock(h,db.a,FALSE);
          IF rs <> Ok THEN ShowStatus(rs); rs := Close(h,FALSE); RETURN END;
          END;
        EnterFile(0,0,230H,"BOOTBLOCKS.SYS");
        EnterFile(1,2,210H,"HOMEBLOCK.SYS");
        EnterFile(2,3,210H,"ALLOCATION.SYS");
        EnterFile(3,fs+4,375H,"DIRECTORY.SYS");
        rs := Seek(h,fs+4);
        IF rs <> Ok THEN ShowStatus(rs); rs := Close(h,FALSE); RETURN END;
        rs := WriteBlock(h,db.a,FALSE);
        IF rs <> Ok THEN ShowStatus(rs); rs := Close(h,FALSE); RETURN END;
        ShowStatus(Close(h,FALSE));
END InitialiseVolume;

PROCEDURE  MountVolume;

VAR     dev, lab, opt : LineType;
        priv : BOOLEAN;

BEGIN   Component(Line,dev);
        Component(Line,lab);
        Component(Line,opt);
        IF dev[0] = 0C THEN Err("No device specified"); RETURN END;
        UpCase(opt);
        priv := (Find(opt,"PRIVATE") = 0) AND (opt[0] <> 0C);
        ShowStatus(Mount(dev,lab,priv));
END MountVolume;

PROCEDURE  DismountVolume;

VAR     dev : LineType;

BEGIN   Component(Line,dev);
        IF dev[0] = 0C THEN Err("Device name required"); RETURN END;
        ShowStatus(Dismount(dev));
END DismountVolume;

PROCEDURE  FreeSpaceVolume;

VAR     dev, t : LineType;
        s : CARDINAL;
        rs : Status;

BEGIN   Component(Line,dev);
        rs := Translate(dev,t);
        s := Find(Hyphen,t);
        Remove(t,s,78);
        rs := FreeSpace(t,s);
        IF rs <> Ok THEN ShowStatus(rs); RETURN END;
        WriteString('Device ');
        WriteString(t);
        WriteString(' has ');
        WriteCard(s,1);
        WriteString(' blocks free.');
        WriteLn;
END FreeSpaceVolume;

PROCEDURE  MakeDirectory;

VAR     spec : LineType;
        h : Handle;
        rs : Status;

BEGIN   Component(Line,spec);
        IF spec[0] = 0C THEN Err("Directory name required"); RETURN END;
        rs := Create(spec,FileMode(1F0H),h);
        ShowStatus(rs);
        rs := Close(h,TRUE);
END MakeDirectory;

PROCEDURE  TimeCommand;   (* display / set the system time *)

VAR     rs : Status;
        tim : Time;
        s, t : LineType;

BEGIN   rs := GetTime(tim);
        IF rs <> Ok THEN ShowStatus(rs); RETURN END;
        Component(Line,s);
        IF s[0] <> 0C THEN
          Component(Line,t);
          Insert("  ",t,0);
          Insert(s,t,0);
          IF NOT ScanTime(t,tim) THEN Err("Invalid time syntax"); RETURN END;
          rs := SetTime(tim);
          IF rs <> Ok THEN ShowStatus(rs); RETURN END;
          END;
        WriteTime(tim);
        WriteLn;
END TimeCommand;

PROCEDURE  TimeDisplay;

VAR     rs : Status;
        t : Time;

BEGIN   rs := GetTime(t);
        WriteTime(t);
        WriteLn;
        ShowStatus(rs);
END TimeDisplay;

PROCEDURE  SetUserNumber;

VAR     par : LineType;
        rs : Status;
        u : CARDINAL;

BEGIN   Component(Line,par);
        IF (par[0] = 0C) 
          THEN WriteString("User number = ");
               WriteCard(ThisUser,1);
               WriteString("   Process number = ");
               WriteCard(ThisProc,1);
               WriteLn;
          ELSE IF NOT PrivFlag THEN Err("Privilege violation"); RETURN END;
               IF NOT ScanCard(par,10,u) THEN
                 Err("Invalid number"); RETURN; END;
               rs := SetUser(ThisProc,u);
               IF rs = Ok THEN ThisUser := u ELSE ShowStatus(rs) END;
          END;
END SetUserNumber;

PROCEDURE  ShowProcesses;

VAR     p : Process;
        u : CARDINAL;
        n : ARRAY [0..79] OF CHAR;
        rs : Status;

BEGIN   p := 3;
        LOOP rs := GetUser(p,u);
             IF rs = NoSuch THEN EXIT END;
             IF rs = Ok THEN
               WriteString("Process ");
               WriteCard(p,1);
               WriteString("  user ");
               WriteCard(u,1);
               rs := GetName(p,n);
               IF rs = Ok THEN
                 WriteString("  ");
                 WriteString(n);
                 END;
               WriteLn;
               END;
             INC(p);
             END;
END ShowProcesses;

PROCEDURE  BootSystem;

VAR     rs : Status;

BEGIN   IF NOT PrivFlag THEN Err("Privilege violation"); RETURN END;
        rs := ReBoot();
        ShowStatus(rs);
END BootSystem;

PROCEDURE  TypeFile;

VAR     t : LineType;
        tx : ARRAY [0..131] OF CHAR;
        ts : Stream;
        w : BOOLEAN;
        rs : Status;
        n : CARDINAL;
                   
    PROCEDURE DoType( VAR n : ARRAY OF CHAR) : Status;

    BEGIN    OpenStream(ts,n,FALSE);
             IF w THEN WriteString(n); WriteLn; END;
             IF NOT Done THEN RETURN NoFile END;
             IF w THEN WriteLn END;
             LOOP IF EndStream(ts) THEN EXIT END;
                  IF TestCC() THEN
                    CloseStream(ts);
                    RETURN Break;
                    END;
                  ReadLine(ts,tx);
                  WriteString(tx);
                  WriteLn;
                  END;
             CloseStream(ts);
             IF w THEN WriteLn END;
             RETURN Ok;
    END DoType;

BEGIN   Component(Line,t);
        IF t[0] = 0C THEN Err("No files specified"); RETURN END;
        w := TestWild(t);
        IF NOT w THEN 
          ShowStatus(DoType(t));
          RETURN;
          END;
        n := 0;
        rs := SetWild(t);
        IF rs <> Ok THEN ShowStatus(rs); RETURN END;
        LOOP rs := NextWild();
             IF rs = EndFile THEN EXIT END;
             IF rs <> Ok THEN ShowStatus(rs); RETURN END;
             INC(n);
             IF TestCC() THEN ShowStatus(Break); RETURN END;
             rs := DoType(FullName);
             IF rs <> Ok THEN ShowStatus(rs); RETURN END;
             END;
        IF n = 0 THEN Err("No files found"); END;
END TypeFile;

PROCEDURE  CreateFile;

VAR     t : LineType;
        ts, cs : Stream;

BEGIN   Component(Line,t);
        IF t[0] = 0C THEN Err("File name required"); RETURN END;
        CreateStream(cs,t);
        IF NOT Done THEN Err("Cannot create file"); RETURN END;
        OpenStream(ts,"$TI",FALSE);
        IF NOT Done THEN CloseStream(cs); Err("Cannot open $TI"); RETURN END;
        WHILE NOT EndStream(ts) DO
          ReadLine(ts,t);
          WriteLine(cs,t);
          END;
        CloseStream(cs);
        CloseStream(ts);
END CreateFile;

PROCEDURE  PrintFile;

VAR     t, opt : LineType;
        rs, xs : Status;
        qh, xh : Handle;
        w, wn : BOOLEAN;
        c : CHAR;
        n : CARDINAL;

BEGIN   rs := Translate("$QUEUE",t);     (* Does the queue manager exist ? *)
        IF rs <> Ok THEN DirectPrint; RETURN END;
        Component(Line,t);
        IF t[0] = 0C THEN Err("No files specified"); RETURN END;
        Component(Line,opt);
        c := CAP(opt[0]);
        wn := TestWild(t);
        w := (wn OR (c = 'L')) AND (c <> 'N');
        rs := Open("$QUEUE",TRUE,qh);
        IF rs <> Ok THEN ShowStatus(rs); RETURN END;
        IF wn 
          THEN rs := SetWild(t);
               n := 0;
               IF rs <> Ok THEN 
                 ShowStatus(rs); 
                 xs := Close(qh,FALSE);
                 RETURN 
                 END;
               LOOP rs := NextWild();
                    IF rs <> Ok THEN EXIT END;
                    INC(n);
                    IF TestCC() THEN rs := Break; EXIT END;
                    rs := WriteBlock(qh,FullName,FALSE);
                    IF rs <> Ok THEN EXIT END;
                    IF w THEN
                      WriteString(FullName);
                      WriteLn;
                      END;
                    END;
          ELSE rs := Translate(t,opt);
               IF rs <> Ok THEN 
                 ShowStatus(rs); 
                 xs := Close(qh,FALSE);
                 RETURN 
                 END;
               rs := Open(opt,FALSE,xh);
               IF rs <> Ok THEN 
                 ShowStatus(rs); 
                 xs := Close(qh,FALSE);
                 RETURN 
                 END;
               xs := Close(xh,TRUE);
               rs := WriteBlock(qh,opt,FALSE);
               n := 1;
          END;
        xs := Close(qh,FALSE);
        IF rs = EndFile THEN rs := Ok END;
        IF rs <> Ok THEN ShowStatus(rs); RETURN END;
        IF n = 0 THEN Err("No files found"); END;
END PrintFile;

PROCEDURE  DirectPrint;

VAR     t : LineType;
        fs, ps : Stream;

BEGIN   Component(Line,t);
        IF t[0] = 0C THEN Err("File name required"); RETURN END;
        OpenStream(fs,t,FALSE);
        IF NOT Done THEN Err("Cannot open file"); RETURN END;
        CreateStream(ps,"#P0");
        IF NOT Done THEN 
          CloseStream(fs); 
          Err("Cannot access printer"); 
          RETURN;
          END;
        WHILE NOT EndStream(fs) DO
          ReadLine(fs,t);
          WriteLine(ps,t);
          END;
        WriteChar(ps,14C); (* Form feed *)
        CloseStream(fs);
        CloseStream(ps);
END DirectPrint;

PROCEDURE  ChangeMode;

VAR     rs, xs : Status;
        h : Handle;
        fn, ms : LineType;  
        tim : Time;
        ownditto, modeditto, log, wld : BOOLEAN;
        own, oown, n : CARDINAL;
        mode, omode : FileMode;
        c : CHAR;

  PROCEDURE  DoChMod( VAR n : ARRAY OF CHAR);

  BEGIN      rs := Open(n,TRUE,h);
             IF rs = Ok THEN
               rs := GetMode(h,tim,omode,oown);
               IF rs = Ok THEN 
                 IF modeditto THEN mode := omode END;
                 IF ownditto THEN own := oown END;
                 rs := SetMode(h,mode,own);
                 END;
               xs := Close(h,TRUE);
               END;
             IF rs = Ok THEN rs := xs END;
             IF ((rs <> Ok) AND wld) OR log THEN WriteString(n); WriteLn; END;
             IF (rs <> Ok) THEN ShowStatus(rs) END;
  END DoChMod;

BEGIN   Component(Line,fn);
        Component(Line,ms);
        IF ms[0] = 0C THEN 
          Err("File specification and mode required"); 
          RETURN;
          END;
        modeditto := (ms[0] = '_');
        IF NOT modeditto THEN 
          IF NOT ScanCard(ms,16,own) THEN
            Err("Mode must be HEX numeral"); 
            RETURN;
            END;
          mode := FileMode(own);
          END;
        Component(Line,ms);
        ownditto := (ms[0] = '_') OR (ms[0] = 0C);
        IF NOT ownditto THEN
          IF CAP(ms[0]) = 'S' THEN own := ThisUser
          ELSIF NOT ScanCard(ms,10,own) THEN
            Err("Owner must be DECIMAL numeral"); 
            RETURN;
            END;
          END;
        Component(Line,ms);
        c := CAP(ms[0]);            
        wld := TestWild(fn);
        log := (wld OR (c = 'L')) AND (c <> 'N');
        IF NOT wld THEN
          DoChMod(fn);
          RETURN;
          END;
        rs := SetWild(fn);
        n := 0;
        IF rs <> Ok THEN ShowStatus(rs); RETURN END;
        LOOP rs := NextWild();
             IF rs = EndFile THEN EXIT END;
             IF rs <> Ok THEN ShowStatus(rs); RETURN END;
             INC(n);
             IF TestCC() THEN ShowStatus(Break); RETURN END;
             DoChMod(FullName);
             END;
        IF n = 0 THEN Err("No files found"); END;
END ChangeMode;

PROCEDURE  StopProcess;

VAR     pns : LineType;
        pn : Process;
        u : CARDINAL;
        rs : Status;

BEGIN   Component(Line,pns);
        IF NOT ScanCard(pns,10,pn) THEN Err("Invalid process ID"); RETURN END;
        rs := GetUser(pn,u);
        IF rs <> Ok THEN ShowStatus(rs); RETURN END;
        IF (NOT PrivFlag) AND (u <> ThisUser) 
          THEN Err("Privilege violation"); RETURN END;
        ShowStatus(Kill(pn,Break));
END StopProcess;

PROCEDURE  Options;

VAR     sf : BOOLEAN;
        cc, oc : CARDINAL;

  PROCEDURE  DoOption( VAR v : BOOLEAN; VAR s : ARRAY OF CHAR);

  VAR   temp, opt : LineType;
        k : CARDINAL;
        flag : BOOLEAN;
                                        (* cheap and dirty *)
  BEGIN IF sf 
          THEN IF NOT v THEN WriteString("NO") END;
               WriteString(s); 
               Write(" ");
          ELSE Copy(Line,temp);
               cc := 0;
               LOOP Component(temp,opt);
                    IF opt[0] = 0C THEN EXIT END;
                    INC(cc);
                    UpCase(opt);
                    flag := Find("NO",opt) <> 0;
                    IF NOT flag THEN Remove(opt,0,2) END;
                    IF (Length(opt) <> 0) AND (Find(opt,s) = 0) THEN
                      v := flag;
                      INC(oc);
                      END;
                    END;
          END;
  END DoOption;

BEGIN   sf := Line[0] = 0C;
        IF sf THEN WriteString("Option settings: "); END;
        oc := 0;
        DoOption(BreakFlag,"BREAK");
        DoOption(FatalFlag,"FATAL");
        IF PrivFlag THEN DoOption(PrivSpawnFlag,"PRIVILEGE") END;
        DoOption(VerifyFlag,"VERIFY");
        IF sf THEN WriteLn; END;
        IF (NOT sf) AND (cc <> oc) THEN Err("Unrecognised options") END;
END Options;

PROCEDURE  Echo;

BEGIN   Remove(Line,0,1); (* Swallow one blank *)
        WriteString(Line);
        WriteLn;
END Echo;

PROCEDURE  Suffix( VAR suf, str : ARRAY OF CHAR);

VAR     i, k : CARDINAL;
        dotseen : BOOLEAN;

BEGIN   k := Length(str);
        IF k = 0 THEN RETURN END;
        i := k-1;
        dotseen := FALSE;
        LOOP IF str[i] = '.' THEN
               dotseen := TRUE;
               EXIT;
               END;
             IF (str[i] = '-') OR (i = 0) THEN EXIT END;
             DEC(i);
             END;
        IF NOT dotseen THEN Insert(suf,str,k) END;
END Suffix;

(* PROCEDURE Mail;   (* KGRW *)
VAR
   opf : BOOLEAN;
BEGIN
   opf := PrivSpawnFlag;
   PrivSpawnFlag := TRUE;
   Copy("$LODLIB-MailShell",Command);
   Copy(ThisUserName,Line);
   Spawn0;
   PrivSpawnFlag := opf;
END Mail;

PROCEDURE NewMail;        (* KGRW *)
VAR
   i : CARDINAL;
   rs : Status;
   line : LineType;
BEGIN
   IF Translate("$MAIL",line)<>Ok THEN 
      Err("$MAIL not found"); 
      RETURN;
      END;
   Copy("$MAIL-N.*",line);
   Insert(ThisUserName,line,Length(line)-2);
   rs := SetUser(ThisProc,0);
   IF SetWild(line)=Ok THEN
      i := 0;
      WHILE NextWild()=Ok DO INC(i) END;
      IF i>0 THEN
         WriteString("You have "); WriteCard(i,1);
         WriteString(" new mail message(s)."); WriteLn;
         END;
      END;
   rs := SetUser(ThisProc,ThisUser);
END NewMail;  *)

(* ----  Command line interface section  ---- *)

CONST   AtMax = 3;
        Prompt = "* ";
        ParamFlag = "%";

TYPE    Lptr = POINTER TO LineType;
        Cptr = POINTER TO ComRec;
        ComRec = RECORD Comm : ComName;
                        CASE Int : BOOLEAN OF
                          TRUE   : Rout : PROC |
                          FALSE  : Defn : Lptr
                          END;
                        Next : Cptr;
                 END;

VAR     ComRoot : Cptr;
        Line, Command : LineType;
        FatalFlag, VerifyFlag, PrivFlag, PrivSpawnFlag, BreakFlag : BOOLEAN;
        AtDepth : CARDINAL;
        AtFiles : ARRAY [1..AtMax] OF Stream;
        AtLines : ARRAY [1..AtMax] OF LineType;

PROCEDURE  Indirect; 

VAR     d : CARDINAL;

BEGIN   d := AtDepth+1;
        IF d > AtMax THEN 
          Err("Indirect file nesting depth exceeded"); RETURN END;
        Suffix(".COM",Command);
        OpenStream(AtFiles[d],Command,FALSE);
        IF NOT Done THEN 
          Insert("$COMLIB-",Command,0);
          OpenStream(AtFiles[d],Command,FALSE);
          END;
        IF Done 
          THEN AtDepth := d;
               Copy(Line,AtLines[d]);
          ELSE Err("Cannot open indirect command file"); 
          END;
END Indirect;

PROCEDURE  ReDirect;

VAR     d : CARDINAL;

BEGIN   FOR d := 1 TO AtDepth DO ReleaseStream(AtFiles[d]) END;
        AtDepth := 0;
END ReDirect;

MODULE Historical;

IMPORT ReadString, Status, Result, Write, WriteLn, LineType,
       Copy, Read, WriteString;

EXPORT HistRead;

CONST   MaxHist = 10;

VAR     History : ARRAY [0..MaxHist] OF LineType;
        m : CARDINAL;

PROCEDURE  HistRead( VAR s : ARRAY OF CHAR);

VAR     h, i : CARDINAL;
        d : BOOLEAN;
        c : CHAR;

  PROCEDURE  SetLen;

  BEGIN   i := 0;
          WHILE (s[i] <> 0C) DO INC(i) END;
  END SetLen;

  PROCEDURE  RubOut;

  BEGIN   IF i = 0 THEN RETURN END;
          Write(10C);
          Write(' ');
          Write(10C);
          DEC(i);
  END RubOut;

BEGIN   ReadString(s);
        IF Result() = EscKey THEN
          h := 0;
          d := FALSE;
          WHILE NOT d DO
            SetLen;
            WHILE i > 0 DO RubOut END;
            IF h < m THEN Copy(History[h],s) ELSE s[0] := 0C END;
            WriteString(s);
            SetLen;
            d := TRUE;
            LOOP  Read(c,TRUE);
                  s[i] := 0C;
                  IF c = 177C THEN 
                    RubOut
                  ELSIF c = 32C THEN
                    WriteString('<EOF>');
                    WriteLn;
                    EXIT;
                  ELSIF c = 15C THEN
                    WriteLn;
                    EXIT;
                  ELSIF c = 33C THEN
                    IF h >= m THEN h := 0 ELSE INC(h) END;
                    d := FALSE;
                    EXIT;
                  ELSIF c = 11C THEN
                    IF h >= m THEN h := m-1 ELSE DEC(h) END;
                    d := FALSE;
                    EXIT;
                  ELSIF c = 25C THEN
                    WHILE i > 0 DO RubOut END;
                  ELSIF c = 22C THEN
                    WriteLn;
                    WriteString(s);
                  ELSIF c < ' ' THEN (* ignore *)
                  ELSIF i < HIGH(s) THEN
                    Write(c);
                    s[i] := c;
                    INC(i);
                    END;
                  END;
            END;
          END;
        IF s[0] <> 0C THEN
          FOR i := MaxHist TO 1 BY -1 DO History[i] := History[i-1] END;
          Copy(s,History[0]);
          IF m <= MaxHist THEN INC(m) END;
          END;
END HistRead;

BEGIN   FOR m := 0 TO MaxHist DO History[m][0] := 0C END;
        m := 0;
END Historical;

PROCEDURE  ReadCommand( VAR s : ARRAY OF CHAR);

VAR     dig : CHAR;
        j, k, n : CARDINAL;
        temp, param : LineType;
        pf : ARRAY [0..1] OF CHAR;

BEGIN   IF TestCC() THEN ReDirect END;
        WHILE (AtDepth <> 0) AND EndStream(AtFiles[AtDepth]) DO
          ReleaseStream(AtFiles[AtDepth]);
          DEC(AtDepth);
          END;
        IF AtDepth = 0 
          THEN WriteLn;
               WriteString(Prompt);
               HistRead(s); 
               IF TestCC() THEN END;  (* ignore spurious ^C *)
          ELSE ReadLine(AtFiles[AtDepth],Line);
               pf[0] := ParamFlag; pf[1] := 0C;
               LOOP k := Find(pf,Line);
                    IF k = Length(Line) THEN EXIT END;
                    Copy(AtLines[AtDepth],temp);
                    dig := Line[k+1];
                    n := ORD(dig) - 48;
                    Remove(Line,k,2);
                    IF n < 10 THEN
                      FOR j := 0 TO n DO Component(temp,param) END;
                      Insert(param,Line,k);
                      END;
                    END;
               IF VerifyFlag THEN 
                 WriteString(Prompt);
                 WriteString(Line);
                 WriteLn;
                 END;
          END;
END ReadCommand;

PROCEDURE  Component( VAR s, t : ARRAY OF CHAR);

VAR     a : CARDINAL;

BEGIN   a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] = " ") DO INC(a) END;
        Remove(s,0,a);
        a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] > " ") DO INC(a) END;
        Substring(s,t,0,a);
        IF t[0] = ";" THEN t[0] := 0C; RETURN END;  (* Comments *)
        Remove(s,0,a);
END Component;

PROCEDURE  UpCase( VAR t : ARRAY OF CHAR);

VAR     a : CARDINAL;

BEGIN   FOR a := 0 TO HIGH(t) DO t[a] := CAP(t[a]) END;
END UpCase;

PROCEDURE  DefInt( VAR c : ARRAY OF CHAR; r : PROC);

VAR     p : Cptr;

BEGIN   NEW(p);
        WITH p^ DO
          Copy(c,Comm);
          Int := TRUE;
          Rout := r;
          Next := ComRoot;
          END;
        ComRoot := p;
END DefInt;
          
PROCEDURE  DefExt( VAR c, t : ARRAY OF CHAR);

VAR     p, q : Cptr;

BEGIN   UpCase(c);
        p := ComRoot;
        q := NIL;
        LOOP IF p = NIL THEN EXIT END;
             IF (NOT p^.Int) AND (Compare(p^.Comm,c) = 0) THEN EXIT END;
             q := p;
             p := p^.Next;
             END;
        IF p <> NIL THEN
          IF q = NIL THEN ComRoot := p^.Next ELSE q^.Next := p^.Next END;
          DISPOSE(p^.Defn);
          DISPOSE(p);
          END;
        IF t[0] = 0C THEN RETURN END;
        NEW(p); 
        NEW(p^.Defn);
        WITH p^ DO
          Next := ComRoot;
          Copy(c,Comm);
          Copy(t,Defn^);
          Int := FALSE;
          END;
        ComRoot := p;
END DefExt;

PROCEDURE  Define;

VAR     c, t : LineType;
        x : ComName;
        p : Cptr;
        k : CARDINAL;

BEGIN   Component(Line,c);
        IF c[0] = 0C 
          THEN p := ComRoot;
               WHILE p <> NIL DO
                 WITH p^ DO
                   IF NOT Int THEN
                     x := "                ";
                     Insert(Comm,x,0);
                     WriteString(x);
                     WriteString(" = ");
                     WriteString(Defn^);
                     WriteLn;
                     END;
                   END;
                 p := p^.Next;
                 END;
          ELSE k := 0;
               WHILE Line[k] = ' ' DO INC(k) END;
               Remove(Line,0,k);
               DefExt(c,Line);
          END;
END Define;

PROCEDURE  Lookup;

VAR     t : ComName;
        p, f : Cptr;
        n : CARDINAL;

BEGIN   p := ComRoot;
        LOOP Component(Line,Command);
             Copy(Command,t);
             UpCase(t);
             IF t[0] = 0C THEN RETURN END;
             IF t[0] = "@" THEN
               Remove(Command,0,1);
               Indirect;
               RETURN;
               END;
             f := NIL;
             WHILE p <> NIL DO
               n := Find(t,p^.Comm);
               IF n = 0 THEN
                 IF f = NIL 
                   THEN f := p
                   ELSE Err("Ambiguous command");
                        RETURN;
                   END;
                 END;
               p := p^.Next;
               END;
             IF f = NIL THEN Spawn0; RETURN END;
             IF f^.Int THEN f^.Rout; RETURN END;
             Insert(f^.Defn^,Line,0);
             p := f^.Next;
             END;
END Lookup;

PROCEDURE  Err( VAR s : ARRAY OF CHAR);

BEGIN   Write(15C);
        WriteString("CLI -- ");
        WriteString(s);
        WriteLn;
        IF FatalFlag THEN ReDirect END;
END Err;

PROCEDURE  ShowStatus( rs : Status);

VAR     s : LineType;

BEGIN   IF rs <> Ok THEN
          GetMessage(rs,s);
          Err(s);
          END;
END ShowStatus;

PROCEDURE  InitComTable;

BEGIN   ComRoot := NIL;
        DefInt("SET",SetLogicalF);
        DefInt("SHOW",ShowLogical);
        DefInt("DELETE",DeleteFile);
        DefInt("COPY",CopyFile);
        DefInt("RENAME",RenameFile);
        DefInt("GLOBALSET",SetLogicalT);
        DefInt("FORMAT",FormatVolume);
        DefInt("INITIALISE",InitialiseVolume);
        DefInt("MOUNT",MountVolume);
        DefInt("DISMOUNT",DismountVolume);
        DefInt("FREESPACE",FreeSpaceVolume);
        DefInt("MAKEDIRECTORY",MakeDirectory);
        DefInt("DIRECTORY",DirectoryT);
        DefInt("DATE",TimeCommand);
        DefInt("TIME",TimeCommand);
        DefInt("LOGOUT",Logout);
        DefInt("EXIT",QuietLogout);
        DefInt("USER",SetUserNumber);
        DefInt("REBOOT",BootSystem);
        DefInt("TYPE",TypeFile);
        DefInt("CREATE",CreateFile);
        DefInt("PRINT",PrintFile);
        DefInt("CHANGEMODE",ChangeMode);
        DefInt("OPTIONS",Options);
        DefInt("ECHO",Echo);
        DefInt("RUN",Spawn1);
        DefInt("LIST",DirectoryF);
        DefInt("CURRENT",SetCurrent);
(*        DefInt("MAIL",Mail);
        DefInt("NEWMAIL",NewMail);  *)
        DefInt("STOP",StopProcess);
        DefInt("START",Spawn2);
        DefInt("FILES",WideFiles);
        DefInt("PROCESSES",ShowProcesses);
        DefInt("DEFINE",Define);
END InitComTable;

PROCEDURE CLI;

BEGIN   LOOP ReadCommand(Line);
             Lookup;
             END;
END CLI;

(* ---- Login processor            ---- *)   

PROCEDURE  Disconnect; 
VAR
  unit : CARDINAL;
  rs   : Status;
  s    : ARRAY [0..25] OF CHAR;
BEGIN          
      rs := Translate("$TI",s);
      unit := ORD( s[2] ) - 48;
      IF s[3] <> 0C THEN
        unit := unit * 10 + ORD( s[3] ) - 48;
      END; (* IF *)     
      IF HangUp( unit ) <> Ok THEN                             
        Write(0C); Write(0C); Write(0C); Write(0C); Write(377C); 
      END; (* IF *)
END Disconnect;

PROCEDURE  Logout;

BEGIN   WriteString("Process ");
        WriteCard(ThisProc,1);
        WriteString(" terminated at ");
        TimeDisplay;
        Disconnect;
        Exit(Ok);
END Logout;

PROCEDURE  QuietLogout;

BEGIN   Disconnect;
        Exit(Ok);
END QuietLogout;

PROCEDURE  GetPassword( VAR pw : ARRAY OF CHAR);

VAR     i : CARDINAL;

BEGIN   i := 0;
        LOOP Read(pw[i],TRUE);
             IF pw[i] = 25C THEN i := 0
               ELSIF pw[i] = 15C THEN EXIT
               ELSIF (pw[i] = 177C) THEN
                 IF (i > 0) THEN DEC(i) END
               ELSIF i = HIGH(pw) THEN EXIT
               ELSE INC(i)
               END;
             END;
        IF pw[i] = 15C THEN pw[i] := 0C END;
        WriteLn;
END GetPassword;
          
CONST   pwlck = "passwords.lock";

PROCEDURE  Login() : BOOLEAN;

VAR     temp, pass : LineType;
        pwf : Stream;
        quiet, found : BOOLEAN;
        rs : Status;
        attempts : CARDINAL;

BEGIN   WriteLn;    
        WriteLn;           
        attempts := 0;
        OpenStream(pwf,"$SysDisk-NoLogins",FALSE);
        IF Done THEN
          LOOP IF EndStream(pwf) THEN EXIT END;
               ReadLine(pwf,Line);
               WriteString(Line);
               WriteLn;
               END;
          ReleaseStream(pwf);
          RETURN FALSE;
          END;
        LOOP WriteString("Username: ");
             ReadString(Line);
             Component(Line,ThisUserName);
             WriteString("Password: ");
             GetPassword(pass); 
             IF NOT LockWait(pwlck,30)
               THEN Err("Waited too long for password file"); RETURN FALSE END;
             OpenStream(pwf,"$SysDisk-PASSWORDS.SYS",FALSE);
             IF NOT Done
               THEN Err("Cannot access PASSWORDS.SYS"); RETURN FALSE END;
             found := FALSE;
             LOOP IF EndStream(pwf) THEN EXIT END;
                  ReadLine(pwf,Line);
                  Component(Line,temp);
                  IF Compare(temp,ThisUserName) = 0 
                    THEN found := TRUE; EXIT END;
                  END;
             Component(Line,temp);
             ReleaseStream(pwf);
             LockClear(pwlck);
             IF temp[0] = '~' THEN 
               Cipher(pass);
               Remove(temp,0,1);
               END;
             found := found AND (Compare(temp,pass) = 0);
             Copy("----------------",pass);
             IF NOT found 
               THEN Err("User authorisation failure");
                    INC(attempts);
                    IF attempts = 3 THEN RETURN FALSE; END;
               ELSE EXIT
               END;                                  
             END; (* attempts loop *)
        Component(Line,temp);
        rs := CreateLogical("CURRENT",temp);
        rs := CreateLogical("HOME",temp);
        Component(Line,temp);
        found := ScanCard(temp,10,ThisUser);
        Component(Line,temp);
        PrivFlag := CAP(temp[0]) = "P";
        Component(Line,temp);
        quiet := CAP(temp[0]) = "Q";
        rs := Translate("$TI",Line);
        IF NOT quiet THEN
          WriteLn;
          WriteString("Process ");
          WriteCard(ThisProc,1);
          WriteString(" started on ");
          WriteString(Line);
          WriteString(" at ");
          TimeDisplay;
          OpenStream(pwf,"$SysNews",FALSE);
          IF Done THEN
            LOOP IF EndStream(pwf) THEN EXIT END;
                 ReadLine(pwf,Line);
                 WriteString(Line);
                 WriteLn;
                 END;
            ReleaseStream(pwf);
            END;
          END;
        rs := SetUser(ThisProc,ThisUser);
        rs := SetName(ThisProc,ThisUserName);
        OpenStream(pwf,"LOGIN.COM",FALSE);
        IF Done THEN
          ReleaseStream(pwf);
          Copy("LOGIN",Command);
          Indirect;
          END;
        OpenStream(pwf,"$SysDisk-SysLOGIN.COM",FALSE);
        IF Done THEN
          ReleaseStream(pwf);
          Copy("$SysDisk-SysLOGIN",Command);
          Indirect;  (* Stack SysLOGIN to go first *)
          END;
        IF TestCC() THEN (* ignore it *) END; 
        RETURN TRUE;
END Login;

VAR     rs : Status;

BEGIN   InitComTable;
        FatalFlag := TRUE;
        VerifyFlag := FALSE;
        PrivFlag := TRUE;
        PrivSpawnFlag := FALSE;
        BreakFlag := TRUE; 
        AtDepth := 0;
        Hyphen[0] := "-"; Hyphen[1] := 0C;
        rs := GetPID(ThisProc);
        rs := GetUser(ThisProc,ThisUser);
        GetCommand(Line);
        IF Line[0] <> 0C 
          THEN Component(Line,Command);
               Indirect;
               CLI;
          ELSE IF Login() THEN CLI END;
          END;
        Disconnect;
END LOGIN.
