MODULE Password;  (* This is the PASSWORD command - it allows users to
                     change their passwords in PASSWORDS.SYS. *)

FROM R2SysCalls IMPORT ReadString, WriteString, Read, WriteLn, Status,
                       SetUser, Rename, GetPID, GetMessage, Exit, 
                       SetCCMode, Delete;
FROM R2NewCalls IMPORT GetName;
FROM R2Streams IMPORT Stream, Done, EndStream, OpenStream, ReadLine,
                      CreateStream, WriteLine, CloseStream, ReleaseStream;
FROM R2Strings IMPORT Remove, Substring, Copy, Compare, Insert, Compress,
                      Length;
FROM LockSync IMPORT LockWait;
FROM OneWay IMPORT Cipher;

CONST   uaf = "$SysDisk-PASSWORDS.SYS";
        tmp = "$SysDisk-passwords.new";
        jnk = "$SysDisk-passwords.old";
        lck = "passwords.lock";

TYPE    string = ARRAY [0..79] OF CHAR;
                                             
VAR     ps, np : Stream;
        Name, oPass, nPass, s, t, n, p : string;
        i, pid : CARDINAL;
        found : BOOLEAN;

PROCEDURE  Error( VAR s : ARRAY OF CHAR);

BEGIN   WriteString("Password -- ");
        WriteString(s);
        WriteLn;
        Copy(jnk,s); Copy(s,oPass); Copy(s,nPass);
        Exit(Status(-1));
END Error;

PROCEDURE  Check( rs : Status);

VAR     s : ARRAY [0..79] OF CHAR;

BEGIN   IF rs = Ok THEN RETURN END;
        GetMessage(rs,s);
        Error(s);
END Check;

PROCEDURE  GetPass( VAR pw : ARRAY OF CHAR);

VAR     i : CARDINAL;

BEGIN   i := 0;
        LOOP Read(pw[i],TRUE);
             IF pw[i] = 25C THEN i := 0
               ELSIF pw[i] = 15C THEN EXIT
               ELSIF (pw[i] = 177C) THEN
                 IF (i > 0) THEN DEC(i) END
               ELSIF i = HIGH(pw) THEN EXIT
               ELSE INC(i)
               END;
             END;
        IF pw[i] = 15C THEN pw[i] := 0C END;
        WriteLn;
END GetPass;

PROCEDURE  Component( VAR s, t : ARRAY OF CHAR);

VAR     a : CARDINAL;

BEGIN   a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] = " ") DO INC(a) END;
        Remove(s,0,a);
        a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] > " ") DO INC(a) END;
        Substring(s,t,0,a);
        IF t[0] = ";" THEN t[0] := 0C; RETURN END;  (* Comments *)
        Remove(s,0,a);
END Component;

BEGIN   Check(GetPID(pid));
        Check(SetUser(pid,0));  (* Fails if not priv'd *)
        Check(GetName(pid,Name));
        WriteString("Changing password for ");
        WriteString(Name);
        WriteLn;
        WriteString("Old password: ");
        GetPass(oPass);
        WriteString("New password: ");
        GetPass(nPass);
        WriteString("Retype new password: ");
        GetPass(s);   
        IF Compare(s,nPass) <> 0 
          THEN Error("Error entering new password"); END;
        Cipher(nPass);
        Insert(" ~",nPass,0);
        IF NOT LockWait(lck,30)
          THEN Error("Waited too long for password file"); END;
        OpenStream(ps,uaf,FALSE);
        IF NOT Done THEN Error("Cannot access password file") END;
        CreateStream(np,tmp);
        IF NOT Done THEN Error("Cannot create temporary file") END;
        found := FALSE;
        SetCCMode(FALSE);  (* No ^C interruptions now please *)
        WHILE NOT EndStream(ps) DO
          ReadLine(ps,s);
          Copy(s,t);
          Component(t,n);
          Component(t,p);
          IF Compare(n,Name) = 0
            THEN IF p[0] = '~' THEN
                   Cipher(oPass);
                   Remove(p,0,1);
                   END;
                 IF Compare(p,oPass) = 0
                   THEN found := TRUE;
                        Insert(nPass,t,0);
                        Insert("  ",t,0);
                        Insert(Name,t,0);
                        WriteLine(np,t);
                   ELSE WriteLine(np,s);
                   END;
            ELSE WriteLine(np,s);
            END;
          END;
        CloseStream(ps);
        CloseStream(np);
        IF found
          THEN Check(Rename(uaf,jnk));
               Check(Rename(tmp,uaf));
               Check(Delete(jnk));
          ELSE Check(Delete(tmp));
               Error("Password record not found, no changes made");
          END; 
END Password.
