MODULE Queue; (* Queue manager for the print spooler *)

FROM SYSTEM IMPORT ADDRESS;
FROM R2SysCalls IMPORT TestMessage, ReadBlock, Open, Close, WriteBlock,
                       GetPID, CreateGlobal, Execute, Status, Process,
                       Exit, GetUser, Handle, GetMessage, WriteLn, WriteString;
FROM R2Storage IMPORT ALLOCATE, DEALLOCATE;
FROM R2Strings IMPORT Copy, Insert;
FROM R2Format IMPORT FormatCard;

TYPE    Line = ARRAY [0..79] OF CHAR;
        Qptr = POINTER TO Qelt;
        Qelt = RECORD Text : Line;
                      Next : Qptr;
               END;
              
VAR     Head, Tail : Qptr;

PROCEDURE  Get( VAR b : ARRAY OF CHAR);

VAR     q : Qptr;

BEGIN   Copy(Head^.Text,b);
        q := Head;
        Head := Head^.Next;
        DISPOSE(q);
END Get;

PROCEDURE  Empty() : BOOLEAN;

BEGIN   RETURN Head = NIL;
END Empty;

PROCEDURE  Put( VAR b : ARRAY OF CHAR);

VAR     q : Qptr;

BEGIN   NEW(q);
        Copy(b,q^.Text);
        q^.Next := NIL;
        IF Head = NIL THEN Head := q ELSE Tail^.Next := q END;
        Tail := q;
END Put;

PROCEDURE  Check( rs : Status);

BEGIN   IF rs <> Ok THEN 
          GetMessage(rs,str);
          WriteString("** Queue manager -- ");
          WriteString(str);
          WriteLn;
          Exit(Status(-1));
          END;
END Check;

VAR     rs : Status;
        thisproc, prn, nxt, usr : CARDINAL;
        str, prns, tmp : Line;
        x : ADDRESS;
        in, out : Handle;
        ready : BOOLEAN;

BEGIN   Head := NIL;   
        ready := TRUE;
        Check(GetPID(thisproc));
        FormatCard(thisproc,10,1,str);
        Insert("#M",str,0);
        Check(CreateGlobal("QUEUE",str));
        Check(Execute("$LODLIB-Printer.LOD","junk","$SysDisk",65535,0,
                      FALSE,FALSE,TRUE,prn,rs,x));
        FormatCard(prn,10,1,prns);
        Insert("#M",prns,0);
        Check(Open("#M0",FALSE,in));
        Check(Open(prns,TRUE,out));
        LOOP nxt := TestMessage(TRUE);
             rs := ReadBlock(in,str);
             IF nxt = prn 
               THEN IF NOT Empty()
                      THEN Get(str);
                           Check(WriteBlock(out,str,FALSE));
                      ELSE ready := TRUE;
                      END
               ELSE IF rs <> EndFile THEN
                      rs := GetUser(nxt,usr);
                      IF rs = Ok THEN
                        tmp[0] := ' '; tmp[1] := 0C;
                        Insert(tmp,str,0);
                        FormatCard(usr,10,1,tmp);
                        Insert(tmp,str,0);
                        IF ready 
                          THEN Check(WriteBlock(out,str,FALSE));
                               ready := FALSE;
                          ELSE Put(str);
                          END;
                        END;
                      END;
               END;
             END;    (* loops forever *)
END Queue.
