MODULE ReadCheck;  (* Find files with bad blocks *)
                               
FROM SYSTEM IMPORT ADDRESS;
FROM R2SysCalls IMPORT GetCommand, Time, Handle, Status, ModeBits, FileMode,
                       Exit, Open, GetMode, ReadBlock, WriteLn, WriteString,
                       GetMessage, Result, ReadString, Translate, FastSize, 
                       Close;
FROM R2Strings IMPORT Copy, Find, Length, Insert;

TYPE    Line = ARRAY [0..79] OF CHAR;
        DirRec = RECORD Name : ARRAY [0..19] OF CHAR;
                        First : CARDINAL;
                        Date : Time;
                        Owner : CARDINAL;
                        Mode : FileMode;
                        END;
        Block = RECORD 
                  CASE BOOLEAN OF
                    TRUE  : a : ARRAY [0..511] OF CHAR  |
                    FALSE : d : ARRAY [0..15] OF DirRec
                    END;
                  END;

VAR     Param, Root : Line;
        Hyphen : ARRAY [0..1] OF CHAR;

PROCEDURE  Scan( VAR fn : ARRAY OF CHAR);  (* Recursive directory scan *)

VAR     h, fh : Handle;
        rs : Status;
        db : Block;
        tm : Time;
        mo : FileMode;
        rsp : Line;
        i, j, k : CARDINAL;
        blk : Block;
        d : BOOLEAN;

BEGIN   Copy(fn,rsp);
        i := Find(Hyphen,rsp);
        j := Length(rsp);
        IF i=j  THEN Insert("-DIRECTORY.SYS",rsp,j) END;
        IF NOT Check(Open(rsp,FALSE,h),'Opening directory: ',rsp) 
          THEN RETURN END;
        IF NOT Check(GetMode(h,tm,mo,i),'Checking mode of: ',rsp)
          THEN RETURN END;
        IF NOT (DIR IN mo) THEN Exit(BadDirect) END; (* Paranoia *)
        LOOP rs := ReadBlock(h,db.a);
             IF rs <> Ok THEN EXIT END;
             FOR i := 0 TO 15 DO
               WITH db.d[i] DO
                 j := Length(Name);
                 IF j <> 0 THEN
                   Copy(Name,rsp);
                   Insert(Hyphen,rsp,0);
                   Insert(fn,rsp,0);         
                   IF Check(FastSize(Root,First,k),'Sizing file: ',rsp) THEN
                     IF Check(Open(rsp,FALSE,fh),'Opening file: ',rsp) THEN
                       j := 1;
                       LOOP IF j > k THEN EXIT END;
                            IF NOT Check(ReadBlock(fh,blk.a),
                                   'Reading file: ',rsp) THEN EXIT END;
                            INC(j);
                            END;
                       d := Check(Close(fh,FALSE),'Closing file: ',rsp);
                       END;
                     END;
                   IF (DIR IN Mode) AND NOT (SYS IN Mode) THEN Scan(rsp) END;
                   END;
                 END;
               END;
             END; 
        IF (rs <> Ok) AND (rs <> EndFile) THEN 
          d := Check(rs,'Reading directory: ',fn) END;
        d := Check(Close(h,FALSE),'Closing directory: ',fn);
END Scan;

PROCEDURE  Check( rs : Status; VAR m, n : ARRAY OF CHAR) : BOOLEAN; 

BEGIN   IF rs = Ok THEN RETURN TRUE END;
        WriteString('ReadCheck -- ');
        GetMessage(rs,Param);
        WriteString(Param);
        WriteString(", ");
        WriteString(m);
        IF HIGH(n) <> 0 THEN WriteString(n) END;
        WriteLn;
        RETURN FALSE;
END Check;

BEGIN   Hyphen[0] := '-'; Hyphen[1] := 0C;
        GetCommand(Param);
        IF Param[0] = 0C THEN 
          WriteString('ReadCheck> ');
          ReadString(Param);
          IF Result() <> Ok THEN Exit(Status(-1)) END;
          END;
        IF Check(Translate(Param,Root),'Translating: ',Param)
          THEN Scan(Root); END;
END ReadCheck.
