MODULE Search;

FROM R2SysCalls IMPORT 
  WriteString, WriteLn, Status, GetCommand, Exit, GetMessage;
FROM R2Format IMPORT
  FormatCard;
FROM R2Strings IMPORT
  Remove, Copy, Length, Insert, Substring;                               
FROM R2Streams IMPORT
  Stream, Done, CreateStream, OpenStream, CloseStream, EndStream,
  WriteLine, ReadLine;
FROM R2Directories IMPORT
  SetWild, NextWild, WildMatch,FullName;
FROM Match IMPORT 
  PatternMatch;
  
TYPE
 
  LineType = ARRAY[0..79] OF CHAR;

PROCEDURE Check( r : Status);
 
VAR
  str : LineType;

BEGIN
  IF r = Ok THEN RETURN END;
  GetMessage(r,str);
  Err(str);
END Check;
    

PROCEDURE  Err( VAR s : ARRAY OF CHAR);

BEGIN
  WriteString("Search -- ");
  WriteString(s);
  WriteLn;
  Exit(Status(-1));
END Err;

PROCEDURE  Component( VAR s, t : ARRAY OF CHAR);

VAR     a : CARDINAL;

BEGIN   a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] = " ") DO INC(a) END;
        Remove(s,0,a);
        a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] > " ") DO INC(a) END;
        Substring(s,t,0,a);
        IF t[0] = ";" THEN t[0] := 0C; RETURN END;  (* Comments *)
        Remove(s,0,a);
END Component;

PROCEDURE SearchFiles;
            
VAR
  Line, temp, buffer, filepat, searchpat, outfilename : LineType;
  rs : Status;                 
  outstream, instream : Stream;          
  i,j,lc,bc : CARDINAL;
  delimiter : CHAR;
  first,found : BOOLEAN;

BEGIN                                                         
  GetCommand(Line);
  Component(Line,filepat);
  IF filepat[0] = 0C THEN
    Err("No files specified");
  END;                
  i := 0;
  WHILE ( i < HIGH(Line)) AND ((Line[i] <> 0C) AND (Line[i] = " " )) DO
    INC(i);
  END;
  delimiter := Line[i];
  INC(i);                                   
  j := i;
  WHILE ( j < HIGH(Line)) AND ((Line[j] <> 0C) AND (Line[j] <> delimiter )) DO
    INC(j);
  END;
  IF (Line[j] <> delimiter) THEN
    Err("Missing Delimiter");
  END;                     
  IF delimiter = 0C THEN
    Err("No pattern specified");
  END;
  Substring(Line,searchpat,i,j-i);
  Remove(Line,i-1,j-i+2);
  Component(Line,outfilename);  
  IF outfilename[0] = 0C THEN Copy("$TI",outfilename) END;
  CreateStream(outstream,outfilename);
  IF NOT Done THEN
    Err("Unable to open output file");
  END;
  Check(SetWild(filepat));
  found := FALSE;
  LOOP 
    rs := NextWild();
    IF rs <> Ok THEN 
      EXIT;
    ELSE
      found := TRUE; (* At least one file matches specification *)
    END;
    OpenStream(instream,FullName,FALSE);
    IF NOT Done THEN
      Copy("Unable to open file ",Line);
      Insert(FullName,Line,20);
      WriteLine(outstream,Line);
    ELSE
      first := TRUE; (* First time file is used, print name *)
      lc := 1;       (* Count of lines in the file          *)
      WHILE NOT EndStream(instream) DO
        ReadLine(instream,Line);
        IF PatternMatch(searchpat,Line,TRUE) THEN
          IF first THEN
            WriteLine(outstream,FullName);
            first := FALSE;
          END;
          buffer[0] := 0C;
          bc := 0;
          FormatCard(lc,10,5,temp);
          Insert(temp,buffer,bc);
          INC(bc,5);
          Insert(": ",buffer,bc);
          INC(bc,2);
          Insert(Line,buffer,bc);
          WriteLine(outstream,buffer);
        END;(*IF*)
        INC(lc); (* Line Counter *)
      END;(*WHILE*)
      CloseStream(instream);
    END;(*IF*)
  END;(*LOOP*)

  IF (rs <> Ok) AND (rs <> EndFile) THEN
    Check(rs);
  END;                                                          
  IF NOT found THEN
    Err("No matching files");
  END;
  CloseStream(outstream);                                        
END SearchFiles;

BEGIN
  SearchFiles;
END Search.
