MODULE Sort;

FROM R2SysCalls IMPORT 
  ReadString, WriteString, WriteLn, Translate, Status, Write,
  GetCommand, Exit, GetMessage, Result;
FROM R2Strings IMPORT
  Remove, Length, Insert, Substring, Compare;
FROM R2Streams IMPORT
  Stream, Done, OpenStream, ReadLine, CreateStream, CloseStream, 
  WriteLine, EndStream;
FROM R2Storage IMPORT
  ALLOCATE, DEALLOCATE;

CONST 
  maxlines = 2499; 

TYPE
 
  LineType = ARRAY[0..131] OF CHAR;
  LinePtr  = POINTER TO LineType;
  LineArray = ARRAY[0..maxlines] OF LinePtr;

PROCEDURE  Copy( VAR s, t : ARRAY OF CHAR);

VAR     i, k : CARDINAL;

BEGIN   k := HIGH(s);
        IF k > HIGH(t) THEN k := HIGH(t) END;
        i := 0;
        WHILE (i <= k) AND (s[i] <> 0C) DO
          t[i] := s[i];
          INC(i);
          END;
        IF i <= HIGH(t) THEN t[i] := 0C END;
END Copy;

PROCEDURE Check(
                   r : Status);
 
VAR
  str : LineType;

BEGIN
  IF r = Ok THEN RETURN END;
  GetMessage(r,str);
  Err(str);
END Check;
    

PROCEDURE  Err(
               VAR s : ARRAY OF CHAR);

BEGIN
  WriteString("Sort -- ");
  WriteString(s);
  WriteLn;
  Exit(Status(-1));
END Err;

PROCEDURE  Component( VAR s, t : ARRAY OF CHAR);

VAR     a : CARDINAL;

BEGIN   a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] = " ") DO INC(a) END;
        Remove(s,0,a);
        a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] > " ") DO INC(a) END;
        Substring(s,t,0,a);
        IF t[0] = ";" THEN t[0] := 0C; RETURN END;  (* Comments *)
        Remove(s,0,a);
END Component;
                  
PROCEDURE Sort(VAR data : LineArray; cnt : CARDINAL);
        
  PROCEDURE CompareLines(p1,p2 : LinePtr) : INTEGER;

  BEGIN    
    RETURN Compare(p1^,p2^); (* Returns a negative number if less *)
  END CompareLines;

  PROCEDURE Swap(VAR p1,p2 : LinePtr);

  VAR
    t : LinePtr;
      
  BEGIN
    t := p2;
    p2 := p1;
    p1 := t;
  END Swap;

  PROCEDURE qs(l,r : INTEGER; VAR dat : LineArray);
    VAR
      i,j : INTEGER;
      x,y : LinePtr;
  BEGIN
    i := l; j := r;
    x := dat[(l+r) DIV 2];
    REPEAT
      WHILE CompareLines(dat[i],x) < 0 DO
        INC(i);
      END;
      WHILE CompareLines(x,dat[j]) < 0 DO
        DEC(j);
      END;                            
      IF i <= j THEN
        Swap(dat[i],dat[j]);
        INC(i);DEC(j);
      END;
    UNTIL i > j;
    IF l < j THEN qs(l,j,dat) END;
    IF l < r THEN qs(i,r,dat) END;
  END qs;
BEGIN
  qs(0,cnt,data);
END Sort; 
  
VAR
  Line, infilename, outfilename, temp, buffer : LineType;
  rs : Status;                 
  instream, outstream : Stream;
  k, count, len  : CARDINAL;
  filelines : LineArray;

BEGIN                                                         
  GetCommand(Line);
  Component(Line,temp);
  IF temp[0] = 0C THEN
    LOOP
      WriteString("Sort - Input File> ");
      ReadString(temp);
      CASE Result() OF
    
        EndFile,EscKey : Exit(Status(-1)); |
      
        Ok             : IF temp[0] <> 0C THEN EXIT END;

      ELSE
      END;
    END;
  END;
  Check(Translate(temp,infilename));
  OpenStream(instream,infilename,FALSE);
  IF NOT Done THEN
    Err("Unable to open input file");
  END;
  Component(Line,temp);
  IF temp[0] = 0C THEN Copy("$TI",temp) END;  (* Output to Screen *)
  Check(Translate(temp,outfilename));
  CreateStream(outstream,outfilename);
  IF NOT Done THEN
    Err("Unable to open output file");
  END;      
  count := 0;
  WHILE NOT EndStream(instream) DO  
    buffer[0] := 0C;
    ReadLine(instream,buffer);
    len := Length(buffer) + 1;
    ALLOCATE(filelines[count],len);
    IF filelines[count] = NIL THEN
      Err("Allocation Error");
    END;
    Copy(buffer,filelines[count]^);
    INC(count);
  END;
  CloseStream(instream);
  IF count > 0 THEN (* Some records were read *)
    DEC(count);     
    Sort(filelines,count); 
    FOR k := 0 TO count DO
      WriteLine(outstream,filelines[k]^);
    END;
  END;
  CloseStream(outstream);  
END Sort.
