MODULE View;

FROM SYSTEM IMPORT
  ADDRESS;
FROM R2SysCalls IMPORT 
  Write, ReadString, WriteString, WriteLn, Translate, Status, Read,
  Handle, Open, Close, GetCommand, Exit, GetMessage, Result, Sleep;
FROM R2Strings IMPORT
  Remove, Copy, Length, Insert, Substring;
FROM R2Streams IMPORT
  Stream, Done, CloseStream, WriteLine, WriteChar,
  GetPos, SetPos, EndStream, ReadLine, OpenStream;      
FROM R2Format IMPORT 
  FormatCard, ScanCard;       
FROM Match IMPORT 
  PatternMatch;

CONST

  maxpage = 255;
  pagelen = 22;
prompt  = "<SPACE>=Next <ESC>=Backup E=End G=Goto S=Search F=FindNext Q=Quit ";
spaces  = "                                                                  ";
  dashes  = "----------------------------------------";

TYPE
  ShortLine = ARRAY[0..39] OF CHAR; 
  LineType  = ARRAY[0..79] OF CHAR;
  pagetype  = ARRAY[0..pagelen] OF LineType;
  pagearray = ARRAY[0..maxpage] OF ADDRESS;

VAR
  currentpage,pagenum,totalpages : CARDINAL;
  page     : pagearray;
  searchpat : ShortLine;

PROCEDURE ClearTypeAhead;

VAR
  ch : CHAR;

BEGIN  
  REPEAT
    Read(ch,FALSE);
  UNTIL Result() = EmptyBuff;
END ClearTypeAhead;

PROCEDURE Check(
                   r : Status);
 
VAR
  str : LineType;

BEGIN
  IF r = Ok THEN RETURN END;
  GetMessage(r,str);
  Err(str);
END Check;
    

PROCEDURE  Err(
               VAR s : ARRAY OF CHAR);

BEGIN
  WriteString("View -- ");
  WriteString(s);
  WriteLn;
  Exit(Status(-1));
END Err;

PROCEDURE  Component( VAR s, t : ARRAY OF CHAR);

VAR     a : CARDINAL;

BEGIN   a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] = " ") DO INC(a) END;
        Remove(s,0,a);
        a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] > " ") DO INC(a) END;
        Substring(s,t,0,a);
        IF t[0] = ";" THEN t[0] := 0C; RETURN END;  (* Comments *)
        Remove(s,0,a);
END Component;

PROCEDURE GetPage(VAR target  : CARDINAL;
                  VAR lines   : pagetype);
(*
    Goes from current to target page.  The contents of the page are
    contained in lines. If target is past EOF then last page is returned.
*)

VAR 
  i   : CARDINAL;               
  buf : LineType;         
  pos : ADDRESS;

BEGIN                
  IF target > totalpages THEN (* We've not been here before *)
    SetPos(instream,page[totalpages]); (* Go as far as possible *)
    LOOP
      FOR i := 0 TO pagelen-1 DO
        ReadLine(instream,lines[i]);
      END;
      IF EndStream(instream) THEN (* This is the last page *)
        target := totalpages;
        EXIT ;
      END;
      GetPos(instream,pos);                                
      INC(totalpages);
      IF totalpages > maxpage THEN
        target := maxpage;
        EXIT;
      END;
      page[totalpages] := pos;
      IF totalpages = target THEN
        EXIT;
      END;
    END;
  END;
  SetPos(instream,page[target]);
  FOR i := 0 TO pagelen-1 DO
    ReadLine(instream,lines[i]);
  END;
END GetPage;

PROCEDURE DisplayPrompt( pg : CARDINAL);
                                      
VAR
  buf : LineType;

BEGIN
  Write(15C);
  WriteString(prompt);
  WriteString("  Page ");
  FormatCard(pg+1,10,4,buf);WriteString(buf);
END DisplayPrompt;

PROCEDURE DisplayPage(VAR pg : CARDINAL);
   
VAR
  i,j : CARDINAL;
  buf : LineType;                
  pos : ADDRESS;
  display : pagetype;  

BEGIN            
  GetPage(pg,display);
  WriteLn;
  FOR i := 0 TO pagelen-1 DO
    buf := display[i];
    j := 0;
    WHILE (j <= HIGH(buf)) AND (buf[j] <> 0C) DO
      IF (buf[j] < " ") OR (buf[j] > "~") THEN buf[j] := 200C END;
      INC(j);
    END; 
    WriteString(buf);WriteLn;
  END;        
  WriteString(dashes);WriteString(dashes);WriteLn;
  DisplayPrompt(pg);
END DisplayPage;
               
PROCEDURE FindPat(VAR pat : ARRAY OF CHAR;
                  VAR pg : CARDINAL) : BOOLEAN;
 
VAR
  pos : ADDRESS;
  tar, cur : CARDINAL;
  buf : LineType;             
  found : BOOLEAN;
  searchpage : pagetype;

BEGIN
  found := FALSE;
  cur := pg;
  LOOP    
    tar := cur + 1;
    GetPage(tar,searchpage);     (* Start looking on next page  *)
    IF tar = cur THEN            (* Reached the end of the file *)
      found := FALSE;
      EXIT;
    ELSE
      FOR i := 0 TO pagelen-1 DO (* Search the page            *)
        found := found OR PatternMatch(searchpat,searchpage[i],TRUE);
      END;
      IF found THEN 
        pg := tar;
        EXIT;
      END;                                                      
    END;                                  
    INC(cur);
  END;(*LOOP*)
                             
  RETURN found;

END FindPat;      
  
PROCEDURE ReadStringNoCR(VAR s : ARRAY OF CHAR);
                            
VAR
  i : CARDINAL;

BEGIN
  i := 0;
  LOOP
    Read(s[i],TRUE);
    IF s[i] = 177C THEN
      IF i > 0 THEN
        Write(10C);Write(" ");Write(10C);
        DEC(i);
      END;
    ELSIF s[i] = 15C THEN
      s[i] := 0C;
      EXIT;
    ELSIF s[i] < " " THEN (* Ignore *)
    ELSIF i < HIGH(s) THEN
      Write(s[i]);
      INC(i);
    END;
  END;
END ReadStringNoCR;
  
VAR
  Line, infilename, temp, buffer : LineType;
  response : ShortLine;
  rs : Status;                 
  instream : Stream;          
  i : CARDINAL;
  ch : CHAR; 
  repaint,searchon : BOOLEAN;
  
BEGIN                                                         
  GetCommand(Line);
  Component(Line,temp);
  IF temp[0] = 0C THEN
    LOOP
      WriteString("View> ");
      ReadString(temp);
      CASE Result() OF
      
        EndFile,EscKey : Exit(Status(-1)); |
        
        Ok             : IF temp[0] <> 0C THEN EXIT END;
  
      ELSE
      END;
    END;
  END;
  Check(Translate(temp,infilename));
  OpenStream(instream,infilename,FALSE);
  IF NOT Done THEN
    Err("Cannot open input file");
  END;

  pagenum  := 0;                 (* Initialize page number to 0            *)
  currentpage := 0;              (* This is also the last page looked at   *)
  GetPos(instream,page[pagenum]);(* Enter 1st entry in page table          *)
  totalpages := 0;               (* Number of farthest page read into file *)
  repaint := TRUE;              
  searchon := FALSE;
 
  LOOP              

    ClearTypeAhead;
    IF NOT repaint THEN
      DisplayPrompt(pagenum);
    ELSE  
      DisplayPage(pagenum);
    END;

    currentpage := pagenum;
    Read(ch,TRUE);                    
    repaint := TRUE;

    CASE CAP(ch) OF

       33C : (* EscKey *)
            IF pagenum > 0 THEN DEC(pagenum) END;
            
     | 32C,"Q" : (* Ctrl-Z *)
            EXIT;      
            
     | "E" : (* Go to End of File *)
            pagenum := maxpage;         

     | "G" : (* Read a page number from terminal *)
            LOOP
              Write(15C); 
              WriteString(spaces);
              Write(15C);
              WriteString("Go to page: ");
              ReadStringNoCR(response);
              IF ScanCard(response,10,i) THEN
                IF i > 0 THEN 
                  pagenum := i - 1; 
                  EXIT; 
                END;
              END;                                                        
           END;

     | "S" : (* Read a search pattern from terminal *)
            Write(15C); 
            WriteString(spaces);
            Write(15C);
            WriteString("Enter search pattern: ");
            ReadStringNoCR(response);
            IF response[0] <> 0C THEN
              searchpat := response;
              searchon:= TRUE;
            END;
            IF searchon THEN
              IF NOT FindPat(searchpat,pagenum) THEN
                Write(15C); 
                WriteString(spaces);
                Write(15C);
                WriteString("Search pattern not found.");
                Check(Sleep(111));
                repaint := FALSE;
              END;             
            ELSE
            repaint := FALSE;
            END;

     | "F" : (* Find next occurence of search pattern *)
            IF searchon THEN
              IF NOT FindPat(searchpat,pagenum) THEN
                Write(15C); 
                WriteString(spaces);
                Write(15C);
                WriteString("Search pattern not found.");
                Check(Sleep(111));
                repaint := FALSE;
              END;       
            ELSE
              repaint := FALSE;
            END;
    ELSE
      INC(pagenum);
    END(*CASE*);
  END;(*LOOP*)
  WriteLn;
  CloseStream(instream);

END View.
