MODULE Zap;

FROM SYSTEM IMPORT
  ADDRESS ;
FROM R2SysCalls IMPORT 
  ReadString, Write, WriteString, WriteLn, Status, 
  GetCommand, Exit, GetMessage, Result;
FROM R2Format IMPORT
  FormatCard, ScanCard, ScanAddr, FormatAddr;
FROM R2Strings IMPORT
  Remove, Copy, Length, Insert, Substring, Compress;
FROM ZapFiles IMPORT
  StartFile, PutByte, GetByte, FinishFile;

CONST
  base = 16;

TYPE
 
  LineType = ARRAY[0..79] OF CHAR;
  Block = ARRAY[0..511] OF CHAR;
  ModeType = (word, character, byte, none);

PROCEDURE Check(
                   r : Status);
 
VAR
  str : LineType;

BEGIN
  IF r = Ok THEN RETURN END;
  GetMessage(r,str);
  Fatal(str);
END Check;

PROCEDURE Error( VAR s : ARRAY OF CHAR);

BEGIN
  WriteString("Zap -- ");
  WriteString(s);
  WriteLn;
END Error;    

PROCEDURE  Fatal( VAR s : ARRAY OF CHAR);

BEGIN
  Error(s);
  Exit(Status(-1));
END Fatal;

PROCEDURE  Component( VAR s, t : ARRAY OF CHAR);

VAR     a : CARDINAL;

BEGIN   a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] = " ") DO INC(a) END;
        Remove(s,0,a);
        a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] > " ") DO INC(a) END;
        Substring(s,t,0,a);
        IF t[0] = ";" THEN t[0] := 0C; RETURN END;  (* Comments *)
        Remove(s,0,a);
END Component;
  
PROCEDURE Display(    address : ADDRESS;
                      mode    : ModeType);
                             
VAR
  cardval ,i : CARDINAL;
  charval    : CHAR;
  str        : LineType;

BEGIN     
  FormatAddr(address,base,8,str);
  FOR i := 1 TO 7 DO
    IF str[i] = " " THEN
      str[i] := "0";
    END;
  END;
  WriteString(str);
  WriteString(": ");
  
  CASE mode OF
 
    word      : cardval := 256*GetByte(address);
                INC(address);
                INC(cardval,GetByte(address));
                FormatCard(cardval,base,4,str);
                FOR i := 0 TO 2 DO
                  IF str[i] = " " THEN str[i] := "0" END;
                END;
                WriteString(str); |

    byte      : cardval := GetByte(address);
                FormatCard(cardval,base,2,str);
                IF str[0] = " " THEN str[0] := "0" END;
                WriteString(str); |
   
    character : charval := CHR(GetByte(address));
                cardval := ORD(charval) MOD 128;
                IF (cardval >= 32) AND (cardval < 127) THEN
                  Write('"');
                  Write(charval);
                  Write('"');Write(" ");
                ELSE
                  Write("<");
                  FormatCard(ORD(charval),base,2,str);
                  IF str[0] = " " THEN str[0] := "0" END;
                  WriteString(str);
                  Write(">");
                END;        
 
    END;(*CASE*)
END Display;
               
PROCEDURE Assign(VAR address : ADDRESS;
                     mode    : ModeType;
                     str     : LineType);
            
VAR
  value,i : CARDINAL;

BEGIN  

  CASE mode OF
   
    word      : IF NOT ScanCard(str,base,value) THEN
                  Error("Invalid Input");
                ELSIF value <= 65535 THEN                                   
                  PutByte(address,value DIV 256);
                  INC(address);
                  PutByte(address,value MOD 256);
                  INC(address);
                ELSE
                  Error("Value too large ");
                END;
                |

    byte      : IF NOT ScanCard(str,base,value) THEN
                  Error("Invalid Input");
                ELSIF value <= 255 THEN
                  PutByte(address,value);
                  INC(address);
                ELSE
                  Error("Value too large ");
                END;
                |

    character    : i := 0;
                WHILE (i < HIGH(str)) AND (str[i] <> 0C) DO
                  PutByte(address,ORD(str[i]));
                  INC(address);
                  INC(i);
                END;

  END;

END Assign;
            
VAR
  Line, temp, adds ,command: LineType;
  mode : ModeType;
  rs : Status;                 
  i, modesize : CARDINAL;
  address, a : ADDRESS;

BEGIN                                                         
  GetCommand(Line);
  Component(Line,temp);
  IF temp[0] = 0C THEN
    LOOP
      WriteString("Zap> ");
      ReadString(temp);
      CASE Result() OF
  
        EndFile,EscKey : Exit(Status(-1)) |
  
        Ok      : IF temp[0] <> 0C THEN EXIT END;
  
      ELSE
      END;   
    END;
  END;
  Check(StartFile(temp));
  mode := none;
  address := 0;
  LOOP
    WriteString(". ");
    ReadString(Line);
    IF Result() = EndFile THEN EXIT END; (* End on ^Z   *)
    WriteLn;
    Component(Line,command);    (* Mode                 *)
    Component(Line,adds);       (* Address, as a string *)
    Compress(adds);
    IF (adds[0] <> 0C) AND ScanAddr(adds,base,a) THEN 
        address := a; 
    END;
    
      CASE CAP(command[0]) OF

        "W" : mode := word;          
              modesize := 2; |

        "B" : mode := byte;          
              modesize := 1; |
                        
        "C" : mode := character;          
              modesize := 1; |

        0C  : mode := none;  |
    
        "E" : EXIT;
 
       ELSE                                         
         Error("Invalid command");
         mode := none;
       END;
                     
       IF mode <> none THEN
         LOOP
           Display(address,mode);
           WriteString(" _");
           ReadString(command);   
           CASE Result() OF

             EndFile : WriteLn; EXIT; |

             EscKey  : IF ((mode <> word) AND (address <> 0)) OR
                          ((mode = word)  AND (address >= 2)) THEN 
                         DEC(address,modesize);           
                       END;
                       WriteLn; 
           ELSE
             IF mode <> character THEN
               Compress(command);   
             END;
             IF command[0] = 0C THEN (* Carriage Return *)
               INC(address,modesize);
             ELSE (* Replace with value *)
               Assign(address,mode,command);
             END;
           END;             
         END;(*Inner LOOP*)
       END;
  END;(*Outer LOOP*)
  Check(FinishFile());
END Zap.
