IMPLEMENTATION MODULE ZapFiles;

FROM SYSTEM IMPORT
  ADDRESS;
FROM R2SysCalls IMPORT
  Status, Open, Close, Translate, ReadBlock, WriteBlock, 
  Handle, Seek, GetMessage, WriteString, WriteLn, Exit;
                                    
TYPE
  LineType = ARRAY[0..79] OF CHAR;
  Block    = ARRAY[0..511] OF CHAR;

VAR
  infilename : LineType;              
  infile     : Handle;
  block      : Block;
  blocknum   : CARDINAL;
  dirtyblock : BOOLEAN;

PROCEDURE Check( r : Status);

VAR 
  str : LineType;

BEGIN
  IF r = Ok THEN RETURN END;
  GetMessage(r,str);
  WriteString("Zap -- ");
  WriteString(str);
  WriteLn;
  Exit(Status(-1));
END Check;

PROCEDURE WriteOldBlock;

BEGIN
  Check(Seek(infile,blocknum));
  Check(WriteBlock(infile,block,FALSE));
END WriteOldBlock;

PROCEDURE AddrToBlock( address : ADDRESS) : CARDINAL;

BEGIN
  RETURN CARDINAL(address DIV 512);
END AddrToBlock;
  
PROCEDURE GetBlock(    address : ADDRESS);

BEGIN                                                 
  IF dirtyblock THEN WriteOldBlock END;
  dirtyblock := FALSE;
  blocknum := AddrToBlock(address);
  Check(Seek(infile,blocknum));
  Check(ReadBlock(infile,block));
END GetBlock;

PROCEDURE inblock(    address : ADDRESS) : BOOLEAN;

BEGIN
  RETURN AddrToBlock(address) = blocknum;
END inblock;

PROCEDURE GetByte(    address : ADDRESS) : CARDINAL;
BEGIN
  IF NOT inblock(address) THEN
    GetBlock(address);
  END;
  RETURN ORD(block[CARDINAL(address MOD 512)]);
END GetByte;

PROCEDURE PutByte(VAR address : ADDRESS;
                      value   : CARDINAL);
BEGIN
  IF NOT inblock(address) THEN
    GetBlock(address);
  END;
  dirtyblock := TRUE;
  block[CARDINAL(address MOD 512)] := CHR(value);
END PutByte;

PROCEDURE StartFile( VAR filename : ARRAY OF CHAR) : Status;
 
VAR
  rs : Status;
 
BEGIN                                                         
  rs := Translate(filename,infilename);
  IF rs = Ok THEN
    rs := Open(infilename,TRUE,infile);
    IF rs = Ok THEN
      dirtyblock := FALSE;
      blocknum := 0;
      GetBlock(0);
    END;
  END;
  RETURN rs;
END StartFile;   
     
PROCEDURE FinishFile() : Status;

BEGIN      
  IF dirtyblock THEN WriteOldBlock END;
  RETURN(Close(infile,TRUE));
END FinishFile;

END ZapFiles.
